#include "TPTP_parser.h"
#ifndef __PR_ABSTRACT_SYNTAX_H
 #include "PR_PARSER/pr_abstract_syntax.h"
#endif
#ifndef __PR_PARSER_H
 #include "PR_PARSER/pr_parser.h"
#endif
#ifndef __TPTP_SYNTAX_H
 #include "TPTP_syntax.h"
#endif
#ifndef __OPTIONS_H
 #include "options.h"
#endif

#include <iostream.h>

#ifndef __WEIGHTING_H 
 #include "weighting.h"
#endif

#include <stdlib.h>

extern HEADER_WEIGHT_LIST* HWList;

//**********************************************************************

long ParseLongInt(const char* const str)
 {
  char* tail = (char*)0;
  long val = strtol(str,&tail,0);
  return val;
 }; // long ParseLongInt(const char* const str)

HEADER_WEIGHT* CreateHW(PR_LITERAL* item)
 {
  if (PR_TERM_LIST::Length(item->Arguments()) == 2)
   {
    PR_TERM* arg1 = PR_TERM_LIST::Nth(item->Arguments(),1);
    PR_TERM* arg2 = PR_TERM_LIST::Nth(item->Arguments(),0);
    TERM header;
    if (TPTP_SYNTAX::CreateHeader(header,arg1)) 
     {
      if (arg2->Tag() == PR_TERM::ATOM_NUMBER)
       {
        return new HEADER_WEIGHT(header,ParseLongInt(arg2->Symbol()->Symbol()));
       }
      else
       {
        if (arg2->Tag() == PR_TERM::ATOM_FUNCTOR)
         {
          if (!(strcmp(arg2->Symbol()->Symbol(),"min"))) 
           {
            return new HEADER_WEIGHT(header,HEADER_WEIGHT::MIN);
           };
          if (!(strcmp(arg2->Symbol()->Symbol(),"max"))) 
           {
            return new HEADER_WEIGHT(header,HEADER_WEIGHT::MAX);
           };
         };
       };
     };
   };
  return (HEADER_WEIGHT*)0;
 }; // HEADER_WEIGHT* CreateHW(PR_LITERAL* item)

bool ProcessInputItem(PR_LITERAL* item,const char* TPTPDir,const char* problemName)
 {
  //cout << "ITEM\n     ";
  //item->Output(); cout << "\n";
     
  switch (TPTPClassify(item))
   {
    case TPTP_INCLUDE_DIRECTIVE :
     if (PR_TERM_LIST::Length(item->Arguments()) == 1)
      {
       if (item->Arguments()->Term()->Tag() == PR_TERM::ATOM_STRING)
        {
         const char* incl_file_name = item->Arguments()->Term()->Symbol()->Symbol();
         char file_name[strlen(TPTPDir) + strlen(incl_file_name) + 2];
         strcpy(file_name,TPTPDir);
         file_name[strlen(TPTPDir)] = '/';
         strcpy(file_name + strlen(TPTPDir) + 1,incl_file_name);
         file_name[strlen(TPTPDir)  + strlen(incl_file_name) + 1] = (char)0;
         return ReadFile(TPTPDir,file_name);
        }
       else
        {
         cout << "Error in " << problemName << " ";
         cout << "[" << item->Header()->LineNum() << "/" << item->Header()->CharNum() << "] : ";
         cout << "Wrong argument. " << "\n";
         return false;
        };
      }
     else 
      {
       cout << "Error in " << problemName << " ";
       cout << "[" << item->Header()->LineNum() << "/" << item->Header()->CharNum() << "] : ";
       cout << "Wrong arity. " << "\n";
       return false; 
      };
    case TPTP_INPUT_CLAUSE : 
     if (TPTP_SYNTAX::CreateClause(item))
      {
      
       return true;
      }
     else
      {
       cout << "Error in " << problemName << " ";
       cout << "[" << item->Header()->LineNum() << "/" << item->Header()->CharNum() << "] : ";
       cout << "Can not read an input clause. " << "\n";
       return false;
      };
     
    case TPTP_INPUT_FORMULA :
     cout << "************* FOF syntax is not imlemented yet! **************\n";
     return false;
    case EXT_TPTP_SET_OPTION_DIRECTIVE : 
     if (PR_TERM_LIST::Length(item->Arguments()) == 2)
      {
       if (item->Arguments()->Next()->Term()->Tag() == PR_TERM::ATOM_FUNCTOR)
        {
         char* option_name = (char*)(item->Arguments()->Next()->Term()->Symbol()->Symbol());
         char* option_value = (char*)(item->Arguments()->Term()->Symbol()->Symbol());
         char* expl = (char*)0;
         
         if (!OPTION::Set(option_name,option_value,expl))
          {
           cout << "Warning: can not set option \"" << option_name << "\",";
           cout << expl << ".\n";
           cout << "in " << problemName;
           cout << " [" << item->Arguments()->Next()->Term()->Symbol()->LineNum();
           cout << "/" << item->Arguments()->Next()->Term()->Symbol()->CharNum() << "]\n";
          };
         return true;
        }
       else
        {
         cout << "Error in " << problemName << " ";
         cout << "[" << item->Header()->LineNum() << "/" << item->Header()->CharNum() << "] : ";
         cout << "Wrong option name. " << "\n";
         return false;
        };
      }
     else 
      {
       cout << "Error in " << problemName << " ";
       cout << "[" << item->Header()->LineNum() << "/" << item->Header()->CharNum() << "] : ";
       cout << "Wrong arity. " << "\n";
       return false; 
      };
    case EXT_TPTP_HEADER_WEIGHT : 
     {
      HEADER_WEIGHT* hw = CreateHW(item);
      if (hw)
       {
        HEADER_WEIGHT_LIST::Add(hw,HWList);
        return true;
       }
      else
       {
        cout << "Error in " << problemName << " ";
        cout << "[" << item->Header()->LineNum() << "/" << item->Header()->CharNum() << "] : ";
        cout << "Can not read header weight. " << "\n";
        return false;
       };
     };
    case TPTP_UNCKNOWN_ITEM_SORT : 
     cout << "Error in " << problemName << " ";
     cout << "[" << item->Header()->LineNum() << "/" << item->Header()->CharNum() << "] : ";
     cout << "Uncknown input item type. " << "\n";
     return false;
     
   };
  return false; // to surpress warnings;
 }; // bool ProcessInputItem(PR_LITERAL* item,const char* TPTPDir,const char* problemName)

bool ReadFile(const char* TPTPDir, const char* problemName)
 {
  PR_FILE_PARS_RESULT file_pars_res;
  ParseFile(problemName,file_pars_res);
  if (file_pars_res.Error())
   {
    cout << "Error in " << problemName << " ";
    cout << "[" << file_pars_res.ErrorLineNum() << "/" << file_pars_res.ErrorCharNum() << "] : ";
    cout << file_pars_res.ErrorMessage() << "\n";
    return false;
   };
  
  for (PR_LITERAL_LIST* lst = file_pars_res.Literals(); lst; lst = lst->Next())
   {
    if (!(ProcessInputItem(lst->Literal(),TPTPDir,problemName)))
     {
      return false;
     };
   };

  return true;   
 }; // bool ReadFile(const char* TPTPDir, const char* problemName)

//**********************************************************************