#define __ANCESTORS_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif

#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif

#ifndef __MEMORY_H
 #include "memory.h"
#endif 

#ifndef __STACK_H
 #include "stack.h"
#endif

#ifndef __OUTPUT_H
 #include "output.h"
#endif  

class Clause;

//*************************************************************************

class ANCESTORS
 {
  public:
   enum BACKGR { UNCKNOWN, AXIOM, CONJECTURE, HYPOTHESIS, PREPRO, RESOL, SIMPL };
  private:
   BACKGR background;
   int number;
  public:
   ANCESTORS(int num) : background(UNCKNOWN), number(num) {}; 
   inline BACKGR Background() const { return background; };
   inline void SetBackground(BACKGR b) { background = b; };
   inline int Number() const { return number; }; 
   inline Clause** Clauses() const { return (Clause**)(this + 1); };
   inline bool PreprocessedInput() const { return (background == PREPRO); };
   inline bool Input() const
    {
     switch (background)
      {
       case AXIOM :
       case CONJECTURE :
       case HYPOTHESIS :
        return true;
       default : return false; 
      };
    };
   
   inline void* operator new(size_t,int len)
    {
     return Memory::Allocate((sizeof(ANCESTORS)/4) + len);
    };
   inline void operator delete(void* obj)
    { 
     Memory::Deallocate(obj,(sizeof(ANCESTORS)/4) + ((ANCESTORS*)obj)->Number());
    };  
    
   inline bool OneOfAncestors(const Clause* cl)
    {
     for (int i = 0; i < number; i++)
      if (((Clauses())[i]) == cl) return true;
     return false;
    };
     
   static inline const char* BackgroundStr(BACKGR b) 
    {
     switch (b)
      {
       case UNCKNOWN : return " ? ";
       case AXIOM : return "axiom";
       case CONJECTURE : return "conjecture";
       case HYPOTHESIS : return "hypothesis";
       case PREPRO : return "prepro ";
       case RESOL : return "";
       case SIMPL: return "sim. ";
       default : return " ? ";
      };
    };
 }; // class ANCESTORS

const int MAX_NUMBER_OF_ANCESTORS = 1024;
class TMP_ANCESTORS
 {
  private:
   ANCESTORS::BACKGR background;
   STACK_WITH_ITERATOR<Clause*,MAX_NUMBER_OF_ANCESTORS> ancestors;
  public:
   TMP_ANCESTORS() 
    : background(ANCESTORS::UNCKNOWN), 
      ancestors("MAX_NUMBER_OF_ANCESTORS in TMP_ANCESTORS::ancestors")
     {};
   inline void Reset() { background = ANCESTORS::UNCKNOWN; ancestors.Reset(); };
   inline void SetBackground(ANCESTORS::BACKGR b) { background = b; };
   inline void Push(Clause* cl) { ancestors.Push(cl); };
   inline ANCESTORS* NewAncestors() 
    {
     int n = ancestors.NumberOfElements();
     ANCESTORS* res = new(n) ANCESTORS(n);
     res->SetBackground(background);
     Clause** cl = res->Clauses() + n - 1;
     ancestors.ResetIterator();
     for(Clause** anc = ancestors.Next(); anc; anc = ancestors.Next())
      {
       *cl = *anc;
       cl--;
      };
     return res;
    };  
 }; // TMP_ANCESTORS


//**************************************************************************
