#define __B_MATCHING_H
#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif
#ifndef __TERM_H
 #include "term.h"
#endif
#ifndef __SHARING_H
 #include "sharing.h"
#endif
#ifndef __RESOLVENT_H
 #include "resolvent.h"
#endif

//*************************************************************************************

extern RESOLVENT Resolvent;

typedef SHARING_TREE_NODE ST_NODE; 

class ST_PATH
 {
  private:
   ST_NODE** nodes;
   ST_NODE** last;  
   long holes;
   ST_NODE** iter;
  public:
   ST_PATH() {};
   inline void SetMemory(ST_NODE** nmem) { nodes = nmem; };
   inline void ExtractLeftMost(ST_NODE* node);
   inline void ExtractLeftMost(ST_NODE**& memory,ST_NODE*& node)
    {
     SetMemory(memory);
     ExtractLeftMost(node);
     memory = FreeMem();
     node = LastNode()->Below();
    };
   inline bool Alter();
   inline bool Alter(ST_NODE**& memory,ST_NODE*& node)
    {
     if (Alter())
      {
       memory = FreeMem();
       node = LastNode()->Below();
       return true;
      }
     else return false; 
    };
   inline bool SkipEqual(ST_NODE*& node);
   inline ST_NODE* LastNode() const { return *last; }; 
   inline ST_NODE** FreeMem() const { return last + 1; };
   inline void ResetIter() { iter = nodes; };
   inline TERM IterCurrSym() const { return (*iter)->Tag(); };
   inline bool IterNext() { iter++; return (iter <= last); };  
   inline bool Collect()
    {
     ResetIter();
     do
      {
       if (!Resolvent.PushSymbol(IterCurrSym())) return false;
      }
     while (IterNext());
     return true;
    };      
 }; // class ST_PATH

class BM_INSTR
 {
  public:
   enum TAG { START, FUNC, VAR_FIRST_OCC, VAR, SUCCESS };
  private:
   TAG tag;
   TERM symbol;
   BM_INSTR* backtrack_instr;
  public:
   BM_INSTR() {};
   inline TAG& Tag() { return tag; };
   inline TERM& Symbol() { return symbol; };
   inline BM_INSTR*& BacktrackInstr() { return backtrack_instr; };
   inline void Output(BM_INSTR* code_start)
    {
     ::Output("["); ::Output((int)(this - code_start)); ::Output("] ");
     switch (Tag())
      {
       case START:         ::Output("START         "); ::Output(Symbol()); break;         
       case FUNC:          ::Output("FUNC          "); ::Output(Symbol()); break;   
       case VAR_FIRST_OCC: ::Output("VAR_FIRST_OCC "); ::Output(Symbol()); break; 
       case VAR:           ::Output("VAR           "); ::Output(Symbol()); break; 
       case SUCCESS:       ::Output("SUCCESS"); break;
      }; 
     ::Output(" backtr["); ::Output((int)(BacktrackInstr() - code_start)); ::Output("] ");
    };
   static inline void OutputCode(BM_INSTR* code)
    {
     BM_INSTR* start = code;
     while (code->Tag() != SUCCESS) { code->Output(start); ::Output("\n"); code++; };
     code->Output(start); ::Output("\n");  
    };
 }; // class BM_INSTR 


class B_MATCHING
 {
  private:
   SHARING_INDEX* index;
   BM_INSTR code[MAX_TERM_WEIGHT];
   BM_INSTR* curr_instr; 
   VAR_MULTISET var_counter;
   BM_INSTR* last_backtrack_instr;
   TERM* query;
   TERM* match;
   ST_PATH subst[MAX_NUMBER_OF_VARS];
   ST_NODE* subst_memory[MAX_TERM_WEIGHT];
   ST_NODE** subst_free_mem;
  public:
   B_MATCHING(SHARING_INDEX* ind) : index(ind), var_counter() {};
   inline void PushStart(const TERM& fun)
    {
     curr_instr = code;
     var_counter.Reset(); 
     curr_instr->Tag() = BM_INSTR::START;
     curr_instr->Symbol() = fun;
     curr_instr->BacktrackInstr() = curr_instr;
     last_backtrack_instr = curr_instr;
     curr_instr++;
     subst_free_mem = subst_memory;
    };
   inline void PushSymbol(const TERM& sym)
    {
     curr_instr->Symbol() = sym;
     curr_instr->BacktrackInstr() = last_backtrack_instr;
     if (sym.IsVariable())
      {
       int var = sym.Var();
       if (var_counter.Occurences(var))
        {
         curr_instr->Tag() = BM_INSTR::VAR;
        }
       else // first occurence of the variable
        {
         curr_instr->Tag() = BM_INSTR::VAR_FIRST_OCC;
         last_backtrack_instr = curr_instr;
        };
       var_counter.Register(var);
      }
     else // functor
      {
       curr_instr->Tag() = BM_INSTR::FUNC;
      };
     curr_instr++;
    };
  inline void PushSuccess() 
   {
    curr_instr->Tag() = BM_INSTR::SUCCESS;
    curr_instr->BacktrackInstr() = last_backtrack_instr;
    curr_instr++;
   };
  inline void SetQuery(TERM* q)
   {
    CHECK21(q->IsComplex(),"A0 in  B_MATCHING::SetQuery(TERM* q)");
    query = q;
    TERM::ITERATOR iter;
    iter.Reset(q); 
    PushStart(*q);
    while (iter.Next()) PushSymbol(iter.SymbolRef());
    PushSuccess();
    curr_instr = code; 
   }; 
  inline bool NextMatch();
  inline TERM* Query() const { return query; };
  inline TERM* Match() const { return match; };
  inline bool CollectSubst(int var) { return subst[var].Collect(); };
  inline void OutputCode() { BM_INSTR::OutputCode(code); };
 }; // class B_MATCHING

//******************** Definitions for B_MATCHING: ************************

inline bool B_MATCHING::NextMatch()
 {
  DCONTROL21("BEGIN in B_MATCHING::NextMatch()");
  ST_NODE* curr_node;
  TERM sym;
  int var;     
  switch (curr_instr->Tag())
   {
    case BM_INSTR::START:
     DCONTROL21("START in B_MATCHING::NextMatch()");

     if (!(curr_instr->Symbol().Arity()))
      {
       match = index->Const(curr_instr->Symbol());
       curr_instr++;       
       return true;
      };
    
     DCONTROL21("START.10");

     curr_node = *(index->Tree(curr_instr->Symbol()));
     
     DCONTROL21("START.20");

     if (!curr_node) return false;
     curr_instr++;
     goto perform_instr;

    case BM_INSTR::SUCCESS: goto backtrack;
    #ifdef _DEBUG21_
     default: DebugError("E0 in B_MATCHING::NextMatch()"); return false;
    #else
     #ifdef _SUPPRESS_WARNINGS_
      default: return false;
     #endif
    #endif
   };  
  
  perform_instr:
   switch (curr_instr->Tag())
    {
     case BM_INSTR::FUNC:
      CHECK21(!curr_node->IsLeaf(),"A10 in B_MATCHING::NextMatch()");
      sym = curr_instr->Symbol();
      while (curr_node->Tag() != sym)
       {
        curr_node = curr_node->Alternative();
        if (!curr_node) goto backtrack;
       };      
      curr_node = curr_node->Below();
      curr_instr++;
      goto perform_instr;

     case BM_INSTR::VAR_FIRST_OCC:
      var = curr_instr->Symbol().Var();
      subst[var].ExtractLeftMost(subst_free_mem,curr_node);
      curr_instr++;       
      goto perform_instr;

     case BM_INSTR::VAR:
      var = curr_instr->Symbol().Var();
      if (subst[var].SkipEqual(curr_node))
       {
        curr_instr++;       
        goto perform_instr;
       }       
      else goto backtrack; 
 
     case BM_INSTR::SUCCESS: 
      // curr_node must be a leaf here
      CHECK21(curr_node->IsLeaf(),"A50 in B_MATCHING::NextMatch()");
      match = curr_node->Term(); 
      return true;      

     #ifdef _DEBUG21_
      default: DebugError("E1 in B_MATCHING::NextMatch()"); return false;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default: return false;
      #endif
     #endif    
    };  
  
  backtrack:
   curr_instr = curr_instr->BacktrackInstr();
   switch (curr_instr->Tag())
    {
     case BM_INSTR::START: return false;

     case BM_INSTR::VAR_FIRST_OCC:
      var = curr_instr->Symbol().Var();
      if (subst[var].Alter(subst_free_mem,curr_node))
       {
        curr_instr++;       
        goto perform_instr;
       }       
      else goto backtrack;

     #ifdef _DEBUG21_
      default: DebugError("E2 in B_MATCHING::NextMatch()"); return false;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default: return false;
      #endif
     #endif   
    };

 }; // bool B_MATCHING::NextMatch()


//********************* Definitions for ST_PATH: **********************************

inline bool ST_PATH::Alter()
 {
  CHECK21(!holes,"A0 in ST_PATH::Alter()");
  ST_NODE* node;
  try_last:
   if (last < nodes) return false;
   node = *last;   
   last--;
   holes = holes - node->Tag().Arity() + 1;
   if (node->Alternative())
    {
     CHECK21(holes > 0,"A1 in ST_PATH::Alter()");        
     // now extract leftmost continuation    
     node = node->Alternative();
     do
      {
       last++;
       *last = node;
       holes = holes + node->Tag().Arity() - 1;  
       node = node->Below();
      }
     while (holes);
     return true;
    }  
   else goto try_last; 
 }; // bool ST_PATH::Alter() 

inline void ST_PATH::ExtractLeftMost(ST_NODE* node)
 { 
  *nodes = node;
  last = nodes;
  holes = node->Tag().Arity();
  while (holes)
   {
    last++;
    node = node->Below();
    *last = node;
    holes = holes + node->Tag().Arity() - 1;  
   };
 }; // void ST_PATH::ExtractLeftMost(ST_NODE* node) 

inline bool ST_PATH::SkipEqual(ST_NODE*& node)
 {
  for (ST_NODE** n = nodes; n <= last; n++)
   {
    TERM sym = (*n)->Tag();    
    try_curr_node:
     if (node->Tag() == sym)
      {
       node = node->Below();
      }
     else
      {
       node = node->Alternative();
       if (node) goto try_curr_node;
       return false; 
      }; 
   }; 
  return true; 
 }; // bool ST_PATH::SkipEqual(ST_NODE*& node) 





//*************************************************************************************
