#define __BACK_DEMODULATION_H
#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif
#ifndef __B_MATCHING_H
 #include "b_matching.h"
#endif
#ifndef __RESOLVENT_H
 #include "resolvent.h"
#endif
#ifndef __HASH_TABLE_H
 #include "hash_table.h"
#endif
//******************************************************************************** 

extern void AnotherBackwardSimplified(Clause* simplified_cl,Clause* simplifier);
extern RESOLVENT Resolvent;

class BACK_DEMOD
 {
  private:
   class CL_LIST;
   class CLAUSES;
  private:
   B_MATCHING matching;
   CLAUSE_QUEUE query_queue;
  public:
   BACK_DEMOD(SHARING_INDEX* ind) : matching(ind), query_queue() {};
   inline bool EnqueueQuery(Clause* cl)
    {
     if (cl->Unit())
      {
       TERM* lit = cl->Literals()->Args()->First();
       if (lit->IsEquality() 
           && lit->Positive() 
           && lit->IsOrderedEq()
           && !(lit->Args()->IsVariable()))
        {         
         query_queue.Enqueue(cl);
         return true;
        };
      };
     return false;
    };
  private:
   inline bool EligibleAsSimplifier(Clause* cl)
    {
     return (cl->Active()); 
    };
   inline bool NextMatch();
   inline void SimplifyBy(Clause* cl);
   inline bool CollectSimplified(Clause* simplified_cl,
                                 Clause* simplifier,
                                 int lit_num,
                                 TERM* redex,
                                 const TERM& replacing_term);
  public:
   inline void Simplify()
    {
     
     DCONTROL21("BEGIN in BACK_DEMOD::Simplify()"); 
     Clause* simplifier;
     while (query_queue.Dequeue(simplifier))
      {
       if (EligibleAsSimplifier(simplifier)) SimplifyBy(simplifier);
      };
     DCONTROL21("END in BACK_DEMOD::Simplify()"); 
     
    };
   void IntegrateIntoIndex(Clause* cl);   
   void RemoveFromIndex(Clause* cl); 
 }; // class BACK_DEMOD 

class BACK_DEMOD::CL_LIST
 {
  private:
   Clause* clause;
   int lit_num;   
   CL_LIST* next;
  public:
   CL_LIST(Clause* cl,int ln,CL_LIST* n) : clause(cl), lit_num(ln), next(n) {};
   inline void* operator new(size_t)
    {
     return Memory::Allocate(sizeof(CL_LIST)/4);
    };
   inline void operator delete(void* obj)
    {
     Memory::Deallocate(obj,sizeof(CL_LIST)/4);
    };   

   inline Clause* TheClause() const { return clause; }; 
   inline int LitNum() const { return lit_num; };
   inline CL_LIST*& Next() { return next; };
  
   inline CL_LIST* Insert(Clause* cl,int ln)
    {
     long cl_num = cl->Number();
     if ((!this) || TheClause()->Number() <= cl_num)
      {     
       return new CL_LIST(cl,ln,this); 
      };
     CL_LIST* res = this;
     CL_LIST* lst = res;
     CL_LIST* lst_next = lst->Next();
     while (lst_next && lst_next->TheClause()->Number() > cl_num)
      {
       lst = lst_next;
       lst_next = lst->Next();
      };     
     // here either lst_next == 0 or lst_next->TheClause()->Number() <= cl_num

    
     lst->Next() = new CL_LIST(cl,ln,lst_next);
     return res;
    };  
   
   inline CL_LIST* Remove(Clause* cl,long& num_of_removed)
    {
     num_of_removed = 0;
     long cl_num = cl->Number();
     CL_LIST* res = this;     
     CL_LIST** lst = &res;     
     check_lst:
      if ((!(*lst)) || ((*lst)->TheClause()->Number() < cl_num)) return res; 
      if ((*lst)->TheClause() == cl)
       {
        num_of_removed++;
        CL_LIST* tmp = *lst;
        *lst = tmp->Next();
        delete tmp;
       }
      else lst = &((*lst)->Next());
      goto check_lst;
    };  
  inline void Output()
   {
    ::Output("<"); ::Output(clause->Number()); ::Output(","); ::Output(lit_num); ::Output(">");
   }; 
  static inline void Output(CL_LIST* lst)
   {
    ::Output("{*");
    while (lst) { lst->Output(); lst = lst->Next(); };
    ::Output("*}");
   }; 
 }; // class BACK_DEMOD::CL_LIST

#define BACK_DEMOD_SIZE_OF_HASH_TABLE 11

class BACK_DEMOD::CLAUSES
 {
  public: 
   class ITER;
  private:
   CL_LIST* table[BACK_DEMOD_SIZE_OF_HASH_TABLE];
   long num_of_elements;
  public:
   CLAUSES() : num_of_elements(0)
    {
     for (int i = 0; i < BACK_DEMOD_SIZE_OF_HASH_TABLE; i++) table[i] = (CL_LIST*)0;
    };
   inline void Insert(Clause* cl,int ln)
    {     
     CL_LIST** lst_addr = table + (cl->Number() % BACK_DEMOD_SIZE_OF_HASH_TABLE);
     *lst_addr = (*lst_addr)->Insert(cl,ln);
     num_of_elements++;
    };
   inline void Remove(Clause* cl)
    {
     CL_LIST** lst_addr = table + (cl->Number() % BACK_DEMOD_SIZE_OF_HASH_TABLE);
     long removed;
     *lst_addr = (*lst_addr)->Remove(cl,removed);
     num_of_elements -= removed;
    };
   inline long NumberOfElements() const { return num_of_elements; };
   inline void* operator new(size_t)
    {
     return Memory::Allocate(sizeof(CLAUSES)/4);
    };
   inline void operator delete(void* obj)
    {
     Memory::Deallocate(obj,sizeof(CLAUSES)/4);
    };   
   inline void Output()
    {
     for (int i = 0; i< BACK_DEMOD_SIZE_OF_HASH_TABLE; i++) 
      {
       ::Output("[* "); ::Output(i); ::Output(" *] "); CL_LIST::Output(table[i]); ::Output("\n");
      }; 
    }; 
 }; // class BACK_DEMOD::CLAUSES

class BACK_DEMOD::CLAUSES::ITER
 {
  private:
   CLAUSES* clauses;
   int next_list;
   CL_LIST* next_elem; 
  public:
   ITER() {};
   ITER(CLAUSES* cls) { Reset(cls); };
   inline void Reset(CLAUSES* cls)
    {
     if ((!cls) || (!cls->NumberOfElements()))  { next_list = BACK_DEMOD_SIZE_OF_HASH_TABLE; return; };
     clauses = cls;
     next_list = 0;
     while (!(clauses->table[next_list]))
      {
       next_list++;
       if (next_list == BACK_DEMOD_SIZE_OF_HASH_TABLE) return;      
      };
     next_elem = clauses->table[next_list]; 
    };
   inline bool Next(CL_LIST*& el) 
    {
     if (next_list == BACK_DEMOD_SIZE_OF_HASH_TABLE) return false;
     el = next_elem;
     next_elem = next_elem->Next();
     if (!next_elem)
      {
       next_list++;
       while ((next_list < BACK_DEMOD_SIZE_OF_HASH_TABLE) && (!(clauses->table[next_list])))
        {
         next_list++;
        };
       if (next_list < BACK_DEMOD_SIZE_OF_HASH_TABLE) next_elem = clauses->table[next_list];  
      };
     return true;    
    }; 
 }; // class BACK_DEMOD::CLAUSES::ITER 


//*********************** Definitions for BACK_DEMOD: ***************************

inline bool BACK_DEMOD::CollectSimplified(Clause* simplified_cl,
                                          Clause* simplifier,
                                          int lit_num,
                                          TERM* redex,
                                          const TERM& replacing_term)
 {    
  Resolvent.Reset();
  Resolvent.PushAncestor(simplifier);
  Resolvent.PushAncestor(simplified_cl);
  Resolvent.SetBackground(ANCESTORS::SIMPL);
  TERM* lits = simplified_cl->Literals();
  for (int ln = 0; ln < simplified_cl->Length(); ln++)
   {
    TERM* lit = lits->NthArg(ln)->First();
    if (ln == lit_num) // this literal must be rewritten
     {
      CHECK21(lit->Arity(),"A0 in BACK_DEMOD::CollectSimplified(..)");
      TERM header = *lit;
      if (header.IsEquality()) header.MakeUnordered();
      if (!Resolvent.OpenNonpropLiteral(header)) return false;
      TERM::ITERATOR iter; 
      iter.Reset(lit);    
      iter.Next();
      check_position:        
       if (iter.CurrPos() == redex)
        {
         CHECK21(iter.SymbolRef().IsComplex(),"A10 in BACK_DEMOD::CollectSimplified(..)");
         if (replacing_term.IsVariable())
          {
           if (!matching.CollectSubst(replacing_term.Var())) return false;
          }
         else
          {
           TERM::ITERATOR rt_iter;
           rt_iter.Reset(replacing_term.First());
           do
            {
             if (rt_iter.SymbolRef().IsVariable())
              {
               if (!matching.CollectSubst(rt_iter.SymbolRef().Var())) return false;
              }
             else
              {
               if (!Resolvent.PushSymbol(rt_iter.SymbolRef())) return false;
              }; 
            }  
           while (rt_iter.Next());
          };
         if (iter.SkipTerm()) goto check_position;
        }
       else
        {
         if (!Resolvent.PushSymbol(iter.SymbolRef())) return false;
         if (iter.Next()) goto check_position;
        };
        
      if (!Resolvent.CloseLiteral()) return false;   
     }
    else
     {
      if (!Resolvent.PushLiteral(lit)) return false;       
     }; 
   };
  return true;   
 }; // bool BACK_DEMOD::CollectSimplified(Clause* simplified_cl,...





inline void BACK_DEMOD::SimplifyBy(Clause* cl)
 {
  TERM* lit = cl->Literals()->Args()->First();   
  TERM* t = lit->Args()->First();
  matching.SetQuery(t);

  while (NextMatch())
   {
    TERM* redex = matching.Match();
    CLAUSES* clauses = (CLAUSES*)(redex->Info()->First());  
        
    //clauses->Output(); cout << "\n\n"; // debug
    
    CLAUSES::ITER iter(clauses);    
    CL_LIST* next_pair;
   
       

    while (iter.Next(next_pair))
     {    
      //cout << (long)next_pair << "\n";  // debug
      //next_pair->Output(); cout << "\n\n"; // debug     

      Clause* curr_clause = next_pair->TheClause();
      if ((curr_clause != cl) && (curr_clause->Active()))
       {
        int curr_lit_num = next_pair->LitNum();
                    
        AnotherBackwardSimplified(curr_clause,cl);
       
        TERM replacing_term = *(lit->Args()->Next()); 
        if (CollectSimplified(curr_clause,cl,curr_lit_num,redex,replacing_term))
         {        
          //::Output("CLAUSE "); ::Output(cl); // debug
          //::Output("\n\nSIMPLIFIES "); ::Output(curr_clause); ::Output("\n\n"); // debug
          //::Output("INTO "); Resolvent.Output(); ::Output("\n\n\n\n"); // debug
          

          Resolvent.ProcessCollected();    
         };             
       };
     };
      
   };
   
 }; // void BACK_DEMOD::SimplifyBy(Clause* cl) 




inline bool BACK_DEMOD::NextMatch()
 {
  if (matching.NextMatch())
   {
    //::Output("QUERY "); ::Output(matching.Query()); ::Output("\n"); // debug
    //::Output("MATCH "); ::Output(matching.Match()); ::Output("\n\n\n"); // debug    

    return true;
   }
  else return false;
 }; // bool BACK_DEMOD::NextMatch()
  



//******************************************************************************** 
