#include "binary_resolution.h"

#ifndef __DEBUG_H
 #include "debug.h"
#endif 
#ifndef __RESOLVENT_H
 #include "resolvent.h"
#endif

//************************************************************************

BIN_ORD_RES OrderedResolution;

extern RESOLVENT Resolvent; 

bool BIN_ORD_RES::NextLiteral()
 {
  if (currentLit > premise1.LastMax()) { return false; };
   
  /***** Compilation: *****/
  
  ORMachine.CompileLiteral((premise1.MaxLiterals())[currentLit]);
  ORMachine.InversePolarity();
  
  /******** Literal compiled. ***********/
  
  premise1.Cut(currentLit);
  
  currentLit++;
  
  GlobalSubst.UnbindAll();
  RES_PAIR_LIST* candidate;
  
  if (ORMachine.FindFirstLeaf())  
   {
    process_leaf: 
     DCONTROL11("A5 in BIN_ORD_RES::NextLiteral()"); 
     ORMachine.ResetCandidatesFromLeaf();
     try_next_cand:
      candidate  = ORMachine.NextCandidate();
      if (!candidate) goto try_next_leaf;
      if (candidate->TheClause()->MinResWeight() 
          + FirstPremise()->MinResWeight() 
          > Resolvent.MaxWeight())
       {
        Resolvent.RegisterTooHeavy();
        goto try_next_cand;
       };   
            
      
      
      DCONTROL11("A7 in BIN_ORD_RES::NextLiteral()");
     
       
      //cout << "CAND\n"; ::Output((long)(candidate->Redex())); cout << "\n"; // debug
     
     
       
      premise2.LoadOptimized(index2,candidate);

      // first resolvent without factoring
       
      DCONTROL11("A9 in BIN_ORD_RES::NextLiteral()");
                     
      MakeNewResolvent(); 
       
      DCONTROL11("A10 in BIN_ORD_RES::NextLiteral()");
     
     
      // if (Resolvent.WasTautology()) { goto no_more_factors; };
       
      DCONTROL11("A11 in BIN_ORD_RES::NextLiteral()");
     
      ResetFactoringState();
       
      DCONTROL11("A12 in BIN_ORD_RES::NextLiteral()");
       
      while (NextFactor())
       {
        DCONTROL11("A15 in BIN_ORD_RES::NextLiteral()");
         
        MakeNewResolvent(); 
         
        DCONTROL11("A16 in BIN_ORD_RES::NextLiteral()");
       }; 
       
      // no_more_factors:
       DCONTROL11("A14 in BIN_ORD_RES::NextLiteral()");
       premise2.Unload();
       DCONTROL11("A14.1 in BIN_ORD_RES::NextLiteral()");
       goto try_next_cand;
     
     // no more candidates in this leaf, try to find another one
     try_next_leaf:
      DCONTROL11("A20 in BIN_ORD_RES::NextLiteral()");
      if (ORMachine.FindNextLeaf()) { goto process_leaf; };
   };
   
  // no more resolvents for this literal
  premise1.UnmarkCut();
  
  DCONTROL11("END in BIN_ORD_RES::NextLiteral()");
  
  return true;
 }; // bool BIN_ORD_RES::NextLiteral()
 
void BIN_ORD_RES::ResetFactoringState()
 {
  premise1.ResetFactoring();
  premise2.ResetFactoring();
 }; // void BIN_ORD_RES::ResetFactoring() 
 
bool BIN_ORD_RES::NextFactor()
 {
  if (premise2.WasFactored())
   {
    if (premise2.GetNextFactor())
     {
      return true;
     }
    else 
     {
      premise2.ResetFactoring();
      if (premise1.WasFactored())
       {
        return premise1.GetNextFactor();
       }
      else
       {
        return premise1.GetFirstFactor(premise1.CutLit()+1);
       };
     };
   }
  else 
   {
    if (premise2.GetFirstFactor(0))
     {
      return true;
     }
    else
     {
      if (premise1.WasFactored())
       {
        return premise1.GetNextFactor();
       }
      else
       {
        return premise1.GetFirstFactor(premise1.CutLit()+1);
       };
     };
   };
 }; // bool BIN_ORD_RES::NextFactor()
 

void BIN_ORD_RES::MakeNewResolvent()
 {
  DCONTROL11("BEGIN in BIN_ORD_RES::MakeNewResolvent()");
  
  if (!CollectResult()) return;
  
  DCONTROL11("A1 in BIN_ORD_RES::MakeNewResolvent()");
  
  // resolvent collected in Resolvent
  Resolvent.SetBackground(ANCESTORS::RESOL);
  Resolvent.PushAncestor(FirstPremise());
  Resolvent.PushAncestor(SecondPremise());
  
  DCONTROL11("A5 in BIN_ORD_RES::MakeNewResolvent()");
  
  Resolvent.ProcessCollected(); 
  
  DCONTROL11("END in BIN_ORD_RES::MakeNewResolvent()");
 }; // void BIN_ORD_RES::MakeNewResolvent() 
 
//************************************************************************ 
 
 
 
 
 
 
 
 
 
