#define __BINARY_RESOLUTION_H

#ifndef __CLAUSE_H
 #include "clause.h"
#endif
#ifndef __PREMISE_FACTORING_H
 #include "premise_factoring.h"
#endif
#ifndef __IU_COMMAND_H
 #include "iu_command.h"
#endif
#ifndef __OR_MACHINE_H
 #include "or_machine.h"
#endif
#ifndef __READCMD_H
 #include "readcmd.h"
#endif
#ifndef __RESOLVENT_H
 #include "resolvent.h"
#endif
#ifndef __WEIGHT_FORMULA_H
 #include "weight_formula.h"
#endif

extern RESOLVENT Resolvent; 

//*************************************************************

class BIN_ORD_RES;
extern BIN_ORD_RES OrderedResolution;

//*************************************************************
class BIN_ORD_RES
 {
  private:
   OR_PREMISE_WITH_FACTORING premise1,premise2;
   const int index1,index2;
   int currentLit;
   OR_MACH ORMachine;
   WEIGHT_FORMULA weight_form;
  private:
   void ResetFactoringState();
   bool NextFactor();
   void MakeNewResolvent();
  public:
   BIN_ORD_RES() 
    : premise1(), 
      premise2(),
      index1(0),
      index2(1),
      weight_form()
    {
     weight_form.SetIndex(index1);
     //ORMachine.SetIndex(index2);
    };
   inline void CollectWeightForm(Clause* c);
   inline long MinResWeight() { return weight_form.Compute(); };
   inline void Load(Clause* c);
   bool NextLiteral();
   inline Clause* FirstPremise() { return premise1.TheClause(); };
   inline Clause* SecondPremise() { return premise2.TheClause(); };
   inline bool CollectResult() 
    { 
     if (MinResWeight() + premise2.TheClause()->MinResWeight()
         > Resolvent.MaxWeight())
      {
       Resolvent.RegisterTooHeavy();
       return false;
      }; 
     Resolvent.Reset();
     return premise1.CollectNonmarkedLiterals() && premise2.CollectNonmarkedLiterals();
    }; 
 }; // class BIN_ORD_RES 
  
inline void BIN_ORD_RES::CollectWeightForm(Clause* c)
 {
  weight_form.Reset();
  TERM* lits = c->Literals();
  for (const TERM* lit_ref = lits->NthArg(lits->MaxLiterals());!lit_ref->IsNull();lit_ref = lit_ref->Next())
   weight_form.Compile(lit_ref->First()); 
 }; // inline void CollectWeightForm() 
 
 
inline void BIN_ORD_RES::Load(Clause* c)
 {
  // reset everithing here
  ORMachine.SetIndex(index2);
  premise1.Load(index1,c);
  CollectWeightForm(c);
  currentLit = 0;
 }; // inline void BIN_ORD_RES::Load(Clause* c)   
 
//************************************************************************* 