#define __BS_COMMAND_H

#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif

#ifndef __DEBUG_H
 #include "debug.h"
#endif

#ifndef __DTREE_H
 #include "dtree.h"
#endif

#ifndef __VARIABLES_H
 #include "variables.h"
#endif

#ifndef __BS_PRIMITIVES_H
 #include "bs_primitives.h"
#endif

//*********** Temporary version for debugging: ***************

extern CLAUSE_INDEX ClauseIndex; 

class BACK_SUBSUMPTION
 {
  private:
   BS_VARS subst;
   BS_PROCESSOR proc;   
  public:
   BACK_SUBSUMPTION() : subst(), proc() {};
   inline void Subsume(const Clause* cl);
 }; // class BACK_SUBSUMPTION

inline void BACK_SUBSUMPTION::Subsume(const Clause* cl)
 {

  TERM* lit = cl->Literals()->Args()->First();
  
  if (lit->Arity() == 2)
   {
    DTREE_NODE* tree = *(ClauseIndex.Tree(*lit));
    if (tree)
     {
      //cout << "*******************************************\n"; // debug 
     
      proc.Reset(*(ClauseIndex.Tree(*lit)));
    
      DTREE_TERM* v1 = subst.BSVar(0);
      DTREE_TERM* v2 = subst.BSVar(1);
      
      proc.Assign(v1); 
      
      track:
       cout << "TERM\n";
       v1->Output();
       cout << "TRACKING:\n"; // debug
       if (proc.Track(v1)) 
        {         
         cout << "SUCCESS\n"; // debug
         CHECK2(proc.CurrentNode()->Next()->IsLeaf(),"A1 in BACK_SUBSUMPTION::Subsume(..)");
        }
       else
        {
         cout << "FAILURE\n"; // debug
        }; 
        
      
       if (proc.Alter(v1)) goto track;
      
      /***********
      new_v2:     
      proc.Assign(v2);            
      
      
      
      //cout << "(((((((((((((((  V1 )))))))))))))))))))\n";  
      //v1->Output();      
      //cout << "         (((((((((((((((  V2 )))))))))))))))))))\n";  
      //v2->Output();       
      
      CHECK2(proc.CurrentNode()->IsLeaf(),"A1 in BACK_SUBSUMPTION::Subsume(..)");
      
      while (proc.Alter(v2)) 
       {
        //cout << "(((((((((((((((  V1 )))))))))))))))))))\n";  
        //v1->Output();      
        //cout << "         (((((((((((((((  V2 )))))))))))))))))))\n";  
        //v2->Output(); 
        
        CHECK2(proc.CurrentNode()->IsLeaf(),"A2 in BACK_SUBSUMPTION::Subsume(..)");
      
        
       };
      if (proc.Alter(v1)) goto new_v2; 
      
      ******************/
      
     };
   };
  
 
 }; // void BACK_SUBSUMPTION::Subsume(const Clause* cl)


//*************************************************************************

enum BS_TAG
 {
  BS_START,
  BS_END,
  BS_VAR,
  BS_FUNC
 }; // enum BS_TAG

class BS_COMMAND_ARG
 {
  private:
   TERM content;
  public:
   BS_COMMAND_ARG() {};
   inline DTREE_TERM* Var() const { return (DTREE_TERM*)(content.First()); };
   inline void SetVar(DTREE_TERM* var) { content.SetFirst((TERM*)var); };
   inline TERM Func() const { return content; };
   inline void SetFunc(const TERM f) { content = f; };
   inline void InversePolarity() { content.InversePolarity(); };
 }; // class BS_COMMAND_ARG
 
class BS_COMMAND
 {
  private:
   BS_TAG tag;
   BS_COMMAND_ARG arg;
  public:
   BS_COMMAND() {};
   inline BS_TAG& Tag() { return tag; };
   inline BS_COMMAND_ARG& Arg() { return arg; };
   inline void Set(const BS_TAG t,DTREE_TERM* var) 
    {
     tag = t;
     arg.SetVar(var);
    };
   inline void Set(const BS_TAG t,const TERM f) 
    {
     tag = t;
     arg.SetFunc(f);
    };
  public:
   static void OutputTag(BS_TAG t) 
    {
     switch (t)
      {
       case BS_START: ::Output("BS_START"); break;
       case BS_END: ::Output("BS_END"); break;
       case BS_VAR: ::Output("BS_VAR"); break;
       case BS_FUNC: ::Output("BS_FUNC"); break;
      };
    };
   inline void Output(BS_VARS& subst)
    {
     OutputTag(tag);
     ::Output(" ");
     switch (tag)
      {
       case BS_START: ::Output(arg.Func()); break;
       case BS_END: break;
       case BS_VAR: ::OutputVar(subst.VarNum(arg.Var())); break;
       case BS_FUNC: ::Output(arg.Func()); break;
      };
    };
 }; // class BS_COMMAND

class BS_LIT_CODE
 {
  private:
   VAR_MULTISET vars;
   BS_COMMAND code[MAX_TERM_WEIGHT+1];
   BS_COMMAND* command;   
  public:
   BS_LIT_CODE() : vars(), command(code) {};
   inline void ResetWriting() { command = code; vars.Reset(); };
   inline BS_COMMAND* CurrentCommand() const { return command; };
   inline void NextCommand() { command++; };
   inline VAR_MULTISET& Vars() { return vars; };
  public: // for debugging
   inline void Output(BS_VARS& subst)
    {
     BS_COMMAND* c = code;
     do
      {
       c->Output(subst);
       ::Output('\n');
       c++;
      }
     while (c->Tag() != BS_END);
     ::Output("VARIABLES ");
     vars.Output();
     ::Output('\n');
    };
 }; // class BS_LIT_CODE

class BS_CLAUSE_CODE
 {
  private:
   BS_VARS subst;
   BS_LIT_CODE literal[MAX_NUMBER_OF_LITERALS];
   int next_lit;
  public:
   BS_CLAUSE_CODE() : subst(), next_lit(0)
    {
     for (int i = 1; i < MAX_NUMBER_OF_LITERALS; i++)
      {
       literal[i].BS_LIT_CODE();
      };
    };
   inline void Reset() { next_lit = 0; }; 
   inline int NextLiteral() { return next_lit; };
   inline void CloseLit() { next_lit++; };
   inline int NumOfLiterals() const { return next_lit; };
   inline BS_LIT_CODE& LitCode(const int n) { return literal[n]; };
   inline BS_VARS& Subst() { return subst; };
  public: // for debugging
   inline void Output()
    {
     for (int lit = 0; lit < NumOfLiterals(); lit++)
      {
       LitCode(lit).Output(subst);
       ::Output('\n');
      };    
    };   
 }; // class BS_CLAUSE_CODE

class BS_COMPILER
 {
  public:
   static inline void Compile(Clause* cl,BS_CLAUSE_CODE& code)
    {
     code.Reset();
     for (TERM* lit_ref = cl->Literals()->Args(); !(lit_ref->IsNull()); lit_ref = lit_ref->Next())
      {
       Compile(lit_ref->First(),code.LitCode(code.NextLiteral()),code.Subst());
       code.CloseLit();
      };
    }; // void Compile(Clause* cl,BS_CLAUSE_CODE& code)
    
   static inline void Compile(TERM* lit,BS_LIT_CODE& code,BS_VARS& subst) 
    {
     code.ResetWriting();
     TERM header = *lit;
     code.CurrentCommand()->Set(BS_START,header);
     code.NextCommand();
     if (header.Arity())
      {
       TERM::ITERATOR iter;
       iter.Reset(lit);
       iter.Next();
       do
        {
         if (iter.CurrentSymIsVar())
          {
           int var = iter.Symbol().Var();
           code.Vars().Register(var);
           code.CurrentCommand()->Set(BS_VAR,subst.BSVar(var));          
          }
         else
          {
           code.CurrentCommand()->Set(BS_FUNC,iter.Symbol());
          };
         code.NextCommand();
        }
       while (iter.Next());     
      };
     code.CurrentCommand()->Tag() = BS_END;
    }; // void void Compile(TERM* lit,BS_LIT_CODE& code,BS_VARS& subst) 
 }; // class BS_COMPILER


//*************************************************************************


























