#include "bs_machine.h"

//***********************************************************************

//****************** Definitions for BS_MACHINE: *********************

void BS_MACHINE::BSubsume(const Clause* cl)
 {
  DCONTROL14("BEGIN in BS_MACHINE::BSubsume(..)");
  
  //cout << "QUERY "; ::Output(cl); cout << "\n"; // debug
  
  DCONTROL14("A0 in BS_MACHINE::BSubsume(..)");
  
  if (candidates.RetrieveCandidates(cl)) 
   {        
   
    DCONTROL14("A1 in BS_MACHINE::BSubsume(..)");
   
    //candidates.Output(); cout << "\n"; // debug
  
    if ((candidates.CollectUniqueVars()) 
        && (candidates.FilterSupportedVars()))
     {   
      candidates.Squeeze();
     }; 
    
    candidates.CompileJoin();
        
    candidates.LoadFirstClause();
    
    while (candidates.EqualizeClause())  
     {
      
      if (candidates.FindFirstLocalSubstitutions())
       { 
        candidates.FindOtherLocalSubstitutions();
        
        if (candidates.Join() && SubsumptionAllowed(cl,candidates.CurrentClause()))
         {
          
          AnotherBackwardSubsumed((Clause*)(candidates.CurrentClause()));
          
         };    
       };
      if (!(candidates.LoadNextClause())) 
       {
        DCONTROL14("END2 in BS_MACHINE::BSubsume(..)"); 
        return;
       }; 
     };    
   };
  DCONTROL14("END1 in BS_MACHINE::BSubsume(..)"); 
 }; // void BS_MACHINE::BSubsume(const Clause* cl)



//****************** Definitions for BS_CANDIDATES: *********************

bool BS_CANDIDATES::RetrieveCandidates(const Clause* cl) 
 {
  DCONTROL14("BEGIN in BS_CANDIDATES::RetrieveCandidates(const Clause* cl)");
  NewQueryClause(cl);

  const TERM* literals = cl->Literals();
  int length = literals->Arity();  
  const TERM* lit_ref = literals->Args(); 
  for (int lit_num = 0; lit_num < length; lit_num++)
   {
    NewLiteral(lit_num,lit_ref->First());
           
    if (!(RetrieveCandidates(lit_ref->First(),false,path_index))) return false;
      
    EndOfLiteral();
     
    /********* experimental **********   
    if (LastLitCand()->Ground())
     {
      return RetrieveCandidatesFromSmallIndex(lit_ref);
     }; 
    
    *********************************/
     
    lit_ref = lit_ref->Next();   
   };
  DCONTROL14("END in BS_CANDIDATES::RetrieveCandidates(const Clause* cl)"); 
  return true; 
 }; // bool BS_CANDIDATES::RetrieveCandidates(const Clause* cl) 
   
bool BS_CANDIDATES::RetrieveCandidatesFromSmallIndex(const TERM* ground_lit_ref) 
 {
  DCONTROL14("BEGIN in BS_CANDIDATES::RetrieveCandidatesFromSmallIndex(..)");
  
  Reset();
  SMALL_PATH_INDEX* small_path_index = SMALL_PATH_INDEX::SpecIndex(ground_lit_ref->First());
  if (!small_path_index) return false;
  
  if (query_clause->Unit())
   {
     
    NewLiteral(0,ground_lit_ref->First());
       
    PI_C_LIST* c_list = small_path_index->AllClauses();
       
    if (!(c_list)) return false;
       
    NewCandidatesProp(c_list);
       
    EndOfLiteral();
       
    DCONTROL14("END3 in BS_CANDIDATES::RetrieveCandidatesFromSmallIndex(..)"); 
       
    return true;
   }
  else // nonunit
   {
    DCONTROL14("A4 in BS_CANDIDATES::RetrieveCandidatesFromSmallIndex(..)");
    
    const TERM* literals = query_clause->Literals();
    int length = literals->Arity();  
    const TERM* lit_ref = literals->Args(); 
    for (int lit_num = 0; lit_num < length; lit_num++)
     {
      if (lit_ref != ground_lit_ref)
       {
        NewLiteral(lit_num,lit_ref->First());
       
        if (!(RetrieveCandidates(lit_ref->First(),false,small_path_index))) 
         {
          DCONTROL14("END2 in BS_CANDIDATES::RetrieveCandidatesFromSmallIndex(..)"); 
          return false;
         }; 
       
        EndOfLiteral();
       }; 
      lit_ref = lit_ref->Next();   
     };
   
    DCONTROL14("END1 in BS_CANDIDATES::RetrieveCandidatesFromSmallIndex(..)"); 
    return true; 
   }; 
 }; // bool BS_CANDIDATES::RetrieveCandidatesFromSmallIndex(const TERM* ground_lit_ref) 

bool BS_CANDIDATES::RetrieveCandidates(const TERM* lit,bool inverse_polarity,PATH_INDEX* path_index)
 { 
 
  DCONTROL14("BEGIN in BS_CANDIDATES::RetrieveCandidates(const TERM* lit,bool inverse_polarity,PATH_INDEX* path_index)");
  TERM header = *lit;
  
  DCONTROL14("A0 in BS_CANDIDATES::RetrieveCandidates(const TERM* lit,bool inverse_polarity,PATH_INDEX* path_index)");
    
  
  if (inverse_polarity) header.InversePolarity();
  int arity = header.Arity();
  if (arity)
   {
    if (header.IsEquality()) header.MakeUnordered();

    DCONTROL14("A1 in BS_CANDIDATES::RetrieveCandidates(const TERM* lit,bool inverse_polarity,PATH_INDEX* path_index)");
    
    
    PI_LEVEL* lev = path_index->Tree(header)->Subterms();
    
    DCONTROL14("A2 in BS_CANDIDATES::RetrieveCandidates(const TERM* lit,bool inverse_polarity,PATH_INDEX* path_index)");
    
    for (const TERM* arg_ref = lit->Args(); !(arg_ref->IsNull()); arg_ref = arg_ref->Next())
     {
           
     
      DCONTROL14("A3 in BS_CANDIDATES::RetrieveCandidates(const TERM* lit,bool inverse_polarity,PATH_INDEX* path_index)");
    
     
      if (RetrieveCandidates(arg_ref,lev)) { lev++; }
      else 
       {
        DCONTROL14("END2 in BS_CANDIDATES::RetrieveCandidates(const TERM* lit,bool inverse_polarity,PATH_INDEX* path_index)"); 
        return false; 
       };
      DCONTROL14("A4 in BS_CANDIDATES::RetrieveCandidates(const TERM* lit,bool inverse_polarity,PATH_INDEX* path_index)");      
     };
   }
  else // propositional 
   {
    DCONTROL14("A5 in BS_CANDIDATES::RetrieveCandidates(const TERM* lit,bool inverse_polarity,PATH_INDEX* path_index)");   
   
    PI_C_LIST* c_lst = path_index->PropOccurences(*lit,inverse_polarity); 
    if (!c_lst) 
     {
      DCONTROL14("END3 in BS_CANDIDATES::RetrieveCandidates(const TERM* lit,bool inverse_polarity,PATH_INDEX* path_index)"); 
      return false;
     }; 
    NewCandidatesProp(c_lst);
   };  
  DCONTROL14("END1 in BS_CANDIDATES::RetrieveCandidates(const TERM* lit,bool inverse_polarity,PATH_INDEX* path_index)"); 
  return true;   
 }; // bool BS_CANDIDATES::RetrieveCandidates(const TERM* lit,bool inverse_polarity,PATH_INDEX* path_index)


bool BS_CANDIDATES::RetrieveCandidates(const TERM* lit,
                                       bool inverse_polarity,
                                       SMALL_PATH_INDEX* small_path_index)
 {
 
  DCONTROL14("BEGIN in BS_CANDIDATES::RetrieveCandidates(..,SMALL_PATH_INDEX* small_path_index)");
 
  TERM header = *lit;
  if (inverse_polarity) header.InversePolarity();
  int arity = header.Arity();    
  if (arity)
   {
    PI_NODE* node = small_path_index->Tree(header);
    if (!node) return false;
    PI_LEVEL* lev = node->Subterms();
    
    for (const TERM* arg_ref = lit->Args(); !(arg_ref->IsNull()); arg_ref = arg_ref->Next())
     {
      if (RetrieveCandidates(arg_ref,lev)) { lev++; }
      else return false; 
     };
   
   }
  else // propositional 
   {
    PI_C_LIST* c_lst = small_path_index->PropOccurences(*lit,false); 
    if (!c_lst) return false;
    NewCandidatesProp(c_lst);
   };  
  return true;   
 }; // bool BS_CANDIDATES::RetrieveCandidates(const TERM* lit,bool inverse_polarity,SMALL_PATH_INDEX* small_path_index)


bool BS_CANDIDATES::RetrieveCandidates(const TERM* t_ref,PI_LEVEL* lev)
 { 
  
  DCONTROL14("BEGIN in BS_CANDIDATES::RetrieveCandidates(const TERM* t_ref,PI_LEVEL* lev)");
  
  if (!(lev->Occurences())) 
   {
    DCONTROL14("END7 in BS_CANDIDATES::RetrieveCandidates(const TERM* t_ref,PI_LEVEL* lev)");

    return false;
   }; 
  DCONTROL14("A0 in BS_CANDIDATES::RetrieveCandidates(const TERM* t_ref,PI_LEVEL* lev)");
 
 
  if (t_ref->IsVariable())
   {
   
    DCONTROL14("A1 in BS_CANDIDATES::RetrieveCandidates(const TERM* t_ref,PI_LEVEL* lev)");
 
    NewCandidatesVar(t_ref->Var(),lev->Occurences());
    
    DCONTROL14("END6 in BS_CANDIDATES::RetrieveCandidates(const TERM* t_ref,PI_LEVEL* lev)");

    
    return true;   
   }
  else // reference
   {
   
    DCONTROL14("A2 in BS_CANDIDATES::RetrieveCandidates(const TERM* t_ref,PI_LEVEL* lev)");
 
    PI_NODE* node = lev->ComplexSubterms();
    TERM* trm = t_ref->First();
    TERM header = *trm;
    while ((node) && (node->Header() != header)) node = node->Alternative();
    if (node)
     {
      int arity = header.Arity();
      if (arity)
       {
        lev = node->Subterms();
        
        for (const TERM* arg_ref = trm->Args(); !(arg_ref->IsNull()); arg_ref = arg_ref->Next())
         {
          if (RetrieveCandidates(arg_ref,lev)) { lev++; }
          else 
           {
            DCONTROL14("END5 in BS_CANDIDATES::RetrieveCandidates(const TERM* t_ref,PI_LEVEL* lev)");

            return false; 
           };
         };
        DCONTROL14("END4 in BS_CANDIDATES::RetrieveCandidates(const TERM* t_ref,PI_LEVEL* lev)");
 
        return true; 
       }
      else // constant  
       if (node->ConstOccurences())
        {
         NewCandidatesConst(node->ConstOccurences());
         DCONTROL14("END3 in BS_CANDIDATES::RetrieveCandidates(const TERM* t_ref,PI_LEVEL* lev)");

         return true;
        }
       else
        {
         DCONTROL14("END2 in BS_CANDIDATES::RetrieveCandidates(const TERM* t_ref,PI_LEVEL* lev)");

         return false;
        };  
     }
    else
     {
      DCONTROL14("END1 in BS_CANDIDATES::RetrieveCandidates(const TERM* t_ref,PI_LEVEL* lev)");
      return false; 
     };    
   };    
 }; // bool BS_CANDIDATES::RetrieveCandidates(const TERM* t_ref,PI_LEVEL* lev)


void BS_CANDIDATES::Output()
 {
  ::Output("<<< ");
  for (BS_LIT_CANDIDATES* l_cand = lit_candidates; l_cand < next_lit_cand; l_cand++) 
   {
    l_cand->Output(); ::Output(' ');
   };
  ::Output(" >>>");
 }; // void BS_CANDIDATES::Output() 
    
void BS_CANDIDATES::OutputLocalSubstitutions()
 {
  for (BS_LIT_CANDIDATES* l_cand = lit_candidates; l_cand < next_lit_cand; l_cand++) 
   {
    l_cand->OutputLocalSubstitutions(); ::Output(' ');
   };
 }; // void BS_CANDIDATES::OutputLocalSubstitutions() 


//********************* Definitions for BS_LIT_CANDIDATES: *************

void BS_LIT_CANDIDATES::Output()
 {
  //::Output("lit"); ::Output(lit_num); 
  ::Output("{");
  for (BS_CAND* cand = first_cand; cand <= last_cand; cand++)
   {
    cand->Output();
    if (cand < last_cand) ::Output(" ");
   };
  ::Output("}"); 
 }; // void BS_LIT_CANDIDATES::Output()
     
void BS_LIT_CANDIDATES::OutputLocalSubstitutions()
 {     
  ::Output("lit"); ::Output(lit_num); ::Output("{");
  if ((!propositional) && (variables.Size()))
   {
    local_subst_list.Output();
   };
  ::Output("}");
 }; // void BS_LIT_CANDIDATES::OutputLocalSubstitutions() 


//********************* Definitions for BS_CAND: ***********************

void BS_CAND::Output()
 {  
  /******
  switch (tag)
   {
    case PROP_LIT :
     ::Output("<p>");    
     break;
    case CONST :
     ::Output("<c>");
     break;
    case VAR_FIRST_OCC :
     ::Output("<vfo>");
     break; 
    case VAR :
     ::Output("<v>");
     break;
   };
   
  if (NonEmpty()) 
   {   
    switch (tag)
   {
    case PROP_LIT :
     CList()->Output();
     break;
    case CONST :
     CLList()->Output();
     break;
    case VAR_FIRST_OCC : 
    case VAR :
     CLTList()->Output();
     break;
   };
   } 
  else ::Output("{empty}");
  *********/
  ::Output(PI_C_LIST::Length(CList()));
  
 }; // void BS_CAND::Output() 


//********************* Definitions for BS_JOIN_COMMAND: *****************

void BS_JOIN_COMMAND::Output()
 {
  switch (tag)
   {
    case NEXT : ::Output("NEXT"); break;
    case MOVE : 
     ::Output("MOVE ");
     ::Output(arg.var); 
     break; 
    case COMPARE : 
     ::Output("COMPARE ");
     ::Output(arg.var);     
     break;
    case SUCCESS : ::Output("SUCCESS"); break;   
   };
 }; // void BS_JOIN_COMMAND::Output()

//********************* Definitions for BS_JOIN_MACHINE: *****************

void BS_JOIN_MACHINE::OutputCode()
 {
  BS_JOIN_COMMAND* comm = code;
  while (comm->Tag() != BS_JOIN_COMMAND::SUCCESS)
   {
    comm->Output();
    ::Output("  ");
    comm++;
   };     
  comm->Output();
 }; // void BS_JOIN_MACHINE::OutputCode()


//***********************************************************************
