
#define __BS_MACHINE_H
#ifndef __DEBUG_H
 #include "debug.h"
#endif 
#ifndef __TERM_H
 #include "term.h"
#endif
#ifndef __CLAUSE_H
 #include "clause.h"
#endif
#ifndef __STACK_H
 #include "stack.h"
#endif  
#ifndef __PATH_INDEX_H
 #include "path_index.h"
#endif
#ifndef __VARIABLES_H
 #include "variables.h"
#endif 

extern void AnotherBackwardSubsumed(Clause* const cl);


//#ifndef __TIMER_H
// #include "timer.h"
//#endif 

//***************************************************************************

class BS_SUBST
 {
  private:
   TERM term[MAX_NUMBER_OF_VARS];
   int lit_num;
  public:
   inline void Put(int var,TERM t) { term[var] = t; };
   inline bool Check(int var,TERM t) { return (term[var] == t); }; 
   inline TERM Term(int var) const { return term[var]; };
   inline int& LitNum() { return lit_num; };
 }; // class BS_SUBST

class BS_LOCAL_SUBST_LIST 
 {
  private: // structure
   BS_SUBST local_subst[MAX_NUMBER_OF_LITERALS];
   BS_SUBST* next_local_subst; 
   BS_SUBST* cursor;   
  public:
   inline void Reset() { next_local_subst = local_subst; };
   inline void Next() { next_local_subst++; };
   inline BS_SUBST* Current() { return next_local_subst; };
   inline void ResetCursor() { cursor = local_subst - 1; };
   inline bool ShiftCursor() { cursor++; return (cursor < next_local_subst); };
   inline BS_SUBST* Cursor() { return cursor; };
  public: // for debugging
   inline void Output() 
    { 
     ::Output(next_local_subst - local_subst);
     ::Output(" l.s. ");
    };    
 }; // class BS_LOCAL_SUBST_LIST 

class BS_JOIN_COMMAND
 {
  public: 
   enum TAG { NEXT, MOVE, COMPARE, SUCCESS };
  private:
   TAG tag;
   union
    {     
     int var;
     BS_LOCAL_SUBST_LIST* subst_list; 
    } arg;
  public:
   inline TAG Tag() const { return tag; };
   inline void MakeNEXT(BS_LOCAL_SUBST_LIST* lsl)
    {
     tag = NEXT;
     arg.subst_list = lsl;
    };  
   inline BS_LOCAL_SUBST_LIST* SubstList() const { return arg.subst_list; }; 
   inline void MakeMOVE(int var)
    {
     tag = MOVE;
     arg.var = var;
    };
   inline void MakeCOMPARE(int var)
    {
     tag = COMPARE;
     arg.var = var;
    }; 
   inline int Var() const { return arg.var; }; 
   inline void MakeSUCCESS()
    {
     tag = SUCCESS; 
    };  
  public: // for debugging
   void Output();    
 }; // class BS_JOIN_COMMAND

class BS_JOIN_MACHINE
 {
  private: // structure
   BS_JOIN_COMMAND code[MAX_CLAUSE_WEIGHT];
   BS_JOIN_COMMAND* next_command;   
   BS_SUBST global_subst;
   VAR_MULTISET variables;  
   STACK<BS_JOIN_COMMAND*,MAX_NUMBER_OF_LITERALS> backtrack_points;
   bool no_join;
  public: 
   BS_JOIN_MACHINE()
    : backtrack_points("MAX_NUMBER_OF_LITERALS in BS_JOIN_MACHINE::backtrack_points")
    {};   
   inline void ResetCompilation()
    {
     variables.Reset();
     next_command = code;
    };
   inline void CompileLiteral(VAR_MULTISET& lit_vars,BS_LOCAL_SUBST_LIST* subst_list)
    {
     if (lit_vars.Size())
      {
       next_command->MakeNEXT(subst_list);
       next_command++;
       lit_vars.ResetIterator();
       const int* next_var = lit_vars.NextVar();
       do 
        {
         variables.Register(*next_var);
         if (variables.Occurences(*next_var) == 1) // first occurence
          {
           next_command->MakeMOVE(*next_var);
          }
         else
          {
           next_command->MakeCOMPARE(*next_var);
          };
         next_command++;           
        }
       while ((next_var = lit_vars.NextVar())); 
      };
    }; 
   inline void EndOfClause() 
    {
     next_command->MakeSUCCESS(); 
     no_join = (code->Tag() == BS_JOIN_COMMAND::SUCCESS); 
    }; 
   inline bool Run()
    {
     if (no_join) return true;
     backtrack_points.Reset();
     BS_JOIN_COMMAND* current_command = code;
     BS_LOCAL_SUBST_LIST* current_subst_list;
     int var;
     perform_command:
      switch (current_command->Tag())
       {
        case BS_JOIN_COMMAND::NEXT : 
         current_subst_list = current_command->SubstList();
         if (!(current_subst_list->ShiftCursor())) goto backtrack;
         backtrack_points.Push(current_command);
         break;
        case BS_JOIN_COMMAND::MOVE :
         var = current_command->Var();
         global_subst.Put(var,current_subst_list->Cursor()->Term(var));
         break;
        case BS_JOIN_COMMAND::COMPARE :
         var = current_command->Var();
         if (!(global_subst.Check(var,current_subst_list->Cursor()->Term(var)))) 
          goto backtrack;
         break;        
        case BS_JOIN_COMMAND::SUCCESS : return true;
       };
      current_command++; 
      goto perform_command;
     backtrack:
      if (backtrack_points.Empty()) return false;
      current_command = backtrack_points.Pop();
      goto perform_command;
    };
    
  public: // for debug
   void OutputCode();
 }; // class BS_JOIN_MACHINE

class BS_CAND
 {
  public:
   enum TAG { PROP_LIT, CONST, VAR_FIRST_OCC, VAR };
    
  private:
   TAG tag;   
   int var;  
   union
    {
     PI_C_LIST* c_list;
     PI_CL_LIST* cl_list;
     PI_CLT_LIST* clt_list;
    } candidates; 
  public:
   BS_CAND() {}; 
   inline BS_CAND& operator= (BS_CAND& cand)
    {
     tag = cand.tag;
     var = cand.var;
     candidates.c_list = cand.candidates.c_list;
     return (*this);
    };
   inline TAG& Tag() { return tag; };
   inline int& Var() { return var; };
   inline PI_C_LIST*& CList() { return candidates.c_list; };
   inline PI_CL_LIST*& CLList() { return candidates.cl_list; };
   inline PI_CLT_LIST*& CLTList() { return candidates.clt_list; };   
  public:
   inline const Clause* TopClause() const;
   inline long TopClauseNumber() const { return candidates.c_list->ClauseNumber(); };
   inline bool NonEmpty() const { return (bool)candidates.c_list; }; 
   inline void SkipYounger(long cl_num); 
   inline void SkipYoungerOrEqual(long cl_num);  
  public: // methods for debugging
   void Output();
 }; // class BS_CAND
 
class BS_LIT_CANDIDATES
 { 
  private: // structure
   VAR_MULTISET variables;   
   bool propositional;
   int lit_num;
   const TERM* term;
   BS_CAND* first_cand;
   BS_CAND* last_cand; 
  private: // substitutions
   BS_LOCAL_SUBST_LIST local_subst_list;   
  public:
   BS_LIT_CANDIDATES() : variables() {};
   inline VAR_MULTISET& Variables() { return variables; };
   inline const TERM* Term() { return term; };
   inline BS_CAND* LastCand() const { return last_cand; };
   inline BS_LOCAL_SUBST_LIST* LocalSubstList() { return &local_subst_list; };
   inline void Open(int ln,const TERM* lit,BS_CAND* cand) 
    { 
     lit_num = ln; 
     term = lit;
     first_cand = cand; 
     propositional = false; 
     variables.Reset();
    };   
   inline bool NonGround() const { return (bool)variables.Size(); }; 
   inline bool Ground() const { return !(variables.Size()); }; 
   inline void Close(BS_CAND* cand) { last_cand = cand; }; 
   inline bool RegisterVar(int var) 
    {
     variables.Register(var); 
     return (variables.Occurences(var) == 1);
    };
   inline void MarkAsPropositional() { propositional = true; };    
   
   inline void ShiftAndSqueeze(BS_CAND*& to,VAR_MULTISET& supp_vars)
    {
     BS_CAND* cand = first_cand; 
     first_cand = to;
     check_cand:
      if (cand > last_cand) { last_cand = to-1; return; };
      if (cand->Tag() == BS_CAND::VAR_FIRST_OCC) 
       {
        if (supp_vars.Occurences(cand->Var())) // supported variable, skip it
         {
          cand++;
          goto check_cand;
         }; 
       };
      if (cand != to)
       {
        *to = *cand;
       };        
      to++;
      cand++;
      goto check_cand;     
    };
   
   inline bool FindFirstLocalSubst(int cl_num)
    {
     local_subst_list.ResetCursor();
     local_subst_list.Reset();     
     return FindLocalSubst(cl_num,local_subst_list.Current());
    };
   inline bool FindNextLocalSubst(int cl_num)
    {
     if (propositional) return false;
     local_subst_list.Next();     
     CHECK3(first_cand->CList(),"A1 in BS_LIT_CANDIDATES::FindNextLocalSubst(..)");
     first_cand->CList() = first_cand->CList()->Next();
     if (!(first_cand->CList())) return false;
     return FindLocalSubst(cl_num,local_subst_list.Current());
    };
   
   inline void FindAllOtherLocalSubst(int cl_num)
    {
     while (FindNextLocalSubst(cl_num)) {};     
    };
      
   
   inline bool FindLocalSubst(int cl_num,BS_SUBST* loc_subst)
    {
     if (propositional) return true;
     
     int current_lit_num = first_cand->CLList()->LitNum();     
     
     BS_CAND* cand = first_cand;
     check_cand:
      if (cand > last_cand) { loc_subst->LitNum() = current_lit_num; return true; };
      equalize_lit:
      
       while ((cand->CLList()) 
              && (cand->TopClause()->Number() == cl_num)
              && (cand->CLList()->LitNum() > current_lit_num))
        {      
         cand->CLList() = cand->CLList()->Next();
        };
       if ((cand->CLList()) 
            && (cand->TopClause()->Number() == cl_num))
        {
         if (cand->CLList()->LitNum() == current_lit_num)
          {
           switch (cand->Tag())
            {             
             case BS_CAND::VAR_FIRST_OCC : 
              loc_subst->Put(cand->Var(),cand->CLTList()->Term());
              break;
             case BS_CAND::VAR :
              if (!(loc_subst->Check(cand->Var(),cand->CLTList()->Term())))
               {
                first_cand->CLList() = first_cand->CLList()->Next();
                if ((first_cand->CLList()) && (first_cand->TopClause()->Number() == cl_num))
                 {
                  cand = first_cand;
                  current_lit_num = first_cand->CLList()->LitNum();     
                  goto equalize_lit;
                 }
                else return false; 
               };
             
              break;
             default : break; 
            };
                      
           cand++;
           goto check_cand;
          }
         else // new value for current_lit_num
          {
           current_lit_num = cand->CLList()->LitNum();
           cand = first_cand;
           goto equalize_lit;
          };
        }     
       else return false; 
    }; //  bool FindLocalSubst(int cl_num,BS_SUBST* loc_subst)
    
  public: // methods for debugging
   void Output();
   void OutputLocalSubstitutions();   
 }; // class BS_LIT_CANDIDATES


//extern TIMER EqClauseTimer; // debug


class BS_CANDIDATES
 { 
  private: // structure
   PATH_INDEX* path_index;
   const Clause* query_clause;
   BS_CAND candidates[MAX_CLAUSE_WEIGHT];
   VAR_MULTISET all_vars;
   VAR_MULTISET supported_vars;
   BS_CAND* next_cand;
   BS_LIT_CANDIDATES lit_candidates[MAX_NUMBER_OF_LITERALS];
   BS_LIT_CANDIDATES* next_lit_cand;   
  private: // registers
   long current_clause_num;
   const Clause* current_clause;
  private: // joins
   BS_JOIN_MACHINE join_machine;  
  public: 
   BS_CANDIDATES(PATH_INDEX* pi) : path_index(pi), all_vars(), supported_vars(), join_machine() { Reset(); };   
   inline void Reset() 
    {
     next_cand = candidates;
     next_lit_cand = lit_candidates;
     all_vars.Reset();
    };
   inline const Clause* CurrentClause() const { return current_clause; };
    
   inline void NewQueryClause(const Clause* cl) 
    {
     Reset();
     query_clause = cl;
    }; 
   inline void MakeCurrentClause(const Clause* cl)
    {
     current_clause = cl;
     current_clause_num = cl->Number();
    }; 
   inline void NewLiteral(int lit_num,const TERM* lit) 
    {
     next_lit_cand->Open(lit_num,lit,next_cand);
    };
   inline void EndOfLiteral() 
    {
     next_lit_cand->Close(next_cand - 1);
     next_lit_cand++;
    };  
   
   
   inline BS_LIT_CANDIDATES* LastLitCand() const { return next_lit_cand - 1; }; 
    
   inline void NewCandidatesVar(int var,PI_CLT_LIST* lst) 
    {
     all_vars.Register(var);
     if (next_lit_cand->RegisterVar(var)) // first occ. of the variable in the literal
      { next_cand->Tag() = BS_CAND::VAR_FIRST_OCC; }
     else 
      { next_cand->Tag() = BS_CAND::VAR; };
     next_cand->Var() = var;
     next_cand->CLTList() = lst;
     next_cand++;
     //OutputVar(var); cout << " "; 
    };
   inline void NewCandidatesConst(PI_CL_LIST* lst) 
    {
     next_cand->Tag() = BS_CAND::CONST;
     next_cand->CLList() = lst;
     next_cand++;
     //cout << "c ";
    };
   inline void NewCandidatesProp(PI_C_LIST* lst) 
    {
     next_lit_cand->MarkAsPropositional();
     next_cand->Tag() = BS_CAND::PROP_LIT;
     next_cand->CList() = lst;
     next_cand++;
     //cout << "c ";
    };      
        
   bool RetrieveCandidates(const Clause* cl); 
   bool RetrieveCandidatesFromSmallIndex(const TERM* ground_lit_ref);
    
   inline bool CollectUniqueVars()
    {
     supported_vars.Reset();
     if (all_vars.Size())
      {
       all_vars.ResetIterator();
       for (const int* next_var = all_vars.NextVar(); next_var; next_var = all_vars.NextVar())
        {
         if (all_vars.Occurences(*next_var) == 1) // unique variable
          supported_vars.Register(*next_var);
        };
      };
     return (bool)supported_vars.Size();   
    };
   
   inline void FilterSupportedVars(const TERM* term)
    {
     // term must be complex here
     if (term->Arity())
      {       
       const TERM* arg = term->Args();
       if (term->Arity() == 1)
        {
         if (arg->IsVariable())
          {
           if (supported_vars.Occurences(arg->Var())) supported_vars.Unregister(arg->Var());
          }
         else
          {
           FilterSupportedVars(arg->First());
          }; 
        }
       else
        {
         bool args_supported = false;
         while (!(arg->IsNull()))
          {
           if (arg->IsVariable())
            {
             if (!(supported_vars.Occurences(arg->Var()))) args_supported = true;             
            }
           else
            {
             args_supported = true;
             FilterSupportedVars(arg->First());
            };            
           arg = arg->Next();
          };                 
         if (!args_supported) // one variable must be deleted from supported_vars
          {
           supported_vars.Unregister(term->Args()->Var());
          };            
        };         
      }; 
    }; // void FilterSupportedVariables(const TERM* term)
   
   inline bool FilterSupportedVars()
    {
     for (BS_LIT_CANDIDATES* l_cand = lit_candidates; l_cand < next_lit_cand; l_cand++) 
      {
       if (!(l_cand->Ground()))
        {
         FilterSupportedVars(l_cand->Term());         
        };
      }; 
     return (bool)supported_vars.Size();      
    };
   
   inline void Squeeze()
    {
     BS_CAND* cand = candidates;
     for (BS_LIT_CANDIDATES* l_cand = lit_candidates; l_cand < next_lit_cand; l_cand++) 
      {
       l_cand->ShiftAndSqueeze(cand,supported_vars);       
      };
    };
        
   inline void LoadFirstClause()
    {
     current_clause_num = candidates->TopClauseNumber();     
    };  
    
   inline bool LoadNextClause()
    {
     candidates->SkipYoungerOrEqual(current_clause_num);
     if (candidates->NonEmpty())
      { 
       current_clause_num = candidates->TopClauseNumber();
       return true;
      }
     else return false;
    };    
    
   inline bool FindFirstLocalSubstitutions()
    {
     for (BS_LIT_CANDIDATES* l_cand = lit_candidates; l_cand < next_lit_cand; l_cand++) 
      {
       if (!(l_cand->FindFirstLocalSubst(current_clause_num))) return false;
      }; 
     return true;
    }; 
   
   inline void FindOtherLocalSubstitutions()
    {
     for (BS_LIT_CANDIDATES* l_cand = lit_candidates; l_cand < next_lit_cand; l_cand++)      
      {
       l_cand->FindAllOtherLocalSubst(current_clause_num);
      }; 
    }; 
   inline bool EqualizeClause(); 
  public: // joins
   inline void CompileJoin()
    {
     join_machine.ResetCompilation();
     for (BS_LIT_CANDIDATES* l_cand = lit_candidates; l_cand < next_lit_cand; l_cand++)
      {
       if (l_cand->NonGround())
        {
         join_machine.CompileLiteral(l_cand->Variables(),l_cand->LocalSubstList());         
        };
      };
     join_machine.EndOfClause(); 
    };
   inline bool Join() { return join_machine.Run(); };
  private: // methods for retrieval
   bool RetrieveCandidates(const TERM* t_ref,PI_LEVEL* lev);
  public: // methods for retrieval 
   bool RetrieveCandidates(const TERM* lit,bool inverse_polarity,PATH_INDEX* path_index);      
   bool RetrieveCandidates(const TERM* lit,bool inverse_polarity,SMALL_PATH_INDEX* small_path_index);         
  public: // methods for debugging
   void Output();
   void OutputLocalSubstitutions();
   inline void OutputJoinCode() { join_machine.OutputCode(); };
   inline void OutputSuppVars() { supported_vars.Output(); }; 
 }; // class BS_CANDIDATES





//*************************************************************************** 

class BS_MACHINE
 {
  private:
   BS_CANDIDATES candidates;
  private:
   inline bool SubsumptionAllowed(const Clause* subsuming,const Clause* subsumed)
    {
     return subsuming->NumOfPosEq() <= subsumed->NumOfPosEq();
    };
  public: 
   BS_MACHINE(PATH_INDEX* pi) : candidates(pi) {}; 
   void BSubsume(const Clause* cl);
 }; // class BS_MACHINE 

 
//*********** Definitions for BS_CANDIDATES: ****************

inline bool BS_CANDIDATES::EqualizeClause()
 {
     
  //EqClauseTimer.Start(); // debug
     
  DCONTROL14("BEGIN in BS_CANDIDATES::EqualizeClause()");
  BS_CAND* curr_cand = candidates;
  check_cand:
   if (curr_cand == next_cand)
    {
     DCONTROL14("A1 in BS_CANDIDATES::EqualizeClause()");
       
     //EqClauseTimer.Stop(); // debug
     
     current_clause = candidates->TopClause();
        
     DCONTROL14("END(1) in BS_CANDIDATES::EqualizeClause()");
     return true;
    }; 
   equalize_curr:
    //if (!(curr_cand->Active())) { curr_cand++; goto check_cand; };
    curr_cand->SkipYounger(current_clause_num); 
    if (curr_cand->NonEmpty())
     {
      if (curr_cand->TopClauseNumber() == current_clause_num)
       {
        curr_cand++; goto check_cand;
       }
      else
       {
        //MakeCurrentClause(curr_cand->TopClause());
         
        current_clause_num = curr_cand->TopClauseNumber(); // debug
           
        curr_cand = candidates;
        goto equalize_curr;
       }; 
     }
    else 
     {
      DCONTROL14("END(2) in BS_CANDIDATES::EqualizeClause()");
         
      //EqClauseTimer.Stop(); // debug 
          
      return false; 
     }; 
       
 }; // bool BS_CANDIDATES::EqualizeClause()





//*********** Definitions for BS_CAND: ****************

inline const Clause* BS_CAND::TopClause() const 
 {
  return candidates.c_list->TheClause(); 
 }; // const Clause* BS_CAND::TopClause() const  

inline void BS_CAND::SkipYounger(long cl_num)
 {
  PI_C_LIST::SkipYounger(cl_num,candidates.c_list);  
 }; // void BS_CAND::SkipYounger(long cl_num) 

inline void BS_CAND::SkipYoungerOrEqual(long cl_num)
 {
  PI_C_LIST::SkipYoungerOrEqual(cl_num,candidates.c_list);
 }; // void BS_CAND::SkipYoungerOrEqual(long cl_num) 
  

//***********************************************************************
 
