#define __BS_PRIMITIVES_H

#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif

#ifndef __DEBUG_H
 #include "debug.h"
#endif

#ifndef __DTREE_H
 #include "dtree.h"
#endif

//******************************************************************************

inline int Log2(int x) 
 {
  int a = 1;
  int l = 0;
  while (a < x) { a = 2*a; l++; };
  return l;
 }; 
 
const int MAX_NUMBER_OF_DTREE_ALTERNATIVES = MAX_NUMBER_OF_FUNCTORS + MAX_NUMBER_OF_VARS;

const int MAX_NUMBER_OF_DTREE_STORED_ALTS = Log2(MAX_NUMBER_OF_DTREE_ALTERNATIVES) * MAX_TERM_WEIGHT;

class DTREE_PATH
 {
  private: 
   DTREE_NODE* path[MAX_TERM_WEIGHT];
   DTREE_NODE** next;
   DTREE_NODE** alternative;
   DTREE_NODE** next_alternative;
  public:
   DTREE_PATH() 
    { 
     alternative = new (DTREE_NODE*)[MAX_NUMBER_OF_DTREE_STORED_ALTS]; 
     Reset(); 
    };
   inline void Reset() 
    {
     next = path; 
     next_alternative = alternative;
    };
   inline void PushFirst(DTREE_NODE* node) 
    { 
     *next = node; 
     next++;
     *next_alternative = (DTREE_NODE*)0;
     next_alternative++;
    };
   inline bool ShiftTop()
    {
     DTREE_NODE* top = *(next - 1);
     if (top->Less())
      {
       if (top->Greater())
        {
         *next_alternative = top->Greater();
         next_alternative++;
        };
       *(next-1) = top->Less();
       return true;
      }
     else
      {
       if (top->Greater())
        {
         *(next-1) = top->Greater();
         return true;
        }
       else
        { 
         next_alternative--;
         if (*next_alternative)
          {
           *(next-1) = *next_alternative;
           return true;
          }
         else // all possible alternatives exhausted
          {
           next--;
           return false;
          };
        };
      }; 
    }; 
   inline DTREE_NODE** NextPosition() const { return next; };
   inline DTREE_NODE* Top() const { return *(next-1); };
 }; // class DTREE_PATH

class DTREE_TERM
 {
  private:
   DTREE_NODE** begin;
   DTREE_NODE** end;
  public:
   DTREE_TERM() : begin((DTREE_NODE**)0) {};
   inline void Reset() { begin = (DTREE_NODE**)0; };
   inline bool Assigned() const { return (bool)begin; };
   inline DTREE_NODE** Begin() const { return begin; };
   inline DTREE_NODE** End() const { return end; };
   inline void Open(DTREE_NODE** b) { begin = b; };
   inline void Close(DTREE_NODE** e) { end = e; }; 
   inline int Weight() { return (End() - Begin()); };
  public: // for debugging
   inline bool CorrectTerm() const
    {
     if (!(Begin())) 
      {
       return false; 
      }; 
     if (Begin() >= End()) { return false; };
     if ((*(Begin()))->IsLeaf()) { return false; };
     unsigned start_depth = (*(Begin()))->Depth();
     for (DTREE_NODE** pos = Begin() + 1; pos < End(); pos++)
      {
       if ((*pos)->IsLeaf()) { return false; };
       if ((*pos)->Depth() <= start_depth) { return false; };
       
      };
     return true;            
    };
   inline void Output() 
    {
     for (DTREE_NODE** node = begin; node < end; node++)
      {
       (*node)->Output(); ::Output("\n");
      };
    };   
 }; // class DTREE_TERM
 
class BS_VARS
 {
  private:
   DTREE_TERM subst[MAX_NUMBER_OF_VARS];
  public:
   BS_VARS()
    { 
     for (int i = 1; i < MAX_NUMBER_OF_VARS; i++)
      {
       subst[i].Reset();
      };
    };
   inline DTREE_TERM* BSVar(const int v) { return (subst + v); };
   inline int VarNum(DTREE_TERM* var) { return (var - subst); };
 }; // class BS_VARS

class BS_PROCESSOR
 {
  private:
   DTREE_PATH path;
   DTREE_NODE* current_node;
  public:
   BS_PROCESSOR() : path() {};
   inline void Reset(DTREE_NODE* tree) { current_node = tree; path.Reset(); };
   inline DTREE_NODE* CurrentNode() const { return current_node; };
   inline bool Track(const TERM func);
   inline void Assign(DTREE_TERM* v);
   inline bool Alter(DTREE_TERM* v);
   inline bool Track(const DTREE_TERM* v);   
 }; // class BS_PROCESSOR

//***************** Definitions for BS_PROCESSOR: *****************************

inline bool BS_PROCESSOR::Track(const DTREE_TERM* v)
 {
  CHECK2(current_node,"A0 in BS_PROCESSOR::Track(DTREE_TERM* v)");
  CHECK2(v->CorrectTerm(),"A1 in BS_PROCESSOR::Track(DTREE_TERM* v)");
  DEBUG2_OP(int depth_diff = current_node->Depth() - (*(v->Begin()))->Depth());
  
  for (DTREE_NODE** pos = v->Begin(); pos < v->End(); pos++)
   {
   
    //current_node->Output(); cout << "\n"; // debug
   
    //CHECK2(depth_diff == ((int)(current_node->Depth()) - (int)((*pos)->Depth())),"A2 in BS_PROCESSOR::Track(DTREE_TERM* v)");
    CHECK2(!(current_node->IsLeaf()),"A3 in BS_PROCESSOR::Track(DTREE_TERM* v)");
    current_node = DTREE::Track((*pos)->Tag(),current_node);            
    if (!current_node) { return false; };
    
    cout << "AGA! "; current_node->Output(); cout << "\n";
    
   };
  return true;
 }; // bool BS_PROCESSOR::Track(DTREE_TERM* v)

inline bool BS_PROCESSOR::Alter(DTREE_TERM* v)
 {
  CHECK2(v->CorrectTerm(),"A1 in BS_PROCESSOR::Alter(DTREE_TERM* v)");
  
  DTREE_NODE** start_pos = v->Begin();
  while (!(path.ShiftTop()))
   {
    if (path.NextPosition() == start_pos) 
     {
      v->Reset();
      return false; 
     };
    CHECK2(!(path.Top()->IsLeaf()),"A2 in BS_PROCESSOR::Alter(DTREE_TERM* v)");
   };
   
  CHECK2(!(path.Top()->IsLeaf()),"A3 in BS_PROCESSOR::Alter(DTREE_TERM* v)"); 
  
  current_node = path.Top()->Next();
  unsigned start_depth = (*start_pos)->Depth();
  check_depth:
   if ((current_node->IsLeaf()) || (current_node->Depth() <= start_depth))
    {
     v->Close(path.NextPosition());
     CHECK2(v->CorrectTerm(),"postcondition 1 in BS_PROCESSOR::Alter(DTREE_TERM* v)");
     return true;
    }
   else
    {
     path.PushFirst(current_node);
     current_node = current_node->Next();
     goto check_depth;
    };
 }; // bool BS_PROCESSOR::Alter(DTREE_TERM* v)

inline void BS_PROCESSOR::Assign(DTREE_TERM* v)
 {
  v->Open(path.NextPosition());
  unsigned start_depth = current_node->Depth();
  next_node:
  
   //current_node->Output(); cout << "\n"; // debug
  
   path.PushFirst(current_node);
   current_node = current_node->Next();
   if ((current_node->IsLeaf()) || (current_node->Depth() <= start_depth))
    {
     v->Close(path.NextPosition());
    }
   else { goto next_node; };
  
  //cout << "ASSIGNED "; v->Output(); cout << "\n"; // debug
   
  CHECK2(v->CorrectTerm(),"postcondition in BS_PROCESSOR::Assign(DTREE_TERM* v)");
 }; // void BS_PROCESSOR::Assign(DTREE_TERM* v)

inline bool BS_PROCESSOR::Track(const TERM func)
 {
  current_node = DTREE::Track(func,current_node);
  return (bool)current_node;
 }; // inline bool BS_PROCESSOR::Track(const TERM func)

//******************************************************************************

































