#define __CLAUSE_INDEX_H

#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif

#ifndef __DEBUG_H
 #include "debug.h"
#endif 

#ifndef __TERM_H
 #include "term.h"
#endif

#ifndef __FLAT_TERM_H
 #include "flat_term.h"
#endif 

#ifndef __CLAUSE_H
 #include "clause.h"
#endif

#ifndef __DTREE_H
 #include "dtree.h"
#endif

#ifndef __ARRAY_H
 #include "array.h"
#endif

#ifndef __SHARING_H
 #include "sharing.h"
#endif

extern SHARING_INDEX GlobalSharingIndex;



//***************************************************************************

class LITERAL_INDEX
 {
  private:
   TERM prop_pos[2*MAX_NUMBER_OF_FUNCTORS];
   TERM prop_neg[2*MAX_NUMBER_OF_FUNCTORS];
   //ARRAY<DTREE_NODE*> positive;
   //ARRAY<DTREE_NODE*> negative;
   DTREE_NODE* positive[MAX_NUMBER_OF_FUNCTORS];
   DTREE_NODE* negative[MAX_NUMBER_OF_FUNCTORS];
  public:
   LITERAL_INDEX()
   // : positive((DTREE_NODE*)0),
   //   negative((DTREE_NODE*)0)
    {
     /**********
     for (int prop_var = 0; prop_var < MAX_NUMBER_OF_FUNCTORS; prop_var++)
      {
       prop_pos[prop_var] = TERM((ARITY)0,(FUNCTOR)prop_var,POL_POS);
       prop_neg[prop_var] = TERM((ARITY)0,(FUNCTOR)prop_var,POL_NEG); 
      };
     **********/
     for (int pred = 0; pred < MAX_NUMBER_OF_FUNCTORS; pred++)
      {
       positive[pred] = (DTREE_NODE*)0;
       negative[pred] = (DTREE_NODE*)0;
      };

     TERM prop_lit;
     for (int prop_var = 0; prop_var < MAX_NUMBER_OF_FUNCTORS; prop_var++)
      {
       prop_lit.MakeComplex((ARITY)0,(FUNCTOR)prop_var,POL_POS);
       *(Share(prop_lit)) = prop_lit;
       *(Share(prop_lit)->InfoField()) = (void*)0;
       
       prop_lit.InversePolarity();
       *(Share(prop_lit)) = prop_lit;
       *(Share(prop_lit)->InfoField()) = (void*)0;
      };  
    }; 
   void Destroy();
    

  private:
   static void DestroyLeafContent(DTREE_NODE* leaf);
   inline TERM* Share(const TERM& prop)
    {
     CHECK2(!(prop.Arity()),"wrong argument in LITERAL_INDEX::Share(..)");
     return (((prop.Polarity() == POL_POS) ? prop_pos : prop_neg) + 2*(prop.Functor()) + 1);
    };
  public:   
   inline DTREE_NODE** Tree(const TERM& header)
    {
     if (header.Polarity() == POL_POS) { return &(positive[header.Functor()]); }
     else { return &(negative[header.Functor()]); };
    };
   inline DTREE_NODE* Integrate(TERM* lit); // returns the leaf
  
   inline DTREE_NODE* Integrate(const FLAT_TERM* lit); // returns the leaf 
   inline DTREE_NODE* IntegrateEq(const TERM& header,
                                  const FLAT_TERM* arg1,
                                  const FLAT_TERM* arg2); // returns the leaf  
   
   inline DTREE_NODE* FindLeaf(TERM* lit) // returns the leaf
    {
     // temporary version
     return Integrate(lit); 
    };
   inline void InfoForDeletion(TERM* lit,
                               DTREE_NODE**& branch_to_destroy,
                               DTREE_NODE*& leaf,
                               DTREE_NODE*& last_fork); // for nonpropositional literals
 }; // class LITERAL_INDEX

class CLAUSE_INDEX 
 {
  private:
   LITERAL_INDEX lit_index;
  private:
   inline void Remove(TERM* lit,Clause* cl,bool remove_from_sharing);
  public:
   CLAUSE_INDEX() : lit_index() {};
   inline void Integrate(Clause* cl);
   inline void Remove(Clause* cl,bool remove_from_sharing);
   inline DTREE_NODE** Tree(const TERM& header) { return lit_index.Tree(header); };
  
   inline TERM* Integrate(const FLAT_TERM* lit,Clause* cl)
    {
     #ifdef _SUPPRESS_WARNINGS_
      cl = cl;
     #endif
    
     DTREE_NODE* leaf = lit_index.Integrate(lit);
     //ORD_CLAUSE_LIST::Insert(cl,&(leaf->Clauses()));
     leaf->NumOfClauses()++;
     return leaf->Term();
    };
   inline TERM* IntegrateEq(const TERM& header,const FLAT_TERM* arg1,const FLAT_TERM* arg2,Clause* cl)
    {
     #ifdef _SUPPRESS_WARNINGS_
      cl = cl;
     #endif
    
     DTREE_NODE* leaf = lit_index.IntegrateEq(header,arg1,arg2);
     //ORD_CLAUSE_LIST::Insert(cl,&(leaf->Clauses()));
     leaf->NumOfClauses()++;
     return leaf->Term();
    }; 
    
   inline void Destroy() { lit_index.Destroy(); };
 }; // class CLAUSE_INDEX

//******************* Definitions for CLAUSE_INDEX: ***********************

inline void CLAUSE_INDEX::Remove(Clause* cl,bool remove_from_sharing)
 {
  //if (remove_from_sharing) { cout << "RM "; ::Output(cl); cout << "\n"; }; // debug


  for (TERM* lit_ref = cl->Literals()->Args(); !(lit_ref->IsNull()); lit_ref = lit_ref->Next())
   {
    Remove(lit_ref->First(),cl,remove_from_sharing);
   };
 }; // void CLAUSE_INDEX::Remove(Clause* cl,bool remove_from_sharing)

inline void CLAUSE_INDEX::Remove(TERM* lit,Clause* cl,bool remove_from_sharing)
 {
  #ifdef _SUPPRESS_WARNINGS_
    cl = cl;
  #endif
  if (lit->Arity())
   {
    DTREE_NODE** branch_to_destroy;
    DTREE_NODE* leaf;
    DTREE_NODE* last_fork;
    lit_index.InfoForDeletion(lit,branch_to_destroy,leaf,last_fork);
    //ORD_CLAUSE_LIST::Remove(cl,&(leaf->Clauses()));  

    CHECK2(leaf->IsLeaf(),"A0 in CLAUSE_INDEX::Remove(TERM* lit,Clause* cl,bool remove_from_sharing)");
    
    if (remove_from_sharing) 
     {
      leaf->NumOfClauses()--;
    
      if (!(leaf->NumOfClauses())) // no more clauses with this literal
       {
        TERM* lit_copy = leaf->Term();

        DTREE::DestroyBranchWithoutForks(*branch_to_destroy);
        if (last_fork)
         {
          DTREE::ReduceFork(last_fork);
         }
        else
         {
          *branch_to_destroy = (DTREE_NODE*)0;
         }; 

     
        GlobalSharingIndex.DestroyLiteralCopy(lit_copy);
        
       }; 
     };
   }
  else // propositional
   {
    //ORD_CLAUSE_LIST::Remove(cl,&((*(lit_index.Tree(*lit)))->Clauses()));
    //((*(lit_index.Tree(*lit)))->NumOfClauses())--;
    if (remove_from_sharing) 
     {
      ((*(lit_index.Tree(*lit)))->NumOfClauses())--;
      // if (!((*(lit_index.Tree(*lit)))->Clauses())) // no more clauses with this literal
      if (!((*(lit_index.Tree(*lit)))->NumOfClauses())) // no more clauses with this literal
       {
        delete *(lit_index.Tree(*lit));
      
        *(lit_index.Tree(*lit)) = (DTREE_NODE*)0;
       };
     };
   };
 }; // void CLAUSE_INDEX::Remove(TERM* lit,Clause* cl,bool remove_from_sharing)

inline void CLAUSE_INDEX::Integrate(Clause* cl)
 {
  DCONTROL5("BEGIN in CLAUSE_INDEX::Integrate(Clause* cl)");
  DTREE_NODE* leaf;
  for (TERM* lit_ref = cl->Literals()->Args(); !(lit_ref->IsNull()); lit_ref = lit_ref->Next())
   {
    leaf = lit_index.Integrate(lit_ref->First());
    
    CHECK2(leaf->IsLeaf(),"A1 in CLAUSE_INDEX::Integrate(Clause* cl)");
    
    CHECK2(TERM::EqualComplexTerms(lit_ref->First(),leaf->Term()),
           "A2 in CLAUSE_INDEX::Integrate(Clause* cl)"); 
    
    lit_ref->SetFirst(leaf->Term());
    //ORD_CLAUSE_LIST::Insert(cl,&(leaf->Clauses()));
    leaf->NumOfClauses()++;
   };
  DCONTROL5("END in CLAUSE_INDEX::Integrate(Clause* cl)"); 
 }; // void CLAUSE_INDEX::Integrate(Clause* cl)

//******************* Definitions for LITERAL_INDEX: ***********************


inline DTREE_NODE* LITERAL_INDEX::Integrate(TERM* lit) 
 // returns the leaf corresponding to the literal
 {
  TERM header = *lit;
  DTREE_NODE** tree = Tree(header);
  
  if (header.Arity())
   {
    TERM::ITERATOR iter;
    iter.Reset(lit);
    DTREE::SURFER surf;
    surf.Reset(tree,1);
   
    if (*tree)
     {
      read_mode_next_sym:
       if (iter.Next())
        {
         if (surf.Read(iter.Symbol()))
          {
           goto read_mode_next_sym;
          }
         else
          {
           surf.Write(iter.Symbol());
           goto write_mode_next_sym;
          }; 
        }
       else // the literal is in the index already
        {
         return surf.Leaf();
        };
     }
    else
     {
      write_mode_next_sym:
       if (iter.Next())
        {
         surf.Write(iter.Symbol());
         goto write_mode_next_sym;
        }
       else // end of the literal, leaf must be created
        {
         surf.CreateLeaf(GlobalSharingIndex.CopyLiteral(lit));
         
         *(surf.Leaf()->Term()->InfoField()) = (void*)0;
         
         return surf.Leaf();
        };
     }; 
   }
  else // propositional
   {
    if (!(*tree)) 
     {
      *tree = new DTREE_NODE(Share(header));
     };
    
    CHECK2((*tree)->IsLeaf(),"A4 in LITERAL_INDEX::Integrate(TERM* lit)");
    
    return *tree;
   };
 }; // DTREE_NODE* LITERAL_INDEX::Integrate(TERM* lit)

inline DTREE_NODE* LITERAL_INDEX::Integrate(const FLAT_TERM* lit) 
 // returns the leaf corresponding to the literal
 {
  //cout << "LIT INDEX INTEGR\n"; 
  //lit->OutputTerm();
  //cout << "\n";
  
  TERM header = lit->Symbol();
  DTREE_NODE** tree = Tree(header);
  
  if (header.Arity())
   {
    FLAT_TERM::ITER iter;
    iter.Reset(lit);
    DTREE::SURFER surf;
    surf.Reset(tree,1);
   
    if (*tree)
     {
      read_mode_next_sym:
       iter.NextSym();
       if (iter.EndOfTerm()) return surf.Leaf(); // the literal is in the index already
       if (surf.Read(iter.CurrPos()->Symbol())) { goto read_mode_next_sym; };
       surf.Write(iter.CurrPos()->Symbol()); 
       goto write_mode_next_sym;
     }
    else
     {
      write_mode_next_sym:
       iter.NextSym();
       if (iter.EndOfTerm()) 
        {
         surf.CreateLeaf(GlobalSharingIndex.CopyFlatLiteral(lit));
         
         *(surf.Leaf()->Term()->InfoField()) = (void*)0;
         
         return surf.Leaf();
        };
       
       surf.Write(iter.CurrPos()->Symbol());
       goto write_mode_next_sym;
     }; 
   }
  else // propositional
   {
    if (!(*tree)) 
     {
      *tree = new DTREE_NODE(Share(header));
     };
    
     /********** debug *********
    
     if (!((*tree)->IsLeaf()))
      {
       cout << "INTEGR "; cout << (long)lit; cout << "\n";
       cout << "NODE "; (*tree)->Output(); cout << "\n\n\n";
      }; 

    ***********************/

    CHECK2((*tree)->IsLeaf(),"A4 in LITERAL_INDEX::Integrate(const FLAT_TERM* lit)");
    
    return *tree;
   };
 }; // DTREE_NODE* LITERAL_INDEX::Integrate(const FLAT_TERM* lit) 

inline DTREE_NODE* LITERAL_INDEX::IntegrateEq(const TERM& header,const FLAT_TERM* arg1,const FLAT_TERM* arg2) 
 // returns the leaf corresponding to the literal
 {
  DTREE_NODE** tree = Tree(header);
  FLAT_TERM::ITER iter;
  iter.Reset(arg1);
  bool first_arg = true;
  DTREE::SURFER surf;
  surf.Reset(tree,1);
  if (*tree)
   {
    read_mode_next_sym:
     if (iter.EndOfTerm()) 
      {
       if (first_arg)
        {// the second argument must be processed
         iter.Reset(arg2);
         first_arg = false;
         goto read_mode_next_sym;
        };
       return surf.Leaf(); // the literal is in the index already
      };
     if (surf.Read(iter.CurrPos()->Symbol())) 
      {
       iter.NextSym();
       goto read_mode_next_sym;
      };
     surf.Write(iter.CurrPos()->Symbol()); 
     iter.NextSym();
     goto write_mode_next_sym;
   }
  else
   {
    write_mode_next_sym:
     if (iter.EndOfTerm()) 
      {
       if (first_arg)
        {// the second argument must be processed
         iter.Reset(arg2);
         first_arg = false;
         goto write_mode_next_sym;
        };
       surf.CreateLeaf(GlobalSharingIndex.CopyFlatEq(header,arg1,arg2));
       *(surf.Leaf()->Term()->InfoField()) = (void*)0;
       return surf.Leaf();
      };
       
     surf.Write(iter.CurrPos()->Symbol());
     iter.NextSym();
     goto write_mode_next_sym;
   }; 
 }; // DTREE_NODE* LITERAL_INDEX::IntegrateEq(const TERM& header,const FLAT_TERM* arg1,const FLAT_TERM* arg2) 

inline void LITERAL_INDEX::InfoForDeletion(TERM* lit,
                                           DTREE_NODE**& branch_to_destroy,
                                           DTREE_NODE*& leaf,
                                           DTREE_NODE*& last_fork)
 {
  // lit can't be propositional here
  CHECK2(lit->Arity(),"A1 in LITERAL_INDEX::InfoForDeletion(..)");
  
  DTREE_NODE** tree = Tree(*lit); // *tree must be nonnull and nonleaf here
    
  CHECK2(*tree,"A2 in LITERAL_INDEX::InfoForDeletion(..)"); 
  CHECK2(!((*tree)->IsLeaf()),"A3 in LITERAL_INDEX::InfoForDeletion(..)");
  
  TERM::ITERATOR iter;
  iter.Reset(lit);
  iter.Next(); // skipping the header
  
  DTREE::InfoForDeletion(iter,tree,branch_to_destroy,leaf,last_fork);
 }; // void LITERAL_INDEX::InfoForDeletion(TERM* lit,...                                               

//**************************************************************************

























