#define __CLAUSE_SET_H
#ifndef __CLAUSE_H
 #include "clause.h"
#endif
#ifndef __OUTPUT_H
 #include "output.h"
#endif

#ifndef __DEBUG_H
 #include "debug.h"
#endif

//*********************************************************************

class CLAUSE_SET 
 {
  private: 
   Clause clauses;
   Clause* current;
   long numOfClauses;
   long removed;
  public:
   inline bool Empty() { return (clauses.Next() == (&clauses)); };
   CLAUSE_SET() : clauses(), numOfClauses(0L), removed(0L) {};
   inline void Insert(Clause* c) { numOfClauses++; Clause::InsertAfter(c,&clauses); };
   inline void Remove(Clause* c) { numOfClauses--; removed++; Clause::Remove(c); };
   inline long Size() const { return numOfClauses; };
   inline long Removed() const { return removed; };
   inline long Total() const { return (Size() + Removed()); };
   inline void ResetIterator() 
    {
     current = Empty() ? ((Clause*)0) : clauses.Next(); 
    };
   inline Clause* Next() 
    {
     Clause* res = current;
     if (current)
      {
       current = current->Next();
       if (current == (&clauses)) { current = (Clause*)0; };
      };
     return res;
    };
   inline Clause* Pop()
    {
     Clause* c = clauses.Next();
     //Clause::Remove(c);
     Remove(c);
     return c;
    };
   
   inline void Purge()
    {
     while (!Empty())
      {
       Pop();
      };
     removed = 0L;
    };
    
   inline void DestroyAllClauses()
    {
     while (!Empty())
      {
       Pop()->Destroy();
      };
     removed = 0L;
    };
   
   inline void OutputClauses()
    {
     ResetIterator();
     for (Clause* c = Next(); c; c = Next()) { Output(c); Output("\n"); };
    };
 }; // class CLAUSE_SET
 
//******************************************************************************










 