#include "command.h"

#ifndef __DEBUG_H
 #include "debug.h"
#endif

#include <string.h> 
#include <stdlib.h>

#ifndef __OPTIONS_H
 #include "options.h"
#endif
#ifndef __SIGNATURE_H
 #include "signature.h"
#endif 

extern SIGNATURE Signature;

#ifndef __READCMD_H
 #include "readcmd.h"
#endif
#ifndef __OUTPUT_H
 #include "output.h"
#endif
#ifndef __TPTP_PARSER_H
 #include "TPTP_parser.h"
#endif

#include <stdlib.h>

inline bool ParseLongInt(char* str,long& result)
 {
  char* tail = (char*)0;
  errno = 0;
  result = strtol(str,&tail,0);
  return (!((errno) || (tail == str)));
 };

//**************************************************************
int first_file_name;
int last_file_name;
char* TPTPDir = "."; // default value
const char* TPTP_DIRECTORY_ENV_VAR = "TPTP_DIR";

extern char *problemFileName;  

long cl_time_limit = MAXLONG; // default
bool cl_output_proof = false; // default

void CLSyntaxHelp()
 {
  cout << "Vampire command line syntax:\n";
  cout << "  vampire -h\n";
  cout << "  vampire [-time_limit <time limit in seconds>] [-output_proof] [<file1> ... <filen>] <problem file> \n";
  cout << "Relevant environment variables:\n";
  cout << "  " << TPTP_DIRECTORY_ENV_VAR << "\n";
 };

void HelpOnOptions()
 {
  char a = 'n';
  cout << "Help on proof search options?(y/n)";
  cin.get(a);
  if ((a == 'y') || (a == 'Y'))
   {
    cout << "HELP ON OPTIONS WILL BE HERE SOON.\n";
   };
 };
 
enum CL_PARSE_RESULT
 {
  CL_ERROR,
  CL_HELP,
  CL_PROBLEM
 };

bool IsFile(char* str)
 {
  char* dummy = str; 
  dummy = 0;
  return true; // temporary
 };

bool IsDirectory(char* str)
 {
  char* dummy = str;
  dummy = 0;
  return true; // temporary
 };

bool ParseFileNames(char** args,int ffn,int lfn)
 {
  for (int fn = ffn; fn <= lfn; fn++)
   {
    if (!(IsFile(args[fn]))) 
     {
      return false;
     };
   };
  return true;
 };

CL_PARSE_RESULT ParseCL(int numarg,char** args)
 {
  if (numarg < 2) { return CL_ERROR; };
  if (!(strcmp(args[1],"-h")))
   {
    if (numarg == 2)
     {
      return CL_HELP;
     }
    else
     {
      return CL_ERROR;
     };
   }; 
  
  first_file_name = 1;
  last_file_name = numarg - 1;
  
  if (!(strcmp(args[1],"-time_limit")))
   {
    first_file_name = 3;
    if (numarg < 4) return CL_ERROR;
    if (!ParseLongInt(args[2],cl_time_limit)) return CL_ERROR;
   };
 
  if (!(strcmp(args[3],"-output_proof")))
   {
    first_file_name = 4;
    if (numarg < 5) return CL_ERROR;
    cl_output_proof = true;
   }; 


  char* tmp = getenv(TPTP_DIRECTORY_ENV_VAR);

  if (tmp)
   {
    TPTPDir = tmp;
   }; 
  

  if (ParseFileNames(args,first_file_name,last_file_name))
   {
    return CL_PROBLEM; 
   }
  else
   {
    return CL_ERROR;
   };
 }; // CL_PARSE_RESULT ParseCL()

bool ReadAllFiles(char** args,int ffn,int lfn)
 {
  for (int fn = ffn; fn <= lfn; fn++)
   {
    if (!(ReadFile(TPTPDir,args[fn]))) { return false; };
   };
  return true;
 };

int main(int numarg,char** args)
 {
  #ifdef _DEBUG_
   InitDebug();
  #endif 
  InitializeCtrlC();
  OPTION::DescribeAllOptions();
  
  // Initializing random number generator
  srand(1u);

  // Initializing substitutions
  //SBST::Initialize();
  //DOUBLE_SBST::Initialize();
 
  //AdjustOutput(); // debug

  switch (ParseCL(numarg,args))
   {
    case CL_ERROR: 
     CLSyntaxHelp();
     break;
    case CL_HELP:
     CLSyntaxHelp();
     HelpOnOptions();
     break;
    case CL_PROBLEM:
     problemFileName = args[last_file_name];
     
     //cout << "*"; // debug
     //cout << problemFileName << "\n";
     
     if (!(ReadAllFiles(args,first_file_name,last_file_name))) { return 1; };  
    
     Signature.MemoLastInputFun();
      
     //return 0; // debug
        
     Prove(problemFileName);
     break;
   }; 
}; // int main(int numarg,char** args)


//************************************************************************
