#define __EQUIVALENCE_RELATION_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif 

//************************************************************************

template <int SIZE_OF_CARRIER>
 class EQ_REL
  {
   private:
    int can_rep[SIZE_OF_CARRIER];
    int registered[SIZE_OF_CARRIER];
    int* next_registered;
   private:
    void Register(int x,int can_x) 
     {
      CHECK12(x != can_x,"ARG1 in EQ_REL<..>::Register(..)");
      CHECK12(x == can_rep[x],"ARG2 in EQ_REL<..>::Register(..)");
      *next_registered = x;
      next_registered++;
      can_rep[x] = can_x; 
     }; 
    void Unregister(int x)
     {
      can_rep[x] = x;
      next_registered--;
      if ((*next_registered) != x)
       {
        for (int* i = registered; i < next_registered; i++)
         if ((*i) == x)
          {
           *i = *next_registered;
           return;
          };
       };
     }; 
   public: 
    EQ_REL() : next_registered(registered)
     {
      for (int i = 0; i < SIZE_OF_CARRIER; i++) can_rep[i] = i;
     };
    inline void Reset() 
     {
      while (next_registered != registered) 
       {
        next_registered--;
        can_rep[*next_registered] = *next_registered;
       };
     };
    inline bool Identity() const { return next_registered == registered; };
    inline int CanRep(int x) const 
     {
      while (can_rep[x] != x) x = can_rep[x];
      return x;
     };
    inline bool Equivalent(int x,int y) const
     {
      return CanRep(x) == CanRep(y);
     };
    inline void MakeEquivalent(int x,int y)
     {
      int can_x = CanRep(x);
      int can_y = CanRep(y);
      if (can_x == can_y) return; // already equivalent
      if (can_x < can_y) { Register(can_y,can_x); }
      else { Register(can_x,can_y); };
     };
    inline bool Split(int& x,int& y)
     { 
      if (next_registered == registered) return false; // identity relation
      next_registered--;
      x = *next_registered;
      y = can_rep[x];
      CHECK12(y == can_rep[y],"A3 in EQ_REL<..>::Split(..)");
      CHECK12(x > y,"A3 in EQ_REL<..>::Split(..)");
      can_rep[x] = x;
      return true;
     };
    
    inline void Subtract(EQ_REL& eq_rel)
     {
      int x;
      int can_x;
      for (int* i = next_registered - 1 ; i>= registered; i--)
       {
        x = *i;
        can_x = can_rep[x];
        if (eq_rel.Equivalent(x,can_x)) { Unregister(x); }
        else
         {
          eq_rel.MakeEquivalent(x,can_x);
         };
       }; 
     }; 
     
    inline bool SplitOn(int x,int y)
     {
      CHECK12(x != y,"A3 in EQ_REL<..>::SplitOn(..)");
      int can_x = CanRep(x);
      int can_y = CanRep(y);
      if (can_x != can_y) return false; // already nonequivalent
      if (x == can_x) { Unregister(y); }
      else Unregister(x);
      return true;
     };
    inline void Normalize()
     {
      for (int* i = next_registered - 1; i >= registered; i--)
       can_rep[*i] = CanRep(*i);
     };
     
   public: // for debugging
    inline void Output()
     {
      for (int* i = registered; i < next_registered; i++)
       {
        ::Output(*i); ::Output(" -> "); ::Output(can_rep[*i]);
        if (i + 1 < next_registered) ::Output(" | ");
       };
     };   
   public:
    class ITER
     {
      private:
       const int* next;
       const int* end;
       const int* can_rep;
      public:
       ITER() {};
       ITER(const EQ_REL& eq_rel) 
        : next(eq_rel.registered), 
          end(eq_rel.next_registered),
          can_rep(eq_rel.can_rep) 
        {};
       inline void Reset(const EQ_REL& eq_rel) 
        {
         next = eq_rel.registered;
         end = eq_rel.next_registered;
         can_rep = eq_rel.can_rep;
        };    
       inline bool Next(int& x,int& y)
        {
         if (next == end) return false;
         x = *next;
         y = can_rep[x];
         CHECK12(x > y,"A3 in EQ_REL<..>::ITER::Next()");
         next++;
         return true;
        }; 
     };  
  }; // class EQ_REL<int SIZE_OF_CARRIER>
 



//*************************************************************************