
         /*********** Forward simplification by unit equalities. ***********/

#define __FORW_UEQ_SIMPLIFICATION_H

#ifndef __M_INDEX_H
 #include "m_index.h"
#endif

//**************************************************************************


class FUEQ_SIMPLIFICATION
 {
  private:
   class INDEX
    {
     private:
      class SUBST
       {
        private:
         int num_of_vars;
        public:
         SUBST(int nv) : num_of_vars(nv) {};
         inline void* operator new(size_t,int nv)
          {
           return Memory::Allocate((sizeof(SUBST)/4) + nv);
          };
         inline void operator delete(void* obj)
          { 
           Memory::Deallocate(obj,(sizeof(SUBST)/4) + ((SUBST*)obj)->NumOfVars());
          }; 
         inline int NumOfVars() const { return num_of_vars; };
        public:
         inline const FLAT_TERM*** Subst() const { return (const FLAT_TERM***)(this + 1); };
         inline const FLAT_TERM*** Subst(int var) const { return Subst() + var; };  
         inline void Output() { ::Output("fues_subst<"); ::Output(NumOfVars()); ::Output(">"); };
       }; // class SUBST
     
      class TERM_WITH_SUBST
       {
        private:
         TERM::ITERATOR term_iter;
         FLAT_TERM::ITER subst_iter;
         const FLAT_TERM*** subst;
         #ifdef _DEBUG20_
          int max_var_num;
         #endif
         bool in_term;
         TERM dummy_term;
        public:
         TERM_WITH_SUBST() : term_iter(),subst_iter(),dummy_term((ARITY)0,(FUNCTOR)0,POL_POS) {};
         TERM_WITH_SUBST(const TERM& t,SUBST* s)
          : term_iter(),
            subst_iter(),
            dummy_term((ARITY)0,(FUNCTOR)0,POL_POS)
          {
           Reset(t,s);
          };
         inline void Reset(const TERM& t,SUBST* s)
          {
           #ifdef _DEBUG20_
            max_var_num = s->NumOfVars() - 1; 
           #endif
           subst = s->Subst();
           if (t.IsVariable())
            {
             in_term = false;
             CHECK20(t.Var() <= max_var_num,"A10 in FUEQ_SIMPLIFICATION::INDEX::TERM_WITH_SUBST::Reset(..)");
             subst_iter.Reset(*(subst[t.Var()]));
             term_iter.Reset(&dummy_term);
             term_iter.Next();
            }
           else // t is reference to a complex term 
            {
             in_term = true;
             term_iter.Reset(t.First());
            };  
          };
        inline bool Next()
         {
          if (in_term)
           {
            next_top_sym:
             if (!term_iter.Next()) return false;
             if (term_iter.SymbolRef().IsVariable()) 
              {
               in_term = false;
               CHECK20(term_iter.SymbolRef().Var() <= max_var_num,"A10 in FUEQ_SIMPLIFICATION::INDEX::TERM_WITH_SUBST::Next()");
               subst_iter.Reset(*(subst[term_iter.SymbolRef().Var()]));
              };
             return true;
           }
          else
           {
            subst_iter.NextSym();
            if (!subst_iter.EndOfTerm()) return true;
            in_term = true;
            goto next_top_sym;
           };
         };
        inline TERM Symbol()
         {
          if (in_term) return term_iter.Symbol();
          return subst_iter.CurrPos()->Symbol();
         };            
       }; // class TERM_WITH_SUBST
      
     private:
      M_INDEX m_index;
     private:
      static void OutputLeafInfo(MI_TREE::NODE* leaf);
     public:
      INDEX() : m_index()
       {
        m_index.SetLeafInfo(OutputLeafInfo);
       };
      inline void Integrate(Clause* cl);
      inline void Remove(Clause* cl);    
      inline MI_TREE::NODE* Match(const FLAT_TERM* query) { return m_index.Match(query); };        
      inline void Output();
    }; // class INDEX
   INDEX index;
   Clause* rule;
   INDEX::TERM_WITH_SUBST rule_rhs;
  public:
   FUEQ_SIMPLIFICATION() : index(), rule_rhs() {};
   inline void Integrate(Clause* cl) { index.Integrate(cl); };
   inline void Remove(Clause* cl) { index.Remove(cl); };

   inline bool Rewrite(const FLAT_TERM* query);
   inline Clause* Rule() const { return rule; }; 
   inline bool RHSNext() { return rule_rhs.Next(); };    
   inline TERM RHSSymbol() { return rule_rhs.Symbol(); };   

   inline void OutputIndex() { index.Output(); };
 }; // class FUEQ_SIMPLIFICATION

//*********************** Definitions for FUEQ_SIMPLIFICATION: *****************

inline bool FUEQ_SIMPLIFICATION::Rewrite(const FLAT_TERM* query)
 {
  CHECK20(query->Symbol().IsComplex(),"A0 in FUEQ_SIMPLIFICATION::Simplify(const FLAT_TERM* query)");
  MI_TREE::NODE* leaf = index.Match(query);
  if (!leaf) return false;  
  rule = (Clause*)(leaf->LeafInfo1());   
  TERM rule_rhs_term = *(rule->Literals()->Args()->First()->NthArg(1)); 
  rule_rhs.Reset(rule_rhs_term,(INDEX::SUBST*)(leaf->LeafInfo2()));  
  return true;
 }; // bool FUEQ_SIMPLIFICATION::Rewrite(const FLAT_TERM* query) 

//*********************** Definitions for FUEQ_SIMPLIFICATION::INDEX: *********

inline void FUEQ_SIMPLIFICATION::INDEX::Integrate(Clause* cl)
 {
  if (cl->Unit())
   {
    TERM* lit = cl->Literals()->Args()->First();
    if (lit->Positive() && lit->IsEquality() && lit->IsOrderedEq())
     {          
      MI_TREE::NODE* leaf = m_index.Integrate(lit->Args()->First());
      if (!leaf->LeafInfo1())
       {
        leaf->LeafInfo1() = (void*)cl;

        /*****/
        int num_of_vars = m_index.IntegrVars().NumOfDifferent();
        SUBST* subst = new(num_of_vars) SUBST(num_of_vars); 
        m_index.IntegrVars().ResetIterator();
        for (const int* v = m_index.IntegrVars().NextVar(); v; v = m_index.IntegrVars().NextVar())
	 {
          *(subst->Subst(*v)) = m_index.IntegrMapVar(*v);
         };
        leaf->LeafInfo2() = (void*)subst;
        /*******/
       }
      else
       {
	
       };
     }; 
   };  
 }; // void FUEQ_SIMPLIFICATION::INDEX::Integrate(Clause* cl) 

inline void FUEQ_SIMPLIFICATION::INDEX::Remove(Clause* cl) 
 {
  DCONTROL20("BEGIN in  FUEQ_SIMPLIFICATION::INDEX::Remove(Clause* cl)");  
  if (cl->Unit())
   {
    TERM* lit = cl->Literals()->Args()->First();
    if (lit->Positive() && lit->IsEquality() && lit->IsOrderedEq())
     {

     // cout << "REMOVE\n"; // debug
     
      DCONTROL20("X10 in  FUEQ_SIMPLIFICATION::INDEX::Remove(Clause* cl)");  
            
      MI_TREE::NODE* leaf;
      MI_TREE::NODE** branch_to_destroy;
      if (m_index.Remove(lit->Args()->First(),leaf,branch_to_destroy))
       { 
        CHECK20(leaf->IsLeaf(),"A50 in FUEQ_SIMPLIFICATION::INDEX::Remove(Clause* cl)");

        CHECK20(((Clause*)(leaf->LeafInfo1()))->Literals()->Args()->First()->Args()->First() == lit->Args()->First(),
                "A60 in FUEQ_SIMPLIFICATION::INDEX::Remove(Clause* cl)");
        DCONTROL20("X20 in  FUEQ_SIMPLIFICATION::INDEX::Remove(Clause* cl)");  
       
        
        //((SUBST*)(leaf->LeafInfo2()))->Output(); ::Output("  "); ::Output((Clause*)(leaf->LeafInfo1())); ::Output("\n\n"); // debug

        delete (SUBST*)(leaf->LeafInfo2());  
        
        DCONTROL20("X30 in  FUEQ_SIMPLIFICATION::INDEX::Remove(Clause* cl)");  
 
        MI_TREE::NODE* fork = (*branch_to_destroy)->Fork();
        (*branch_to_destroy)->SetFork((MI_TREE::NODE*)0);     
        MI_TREE::Destroy(*branch_to_destroy);
        *branch_to_destroy = fork;
       };
     };
   };   
  DCONTROL20("END in  FUEQ_SIMPLIFICATION::INDEX::Remove(Clause* cl)");  
 }; // void FUEQ_SIMPLIFICATION::INDEX::Remove(Clause* cl) 

inline void FUEQ_SIMPLIFICATION::INDEX::Output()
 {  
  m_index.Output();
 }; // void FUEQ_SIMPLIFICATION::INDEX::Output() 


//**********************************************************************************************
