#define __FORWARD_SUBSUMPTION_H
 
#ifndef __DEBUG_H
 #include "debug.h"
#endif

#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif

#ifndef __FS_CODE_H
 #include "fs_code.h"
#endif 
 
#ifndef __STACK_H
 #include "stack.h"
#endif 

//**************************************************************************

class FS_QUERY
 {
  public:
   class LIT
    {    
     private:
      const FLAT_TERM* arg1;
      const FLAT_TERM* arg2;
      LIT* next;
     public:  
      inline const FLAT_TERM* Args() const { return arg1; };
      inline void SetArgs(const FLAT_TERM* a) { arg1 = a; };      
      inline const FLAT_TERM* Arg1() const { return arg1; }; // for equations
      inline const FLAT_TERM* Arg2() const { return arg2; }; // for equations
      inline void SetArgs(const FLAT_TERM* a1,const FLAT_TERM* a2)
       { arg1 = a1; arg2 = a2; };
      inline LIT* Next() const { return next; };
      inline void SetNext(LIT* nl) { next = nl; };
      inline void Set(const FLAT_TERM* a,LIT* nl) { arg1 = a; next = nl; };
      inline void Set(const FLAT_TERM* a1,const FLAT_TERM* a2,LIT* nl) 
       { arg1 = a1; arg2 = a2; next = nl; };
     public: // for debugging
      inline void Output() const
       {
        (Args() - 1)->OutputTerm();
       };  
    }; // class LIT
  private: // structure 
   LIT literal[MAX_NUMBER_OF_LITERALS];
   LIT* next_literal;
   LIT* header_index[MAX_NUMBER_OF_DIFFERENT_HEADERS];
   LIT** const pos_ord_equations;
   LIT** const neg_ord_equations;
   LIT** present_header[MAX_NUMBER_OF_LITERALS];
   LIT*** next_present_header; 
   LIT* last_pos_unord_eq;
   LIT* last_neg_unord_eq;
   long num_of_pos_eq;
  public:
   FS_QUERY() 
    : pos_ord_equations(header_index + POS_ORD_EQ_HEADER.HeaderNum()),
      neg_ord_equations(header_index + NEG_ORD_EQ_HEADER.HeaderNum())
    {
     for (int h = 0; h < MAX_NUMBER_OF_DIFFERENT_HEADERS; h++)
      {
       header_index[h] = (LIT*)0;
      };
     next_present_header = present_header; 
    };
   inline void ResetQuery()
    {
     next_literal = literal;
     while (next_present_header != present_header)
      {
       next_present_header--;
       **next_present_header = (LIT*)0;
      };
     last_pos_unord_eq = (LIT*)0;
     last_neg_unord_eq = (LIT*)0;
     num_of_pos_eq = 0;
    };
   inline void LinkEquations()
    {
     if (last_pos_unord_eq) last_pos_unord_eq->SetNext(*pos_ord_equations);
     if (last_neg_unord_eq) last_neg_unord_eq->SetNext(*neg_ord_equations);
    }; 
   inline void EndOfQuery()
    {
     LinkEquations(); 
    }; 
   inline void PushLit(const FLAT_TERM* lit)
    {
     CHECK7(!lit->Symbol().IsEquality(),"A0 in FS_QUERY::PushLit(..)");
     LIT** lit_lst_ptr = header_index + lit->Symbol().HeaderNum();
     if (*lit_lst_ptr)
      {
       next_literal->Set(lit + 1,*lit_lst_ptr);
      }
     else // no literals with this header yet
      {
       *next_present_header = lit_lst_ptr;
       next_present_header++;
       next_literal->Set(lit + 1,(LIT*)0);
      };
     *lit_lst_ptr = next_literal; 
     next_literal++;  
    };   
    
   inline void PushUnordEq(TERM header,const FLAT_TERM* arg1,const FLAT_TERM* arg2)
    {
     CHECK7(header.IsUnorderedEq(),"ARG in FS_QUERY::PushUnordEq(..)");
     if (header.Positive()) num_of_pos_eq++;
     LIT** lit_lst_ptr = header_index + header.HeaderNum();
     if (*lit_lst_ptr)
      {
       next_literal->Set(arg1,arg2,*lit_lst_ptr);
      }
     else // no literals with this header yet
      {
       *next_present_header = lit_lst_ptr;
       next_present_header++;
       next_literal->Set(arg1,arg2,(LIT*)0);
       if (header.Negative()) { last_neg_unord_eq = next_literal; }
       else // positive
        last_pos_unord_eq = next_literal;
      };
     *lit_lst_ptr = next_literal; 
     next_literal++; 
    };
    
   inline void PushOrdEq(TERM header,const FLAT_TERM* arg1,const FLAT_TERM* arg2)
    {
     CHECK7(header.IsEquality() && header.IsOrderedEq(),"ARG in FS_QUERY::PushOrdEq(..)");     
     if (header.Positive()) num_of_pos_eq++;
     LIT** lit_lst_ptr = header_index + header.HeaderNum();
     if (*lit_lst_ptr)
      {
       next_literal->Set(arg1,arg2,*lit_lst_ptr);
      }
     else // no literals with this header yet
      {
       *next_present_header = lit_lst_ptr;
       next_present_header++;
       next_literal->Set(arg1,arg2,(LIT*)0);
      };
     *lit_lst_ptr = next_literal; 
     next_literal++; 
    }; 
    
   inline LIT* FirstLit(int header_num)  
    {
     return header_index[header_num];
    }; 
   
   inline bool SubsumptionAllowed(Clause* subsuming_cl)
    {
     return subsuming_cl->NumOfPosEq() <= num_of_pos_eq;
    }; 
    
  public: // for debugging
   inline void OutputQuery() const
    {
     for (LIT** const * p = present_header; p < next_present_header; p++)
      for (LIT* lit = **p; lit; lit = lit->Next())
       {
        lit->Output(); ::Output("\n");
       };
    };  
 }; // class FS_QUERY
 
class FS : public FS_CODE, public FS_QUERY
 {
  private:
   const FLAT_TERM* subst[MAX_CLAUSE_WEIGHT];
  public:
   FS() 
    : FS_CODE(subst), 
      FS_QUERY()
    {};
   Clause* Subsume();
 }; // class FS  : public FS_CODE 

//********************************************************************** 
