#include "forward_superposition.h"

//**************** Definitions for FORWARD_SUPERPOSITION:  *************

bool FORWARD_SUPERPOSITION::Load(Clause* cl)
 {
  fpara_premise.Load(cl);
  PREFIX_SYM* lit;
  try_next_lit:
   lit = fpara_premise.NextMaxLit();
   if (!lit) return false;
   if (fpara_machine.LoadLiteral(lit)) return true;
   goto try_next_lit; 
 }; // bool FORWARD_SUPERPOSITION::Load(Clause* cl)
 
bool FORWARD_SUPERPOSITION::NextInference()
 {
  if (fpara_machine.Paramodulate()) return true;  
  PREFIX_SYM* lit;
  try_next_lit:
   lit = fpara_premise.NextMaxLit();
   if (!lit) return false;
   if ((fpara_machine.LoadLiteral(lit)) && (fpara_machine.Paramodulate())) 
    return true;
   goto try_next_lit; 
 }; // bool FORWARD_SUPERPOSITION::NextInference()

bool FORWARD_SUPERPOSITION::NextEqRes()
 {
  PREFIX_SYM* lit; 
  try_next_lit:
   lit = fpara_premise.NextMaxLit();
   if (!lit) return false;
      
   TERM header =  lit->Head().Func();
   if (header.Negative() && header.IsEquality() && header.IsUnorderedEq())
    {
     PREFIX_SYM* arg1 = lit+1;
     PREFIX_SYM* arg2 = arg1->Bracket();
     if (UNIFICATION::Unify(arg1,arg2)) return true;
    };
   goto try_next_lit; 
 }; // bool FORWARD_SUPERPOSITION::NextEqRes()

bool FORWARD_SUPERPOSITION::CollectResult()
 {
  if (fpara_premise.MinResultWeight() 
      + fpara_machine.Candidate()->TheClause()->MinResWeight() 
       > Resolvent.MaxWeight())
   {
    Resolvent.RegisterTooHeavy();
    return false;
   }; 
     
  Resolvent.Reset();
  if (fpara_premise.CollectLiterals() && fpara_machine.CollectResult())
   {
    Resolvent.SetBackground(ANCESTORS::RESOL);
    Resolvent.PushAncestor(fpara_premise.Query());
    Resolvent.PushAncestor(fpara_machine.Candidate()->TheClause());
    return true;
   }
  else return false; 
 }; // bool FORWARD_SUPERPOSITION::CollectResult()   
 
bool FORWARD_SUPERPOSITION::CollectEqResResult()
 {
  Resolvent.Reset();
  if (fpara_premise.CollectLiterals())
   {
    Resolvent.SetBackground(ANCESTORS::RESOL);
    Resolvent.PushAncestor(fpara_premise.Query());
    return true;
   }
  else return false; 
 }; // bool FORWARD_SUPERPOSITION::CollectEqResResult() 



void FORWARD_SUPERPOSITION::Process(Clause* cl)
 {
  if (Load(cl))
   {
    while (NextInference())
     if (CollectResult())
      {
       //cout << "RESULT:\n"; // debug
       //Resolvent.Output(); cout << "\n\n\n"; // debug 
          
       Resolvent.ProcessCollected(); 
      };
       
     //cout << "EQ. RES. QUERY\n"; // debug
     //::Output(fpara_premise.Query()); cout << "\n"; // debug
       
     if (fpara_premise.EqResApplicable())
      {
       fpara_premise.ResetIter();
       while (NextEqRes())
        {
         if (CollectEqResResult())
          {
           //cout << "EQ. RES. RESULT:\n"; // debug
           //Resolvent.Output(); cout << "\n\n\n"; // debug 
             
           Resolvent.ProcessCollected(); 
          };
         UndoEqRes(); 
        };  
      };
    };   
 }; // void FORWARD_SUPERPOSITION::Process(Clause* cl)

//******************** Definitions for FPARA_PREMISE: ********************

void FPARA_PREMISE::Load(Clause* cl)
 {
  DCONTROL17("BEGIN in FPARA_PREMISE::Load(Clause* cl)");
  eq_res_applicable = false;
  clause = cl;
  TERM* lits = cl->Literals();
  num_of_max_literals = lits->MaxLiterals();
  TERM* lit_ref = lits->Args(); 
  for (int ln = num_of_max_literals; ln; ln--)
   {
    PREFIX_SYM* mem = max_literal[ln-1];
    StandardTermToPrefix(index,lit_ref->First(),mem);
       
    if (lit_ref->First()->Negative()
        && lit_ref->First()->IsEquality() 
        && lit_ref->First()->IsUnorderedEq())
     eq_res_applicable = true;
       
    lit_ref = lit_ref->Next();
   };
  nonmax_literals = lit_ref; 
     
  CreateWeightFormula();
     
  ResetIter();
 }; // void FPARA_PREMISE::Load(Clause* cl) 

bool FPARA_PREMISE::CollectLiterals()
 {
  // all the maximal but the rewritten one
  for (int ln = num_of_max_literals; ln; ln--)
   if ((ln - 1) != next_max_lit)
    if (!Resolvent.PushPrefixLit(max_literal[ln-1])) return false;     
     
  // all nonmaximal 
  for (TERM* nml = nonmax_literals; !nml->IsNull(); nml = nml->Next())
   if (!Resolvent.PushStandardLit(index,nml->First())) return false;
  return true; 
 }; // bool FPARA_PREMISE::CollectLiterals()

//*********************** Definitions for FPARA_MACHINE: ********************

bool FPARA_MACHINE::LoadLiteral(PREFIX_SYM* lit)
 {
  if (!lit->Head().Func().Arity()) return false; // propositional
  DCONTROL16("BEGIN in FPARA_MACHINE::LoadLiteral(..)");
  // here lit must be a plain nonpropositional literal with all brackets set
  literal = lit;
  complex_positions.Reset();
  if (lit->Head().Func().IsEquality())
   {
    if (lit->Head().Func().IsOrderedEq())
     {
      eq_ord_constr.MakeAlwaysValid();
      PREFIX_SYM* end_of_arg = (lit + 1)->Bracket();
      for (PREFIX_SYM* pos = lit + 1; pos < end_of_arg; pos++)
       if (pos->Head().IsFunctor()) complex_positions.Push(pos);
      eq_second_arg = (PREFIX_SYM*)0;
     }
    else // unordered equality
     {         
      PREFIX_SYM* arg1 = lit + 1;
      PREFIX_SYM* arg2 = arg1->Bracket();     
      if (arg1->Head().IsVariable() || arg2->Head().IsVariable())
       {   
        eq_ord_constr.MakeAlwaysValid();
        PREFIX_SYM* end_of_lit = lit->Bracket();
        for (PREFIX_SYM* pos = lit + 1; pos < end_of_lit; pos++)
         if (pos->Head().IsFunctor()) complex_positions.Push(pos);
        eq_second_arg = (PREFIX_SYM*)0;
       }
      else // both arguments are complex 
       {

	 //eq_ord_constr.MakeAlwaysValid(); // debug
  
	 eq_ord_constr.Compile(arg1,arg2);

        PREFIX_SYM* end_of_arg1 = arg1->Bracket();
        for (PREFIX_SYM* pos = arg1; pos < end_of_arg1; pos++)
         if (pos->Head().IsFunctor()) complex_positions.Push(pos);
        eq_second_arg = arg2;
        
       };
     };
   }
  else // noneq. literal
   {
    eq_ord_constr.MakeAlwaysValid();
    PREFIX_SYM* end_of_lit = lit->Bracket();
    for (PREFIX_SYM* pos = lit + 1; pos < end_of_lit; pos++)
     if (pos->Head().IsFunctor()) complex_positions.Push(pos);
    eq_second_arg = (PREFIX_SYM*)0;
   };

  if (complex_positions.Empty()) return false;
  FPARA_MACH::LoadTermToRewrite(complex_positions.Pop());
  return true; 
 }; // bool FPARA_MACHINE::LoadLiteral(PREFIX_SYM* lit)


bool FPARA_MACHINE::LoadRestOfLit()
 {
  if (!eq_second_arg) return false;
  
  eq_ord_constr.Inverse();
  PREFIX_SYM* end_of_arg2 = eq_second_arg->Bracket();
  for (PREFIX_SYM* pos = eq_second_arg; pos < end_of_arg2; pos++)
   if (pos->Head().IsFunctor()) complex_positions.Push(pos);
    
  eq_second_arg = (PREFIX_SYM*)0;
  
  return (!complex_positions.Empty());
 }; // bool FPARA_MACHINE::LoadRestOfLit()


bool FPARA_MACHINE::Paramodulate() 
 {
  DCONTROL16("BEGIN in FPARA_MACHINE::Paramodulate()");
  candidate = FPARA_MACH::NextCandidate();
  if (candidate) { return true; }
   else // try another position
    {
     try_another_position:
      if (complex_positions.Empty()) 
       {
        if (!LoadRestOfLit()) return false;
       };
      
      CHECK16(!complex_positions.Empty(),"A0 in FPARA_MACHINE::Paramodulate()"); 
    
      PREFIX_SYM* term_to_rewrite = complex_positions.Pop();
      FPARA_MACH::LoadTermToRewrite(term_to_rewrite);
      candidate = FPARA_MACH::NextCandidate();
      if (candidate) return true;
      goto try_another_position;
    };
 }; // bool FPARA_MACHINE::Paramodulate() 

bool FPARA_MACHINE::CollectRestOfTheRewritingClause()
 {
  unsigned ln = 0;
  unsigned used_lit_num = candidate->LitNum();
  for (TERM* lit = candidate->TheClause()->Literals()->Args(); 
       !lit->IsNull(); 
       lit = lit->Next())
   {
    if (ln != used_lit_num) 
     if (!Resolvent.PushStandardLit(index,lit->First())) return false;
    ln++;
   }; 
  return true;
 }; // bool FPARA_MACHINE::CollectRestOfTheRewritingClause()

//*********************** Definitions for FPARA_MACH: ********************

RES_PAIR_LIST* FPARA_MACH::NextCandidate() 
 {
  RES_PAIR_LIST* result;
  switch (state)
   {
    case INIT:
     if (para_mach.FindFirstLeaf(*(ORIndex.FParaTree(goal->Head().Func()))))
      {
       state = PARA_FROM_COMPLEX;
       if (ordering_constraint->Violated()) 
        {
	  //ordering_constraint->Output(); ::Output("\n\n"); // debug  
        
	  goto try_next_leaf;
        };
       para_mach.ResetCandidatesFromLeaf();
       result = para_mach.NextCandidateFromTheLeaf();
       CHECK16(result,"A10 in FPARA_MACH::NextCandidate()");
       return result;
      };
     goto try_para_from_var; 
    
     /***
    case IDLE:
     while (para_mach.FindNextLeaf()) // temporary, idiotic but correct version
      {
	cout << "BRRRR\n"; // debug
      };
     return (RES_PAIR_LIST*)0;
     ****/  
  
    case PARA_FROM_COMPLEX:
     result = para_mach.NextCandidateFromTheLeaf();
     if (result) return result;
     try_next_leaf:
      if (para_mach.FindNextLeaf())
       {
         
	 // MakeIdle(); // debug

        if (ordering_constraint->Violated()) 
         {

	   //ordering_constraint->Output(); ::Output("\n\n"); // debug 
   
	   goto try_next_leaf;
         };

        para_mach.ResetCandidatesFromLeaf();
        result = para_mach.NextCandidateFromTheLeaf();
        CHECK16(result,"A20 in FPARA_MACH::NextCandidate()");
        return result;
       };
      goto try_para_from_var;
        
    case PARA_FROM_VAR:
     result = para_mach.NextCandidateFromTheLeaf();
     if (result) return result;
     if (para_mach.FindNextLeaf())
      {
       para_mach.ResetCandidatesFromLeaf();
       result = para_mach.NextCandidateFromTheLeaf();
       CHECK16(result,"A30 in FPARA_MACH::NextCandidate()");
       return result;
      };
     return (RES_PAIR_LIST*)0;
   };
     
  try_para_from_var:
   state = PARA_FROM_VAR;
   para_mach.CompilePseudoLiteral(pseudo_header,goal);
      
   if (para_mach.FindFirstLeaf(*(ORIndex.FParaFromVarTree())))
    {
     para_mach.ResetCandidatesFromLeaf();
     result = para_mach.NextCandidateFromTheLeaf();
     CHECK16(result,"A40 in FPARA_MACH::NextCandidate()");
     return result;
    };
   return (RES_PAIR_LIST*)0; 
 }; // RES_PAIR_LIST* FPARA_MACH::NextCandidate()  



//************************************************************************
