#define __FORWARD_SUPERPOSITION_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif 
#ifndef __PARA_MACHINE_H
 #include "para_machine.h"
#endif 
#ifndef __RESOLVENT_H
 #include "resolvent.h"
#endif
#ifndef __WEIGHT_FORMULA_H
 #include "weight_formula.h"
#endif 

extern RESOLVENT Resolvent;

//*************************************************************************


  
class FPARA_MACH 
 {
  private:
   enum STATE { INIT, 
                //IDLE, 
                PARA_FROM_COMPLEX, PARA_FROM_VAR };
  private:   
   STATE state;
   PARA_MACH para_mach;
   PREFIX_SYM* goal;
   TERM pseudo_header; 
   ORD_CONSTR* ordering_constraint;
  public: 
   FPARA_MACH(ORD_CONSTR* constr) 
    : para_mach(),
      pseudo_header((ARITY)1,(FUNCTOR)0,POL_POS),
      ordering_constraint(constr)
    {};
   inline void SetIndex(int ind) { para_mach.SetIndex(ind); };
   inline int GetIndex() const { return para_mach.GetIndex(); };
   inline PREFIX_SYM* Goal() const { return goal; };
   inline void LoadTermToRewrite(PREFIX_SYM* t)
    {
     // t must be a plain complex term with all brackets set
     goal = t;
     para_mach.CompileComplexTerm(goal);
     state = INIT;
    };
    
   RES_PAIR_LIST* NextCandidate();
   //inline void MakeIdle() { state = IDLE; };
 }; // class FPARA_MACH 
  
 
class FPARA_MACHINE : public FPARA_MACH
 {
  private:
   STACK<PREFIX_SYM*,MAX_TERM_WEIGHT> complex_positions;
   PREFIX_SYM* literal;
   RES_PAIR_LIST* candidate;
   int index;
   ORD_CONSTR eq_ord_constr; // experimental
   PREFIX_SYM* eq_second_arg;
  public:
   FPARA_MACHINE() 
    : FPARA_MACH(&eq_ord_constr), 
      complex_positions("MAX_TERM_WEIGHT in FPARA_MACHINE::complex_positions"),
      eq_ord_constr()
    {};
   inline void SetIndex(int ind) { index = ind; FPARA_MACH::SetIndex(ind); };
   inline int GetIndex() const { return index; };
   inline RES_PAIR_LIST* Candidate() const { return candidate; };
   bool LoadLiteral(PREFIX_SYM* lit);
   bool LoadRestOfLit();
   bool Paramodulate();

   inline bool CollectResult()
    {
     return CollectRewrittenLiteral() && CollectRestOfTheRewritingClause();
    };
  private:
   inline bool CollectRewrittenLiteral()
    {
     return Resolvent.PushRewrittenLit(literal,
                                       FPARA_MACH::Goal(),
                                       candidate->ReplacingTerm(),
                                       index);
    };  
    
   bool CollectRestOfTheRewritingClause();
 }; // class FPARA_MACHINE : public FPARA_MACH 
 

class FPARA_PREMISE
 {
  private:
   Clause* clause;
   int index;
   PREFIX_SYM max_literal[MAX_TERM_WEIGHT][MAX_NUMBER_OF_LITERALS];
   int num_of_max_literals;
   TERM* nonmax_literals;
   int next_max_lit;
   bool eq_res_applicable;
   WEIGHT_FORMULA weight_form;
  public:
   FPARA_PREMISE() : weight_form() {};
   inline void SetIndex(int ind) { index = ind; weight_form.SetIndex(ind); };
   inline Clause* Query() const { return clause; };
   inline void ResetIter() { next_max_lit = num_of_max_literals; };
   inline bool EqResApplicable() const { return eq_res_applicable; };
   inline void CreateWeightFormula()
    {
     weight_form.Reset();
     for (TERM* lit_ref = nonmax_literals;!lit_ref->IsNull();lit_ref = lit_ref->Next())
      weight_form.Compile(lit_ref->First());
     
     ResetIter();
     for (PREFIX_SYM* m_lit = NextMaxLit(); m_lit; m_lit = NextMaxLit())
      {
       TERM header = m_lit->Head().Func();
       if (header.IsEquality() && header.IsOrderedEq()) weight_form.Compile(m_lit->Arg2()); 
      };
      
    };
   
   void Load(Clause* cl);

   inline PREFIX_SYM* NextMaxLit()
    {
     if (next_max_lit)
      {
       next_max_lit--;
       return max_literal[next_max_lit];
      };
     return (PREFIX_SYM*)0;
    }; 
   
   inline long MinResultWeight()
    {
     return weight_form.Compute();
    };
    
   bool CollectLiterals();
 }; // class FPARA_PREMISE

class FORWARD_SUPERPOSITION
 {
  private:
   FPARA_MACHINE fpara_machine;
   FPARA_PREMISE fpara_premise;
  public:
   FORWARD_SUPERPOSITION()
    : fpara_machine(),
      fpara_premise()
    {
     fpara_premise.SetIndex(0);
     fpara_machine.SetIndex(1);
    }; 
   bool Load(Clause* cl);
   bool NextInference();
   inline void UndoEqRes()
    {
     GlobalSubst.UnbindLastSection();
    }; 
   bool NextEqRes();

   bool CollectResult();
   bool CollectEqResResult();
   void Process(Clause* cl);
 }; // class FORWARD_SUPERPOSITION
 
 
//**************************************************************************** 
