#include "fs_code.h"


//#define DO(s) { if (new_clause->Number() == 697) { s; }; );
//#define DO(s) { };
#undef DO


//*************** Definitions for FS_CODE: *********************************

void FS_CODE::Remove(Clause* cl) 
 {
  if (!tree) return;
  removal.Reset(&tree,cl);
  TERM* lit_ref = cl->Literals()->Args();
  TERM* lit;
  next_lit:
   if (lit_ref->IsNull()) 
    {
     // end of clause
     if (removal.FindSuccess())
      {
       removal.DestroyBranch();
      }
     else
      {
       #ifdef _DEBUG_
        DebugWarning("(0)FS_CODE::Remove(Clause* cl): The clause is not found in the index.");
        ::Output("CLAUSE: "); ::Output(cl); ::Output("\n");
        
        //OutputCode(); // debug
        //RuntimeError(""); // debug
        
       #endif
      };
     return;
    };
   lit = lit_ref->First();
   lit_ref = lit_ref->Next();
   if (removal.SkipLiteral(lit)) goto next_lit;
   #ifdef _DEBUG_
     DebugWarning("(1)FS_CODE::Remove(Clause* cl): The clause is not found in the index.");
     ::Output("CLAUSE: "); ::Output(cl); ::Output("\n");
     
     //OutputCode(); // debug
     //RuntimeError(""); // debug
     
   #endif
 }; // void FS_CODE::Remove(Clause* cl) 


void FS_CODE::Integrate(Clause* cl) 
 {
  integrator.Reset(&tree,cl);
  TERM* lit_ref = cl->Literals()->Args();
  TERM* lit;
  if (!tree) goto write_mode_next_lit;
  
  read_mode_next_lit:
   if (lit_ref->IsNull()) 
    {
     // end of clause
     integrator.InsertSuccess(cl);
     DCONTROL7("END1 in FS_CODE::Integrate(Clause* cl)");
     return;
    }; 
   lit = lit_ref->First();
   lit_ref = lit_ref->Next();
   if (integrator.ReadLiteral(lit)) 
    {
     goto read_mode_next_lit;
    }
   else
    {
     CHECK7(!integrator.CurrInstr(),"A10 in FS_CODE::Integrate(Clause* cl)");
     goto write_mode_next_lit;
    }; 
    
  write_mode_next_lit:
   DCONTROL7("Y1 in FS_CODE::Integrate(Clause* cl)");
   if (lit_ref->IsNull()) 
    {
     // end of clause
     integrator.WriteSuccess(cl);
     DCONTROL7("END2 in FS_CODE::Integrate(Clause* cl)");
     return;
    }; 
   lit = lit_ref->First();
   lit_ref = lit_ref->Next();
   integrator.WriteLiteral(lit);
   goto write_mode_next_lit;
 }; // void FS_CODE::Integrate(Clause* cl)  


//*************** Definitions for FS_CODE_INTEGRATOR: **********************

void FS_CODE_INTEGRATOR::WriteEqLiteral(TERM* lit)
 {
  CHECK7(lit->IsEquality(),"ARG in FS_CODE_INTEGRATOR::WriteEqLiteral(..)"); 
  TERM* args = lit->Args();
  TERM arg1 = *args;
  TERM arg2 = *(args->Next());
  
  if (lit->IsOrderedEq())
   {
    WriteOrdEq(*lit); 
   }
  else
   {
    WriteUnordEq(*lit); 
   }; 
  
  WriteTerm(arg1);
  WriteSecondEqArg();
  WriteTerm(arg2);
 }; // void FS_CODE_INTEGRATOR::WriteEqLiteral(TERM* lit)

void FS_CODE_INTEGRATOR::WriteLiteral(TERM* lit)
 { 
  if (lit->IsEquality()) { WriteEqLiteral(lit); return; };
  TERM::ITERATOR iter;
  TERM sym;

  if (lit->Arity())
   {
    WriteHeader(*lit);
    
    iter.Reset(lit);
    write_mode_next_sym:
     DCONTROL7("Y8 in FS_CODE_INTEGRATOR::WriteLiteral(..)");
      
     if (!(iter.Next())) return;
               
     DCONTROL7("Y9 in FS_CODE_INTEGRATOR::WriteLiteral(..)");
      
     sym = iter.Symbol();
     if (iter.CurrentSymIsVar())
      {
       WriteVar(sym.Var());
      }
     else // functor
      {
       WriteFunc(sym);
      }; 
     goto write_mode_next_sym; 
   }
  else // propositional
   {
    WritePropLit(*lit);
    return;
   }; 
 }; // void FS_CODE_INTEGRATOR::WriteLiteral(TERM* lit)

bool FS_CODE_INTEGRATOR::ReadVar(int var)
 {
  if (!(ReadVar())) 
   { 
    InsertVar(fs_var);
    if (vars.Occurences(var))
     {
      EnterWriteMode(); 
      WriteCompare(first_occ[var],fs_var);
     }
    else // first occurence
     {
      vars.Register(var);
      first_occ[var] = fs_var;
      EnterWriteMode(); 
     }; 
    fs_var++;
    return false; 
   };
          
  if (vars.Occurences(var))
   {
    ReadCompare(first_occ[var],fs_var);
   } 
  else // first occurence of the variable
   {
    vars.Register(var);
    first_occ[var] = fs_var;
   }; 
  fs_var++;
  return true;
 }; // bool FS_CODE_INTEGRATOR::ReadVar(int var)

bool FS_CODE_INTEGRATOR::ReadComplexTerm(TERM* t)
 {
  TERM::ITERATOR iter;
  TERM sym;
  iter.Reset(t);
  if (ReadFunc(*t)) 
   {
    read_mode_next_sym:
     if (!iter.Next()) return true; 
     sym = iter.Symbol();
     if (iter.CurrentSymIsVar())
      {
       if (!(ReadVar(sym.Var()))) goto write_mode_next_sym; 
      }
     else // functor
      {
       if (!(ReadFunc(sym))) 
        { 
         InsertFunc(sym);
         EnterWriteMode(); 
         goto write_mode_next_sym; 
        };
      }; 
     goto read_mode_next_sym;
   };
  InsertFunc(*t);
  EnterWriteMode();  
  
  write_mode_next_sym:
   if (!iter.Next()) return false;
   sym = iter.Symbol();
   if (iter.CurrentSymIsVar())
    {
     WriteVar(sym.Var());
    }
   else // functor
    {
     WriteFunc(sym);
    }; 
   goto write_mode_next_sym; 
 }; // bool FS_CODE_INTEGRATOR::ReadComplexTerm(TERM* t)

bool FS_CODE_INTEGRATOR::ReadEqLiteral(TERM* lit)
 {
  CHECK7(lit->IsEquality(),"ARG in FS_CODE_INTEGRATOR::ReadEqLiteral(..)"); 
  TERM* args = lit->Args();
  TERM arg1 = *args;
  TERM arg2 = *(args->Next());
  
  if (lit->IsOrderedEq())
   {
    if (!ReadOrdEq(*lit)) { WriteOrdEq(*lit); goto write_mode_first_arg; };
   }
  else
   {
    if (!ReadUnordEq(*lit)) { WriteUnordEq(*lit); goto write_mode_first_arg; };
   }; 
  
  // read mode, first argument 
  if (!ReadTerm(arg1)) goto write_mode_second_arg; 
  
  // read mode, second argument
  ReadSecondEqArg(); 
  return (ReadTerm(arg2)) && (ReadCollectedComps()); 
  
  write_mode_first_arg:
   WriteTerm(arg1);
  write_mode_second_arg:
   WriteSecondEqArg();
   WriteTerm(arg2);
   return false;
 }; // bool FS_CODE_INTEGRATOR::ReadEqLiteral(TERM* lit) 

bool FS_CODE_INTEGRATOR::ReadLiteral(TERM* lit)
 {
  if (lit->IsEquality()) return ReadEqLiteral(lit);
  
  TERM::ITERATOR iter;
  TERM sym;
  
  if (lit->Arity())
   {
    DCONTROL7("X5 in FS_CODE_INTEGRATOR::ReadLiteral(..)");
    if (ReadHeader(*lit))
     {
      DCONTROL7("X6 in FS_CODE_INTEGRATOR::ReadLiteral(..)");
      
      iter.Reset(lit);
      read_mode_next_sym:
       if (!(iter.Next())) return ReadCollectedComps();
       sym = iter.Symbol();
       if (iter.CurrentSymIsVar())
        {
         if (!(ReadVar(sym.Var()))) goto write_mode_next_sym; 
        }
       else // functor
        {
         if (!(ReadFunc(sym))) 
          { 
           InsertFunc(sym);
           EnterWriteMode(); 
           CHECK7(!CurrInstr(),"A30 in FS_CODE_INTEGRATOR::ReadLiteral(..)");
           goto write_mode_next_sym; 
          };
        }; 
       goto read_mode_next_sym; 
     } 
    else // the header was not read
     {
      CHECK7(!CurrInstr(),"A40 in FS_CODE_INTEGRATOR::ReadLiteral(..)");
      WriteHeader(*lit);
      goto write_mode_lit_args;
     };  
   }
  else // propositional literal
   {
    if (ReadPropLit(*lit)) return true;
    // the literal was not read
    CHECK7(!CurrInstr(),"A50 in FS_CODE_INTEGRATOR::ReadLiteral(..)");
    WritePropLit(*lit);
    return false;
   }; 
  
  // write mode: 
  write_mode_lit_args: 
   iter.Reset(lit);
   write_mode_next_sym:
    DCONTROL7("Y8 in FS_CODE::Integrate(Clause* cl)");
       
    if (!(iter.Next())) return false;
                
    DCONTROL7("Y9 in FS_CODE::Integrate(Clause* cl)");
       
    sym = iter.Symbol();
    if (iter.CurrentSymIsVar())
     {
      WriteVar(sym.Var());
     }
    else // functor
     {
      WriteFunc(sym);
     }; 
    goto write_mode_next_sym; 
   
 }; // bool FS_CODE_INTEGRATOR::ReadLiteral(TERM* lit)

bool FS_CODE_INTEGRATOR::ReadHeader(TERM header)
 {
  CHECK7(*node_addr,"A0 in FS_CODE_INTEGRATOR::ReadHeader(..)");
  CHECK7(constr_to_shift.Identity(),"A1 in FS_CODE_INTEGRATOR::ReadHeader(..)");
  CHECK7(constr_to_insert.Identity(),"A2 in FS_CODE_INTEGRATOR::ReadHeader(..)");
  CHECK7(!header.IsEquality(),"A2 in FS_CODE_INTEGRATOR::ReadHeader(..)");
  
  FS_COMMAND* curr_instr;
  int header_num = header.HeaderNum();
  
  check_tag:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::LIT_HEADER :
      if (curr_instr->HeaderNum() == header_num) 
       {
        MoveDown(); 
        return true; 
       };
      // different header
      // no break here! go to the next case 
      
     case FS_COMMAND::PROP_LIT :  // no break here! go to the next case 
     case FS_COMMAND::UNORD_EQ :  // no break here! go to the next case 
     case FS_COMMAND::ORD_EQ :  
      // mismatch occured, try fork
      if (curr_instr->Fork())
       {
        MoveRight();
        goto check_tag;
       };
      // no more forks 
      MoveRight();
      return false;
     
     case FS_COMMAND::SUCCESS : 
      
      // curr_instr->TheClause() subsumes the clause currently being integrated
      #ifdef _DEBUG_
       DebugWarning("FS_CODE_INTEGRATOR::ReadHeader(TERM header): The clause being integrated is subsumed by already integrated");
       AccidForwardSubsumed(new_clause,curr_instr->TheClause());
      #endif
      MoveRight();
      if (*node_addr) { goto check_tag; };
      return false;
       
     #ifdef _DEBUG7_ 
      default : 
       DebugError("X0 in FS_CODE_INTEGRATOR::ReadHeader(TERM header)");
       return false;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default : return false;
      #endif 
     #endif  
    };
 }; // bool FS_CODE_INTEGRATOR::ReadHeader(TERM header)
 

bool FS_CODE_INTEGRATOR::ReadPropLit(TERM header)
 {
  CHECK7(*node_addr,"A0 in FS_CODE_INTEGRATOR::ReadPropLit(..)");
  CHECK7(constr_to_shift.Identity(),"A1 in FS_CODE_INTEGRATOR::ReadPropLit(..)");
  CHECK7(constr_to_insert.Identity(),"A2 in FS_CODE_INTEGRATOR::ReadPropLit(..)");
  
  FS_COMMAND* curr_instr;
  int header_num = header.HeaderNum();
  
  check_tag:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::PROP_LIT :
      if (curr_instr->HeaderNum() == header_num) 
       {
        MoveDown(); 
        return true; 
       };
      // different header
      // no break here! go to the next case 
      
     case FS_COMMAND::LIT_HEADER : // no break here! go to the next case 
     case FS_COMMAND::UNORD_EQ : // no break here! go to the next case 
     case FS_COMMAND::ORD_EQ :   
      // mismatch occured, try fork
      if (curr_instr->Fork())
       {
        MoveRight();
        goto check_tag;
       };
      // no more forks 
      MoveRight();
      return false;
     
     case FS_COMMAND::SUCCESS : 
      
      // curr_instr->TheClause() subsumes the clause currently being integrated
      #ifdef _DEBUG_
       DebugWarning("FS_CODE_INTEGRATOR::ReadPropLit(TERM header): The clause being integrated is subsumed by already integrated");
       AccidForwardSubsumed(new_clause,curr_instr->TheClause());
      #endif
      MoveRight();
      if (*node_addr) { goto check_tag; };
      return false;
       
     #ifdef _DEBUG7_ 
      default : 
       DebugError("X0 in FS_CODE_INTEGRATOR::ReadPropLit(TERM header)");
       return false;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default : return false;
      #endif 
     #endif  
    };
 }; // bool FS_CODE_INTEGRATOR::ReadPropLit(TERM header)
 

bool FS_CODE_INTEGRATOR::ReadUnordEq(TERM header)
 {
  CHECK7(*node_addr,"A0 in FS_CODE_INTEGRATOR::ReadUnordEq(..)");
  CHECK7(constr_to_shift.Identity(),"A1 in FS_CODE_INTEGRATOR::ReadUnordEq(..)");
  CHECK7(constr_to_insert.Identity(),"A2 in FS_CODE_INTEGRATOR::ReadUnordEq(..)");
  
  FS_COMMAND* curr_instr;
  int header_num = header.HeaderNum();
  
  check_tag:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::UNORD_EQ :
      if (curr_instr->HeaderNum() == header_num) 
       {
        MoveDown(); 
        return true; 
       };
      // different header
      // no break here! go to the next case 
      
     case FS_COMMAND::LIT_HEADER : // no break here! go to the next case 
     case FS_COMMAND::PROP_LIT : // no break here! go to the next case 
     case FS_COMMAND::ORD_EQ :   
      // mismatch occured, try fork
      if (curr_instr->Fork())
       {
        MoveRight();
        goto check_tag;
       };
      // no more forks 
      MoveRight();
      return false;
     
     case FS_COMMAND::SUCCESS : 
      
      // curr_instr->TheClause() subsumes the clause currently being integrated
      #ifdef _DEBUG_
       DebugWarning("FS_CODE_INTEGRATOR::ReadUnordEq(TERM header): The clause being integrated is subsumed by already integrated");
       AccidForwardSubsumed(new_clause,curr_instr->TheClause());
      #endif
      MoveRight();
      if (*node_addr) { goto check_tag; };
      return false;
       
     #ifdef _DEBUG7_ 
      default : 
       DebugError("X0 in FS_CODE_INTEGRATOR::ReadUnordEq(TERM header)");
       return false;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default : return false;
      #endif 
     #endif  
    };
 }; // bool FS_CODE_INTEGRATOR::ReadUnordEq(TERM header)

bool FS_CODE_INTEGRATOR::ReadOrdEq(TERM header)
 {
  CHECK7(*node_addr,"A0 in FS_CODE_INTEGRATOR::ReadOrdEq(..)");
  CHECK7(constr_to_shift.Identity(),"A1 in FS_CODE_INTEGRATOR::ReadOrdEq(..)");
  CHECK7(constr_to_insert.Identity(),"A2 in FS_CODE_INTEGRATOR::ReadOrdEq(..)");
  
  FS_COMMAND* curr_instr;
  int header_num = header.HeaderNum();
  
  check_tag:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::ORD_EQ :
      if (curr_instr->HeaderNum() == header_num) 
       {
        MoveDown(); 
        return true; 
       };
      // different header
      // no break here! go to the next case 
      
     case FS_COMMAND::LIT_HEADER : // no break here! go to the next case 
     case FS_COMMAND::PROP_LIT : // no break here! go to the next case 
     case FS_COMMAND::UNORD_EQ :   
      // mismatch occured, try fork
      if (curr_instr->Fork())
       {
        MoveRight();
        goto check_tag;
       };
      // no more forks 
      MoveRight();
      return false;
     
     case FS_COMMAND::SUCCESS : 
      
      // curr_instr->TheClause() subsumes the clause currently being integrated
      #ifdef _DEBUG_
       DebugWarning("FS_CODE_INTEGRATOR::ReadOrdEq(TERM header): The clause being integrated is subsumed by already integrated");
       AccidForwardSubsumed(new_clause,curr_instr->TheClause());
      #endif
      MoveRight();
      if (*node_addr) { goto check_tag; };
      return false;
       
     #ifdef _DEBUG7_ 
      default : 
       DebugError("X0 in FS_CODE_INTEGRATOR::ReadOrdEq(TERM header)");
       return false;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default : return false;
      #endif 
     #endif  
    };
 }; // bool FS_CODE_INTEGRATOR::ReadOrdEq(TERM header)

void FS_CODE_INTEGRATOR::ReadSecondEqArg()
 {
  const FLAT_TERM** var1;
  const FLAT_TERM** var2;
  int var_num1;
  int var_num2;
  FS_COMMAND* curr_instr;
  
  check_tag:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::SECOND_EQ_ARG : MoveDown(); return;
     case FS_COMMAND::COMPARE :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      
      CHECK7(!(curr_instr->Fork()),"A5 in FS_CODE_INTEGRATOR::ReadSecondEqArg()");
      
      CHECK7(!constr_to_shift.Equivalent(var_num1,var_num2),"A6 in FS_CODE_INTEGRATOR::ReadSecondEqArg()");
      
      if (constr_to_insert.Equivalent(var_num1,var_num2))
       {
        constr_passed.MakeEquivalent(var_num1,var_num2);
        MoveDown();
       }
      else // COMPARE(var1,var2) must be shifted  
       {
        constr_to_shift.MakeEquivalent(var_num1,var_num2);
        *node_addr = curr_instr->Next();
        delete curr_instr;
       };
      goto check_tag; 
     #ifdef _DEBUG7_ 
      default : 
       DebugError("X0 in FS_CODE_INTEGRATOR::ReadSecondEqArg()");
       return;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default : return;
      #endif 
     #endif 
    };
 }; // void FS_CODE_INTEGRATOR::ReadSecondEqArg()

bool FS_CODE_INTEGRATOR::ReadFunc(TERM func)
 {
  const FLAT_TERM** var1;
  const FLAT_TERM** var2;
  int var_num1;
  int var_num2;
  FS_COMMAND* curr_instr;
  
  check_tag:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::FUNC :
      if (curr_instr->Func() == func) 
       {
        if (curr_instr->Fork()) CopyComps(constr_to_shift,curr_instr->ForkAddr());        
        MoveDown(); 
        return true; 
       };
      // different symbol
      if (curr_instr->Func() > func)
       {
        InsertComps(constr_to_shift,node_addr);
        return false;
       };  
      // curr_instr->Func() < func here
      // no break here! go to the next case 
      
     case FS_COMMAND::VAR :
      // mismatch occured, try fork
      if (curr_instr->Fork())
       {
        CopyComps(constr_to_shift,curr_instr->NextAddr());
        MoveRight();
        goto check_tag;
       };
      // no more forks 
      InsertComps(constr_to_shift,curr_instr->NextAddr());
      MoveRight();
      return false;
     
     case FS_COMMAND::COMPARE :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      
      CHECK7(!(curr_instr->Fork()),"A5 in FS_CODE_INTEGRATOR::ReadFunc(TERM func)");
      
      CHECK7(!constr_to_shift.Equivalent(var_num1,var_num2),"A6 in FS_CODE_INTEGRATOR::ReadFunc(TERM func)");
      
      if (constr_to_insert.Equivalent(var_num1,var_num2))
       {
        constr_passed.MakeEquivalent(var_num1,var_num2);
        MoveDown();
       }
      else // COMPARE(var1,var2) must be shifted  
       {
        constr_to_shift.MakeEquivalent(var_num1,var_num2);
        *node_addr = curr_instr->Next();
        delete curr_instr;
       };
      goto check_tag; 
     #ifdef _DEBUG7_ 
      default : 
       DebugError("X0 in FS_CODE_INTEGRATOR::ReadFunc(TERM func)");
       return false;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default : return false;
      #endif 
     #endif 
    };
 }; // bool FS_CODE_INTEGRATOR::ReadFunc(TERM func)


bool FS_CODE_INTEGRATOR::ReadVar()
 {
  const FLAT_TERM** var1;
  const FLAT_TERM** var2;
  int var_num1;
  int var_num2;
  FS_COMMAND* curr_instr;
  
  check_tag:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::VAR : 
      if (curr_instr->Fork()) CopyComps(constr_to_shift,curr_instr->ForkAddr());
      MoveDown(); 
      return true;
      
     case FS_COMMAND::FUNC :
      // a FUNC can not precede a VAR in a tree
      InsertComps(constr_to_shift,node_addr);
      return false;
     
     case FS_COMMAND::COMPARE :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      CHECK7(!(curr_instr->Fork()),"A5 in FS_CODE_INTEGRATOR::ReadVar()");
      
      CHECK7(!constr_to_shift.Equivalent(var_num1,var_num2),"A6 in FS_CODE_INTEGRATOR::ReadVar()");
      
      if (constr_to_insert.Equivalent(var_num1,var_num2))
       {
        constr_passed.MakeEquivalent(var_num1,var_num2);
        MoveDown();
       }
      else // COMPARE(var1,var2) must be shifted  
       {
        constr_to_shift.MakeEquivalent(var_num1,var_num2);
        *node_addr = curr_instr->Next();
        delete curr_instr;
       };
      goto check_tag;
       
     #ifdef _DEBUG7_ 
      default : 
       DebugError("X0 in FS_CODE_INTEGRATOR::ReadVar()");
       return false;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default : return false;
      #endif 
     #endif 
    };
 }; // bool FS_CODE_INTEGRATOR::ReadVar()

void FS_CODE_INTEGRATOR::ReadCompare(const FLAT_TERM** v1,const FLAT_TERM** v2)
 {
  CHECK7(v1 < v2,"A0 in FS_CODE_INTEGRATOR::ReadCompare(..)");
  int var_num1 = VarNum(v1);
  int var_num2 = VarNum(v2);
  if (constr_to_shift.SplitOn(var_num1,var_num2))
   {
    constr_passed.MakeEquivalent(var_num1,var_num2);
    FS_COMMAND* curr_instr = FS_COMMAND::NewCOMPARE(v1,v2);
    curr_instr->SetNext(*node_addr);
    *node_addr = curr_instr;
    MoveDown();
   }
  else constr_to_insert.MakeEquivalent(var_num1,var_num2); 
 }; // void FS_CODE_INTEGRATOR::ReadCompare(const FLAT_TERM** v1,const FLAT_TERM** v2)

bool FS_CODE_INTEGRATOR::ReadCollectedComps()
 {
  const FLAT_TERM** var1;
  const FLAT_TERM** var2;
  int var_num1;
  int var_num2;
  FS_COMMAND* curr_instr;
  if (constr_to_insert.Identity())
   {
    if (constr_to_shift.Identity())
     {
      both_empty_skip_comps:
       curr_instr = *node_addr;
       if (curr_instr->IsCompare())
        {
         MoveRight();
         if (curr_instr->Fork()) goto both_empty_skip_comps; 
         // no more forks
         CHECK7(!(*node_addr),"S10 in FS_CODE_INTEGRATOR::ReadCollectedComps()");
         return false;
        };
       // !curr_instr->IsCompare() 
       return true; 
     }
    else // constr_to_insert.Identity() and !constr_to_shift.Identity()
     {
      InsertComps(constr_to_shift,node_addr);
      MoveRight();
      CHECK7(!(*node_addr),"S20 in FS_CODE_INTEGRATOR::ReadCollectedComps()");
      return false;
     }; 
   }
  else // !constr_to_insert.Identity()
   {
    if (constr_to_shift.Identity())
     {
      insert_nonempty_check_tag:
       curr_instr = *node_addr;
       if (curr_instr->IsCompare())
        {
         var1 = curr_instr->Var1();
         var2 = curr_instr->Var2();
         var_num1 = VarNum(var1);
         var_num2 = VarNum(var2);
         if (constr_to_insert.Equivalent(var_num1,var_num2))
          {
           constr_passed.MakeEquivalent(var_num1,var_num2);
           MoveDown();
           //if (constr_to_insert.Identity()) goto both_empty_skip_comps;
           goto insert_nonempty_check_tag; 
          }
         else // try fork
          {
           MoveRight();
           if (curr_instr->Fork()) goto insert_nonempty_check_tag; 
           // no fork
           constr_to_insert.Subtract(constr_passed);
           MoveComps(constr_to_insert,node_addr);
           CHECK7(!(*node_addr),"S30 in FS_CODE_INTEGRATOR::ReadCollectedComps()");
           return false;
          };
        }
       else // !curr_instr->IsCompare()
        {
         constr_to_insert.Subtract(constr_passed);
           
         if (constr_to_insert.Identity()) 
          {
           return true;
          };
        
         FS_COMMAND** node_addr_backup = node_addr;
         *node_addr = (FS_COMMAND*)0;
         MoveComps(constr_to_insert,node_addr);
         (*node_addr_backup)->SetFork(curr_instr);
         CHECK7(!(*node_addr),"S40 in FS_CODE_INTEGRATOR::ReadCollectedComps()");
         return false;
        };
     }
    else // !constr_to_insert.Identity() and !constr_to_shift.Identity()
     {
      both_nonempty_check_tag:
       curr_instr = *node_addr;
       if (curr_instr->IsCompare())
        {
         var1 = curr_instr->Var1();
         var2 = curr_instr->Var2();
         var_num1 = VarNum(var1);
         var_num2 = VarNum(var2);
         if (constr_to_insert.Equivalent(var_num1,var_num2))
          {
           constr_passed.MakeEquivalent(var_num1,var_num2);
           
           if (curr_instr->Fork()) CopyComps(constr_to_shift,curr_instr->ForkAddr());
           MoveDown();
            
           /*****
           if (constr_to_insert.Identity()) 
            {
             InsertComps(constr_to_shift,node_addr);
             MoveRight();
             CHECK7(!(*node_addr),"S50 in FS_CODE_INTEGRATOR::ReadCollectedComps()");
             return false;
            };
           *****/
            
           goto both_nonempty_check_tag; 
          }
         else // try fork
          {
           CopyComps(constr_to_shift,curr_instr->NextAddr());
           MoveRight(); 
           if (curr_instr->Fork()) { goto both_nonempty_check_tag; };
           
           // no fork
           
           constr_to_insert.Subtract(constr_passed);
           
           MoveComps(constr_to_insert,node_addr);
           constr_to_shift.Reset();
           CHECK7(!(*node_addr),"S60 in FS_CODE_INTEGRATOR::ReadCollectedComps()");
           return false;
          }; 
        }
       else // !curr_instr->IsCompare()
        {
         constr_to_insert.Subtract(constr_passed);
         
         InsertComps(constr_to_shift,node_addr);
         MoveRight();
         MoveComps(constr_to_insert,node_addr);
         CHECK7(!(*node_addr),"S70 in FS_CODE_INTEGRATOR::ReadCollectedComps()");
         return false;
        };
     }; 
   }; 
 }; // bool FS_CODE_INTEGRATOR::ReadCollectedComps()


//************* Definitions for FS_CODE_REMOVAL: ***********************

bool FS_CODE_REMOVAL::SkipComplexTerm(TERM* t)
 {
  TERM::ITERATOR iter;
  iter.Reset(t);
  skip_sym:
   if (iter.CurrentSymIsVar())
    {
     if (!SkipVar(iter.Symbol().Var())) return false;
    }
   else // functor
    {
     if (!(SkipFunc(iter.Symbol()))) return false;
    };
   if (iter.Next()) goto skip_sym; 
   return true; 
 }; // bool FS_CODE_REMOVAL::SkipComplexTerm(TERM* t)

bool FS_CODE_REMOVAL::SkipEqLiteral(TERM* lit)
 {
  CHECK7(lit->IsEquality(),"ARG in FS_CODE_REMOVAL::SkipEqLiteral(..)");
  if (lit->IsOrderedEq())
   {
    if (!(SkipOrdEq(*lit))) return false;
   }
  else
   {
    if (!(SkipUnordEq(*lit))) return false;
   }; 
  TERM* args = lit->Args();
  TERM arg1 = *args;
  TERM arg2 = *(args->Next());
  
  return (SkipTerm(arg1)) && (SkipSecondEqArg()) && (SkipTerm(arg2));
  
 }; // bool FS_CODE_REMOVAL::SkipEqLiteral(TERM* lit)


bool FS_CODE_REMOVAL::SkipLiteral(TERM* lit)
 {
  if (lit->IsEquality()) return SkipEqLiteral(lit);
 
  TERM::ITERATOR iter;
  TERM sym;
  
  if (lit->Arity())
   {
    if (SkipHeader(*lit))
     {
      iter.Reset(lit);
      next_sym:
       if (!(iter.Next())) return true;
       sym = iter.Symbol();
       if (iter.CurrentSymIsVar())
        {
         if (!SkipVar(sym.Var())) return false;
        }
       else // functor
        {
         if (!(SkipFunc(sym))) return false;
        };
       goto next_sym;   
      }
     else return false;
    }
   else // propositional literal
    {
     return SkipPropLit(*lit);
    };
 }; // bool FS_CODE_REMOVAL::SkipLiteral(TERM* lit)




//*************************************************************************