#define __FS_CODE_H
 
#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif
#ifndef __MEMORY_H
 #include "memory.h"
#endif
#ifndef __TERM_H
 #include "term.h"
#endif
#ifndef __CLAUSE_H
 #include "clause.h"
#endif 
#ifndef __VARIABLES_H
 #include "variables.h"
#endif 
#ifndef __FLAT_TERM_H
 #include "flat_term.h"
#endif 
#ifndef __OUTPUT_H
 #include "output.h"
#endif  
#ifndef __EQUIVALENCE_RELATION_H
 #include "equivalence_relation.h"
#endif  
#ifndef __FS_COMMAND_H
 #include "fs_command.h"
#endif 

//**************************************************************************

typedef EQ_REL<MAX_CLAUSE_WEIGHT> FS_EQ_CONSTR; 

class FS_CODE_INTEGRATOR // must be static - too costly constructor
 {  
  private:
   Clause* new_clause;
   FS_COMMAND** node_addr;
   
   FS_EQ_CONSTR constr_to_insert;
   FS_EQ_CONSTR constr_to_shift;
   FS_EQ_CONSTR constr_passed;
   
   const FLAT_TERM** fs_subst;
  
   const FLAT_TERM** first_occ[MAX_NUMBER_OF_VARS];
   VAR_MULTISET vars; 
   const FLAT_TERM** fs_var;
   
  public:
   FS_CODE_INTEGRATOR(const FLAT_TERM** subst) 
    : constr_to_insert(),
      constr_to_shift(),
      constr_passed(),
      fs_subst(subst),
      vars() 
    {}; 
   inline void Reset(FS_COMMAND** n,Clause* cl) 
    {
     new_clause = cl; 
     node_addr = n; 
     constr_to_insert.Reset();
     constr_to_shift.Reset();
     constr_passed.Reset();
     vars.Reset();
     fs_var = fs_subst;
    };
   inline void MoveDown() { node_addr = (*node_addr)->NextAddr(); };
   inline void MoveRight() { node_addr = (*node_addr)->ForkAddr(); };
   inline int VarNum(const FLAT_TERM** v) const { return v - fs_subst; };
   inline const FLAT_TERM** Var(int var_num) const { return fs_subst + var_num; };

   inline void CopyComps(const FS_EQ_CONSTR& constr,FS_COMMAND** addr)
    {
     int vn1;
     int vn2;
     FS_EQ_CONSTR::ITER iter(constr);
     FS_COMMAND* old_instr = *addr;
     while (iter.Next(vn2,vn1)) 
      { 
       CHECK7(vn1 < vn2,"A0 in FS_CODE_INTEGRATOR::CopyComps(FS_EQ_CONSTR& constr,..)");
       FS_COMMAND* new_instr = FS_COMMAND::NewCOMPARE(Var(vn1),Var(vn2));
       *addr = new_instr;
       addr = new_instr->NextAddr(); 
      };
     *addr = old_instr;
    };
   
   inline void MoveComps(FS_EQ_CONSTR& constr,FS_COMMAND**& addr)
    {
     int vn1;
     int vn2;
     FS_COMMAND* old_instr = *addr;
     while (constr.Split(vn2,vn1)) 
      { 
       CHECK7(vn1 < vn2,"A0 in FS_CODE_INTEGRATOR::MoveComps(FS_EQ_CONSTR& constr,..)");
       FS_COMMAND* new_instr = FS_COMMAND::NewCOMPARE(Var(vn1),Var(vn2));
       *addr = new_instr;
       addr = new_instr->NextAddr(); 
      };
     CHECK7(constr.Identity(),"A10 in FS_CODE_INTEGRATOR::MoveComps(FS_EQ_CONSTR& constr,..)");   
     *addr = old_instr;
    };
   
   inline void InsertComps(FS_EQ_CONSTR& constr,FS_COMMAND** addr)
    {
     int vn1;
     int vn2;
     FS_COMMAND* old_instr = *addr;
     while (constr.Split(vn2,vn1)) 
      { 
       CHECK7(vn1 < vn2,"A0 in FS_CODE_INTEGRATOR::InsertComps(FS_EQ_CONSTR& constr,..)");
       FS_COMMAND* new_instr = FS_COMMAND::NewCOMPARE(Var(vn1),Var(vn2));
       *addr = new_instr;
       addr = new_instr->NextAddr(); 
      };
     CHECK7(constr.Identity(),"A10 in FS_CODE_INTEGRATOR::InsertComps(FS_EQ_CONSTR& constr,..)");   
     *addr = old_instr;
    };
    
   
    // Read mode
    
   bool ReadHeader(TERM header);
   bool ReadPropLit(TERM header);
   bool ReadUnordEq(TERM header);
   bool ReadOrdEq(TERM header);
   void ReadSecondEqArg();
   bool ReadFunc(TERM func);
   bool ReadVar();
   bool ReadVar(int var);
   void ReadCompare(const FLAT_TERM** v1,const FLAT_TERM** v2);
   bool ReadCollectedComps();
   bool ReadLiteral(TERM* lit);
   bool ReadEqLiteral(TERM* lit); 
   bool ReadComplexTerm(TERM* t);
   inline bool ReadTerm(TERM t)
    {
     return (t.IsVariable()) ? ReadVar(t.Var()) : ReadComplexTerm(t.First());
    };
   
    // Write mode
    
   void WriteLiteral(TERM* lit);
   void WriteEqLiteral(TERM* lit);
   
   inline void EnterWriteMode();
   
   inline void WriteHeader(TERM header) 
    { 
     *node_addr = FS_COMMAND::NewLIT_HEADER(header);
     MoveDown();
    }; 
   inline void WritePropLit(TERM header)
    {
     *node_addr = FS_COMMAND::NewPROP_LIT(header);
     MoveDown();
    }; 
   inline void WriteUnordEq(TERM header)
    {
     *node_addr = FS_COMMAND::NewUNORD_EQ(header);
     MoveDown();
    };
   inline void WriteOrdEq(TERM header)
    {
     *node_addr = FS_COMMAND::NewORD_EQ(header);
     MoveDown();
    };
   inline void WriteSecondEqArg()
    {
     *node_addr = FS_COMMAND::NewSECOND_EQ_ARG();
     MoveDown();
    }; 
    
   inline void WriteFunc(TERM func)
    {
     *node_addr = FS_COMMAND::NewFUNC(func);
     MoveDown();
    }; 
   
   inline void InsertFunc(TERM func)
    {
     FS_COMMAND* tmp = *node_addr;
     *node_addr = FS_COMMAND::NewFUNC(func);
     (*node_addr)->SetFork(tmp);
     MoveDown();
    };
    
   inline void WriteVar(const FLAT_TERM** v)
    {
     *node_addr = FS_COMMAND::NewVAR(v);
     MoveDown();
    };  
   
   inline void WriteVar(int var)
    {
     WriteVar(fs_var);
     if (vars.Occurences(var))
      {
       WriteCompare(first_occ[var],fs_var);
      }
     else // first occurence
      {
       vars.Register(var);
       first_occ[var] = fs_var;
      }; 
     fs_var++;
    };
   
   inline void WriteComplexTerm(TERM* t)
    {
     TERM::ITERATOR iter;
     iter.Reset(t);
     
     write_curr_sym:
      if (iter.CurrentSymIsVar())
       {
        WriteVar(iter.Symbol().Var());
       }
      else // functor
       {
        WriteFunc(iter.Symbol());
       };
      if (iter.Next()) goto write_curr_sym;
    }; 
    
   inline void WriteTerm(TERM t)
    {
     if (t.IsVariable()) { WriteVar(t.Var()); }
     else WriteComplexTerm(t.First());
    };
    
   inline void InsertVar(const FLAT_TERM** v)
    {
     FS_COMMAND* tmp = *node_addr;
     *node_addr = FS_COMMAND::NewVAR(v);
     (*node_addr)->SetFork(tmp);
     MoveDown();
    }; 
   inline void WriteCompare(const FLAT_TERM** v1,const FLAT_TERM** v2)
    {
     *node_addr = FS_COMMAND::NewCOMPARE(v1,v2);
     MoveDown(); 
    };
   inline void WriteSuccess(Clause* cl)
    {
     *node_addr = FS_COMMAND::NewSUCCESS(cl);
    };
   inline void InsertSuccess(Clause* cl)
    {
     FS_COMMAND* tmp = *node_addr;
     *node_addr = FS_COMMAND::NewSUCCESS(cl);
     (*node_addr)->SetFork(tmp);
    };
  public: // for debug
   inline const FS_COMMAND* CurrInstr() const { return *node_addr; };  
  private: 
   /****************
     // Processing clauses that are accidentally backward subsumed
   inline void AccidBackSubsumed(Clause* subsumed,Clause* subsuming)
    {
     // temporary
     ::Output("CLAUSE\n"); 
     ::Output(subsuming);
     ::Output("\nACCIDENTALLY BACKWARD SUBSUMED\n");
     ::Output(subsumed);
     ::Output("\n");
    };   
   inline void AccidBackSubsumedSubtree(FS_COMMAND* tree,Clause* subsuming_cl)
    {
     if (tree->Fork()) AccidBackSubsumedSubtree(tree->Fork(),subsuming_cl);
     if (tree->Tag() == FS_COMMAND::SUCCESS) 
      {
       AccidBackSubsumed(tree->TheClause(),subsuming_cl);
       delete tree;
      }
     else
      {
       AccidBackSubsumedSubtree(tree->Next(),subsuming_cl);
       delete tree;
      };  
    }; 
   *****************/ 
     // Processing clauses that are accidentally forward subsumed
   inline void AccidForwardSubsumed(Clause* subsumed,Clause* subsuming)
    {
     // temporary
     ::Output("CLAUSE\n"); 
     ::Output(subsuming); 
     ::Output("\nACCIDENTALLY FORWARD SUBSUMED\n");
     ::Output(subsumed);
     ::Output("\n");
    };
     
 }; // class FS_CODE_INTEGRATOR
 
class FS_CODE_REMOVAL
 {
  private:
   Clause* removed_clause;
   FS_COMMAND** node_addr;
   FS_COMMAND** branch_to_destroy;
   
   FS_COMMAND* comps; // list of instructions, can be changed to a STACK
   
   FS_EQ_CONSTR constr;
   
   const FLAT_TERM** fs_subst;
  
   const FLAT_TERM** first_occ[MAX_NUMBER_OF_VARS];
   VAR_MULTISET vars; 
   const FLAT_TERM** fs_var; 
   
  private:
   inline void MoveDown() { node_addr = (*node_addr)->NextAddr(); };
   inline void MoveRight() { node_addr = (*node_addr)->ForkAddr(); };
   inline int VarNum(const FLAT_TERM** v) const { return v - fs_subst; };
   inline const FLAT_TERM** Var(int var_num) const { return fs_subst + var_num; }; 
  public:
   FS_CODE_REMOVAL(const FLAT_TERM** subst)
    : constr(), 
      fs_subst(subst),
      vars() 
    {};
   inline void Reset(FS_COMMAND** n,Clause* cl) 
    {
     removed_clause = cl;
     node_addr = n; 
     branch_to_destroy = n; 
     comps = (FS_COMMAND*)0;
     constr.Reset();
     vars.Reset();
     fs_var = fs_subst;
    };
   
   bool SkipComplexTerm(TERM* t);
   bool SkipLiteral(TERM* lit); 
   bool SkipEqLiteral(TERM* lit); 
   
   inline bool SkipTerm(TERM t) 
    {
     return (t.IsVariable()) ? SkipVar(t.Var()) : SkipComplexTerm(t.First());
    };
   
    
   inline bool SkipHeader(TERM header);
   inline bool SkipPropLit(TERM header);
   inline bool SkipUnordEq(TERM header);
   inline bool SkipOrdEq(TERM header);
   inline bool SkipSecondEqArg();
   
   inline bool SkipFunc(TERM func);
   inline bool SkipVar();
   inline bool SkipVar(int var)
    {
     if (SkipVar())
      {
       if (vars.Occurences(var))
        {
         SkipCompare(first_occ[var],fs_var);
        } 
       else // first occurence of the variable
        {
         vars.Register(var);
         first_occ[var] = fs_var;
        }; 
       fs_var++;
       return true;
      }
     else return false;
    }; 
   inline void SkipCompare(const FLAT_TERM** v1,const FLAT_TERM** v2); 
   inline bool FindSuccess();
   inline void DestroyBranch() 
    {
     FS_COMMAND* tmp = *branch_to_destroy;
     *branch_to_destroy = tmp->Fork();
     tmp->SetFork((FS_COMMAND*)0);
     FS_COMMAND::DestroyTree(tmp);
    };
  public: // for debugging
   inline FS_COMMAND* Node() const { return *node_addr; }; 
 }; // class FS_CODE_REMOVAL 

class FS_CODE 
 {
  private:
   FS_COMMAND* tree;
   const FLAT_TERM** subst;
   FS_CODE_INTEGRATOR integrator; // quite heavy object (slow constructor), can not be local
   FS_CODE_REMOVAL removal; // quite heavy object (slow constructor), can not be local
  public:
   FS_CODE(const FLAT_TERM** sbst) 
    : tree((FS_COMMAND*)0),
      subst(sbst),
      integrator(subst), 
      removal(subst)  
    {};
   inline FS_COMMAND* Tree() const { return tree; }; 
   inline const FLAT_TERM** Subst() { return subst; };
   void Integrate(Clause* cl);
   void Remove(Clause* cl);
   inline void Destroy()
    {
     if (tree) FS_COMMAND::DestroyTree(tree);
     tree = (FS_COMMAND*)0;
    };
  public: // for debugging
   inline void OutputCode();
 }; // class FS_CODE

//*************** Definitions for FS_CODE: *********************************

inline void FS_CODE::OutputCode()
 {
  if (!tree) { ::Output("<<< Empty FS_CODE >>>\n"); return; };
  ::Output("<<< FS_CODE >>>\n");
  FS_COMMAND::OutputTree(tree,subst); 
  ::Output("<<< End of FS_CODE >>>\n");
 }; // void FS_CODE::Output() 
 
//*************** Definitions for FS_CODE_INTEGRATOR: **********************


inline void FS_CODE_INTEGRATOR::EnterWriteMode()
 {
  // must be used if ReadVar or ReadFunc failed
  CHECK7(constr_to_shift.Identity(),"A0 in FS_CODE_INTEGRATOR::EnterWriteMode()");
  constr_to_insert.Subtract(constr_passed);
  MoveComps(constr_to_insert,node_addr);  
 }; // void FS_CODE_INTEGRATOR::EnterWriteMode()

//**************** Definitions for FS_CODE_REMOVAL: ************************************

inline bool FS_CODE_REMOVAL::SkipHeader(TERM header)
 {
  const FLAT_TERM** var1;
  const FLAT_TERM** var2;
  int var_num1;
  int var_num2;
  FS_COMMAND* curr_instr;
  int header_num = header.HeaderNum();
  CHECK7(*node_addr,"A0 in FS_CODE_REMOVAL::SkipHeader(..)");
  check_instr:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::LIT_HEADER :
      if (curr_instr->HeaderNum() == header_num) 
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        node_addr = curr_instr->NextAddr();
        return true;
       };
      try_fork: 
       if (curr_instr->Fork()) 
        {
         node_addr = curr_instr->ForkAddr();
         branch_to_destroy = node_addr;
         goto check_instr;
        }; 
       return false; 
     case FS_COMMAND::COMPARE :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      if (constr.Equivalent(var_num1,var_num2))
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        node_addr = curr_instr->NextAddr();
        goto check_instr; 
       }
      else { goto try_fork; };
     default : goto try_fork;
    };
 }; // bool FS_CODE_REMOVAL::SkipHeader(TERM header)

inline bool FS_CODE_REMOVAL::SkipPropLit(TERM header)
 {
  const FLAT_TERM** var1;
  const FLAT_TERM** var2;
  int var_num1;
  int var_num2;
 
  FS_COMMAND* curr_instr;
  int header_num = header.HeaderNum();
  CHECK7(*node_addr,"A0 in FS_CODE_REMOVAL::SkipPropLit(..)");
  check_instr:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::PROP_LIT :
      if (curr_instr->HeaderNum() == header_num) 
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        node_addr = curr_instr->NextAddr();
        return true;
       };
      try_fork: 
       if (curr_instr->Fork()) 
        {
         node_addr = curr_instr->ForkAddr();
         branch_to_destroy = node_addr;
         goto check_instr;
        }; 
       return false; 
     case FS_COMMAND::COMPARE :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      
      if (constr.Equivalent(var_num1,var_num2))
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        node_addr = curr_instr->NextAddr();
        goto check_instr; 
       }
      else { goto try_fork; };
     default : goto try_fork;
    };
 }; // bool FS_CODE_REMOVAL::SkipPropLit(TERM header)

inline bool FS_CODE_REMOVAL::SkipUnordEq(TERM header)
 {
  const FLAT_TERM** var1;
  const FLAT_TERM** var2;
  int var_num1;
  int var_num2;
  FS_COMMAND* curr_instr;
  int header_num = header.HeaderNum();
  CHECK7(*node_addr,"A0 in FS_CODE_REMOVAL::SkipUnordEq(..)");
  check_instr:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::UNORD_EQ :
      if (curr_instr->HeaderNum() == header_num) 
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        node_addr = curr_instr->NextAddr();
        return true;
       };
      try_fork: 
       if (curr_instr->Fork()) 
        {
         node_addr = curr_instr->ForkAddr();
         branch_to_destroy = node_addr;
         goto check_instr;
        }; 
       return false; 
     case FS_COMMAND::COMPARE :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2); 
      if (constr.Equivalent(var_num1,var_num2))
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        node_addr = curr_instr->NextAddr();
        goto check_instr; 
       }
      else { goto try_fork; };
     default : goto try_fork;
    };
 }; // bool FS_CODE_REMOVAL::SkipUnordEq(TERM header)


inline bool FS_CODE_REMOVAL::SkipOrdEq(TERM header)
 {
  const FLAT_TERM** var1;
  const FLAT_TERM** var2;
  int var_num1;
  int var_num2;
  FS_COMMAND* curr_instr;
  int header_num = header.HeaderNum();
  CHECK7(*node_addr,"A0 in FS_CODE_REMOVAL::SkipOrdEq(..)");
  check_instr:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::ORD_EQ :
      if (curr_instr->HeaderNum() == header_num) 
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        node_addr = curr_instr->NextAddr();
        return true;
       };
      try_fork: 
       if (curr_instr->Fork()) 
        {
         node_addr = curr_instr->ForkAddr();
         branch_to_destroy = node_addr;
         goto check_instr;
        }; 
       return false; 
     case FS_COMMAND::COMPARE :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      if (constr.Equivalent(var_num1,var_num2))
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        node_addr = curr_instr->NextAddr();
        goto check_instr; 
       }
      else { goto try_fork; };
     default : goto try_fork;
    };
 }; // bool FS_CODE_REMOVAL::SkipOrdEq(TERM header)

inline bool FS_CODE_REMOVAL::SkipSecondEqArg()
 {
  const FLAT_TERM** var1;
  const FLAT_TERM** var2;
  int var_num1;
  int var_num2;
  
  FS_COMMAND* curr_instr;
  check_instr:
   curr_instr = *node_addr;
   CHECK7(!(curr_instr->Fork()),"A5 in FS_CODE_REMOVAL::SkipSecondEqArg()");
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::SECOND_EQ_ARG : 
      node_addr = curr_instr->NextAddr();
      return true;
     case FS_COMMAND::COMPARE :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      
      if (constr.Equivalent(var_num1,var_num2))
       {
        node_addr = curr_instr->NextAddr();
        goto check_instr; 
       }
      else return false; 
     #ifdef _DEBUG7_
      default:
       DebugError("X0 in FS_CODE_REMOVAL::SkipSecondEqArg()");
       return false;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default : return false;
      #endif
     #endif 
    };
 }; // bool FS_CODE_REMOVAL::SkipSecondEqArg()


inline bool FS_CODE_REMOVAL::SkipFunc(TERM func)
 {
  const FLAT_TERM** var1;
  const FLAT_TERM** var2;
  int var_num1;
  int var_num2;
  
  FS_COMMAND* curr_instr;
  CHECK7(*node_addr,"A0 in FS_CODE_REMOVAL::SkipFunc(..)");
  check_instr:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::FUNC :
      if (curr_instr->Func() == func)
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        node_addr = curr_instr->NextAddr();
        return true;
       };
      try_fork: 
       if (curr_instr->Fork()) 
        {
         node_addr = curr_instr->ForkAddr();
         branch_to_destroy = node_addr;
         goto check_instr;
        }; 
       return false;
     case FS_COMMAND::COMPARE :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      
      if (constr.Equivalent(var_num1,var_num2))
       {
        CHECK7(!(curr_instr->Fork()),"A5 in FS_CODE_REMOVAL::SkipFunc(..)");
        node_addr = curr_instr->NextAddr();
        goto check_instr; 
       }
      else return false;
     default : goto try_fork;  
    };
 }; // bool FS_CODE_REMOVAL::SkipFunc(TERM func)

inline bool FS_CODE_REMOVAL::SkipVar()
 {
  const FLAT_TERM** var1;
  const FLAT_TERM** var2;
  int var_num1;
  int var_num2;
 
  FS_COMMAND* curr_instr;
  CHECK7(*node_addr,"A0 in FS_CODE_REMOVAL::SkipVar()");
  check_instr:
   curr_instr = *node_addr;
   switch (curr_instr->Tag())
    {
     case FS_COMMAND::VAR :
      if (curr_instr->Fork()) branch_to_destroy = node_addr;
      node_addr = curr_instr->NextAddr();
      return true;
     case FS_COMMAND::COMPARE :
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      
      if (constr.Equivalent(var_num1,var_num2))
       {
        CHECK7(!(curr_instr->Fork()),"A5 in FS_CODE_REMOVAL::SkipVar()");
        node_addr = curr_instr->NextAddr();
        goto check_instr; 
       }
      else return false;
     default : 
      // try fork 
      if (curr_instr->Fork()) 
       {
        node_addr = curr_instr->ForkAddr();
        branch_to_destroy = node_addr;
        goto check_instr;
       }; 
      return false;
    };
 }; // bool FS_CODE_REMOVAL::SkipVar()

inline void FS_CODE_REMOVAL::SkipCompare(const FLAT_TERM** v1,const FLAT_TERM** v2)
 {
  CHECK7(v1 < v2,"A0 in FS_CODE_REMOVAL::SkipCompare(..)");
  constr.MakeEquivalent(VarNum(v1),VarNum(v2));
 }; // void FS_CODE_REMOVAL::SkipCompare(const FLAT_TERM** v1,const FLAT_TERM** v2)


inline bool FS_CODE_REMOVAL::FindSuccess()
 {
  const FLAT_TERM** var1;
  const FLAT_TERM** var2;
  int var_num1;
  int var_num2;
  
  FS_COMMAND* curr_instr;
  CHECK7(*node_addr,"A0 in FS_CODE_REMOVAL::FindSuccess(..)");
  check_tag:
   curr_instr = *node_addr;
    if (curr_instr->IsCompare())
     {
      var1 = curr_instr->Var1();
      var2 = curr_instr->Var2();
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      if (constr.Equivalent(var_num1,var_num2))
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        MoveDown();
        goto check_tag;
       };
      try_fork: 
       if (curr_instr->Fork())
        {
         node_addr = curr_instr->ForkAddr();
         branch_to_destroy = node_addr;
         goto check_tag;
        };
       return false;  
     };
    // curr_instr->IsCompare()
    if (curr_instr->IsSuccess())
     {
      if (curr_instr->TheClause() == removed_clause)
       {
        if (curr_instr->Fork()) branch_to_destroy = node_addr;
        return true;
       };
     };
    goto try_fork;
 }; // bool FS_CODE_REMOVAL::FindSuccess() 


//**************************************************************************