#define __HYPER_CASCADE_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif 
#ifndef __IU_COMMAND_H
 #include "iu_command.h"
#endif
#ifndef __OR_MACHINE_H
 #include "or_machine.h"
#endif 
#ifndef __PREMISE_FACTORING_H
 #include "premise_factoring.h"
#endif
#ifndef __RESOLVENT_H
 #include "resolvent.h"
#endif

extern RESOLVENT Resolvent;

//******************************************************************

class SINGLE_HYPER_MACHINE : public OR_MACH, public OR_PREMISE_WITH_FACTORING
 {
  private: // "best candidate" strategy
   bool variablesInitiallyAffected;
   bool bestCandidate;
  public:
   SINGLE_HYPER_MACHINE() : OR_MACH(), OR_PREMISE_WITH_FACTORING() {};
   
   static inline void ResetStatic() { OR_MACH::ResetStatic(); };
   
   inline void SetBestCandidate(bool bc) { bestCandidate = bc; };
   inline bool BestCandidate() { return bestCandidate; };
   
   inline bool Start(PREFIX_SYM* lit)
    {
     bool res = OR_MACH::CompileLiteral(lit);
     OR_MACH::MemorizeState();
     SetBestCandidate(false);
     return res; 
    };
   
   inline bool FindFirstLeaf()
    {
     if (OR_MACH::FindFirstLeaf()) 
      {
       variablesInitiallyAffected = OR_MACH::VariablesFromQueryAffected();
       return true;
      };
     return false; 
    };
    
   inline bool FindNextLeaf()
    { 
     if (OR_MACH::FindNextLeaf()) 
      {
       variablesInitiallyAffected = OR_MACH::VariablesFromQueryAffected();
       return true;
      };
     return false; 
    };
    
   inline bool UnitBestCandidate()
    {
     return ((!(variablesInitiallyAffected))
             && (OR_PREMISE_WITH_FACTORING::TheClause()->Unit()));
    };
  public:
   inline void LoadFirstCandidate()
    {
     OR_MACH::ResetCandidatesFromLeaf();
     OR_PREMISE_WITH_FACTORING::LoadOptimized(OR_MACH::GetIndex(),OR_MACH::NextCandidate());   
    };
   inline bool ChangeCandidate()
    {
     OR_PREMISE_WITH_FACTORING::Unload();
     RES_PAIR_LIST* new_candidate = OR_MACH::NextCandidate();
     if (new_candidate)
      {
       OR_PREMISE_WITH_FACTORING::LoadOptimized(OR_MACH::GetIndex(),new_candidate);
       return true;
      }
     else 
      {
       return false; 
      };
    }; // bool ChangeCandidate()
    
   inline void Output()
    {
     OR_MACH::Output();
    };
 }; // class SINGLE_HYPER_MACHINE

/***********************************************************************/

template <int size>
 class HYPER_CASCADE  
  {
   private:
    SINGLE_HYPER_MACHINE ORMachine[size]; 
    int next_machine;
   public:
    HYPER_CASCADE() 
     {
      ResetStatic();
      for (int i = 0; i < size; i++)
       {
        ORMachine[i].SetIndex(i+2);
       };
     };
     
    static inline void ResetStatic() { SINGLE_HYPER_MACHINE::ResetStatic(); }; 
    inline bool LoadLHSLiteral(PREFIX_SYM* lhs_literal,bool wasUsed)
     {
      #ifdef _SUPPRESS_WARNINGS_
       wasUsed = false;
      #endif
      SINGLE_HYPER_MACHINE* machine = &(ORMachine[next_machine]);
      if ((machine->Start(lhs_literal)))
       {
        if ((machine->FindFirstLeaf()))
         {
          next_machine++;
          return true;
         };
       };
      machine->RestoreState();
      return false;
     }; // bool LoadLHSLiteral(PREFIX_SYM* lhs_literal,bool wasUsed)

   inline bool Backtrack(int& unloaded)
    {
     DCONTROL2("A1 in HYPER_CASCADE::Backtrack(..)");
    
     unloaded = 0;
     SINGLE_HYPER_MACHINE* machine = &(ORMachine[next_machine - 1]);
     
     DCONTROL2("A2 in HYPER_CASCADE::Backtrack(..)");
     
     try_next_leaf:
      
      if ((!(machine->BestCandidate())) && (machine->FindNextLeaf()))
       {
        DCONTROL2("A3 in HYPER_CASCADE::Backtrack(..)");
        
        next_machine = (int)(machine - ORMachine) + 1;
        return true;
       }
      else 
       {
        DCONTROL2("A4 in HYPER_CASCADE::Backtrack(..)");
        
        machine->RestoreState();
        unloaded++;
        if (machine == ORMachine) // no more resolvents
         {
          next_machine = 0;
          return false;
         }
        else
         {
          machine--;
          
          DCONTROL2("B1 in HYPER_CASCADE::Backtrack(..)");
          goto try_next_leaf;
         };
       };
    }; // bool Backtrack(int& unloaded) 
    
   inline void LoadFirstCandidates()
    {
     SINGLE_HYPER_MACHINE* lastMachine = &(ORMachine[next_machine - 1]);
     for (SINGLE_HYPER_MACHINE* machine = ORMachine; machine <= lastMachine; machine++)
      {
       machine->LoadFirstCandidate();
       machine->SetBestCandidate(machine->UnitBestCandidate());
      };
    }; // void LoadFirstCandidates()
    
   inline bool ChangeCandidates()
    {
     SINGLE_HYPER_MACHINE* lastMachine = &(ORMachine[next_machine - 1]);
     SINGLE_HYPER_MACHINE* machine = lastMachine;
     while (machine >= ORMachine)
      {
       if ((!(machine->BestCandidate())) && (machine->ChangeCandidate()))
        {
         machine->SetBestCandidate(machine->UnitBestCandidate());
         machine++;
         while (machine <= lastMachine)
          {
           if (!(machine->BestCandidate()))
            {
             machine->LoadFirstCandidate();
            };
           machine++;
          };
         return true;
        };
       machine--;
      };
     return false;
    }; // bool ChangeCandidates()
    
   inline void ResetFactoring() 
    {
     SINGLE_HYPER_MACHINE* lastMachine = &(ORMachine[next_machine - 1]);
     for (SINGLE_HYPER_MACHINE* machine = ORMachine; machine <= lastMachine; machine++)
      {
       if (!(machine->BestCandidate()))
        { 
         machine->ResetFactoring();
        };
      };
    }; // void ResetFactoring()  
    
   inline bool NextFactor() 
    {
     SINGLE_HYPER_MACHINE* lastMachine = &(ORMachine[next_machine - 1]);
     for (SINGLE_HYPER_MACHINE* machine = ORMachine; machine <= lastMachine; machine++)
      {
       if (!(machine->BestCandidate()))
        {
         if (machine->WasFactored())
          {
           if (machine->GetNextFactor())
            {
             return true;
            }
           else
            {
             machine->ResetFactoring();
            };
          }
         else 
          {
           if (machine->GetFirstFactor(0))
            {
             return true;
            };
          };
        };
      };
     return false; // no more factors 
    }; // bool NextFactor() 
    
   inline bool CollectLiterals() 
    {
     SINGLE_HYPER_MACHINE* lastMachine = &(ORMachine[next_machine - 1]);
     for (SINGLE_HYPER_MACHINE* machine = ORMachine; machine <= lastMachine; machine++)
      {
       if (!(machine->CollectNonmarkedLiterals()))
        {
         return false;
        };
      };
     return true;
    }; // bool CollectLiterals()
   inline void CollectAncestors() 
    {
     SINGLE_HYPER_MACHINE* lastMachine = &(ORMachine[next_machine - 1]);
     for (SINGLE_HYPER_MACHINE* machine = ORMachine; machine <= lastMachine; machine++)
      {
       Resolvent.PushAncestor(machine->TheClause());
      };
    }; // void CollectAncestors() 
    
   inline void Output()
    {
    };
  }; // class HYPER_CASCADE<int size>

//************************************************************************  
