#define __HYPER_RULE_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __CLAUSE_H
 #include "clause.h"
#endif
#ifndef __PREFIX_H
 #include "prefix.h"
#endif
#ifndef __VARIABLES_H
 #include "variables.h"
#endif
#ifndef __PREMISE_FACTORING_H
 #include "premise_factoring.h"
#endif
#ifndef __IU_COMMAND_H
 #include "iu_command.h"
#endif
#ifndef __IU_TRANSLATOR_H
 #include "iu_translator.h"
#endif 
#ifndef __UNIFICATION_H
 #include "unification.h"
#endif
#ifndef __RESOLVENT_H
 #include "resolvent.h"
#endif


extern RESOLVENT Resolvent;

//************************************************************************

class HYPER_RULE 
 {
  //long bullshit1[500]; // debug

  // A rule has form: LHS -> RHS
  private: // structure
   static bool LHS_Negative; // determines orientation of rules in hyperresolution
   Clause* clause; // the input clause itself
   bool compiled;
   PREFIX_SYM** RightHandSide; // array of pointers to terms in word representation
   PREFIX_SYM** LeftHandSide;  // array of pointers to terms in word representation
   IU_COMMAND** LeftHandSideCode; 
   PREFIX_SYM** endOfLHS;
   PREFIX_SYM** endOfRHS;
   int sizeOfLHS;
   int sizeOfRHS;
   int number;
   static int lastRule;
   static const int ordinaryRuleIndex = 0;
   static const int unitConflictRuleIndex = MAX_NUMBER_OF_STANDARD_INDICES + 0;
   static bool unit_conflict;
   int ruleIndex;
   PREFIX_SYM* wordRepMemory;
   static IU_TRANSLATOR Compiler;
   static PREFIX_SYM tmpWordMemory[MAX_TERM_WEIGHT];
   IU_COMMAND* codeMemory;
   int memorySize;

   //long bullshit2[500]; // debug

  public: // methods
   HYPER_RULE() 
    {
    };
   ~HYPER_RULE() 
    {
     delete [] wordRepMemory;
     
     if (compiled) 
      { 
       delete [] LeftHandSideCode;
       delete [] codeMemory;
      };
     
     delete [] RightHandSide;
     delete [] LeftHandSide;
     clause->Destroy();
    }; 
   static inline void PositiveHyper(bool flag) { LHS_Negative = flag; };
   static inline bool IsPositiveHyper() { return LHS_Negative; };
   static inline void SetUnitConflict(bool flag) { unit_conflict = flag; };
   inline int Number() { return number; };
   static inline int LastRule() { return lastRule; };
   inline int RuleIndex() { return ruleIndex; };
   inline bool MakeRule(Clause* c,bool compile);
   inline Clause* TheClause() { return clause; };
   inline bool Compiled() const { return compiled; };
   inline PREFIX_SYM** RHS() { return RightHandSide; };
   inline PREFIX_SYM** LHS() { return LeftHandSide; };
   inline PREFIX_SYM** EndOfLHS() { return endOfLHS; };
   inline PREFIX_SYM** EndOfRHS() { return endOfRHS; };
   inline IU_COMMAND** LHSCode() { return LeftHandSideCode; };
   inline int SizeOfLHS() { return sizeOfLHS; };
   inline int SizeOfRHS() { return sizeOfRHS; };
   inline int LongLHS() { return (SizeOfLHS() > 1); };
   inline bool Unit() { return ((SizeOfLHS() == 1) && (!(SizeOfRHS()))); };
   inline void Output(); // for debugging
 }; // class HYPER_RULE

class LHS_LIT_COMPLEXITY
 {
  private:
   int numOfFuncs;
   int numOfVarOccs;
   int numOfFirstVarOccs;
   LARGE_VAR_COUNTER var_counter;
  public:
   inline void Estimate(PREFIX_SYM* lit)
    {
     PREFIX::ITER iter;
     iter.Reset(lit);
     iter.SkipSym();
     int num_of_funcs = 0;
     int num_of_var_occs = 0;
     int num_of_first_var_occs = 0;
     while (!(iter.Empty()))
      {
       if (iter.CurrVar())
        {
         num_of_var_occs++;
         if (!(var_counter.Register(iter.CurrVar()))) 
          {
           num_of_first_var_occs++;
          };
        }
       else
        {
         num_of_funcs++;
        };
       iter.SkipSym();
      };
     numOfFuncs = num_of_funcs;
     numOfVarOccs = num_of_var_occs;
     numOfFirstVarOccs = num_of_first_var_occs;
    }; // void Estimate(PREFIX_SYM* lit)
    
   inline COMPARISON Compare(LHS_LIT_COMPLEXITY& est)
    {
     if (numOfVarOccs - numOfFirstVarOccs > est.numOfVarOccs - est.numOfFirstVarOccs)
      {
       return GREATER;
      }
     else
      {
       if (numOfVarOccs - numOfFirstVarOccs < est.numOfVarOccs - est.numOfFirstVarOccs)
        {
         return LESS;
        }
       else
        {
         if (numOfFuncs >= est.numOfFuncs)
          {
           return GREATER;
          }
         else
          {
           return LESS;
          };
        };
      };
    };
 }; // class LHS_LIT_COMPLEXITY

//********************************************************************

class HYPER_ACTIVE_RULE : public HYPER_RULE
 {
  private:
   static const int currentFactIndex = 1;
   static const int unitConflictSuspectIndex = MAX_NUMBER_OF_STANDARD_INDICES + 1;
  private:
   static OR_PREMISE_WITH_FACTORING currentFact;
   static OR_PREMISE unitConflictSuspect;
   static int cutLHSLiteral;
   static int currentLHSLiteral;
   static bool used[MAX_NUMBER_OF_LITERALS];
   static bool active[MAX_NUMBER_OF_LITERALS];
   static STACK<int,MAX_NUMBER_OF_LITERALS> inactive;
   static int numOfActive;
  private: // dinamic ordering of lhs literals
   static LHS_LIT_COMPLEXITY est1,est2;
   static LHS_LIT_COMPLEXITY* maxEst;
   static LHS_LIT_COMPLEXITY* currEst;
   static bool dinamicLHSOrdering;
  private: // restriction on application of a rule
   static bool empty_RHS_only; 
    // "empty_RHS_only == true" means that only rules with empty RHS are applicable 
  public:
   
  /************ debug ***********
   inline void* operator new(size_t)
    {
     DCONTROL2("BEGIN in new HYPER_ACTIVE_RULE");
     void* res = (void*)(new HYPER_RULE());
     DCONTROL2("BEGIN in new HYPER_ACTIVE_RULE");
     return res;
    };

   inline void operator delete(void* obj)
    {
     
    };
    *****************************/  
  

   static inline void Initialise()
    {
     dinamicLHSOrdering = true; // default
     empty_RHS_only = false;
    }; 
   static inline void SetDinamicLHSOrdering(bool flag) 
    { dinamicLHSOrdering = flag; };
   static inline void EmptyRHSOnly() { empty_RHS_only = true; };
   static inline void SetCurrentFact(Clause* fact)
    {
     currentFact.Load(currentFactIndex,fact);
    };
   static inline void SetUnitConflictSuspect(Clause* fact)
    {
     unitConflictSuspect.Load(unitConflictSuspectIndex,fact);
    };
   static inline Clause* FactClause() { return currentFact.TheClause(); };
   static inline Clause* UnitConflictSuspectClause() 
    { return unitConflictSuspect.TheClause(); };
   static inline void UnloadCurrentFact() { currentFact.Unload(); }; 
   static inline void UnloadUnitConflictSuspect() { unitConflictSuspect.Unload(); };
   inline bool TryToApply()
    {
     if ((empty_RHS_only) && (SizeOfRHS())) { return false; };
     
     GlobalSubst.UnbindAll();
     
     for (int l = 0; l < SizeOfLHS(); l++) { active[l] = true; };
     
     PREFIX_SYM** last_lit = currentFact.MaxLiterals() + currentFact.LastMax();
     
     for (PREFIX_SYM** lhs_lit = LHS(); lhs_lit < EndOfLHS(); lhs_lit++)
      {
       for (PREFIX_SYM** lit = currentFact.MaxLiterals(); lit <= last_lit; lit++)
        {
         if (Compiled() ? Unify(*(LHSCode()+(lhs_lit - LHS())),*lit)
                        : UNIFICATION::UnifyComplex(*lit,*lhs_lit))
          {
           currentFact.UnmarkCut();
           currentFact.Cut((int)(lit - currentFact.MaxLiterals()));
           cutLHSLiteral = (int)(lhs_lit - LHS());
           used[cutLHSLiteral] = true;
           active[cutLHSLiteral] = false;
           numOfActive = SizeOfLHS() - 1;
           for (int l = cutLHSLiteral + 1; l < SizeOfLHS(); l++)
            {
             used[l] = false;
            };
           return true;
          };
        };
      };
     return false; 
    }; // bool TryToApply()
    
   inline bool NextApplication()
    {
     GlobalSubst.UnbindAll();
     PREFIX_SYM** last_lit = currentFact.MaxLiterals() + currentFact.LastMax();
     // first try to resolve the current LHS literal with another literal from the fact
     PREFIX_SYM** first_lit = currentFact.MaxLiterals() + currentFact.CutLit() + 1;
     PREFIX_SYM** lhs_lit = LHS() + cutLHSLiteral;
     
     for (PREFIX_SYM** lit = first_lit; lit <= last_lit; lit++)
      {
       if (Compiled() ? Unify(*(LHSCode()+(lhs_lit - LHS())),*lit)
                      : UNIFICATION::UnifyComplex(*lit,*lhs_lit))
        {
         currentFact.UnmarkCut();
         currentFact.Cut((int)(lit - currentFact.MaxLiterals()));
         return true;
        };
      };
     
     // try other LHS literals
     
     active[cutLHSLiteral] = true;
     lhs_lit++;
     while (lhs_lit < EndOfLHS())
      {
       for (PREFIX_SYM** lit = currentFact.MaxLiterals(); lit <= last_lit; lit++)
        {
         if (UNIFICATION::UnifyComplex(*lit,*lhs_lit))
          {
           currentFact.UnmarkCut();
           currentFact.Cut((int)(lit - currentFact.MaxLiterals()));
           cutLHSLiteral = (int)(lhs_lit - LHS());
           used[cutLHSLiteral] = true;
           active[cutLHSLiteral] = false;
           return true;
          };
        };
       lhs_lit++;
      };
     return false; 
    }; // bool NextApplication()
   
   inline bool Conflict()
    {
     // the rule and the current suspect must be unit here!
     // index of the rule must be the corresponding one
     
     if (Compiled())
      {
       return Unify(*(LHSCode()),*(unitConflictSuspect.MaxLiterals()));
      }
     else
      {
       return UNIFICATION::UnifyComplex(*(unitConflictSuspect.MaxLiterals()),
                                        *(LHS()));  
      };  
     
     //return UNIFICATION::Unify(*(unitConflictSuspect.MaxLiterals()),*(LHS())); 
                     
    };
   
  private:
   inline bool CollectRHSLiterals()
    {
     PREFIX_SYM** end_of_rhs = EndOfRHS();
     for (PREFIX_SYM** rhs_lit = RHS(); rhs_lit < end_of_rhs; rhs_lit++)
      {
       //if (!(TmpClause.PushPrefixLit(*rhs_lit))) { return false; };
       if (!(Resolvent.PushPrefixLit(*rhs_lit))) { return false; };
      };
     return true;
    }; 
  public:
   static inline void ResetFactoring() { currentFact.ResetFactoring(); };
   static inline bool NextFactor() 
    {
     if (currentFact.WasFactored())
      {
       return currentFact.GetNextFactor(); 
      }
     else
      {
       return currentFact.GetFirstFactor(0);
      };
    };
   inline bool CollectLiterals()
    {
     return ((currentFact.CollectNonmarkedLiterals()) && (CollectRHSLiterals()));
    };
  public: // iteration over LHS with respect to the ordering
   static inline void ResetIterOverLHS()
    {
    };
   inline PREFIX_SYM* NextLHSLiteral(bool& wasUsed) 
    {
     if (!numOfActive) { return (PREFIX_SYM*)0; };
     currentLHSLiteral = 0;
     while (!(active[currentLHSLiteral])) { currentLHSLiteral++; }; 
     if ((numOfActive > 1) && (dinamicLHSOrdering))
      // more then one active literal, maximal must be found
      {
       maxEst = &est1;
       currEst = &est2;
       LHS_LIT_COMPLEXITY* tmp;
       
       maxEst->Estimate((LHS())[currentLHSLiteral]);
       int restOfActive = numOfActive - 1;
       int lhs_lit = currentLHSLiteral + 1;
       
       try_next_active:
        // first find next active literal
        while (!(active[lhs_lit])) { lhs_lit++; };
        restOfActive--;
        currEst->Estimate((LHS())[lhs_lit]);
        if (maxEst->Compare(*currEst) == LESS)
         {
          tmp = maxEst;
          maxEst = currEst;
          currEst = tmp;
          currentLHSLiteral = lhs_lit;
         };
        if (restOfActive) 
         {
          lhs_lit++;
          goto try_next_active; 
         };
      }; 
      
     PREFIX_SYM* res = (LHS())[currentLHSLiteral];
     wasUsed = used[currentLHSLiteral];
     active[currentLHSLiteral] = false;
     inactive.Push(currentLHSLiteral);
     numOfActive--;
     return res;
    };
   static inline void Rollback()
    {
     numOfActive++;
     currentLHSLiteral = inactive.Pop();
     active[currentLHSLiteral] = true; 
    };
   static inline void Rollback(int n) { while (n) { Rollback(); n--; }; };
 };
//*********************************************************************

//***************** Definitions for HYPER_RULE: *******************

inline bool HYPER_RULE::MakeRule(Clause* c,bool compile)
 {
  DCONTROL2("BEGIN in HYPER_RULE::MakeRule(Clause* c,bool compile)");
  compiled = compile;
  // first simply count sizes of LHS and RHS
  sizeOfLHS = 0;
  sizeOfRHS = 0;
  
  if (LHS_Negative)
   {
    for (TERM* lit = c->Literals()->Args(); !lit->IsNull(); lit = lit->Next())
     {
      if (lit->First()->Polarity() == POL_NEG) // LHS
       {
        sizeOfLHS++;
       }
      else // RHS
       {
        sizeOfRHS++;
       };
     };
   }
  else // LHS must be positive
   {
    for (TERM* lit = c->Literals()->Args(); !lit->IsNull(); lit = lit->Next())
     {
      if (lit->First()->Polarity() == POL_POS) 
       {
        sizeOfLHS++;
       }
      else // RHS
       {
        sizeOfRHS++;
       };
     };
   };
   
  if (!sizeOfLHS) { return false; }; // the clause is a fact 
  clause = c;


  //cout << "THE CLAUSE\n   "; ::Output(c); cout << "\n"; 


  DCONTROL2("E100 in HYPER_RULE::MakeRule(Clause* c,bool compile)");
  RightHandSide = new (PREFIX_SYM*)[sizeOfRHS];
  DCONTROL2("E110 in HYPER_RULE::MakeRule(Clause* c,bool compile)"); 
  LeftHandSide = new (PREFIX_SYM*)[sizeOfLHS];
  DCONTROL2("E120 in HYPER_RULE::MakeRule(Clause* c,bool compile)");

  if ((unit_conflict) && (Unit()))
   {
    ruleIndex = unitConflictRuleIndex;
   }
  else
   {
    ruleIndex = ordinaryRuleIndex;
   };

  int lhs = 0;
  int rhs = 0;
  memorySize = (c->Weight()); 

  DCONTROL2("E200 in HYPER_RULE::MakeRule(Clause* c,bool compile)");
  wordRepMemory = new PREFIX_SYM[memorySize];
  DCONTROL2("E210 in HYPER_RULE::MakeRule(Clause* c,bool compile)");

  PREFIX_SYM* memory = wordRepMemory; 
  
  if (compiled)
   {
    DCONTROL2("E300 in HYPER_RULE::MakeRule(Clause* c,bool compile)"); 
    LeftHandSideCode = new (IU_COMMAND*)[sizeOfLHS];
    DCONTROL2("E310 in HYPER_RULE::MakeRule(Clause* c,bool compile)");
     
    codeMemory = new (IU_COMMAND)[((MAX_TERM_WEIGHT + 2)*sizeOfLHS)]; // the size can be optimized
    DCONTROL2("E320 in HYPER_RULE::MakeRule(Clause* c,bool compile)");
   }
  else
   {
    LeftHandSideCode = (IU_COMMAND**)0; // just in case
    codeMemory = (IU_COMMAND*)0; // just in case
   };
   
  IU_COMMAND* code_memory = codeMemory;
  PREFIX_SYM* tmp_memory; 
   
  if (LHS_Negative)
   {
    for (TERM* lit = c->Literals()->Args(); !lit->IsNull(); lit = lit->Next())
     {
      if (lit->First()->Polarity() == POL_NEG) // LHS
       {
        LeftHandSide[lhs] = memory;
        
        if (compiled)
         {
          LeftHandSideCode[lhs] = code_memory;
          tmp_memory = tmpWordMemory;
          StandardTermToPrefix(ruleIndex,lit->First(),tmp_memory);
          Compiler.CompileLiteral(tmpWordMemory,code_memory,memory,MAX_TERM_WEIGHT);
          
          LeftHandSideCode[lhs]->InversePolarity();
          
         }
        else
         {
          StandardTermToPrefix(ruleIndex,lit->First(),memory);
         };

        LeftHandSide[lhs]->Head().InversePolarity();
        lhs++;
       }
      else // RHS
       {
        RightHandSide[rhs] = memory;
        rhs++;
        StandardTermToPrefix(ruleIndex,lit->First(),memory);
       };
     };
   }
  else // LHS must be positive
   {
    for (TERM* lit = c->Literals()->Args(); !lit->IsNull(); lit = lit->Next())
     {
      if (lit->First()->Polarity() == POL_POS) // LHS
       {
        LeftHandSide[lhs] = memory;
        
        if (compiled)
         {
          LeftHandSideCode[lhs] = code_memory;
          tmp_memory = tmpWordMemory;
          StandardTermToPrefix(ruleIndex,lit->First(),tmp_memory);
          Compiler.CompileLiteral(tmpWordMemory,code_memory,memory,MAX_TERM_WEIGHT);
          
          LeftHandSideCode[lhs]->InversePolarity();
          
         }
        else
         {
          StandardTermToPrefix(ruleIndex,lit->First(),memory);
         };
        LeftHandSide[lhs]->Head().InversePolarity();
        lhs++;
       }
      else // RHS
       {
        RightHandSide[rhs] = memory;
        rhs++;
        StandardTermToPrefix(ruleIndex,lit->First(),memory);
       };
     };
   };
  
  endOfLHS = LeftHandSide + sizeOfLHS;
  endOfRHS = RightHandSide + sizeOfRHS;
  
  lastRule++;
  number = lastRule;
  return true;
 }; // bool HYPER_RULE::MakeRule(Clause* c,bool compile) 
 
inline void HYPER_RULE::Output()
 {
  ::Output("RULE #"); ::Output(Number()); ::Output(" ");
  //::Output("{"); ::Output(RuleIndex()); ::Output("} ");
  ::Output(TheClause()); 
  ::Output("\n");
  /****
  ::Output("    LHS:\n");
  PREFIX_SYM** lhs = LHS();
  for (int i = SizeOfLHS(); i; i--)
   {
    ::Output("       "); PREFIX::OutputTerm(*lhs); ::Output("\n");
    lhs++;
   };
  ::Output("    RHS:\n");
  PREFIX_SYM** rhs = RHS();
  for (int i = SizeOfRHS(); i; i--)
   {
    ::Output("       "); PREFIX::OutputTerm(*rhs); ::Output("\n");
    rhs++;
   };
   ***/
 }; // void HYPER_RULE::Output()
//*************************************************************************
