#define __HYPERRESOLUTION_H

#ifndef __HYPER_RULE_H
 #include "hyper_rule.h"
#endif
#ifndef __CLAUSE_H
 #include "clause.h"
#endif
#ifndef __READCMD_H
 #include "readcmd.h" 
#endif
#ifndef __HYPER_CASCADE_H
 #include "hyper_cascade.h"
#endif

#ifndef __RESOLVENT_H
 #include "resolvent.h"
#endif

//*****************************************************************

class HYPERRESOLUTION;

extern HYPERRESOLUTION HyperResolution;

//*****************************************************************

extern RESOLVENT Resolvent;

class HYPER_ALL_RULES
 {
  private:
   HYPER_ACTIVE_RULE* newRule;
   CLAUSE_LIST* rules;
   CLAUSE_LIST* unitRules;
   bool unit_conflict;
  private:
   CLAUSE_LIST* restOfRules;
   CLAUSE_LIST* moreRules;
   HYPER_ACTIVE_RULE* currentRule;
   HYPER_ACTIVE_RULE* unitConflictRule;
  public:
   HYPER_ALL_RULES() 
   : newRule(new HYPER_ACTIVE_RULE()),
     rules((CLAUSE_LIST*)0),
     unitRules((CLAUSE_LIST*)0),
     unit_conflict(true) // default
    {
     HYPER_RULE::SetUnitConflict(unit_conflict);
    };
   inline void ResetRules()
    {
     CLAUSE_LIST* tmp = rules;
     while (tmp)
      {
       rules = rules->Next();
       
       delete ((HYPER_ACTIVE_RULE*)(tmp->First()));
       delete tmp;
       tmp = rules;
      };
     tmp = unitRules;
     while (tmp)
      {
       unitRules = unitRules->Next();
       delete ((HYPER_ACTIVE_RULE*)(tmp->First()));
       delete tmp;
       tmp = unitRules;
      };
    };
   inline void SetUnitConflict(bool uc) 
    {
     unit_conflict = uc; 
     HYPER_RULE::SetUnitConflict(uc);
    };
   inline bool UnitConflict() { return unit_conflict; };
   inline bool NewRule(Clause* c,bool compile)
    {
     DCONTROL1("BEGIN in HYPER_ALL_RULES::NewRule(Clause* c,bool compile)");
     if (newRule->MakeRule(c,compile))
      {
       
       //::Output(c); cout << "\n"; // debug
       //newRule->Output(); cout << "\n"; // debug      

       DCONTROL1("A1 in HYPER_ALL_RULES::NewRule(Clause* c,bool compile)"); 
       if ((unit_conflict) && (newRule->Unit()))
        {
         DCONTROL1("A2 in HYPER_ALL_RULES::NewRule(Clause* c,bool compile)"); 
         unitRules = new CLAUSE_LIST((Clause*)newRule,unitRules);
         DCONTROL1("A3 in HYPER_ALL_RULES::NewRule(Clause* c,bool compile)"); 
        }
       else
        {
         DCONTROL1("A4 in HYPER_ALL_RULES::NewRule(Clause* c,bool compile)"); 
         rules = new CLAUSE_LIST((Clause*)newRule,rules);
         DCONTROL1("A5 in HYPER_ALL_RULES::NewRule(Clause* c,bool compile)"); 
        };
     
       //cout << "NEW RULE\n"; // debug     

       newRule = new HYPER_ACTIVE_RULE();

       //cout << "DONE\n"; // debug

       DCONTROL1("END1 in HYPER_ALL_RULES::NewRule(Clause* c,bool compile)"); 
       return true;
      }
     else 
      {
       DCONTROL1("END2 in HYPER_ALL_RULES::NewRule(Clause* c,bool compile)"); 
       return false; 
      };    
    };
   inline HYPER_ACTIVE_RULE* CurrentRule() { return currentRule; };
   inline HYPER_ACTIVE_RULE* UnitConflictRule() { return unitConflictRule; };
   inline bool FindNextRule()
    {
     continue_search:
      while (restOfRules)
       {
        currentRule = (HYPER_ACTIVE_RULE*)(restOfRules->First());
        restOfRules = restOfRules->Next();
        if (currentRule->TryToApply())
         {
          return true;
         };
       };
      if (moreRules)
       {
        restOfRules = moreRules;
        moreRules = (CLAUSE_LIST*)0;
        goto continue_search;
       };
      currentRule = (HYPER_ACTIVE_RULE*)0; // for debugging
      return false;
    };
   inline bool ResetSearch(bool units_eligible)
    {
     restOfRules = rules; 
     if (units_eligible)
      {
       moreRules = unitRules;
      }
     else
      {
       moreRules = (CLAUSE_LIST*)0;
      };
     currentRule = (HYPER_ACTIVE_RULE*)0;
     return FindNextRule();    
    };
   inline Clause* CheckForUnitConflict(Clause* suspect)
    {
     HYPER_ACTIVE_RULE::SetUnitConflictSuspect(suspect);
     CLAUSE_LIST* restOfUnitRules = unitRules;
     while (restOfUnitRules)
      {
       unitConflictRule = (HYPER_ACTIVE_RULE*)(restOfUnitRules->First());
       restOfUnitRules = restOfUnitRules->Next();
       
       //DCONTROL0("X");
       
       if (unitConflictRule->Conflict())
        {
         
         //DCONTROL0("Y");
        
         HYPER_ACTIVE_RULE::UnloadUnitConflictSuspect(); 
         return (unitConflictRule->TheClause());
        };
       
       //DCONTROL0("Z");
       
      };
     HYPER_ACTIVE_RULE::UnloadUnitConflictSuspect();
     return (Clause*)0;
    };
  public: // for debugging
   inline void OutputRules()
    {
     if (unit_conflict)
      {
       Output("-------------- NONUNIT RULES: ---------------------------\n");
      }
     else
      {
       Output("------------------ ALL RULES: ---------------------------\n");
      };
     CLAUSE_LIST* rs = rules;
     while (rs)
      {
       ((HYPER_ACTIVE_RULE*)(rs->First()))->Output();
       Output("\n");
       rs = rs->Next();
      };
     if (unit_conflict)
      {
       Output("------------------ UNIT RULES: --------------------------\n");
       rs = unitRules;
       while (rs)
        {
         ((HYPER_ACTIVE_RULE*)(rs->First()))->Output();
         Output("\n");
         rs = rs->Next();
        };
      };
     Output("------------------ NO MORE RULES: ------------------------\n");
    };
 }; // class HYPER_ALL_RULES

//***********************************************************************

extern void OutputPrefix(PREFIX_SYM*);

class HYPERRESOLUTION : public HYPER_ALL_RULES
 {
  //private: // memory
  // TERM fact_word_memory[MAX_TERM_WEIGHT*2*MAX_NUMBER_OF_LITERALS*MAX_NUMBER_OF_LITERALS];
 
  private:
   HYPER_CASCADE<MAX_NUMBER_OF_LITERALS-1> cascade;
   Clause* the_current_fact;
  private:
   bool FindLeaves();
   bool ChangeLeaves();
   void ProcessLeaves();
   void ProcessCandidates();
   void NewResolvent();
   void ResetFactoring();
   bool NextFactor();
  public:
   HYPERRESOLUTION() 
    : HYPER_ALL_RULES(),
      cascade()
    {
     HYPER_ACTIVE_RULE::Initialise();
    };
   static inline void SetDinamicLHSOrdering(bool flag) 
    { HYPER_ACTIVE_RULE::SetDinamicLHSOrdering(flag); };
   inline void SetUnitConfict(bool uc) 
    { HYPER_ALL_RULES::SetUnitConflict(uc); };
   static inline void EmptyRHSOnly() { HYPER_ACTIVE_RULE::EmptyRHSOnly(); };
   inline void Orient() 
    { HYPER_RULE::PositiveHyper(!(TERM::PositiveIsGreater())); };
   inline bool Load(Clause* fact) // "= false" means that no rule applicable
    {
     // reset everything here
     the_current_fact = fact;
     HYPER_ACTIVE_RULE::SetCurrentFact(fact);
     return ResetSearch( (!(UnitConflict())) || (!(fact->Unit()))); 
    }; 
  
   void ProcessRule();    

   void CollectAncestors(); 
    
   inline bool FindNextRule()
    {
     if (HYPER_ALL_RULES::FindNextRule())
      {
       return true;
      }
     else
      {
       HYPER_ACTIVE_RULE::UnloadCurrentFact();
       return false;
      };
    }; // bool FindNextRule()
   inline Clause* UnitConflictTest(Clause* const c)
    {
     if ((c) 
         && (UnitConflict()) 
         && (c->Unit()))
       { return CheckForUnitConflict(c); }
     else { return (Clause*)0; };
    };

    
   inline void OutputState()
    {
     ::Output("---------- Hyperresolution state: -----------\n");
     if (HYPER_RULE::IsPositiveHyper())
      {
       ::Output("Positive hyperresolution,\n");
      }
     else
      {
       ::Output("Negative hyperresolution,\n");
      };
     //OutputRules();
     
     ::Output("Current rule = "); 
     //::Output(CurrentRule()->TheClause());
     CurrentRule()->Output();
     Output("\n");
     
     ::Output("Current fact = ");
     ::Output(HYPER_ACTIVE_RULE::FactClause());
     ::Output("\n");
     
     ::Output("Current substitution:\n");
     GlobalSubst.Output(OutputPrefix);
     Output("\n");
     
     //cascade.Output();
     ::Output("----------------------------------------------\n");
    };
 }; // class HYPERRESOLUTION 


//************************************************************************* 
 
 
 
 
 
 
 
 
 
 
