#define __IU_COMMAND_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif 
#ifndef __PREFIX_H
 #include "prefix.h"
#endif
#ifndef __VARIABLES_H
 #include "variables.h"
#endif
//#ifndef __OR_INDEX_H
// #include "or_index.h"
//#endif


//****************************************************************************

class IU_COMMAND
 {
  public: 
   enum TAG
    {
     START,
     VAR,
     VAR_FIRST_OCC,
     FUNC,
     FUNC_PLAIN,
     FUNC_GROUND,
     CONST,
     END
    };
  private: // structure of a command
   TAG tag;
   PREF_SYM sym;
   PREFIX_SYM* term;
   IU_COMMAND* nextCommand;
  public:
   IU_COMMAND() {};
   inline TAG& Tag() { return tag; };
   inline PREF_SYM& Symbol() { return sym; };
   inline PREFIX_SYM*& Term() { return term; };
   inline IU_COMMAND*& NextCommand() { return nextCommand; };
   inline IU_COMMAND*& AllVars() 
    {
     CHECK2(Tag() == START,"in IU_COMMAND::AllVars()");
     return nextCommand; 
    };
   inline IU_COMMAND*& NextVar() 
    {
     CHECK2(Tag() == VAR_FIRST_OCC,"in IU_COMMAND::NextVar()");
     return nextCommand; 
    };
   inline void InversePolarity() // applicable to a START command only
    {
     CHECK2(Tag() == START,"in IU_COMMAND::InversePolarity()");
     sym.InversePolarity();
    };
  public:
  public: // for debugging
   static inline void OutputTag(TAG t);
   inline void Output(IU_COMMAND* start);
   inline void OutputCode();
 }; // class IU_COMMAND

//************* Definitions for IU_COMMAND: **********************

inline void IU_COMMAND::OutputTag(TAG t)
 {
  switch (t)
   {
    case START : ::Output("START"); break;
    case VAR : ::Output("VAR"); break;
    case VAR_FIRST_OCC : ::Output("VAR_FIRST_OCC"); break;
    case FUNC : ::Output("FUNC"); break;
    case FUNC_PLAIN : ::Output("FUNC_PLAIN"); break;
    case FUNC_GROUND : ::Output("FUNC_GROUND"); break;
    case CONST : ::Output("CONST"); break;
    case END : ::Output("END"); break;
   };
 }; // void IU_COMMAND::OutputTag(TAG t)
inline void IU_COMMAND::Output(IU_COMMAND* start)
 {
  ::Output("[");
  ::Output(this - start);
  ::Output("] ");
  OutputTag(Tag());
  if (Tag() == END) return;
  ::Output(" ");
  Symbol().Output();
  if ((Tag() == FUNC) || (Tag() == FUNC_GROUND) || (Tag() == FUNC_PLAIN))
   {
    ::Output(" next = [");
    ::Output(NextCommand() - start);
    ::Output("] ");
    PREFIX::OutputTerm(Term());
   };
  if (Tag() == START) 
   {
    ::Output(" variables = [");
    ::Output(AllVars() ? (AllVars() - start) : 0);
    ::Output("] ");
    PREFIX::OutputTerm(Term());
   }; 
  if (Tag() == VAR_FIRST_OCC) 
   {
    ::Output(" next variable = [");
    ::Output(NextVar() ? (NextVar() - start) : 0);
    ::Output("]");
   };
 }; // void IU_COMMAND::Output(IU_COMMAND* start)

inline void IU_COMMAND::OutputCode()
 { 
  IU_COMMAND* com = this;
  do
   {
    com->Output(this);
    ::Output("\n");
    com++;
   }
  while (com->Tag() != END);
  com->Output(this);
  ::Output("\n");
 }; // void IU_COMMAND::OutputCode() 


//****************************************************************************