#define __IU_MACHINE_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __IU_COMMAND_H
 #include "iu_command.h"
#endif 
#ifndef __IU_TRANSLATOR_H
 #include "iu_translator.h"
#endif 
#ifndef __UNIFICATION_H
 #include "unification.h"
#endif
#ifndef __PREFIX_H
 #include "prefix.h"
#endif
#ifndef __VARIABLES_H
 #include "variables.h"
#endif
#ifndef __OR_INDEX_H
 #include "or_index.h"
#endif

//************************************************************************

class IU_MACH
 {
  private:
   enum BACKTRACK_ACTION
    {
     FAILURE,
     ALTER_BINDING_WITH_OC,
     ALTER_BINDING_WITH_OC_SAME_VAR,
     ALTER_BINDING,
     ALTER_UNIFIABLE,
     TRY_ALTERNATIVE_NODE,
     UNBIND_LAST_TRY_ALTERNATIVE_NODE,
     UNBIND_SECTION_TRY_ALTERNATIVE_NODE
    };
  private:
   IU_TRANSLATOR compiler;
   OR_PROCESSOR proc;
   IU_COMMAND code[MAX_TMP_TERM_WEIGHT];
   IU_COMMAND* end_of_code;
   PREFIX_SYM data[MAX_TMP_TERM_WEIGHT];
   IU_COMMAND* backtrack_point[MAX_TMP_TERM_WEIGHT];
   BACKTRACK_ACTION backtrack_action[MAX_TMP_TERM_WEIGHT];
   IU_COMMAND** next_backtrack_point;
   BACKTRACK_ACTION* next_backtrack_action;
   IU_COMMAND* command;
  private: // "sugar"
   VAR*** subst_state;
   RES_PAIR_LIST* candidates;
  public:
   IU_MACH() : compiler(), proc() 
    {
     ResetStatic();
     next_backtrack_point = backtrack_point;
    };
   static inline void ResetStatic() { OR_PROCESSOR::ResetStatic(); };
   inline void SetIndex(const int index) { proc.SetIndex(index); };
   inline int GetIndex() const { return proc.GetIndex(); };
   inline IU_COMMAND* Code() { return code; };
   inline bool CompileLiteral(PREFIX_SYM* lit)
    {
     end_of_code = code;
     PREFIX_SYM* dmem = data;
     bool res = compiler.CompileLiteral(lit,end_of_code,dmem,MAX_TMP_TERM_WEIGHT);
     
     ResetBacktracking();
     return res; 
     //candidates = (RES_PAIR_LIST*)0;
    };
   
   inline void CompilePseudoLiteral(TERM header,PREFIX_SYM* arg)
    {
     end_of_code = code;
     PREFIX_SYM* dmem = data;
     compiler.CompilePseudoLiteral(header,arg,end_of_code,dmem);
     ResetBacktracking(); 
     //candidates = (RES_PAIR_LIST*)0;
    };
    
   inline void InversePolarity()
    {
     code->InversePolarity();
    };
  private:
   inline void ResetBacktracking()
    {
     next_backtrack_point = backtrack_point;
     next_backtrack_action = backtrack_action;
    };
   inline void PushBacktrackPoint(IU_COMMAND* const com)
    {
     *next_backtrack_point = com;
     next_backtrack_point++;
     CHECK13(next_backtrack_point - backtrack_point <= MAX_TMP_TERM_WEIGHT,
            "in IU_MACH::PushBacktrackPoint()");
    };
   inline IU_COMMAND* PopBacktrackPoint() 
    {
     next_backtrack_point--;
     CHECK13(next_backtrack_point >= backtrack_point,"in IU_MACH::PopBacktrackPoint()");
     return *next_backtrack_point;
    }; 
   inline void PushBacktrackAction(const BACKTRACK_ACTION act)
    {
     *next_backtrack_action = act;
     next_backtrack_action++;
    };
   inline BACKTRACK_ACTION PopBacktrackAction()
    {
     CHECK13(next_backtrack_action > backtrack_action,"A1 in IU_MACH::PopBacktrackAction()");
     next_backtrack_action--;
     return *next_backtrack_action;
    };
  private:
   inline bool Backtrack();
   inline bool CompleteSearch();
  public:
   //inline bool FindFirstLeaf();
   inline bool FindFirstLeaf(OR_INDEX_TREE_NODE* tree);
   inline bool FindNextLeaf();
   
  public: // "sugar"
  
   inline bool VariablesFromQueryAffected() 
    {
     IU_COMMAND* vars = code->AllVars();
     while (vars)
      {
       if (vars->Symbol().Var()->Binding()) { return true; };
       vars = vars->NextVar();
      };
     return false;
    };
    
 
   inline bool VariablesFromIndexAffected()
    {
     return GlobalSubst.VariablesAffected(GetIndex());
    };

   inline void MemorizeState() { subst_state = GlobalSubst.MarkState(); };
   inline void RestoreState() { GlobalSubst.UnbindDownto(subst_state); };
   inline OR_INDEX_TREE_NODE* CurrentLeaf() const { return proc.CurrNode(); };
   inline void ResetCandidatesFromLeaf() 
    {
     candidates = CurrentLeaf()->Clauses();
    };
   inline RES_PAIR_LIST* NextCandidate() 
    {
     RES_PAIR_LIST* res = candidates;
     if (candidates) { candidates = candidates->Next(); };
     return res;
    };
  public: // methods for debugging
   inline void Output()
    {
     ::Output("CODE:\n");
     OutputCode();
     ::Output("INDEX = "); ::Output(GetIndex()); 
     ::Output("\n");
    };
   inline void OutputCode()
    { 
     IU_COMMAND* com = code;
     do
      {
       com->Output(code);
       ::Output("\n");
       com++;
      }
     while (com->Tag() != IU_COMMAND::END);
     com->Output(code);
     ::Output("\n");
    }; 
   static inline void OutputBacktrackAction(BACKTRACK_ACTION act)
    {
     switch (act)
      {
       case FAILURE : ::Output("FAILURE"); break;
       case ALTER_BINDING_WITH_OC : ::Output("ALTER_BINDING_WITH_OC"); break;
       case ALTER_BINDING_WITH_OC_SAME_VAR : ::Output("ALTER_BINDING_WITH_OC_SAME_VAR"); break;
       case ALTER_BINDING : ::Output("ALTER_BINDING"); break;
       case ALTER_UNIFIABLE : ::Output("ALTER_UNIFIABLE"); break;
       case TRY_ALTERNATIVE_NODE : ::Output("TRY_ALTERNATIVE_NODE"); break;
       case UNBIND_LAST_TRY_ALTERNATIVE_NODE : ::Output("UNBIND_LAST_TRY_ALTERNATIVE_NODE"); break;
       case UNBIND_SECTION_TRY_ALTERNATIVE_NODE : ::Output("UNBIND_SECTION_TRY_ALTERNATIVE_NODE"); break;
      };
    };
 }; // class IU_MACH


//****************** Definitions: *************************************

inline bool IU_MACH::CompleteSearch()
 {
  DCONTROL13("Z1 in IU_MACH::CompleteSearch()");
 
  IU_COMMAND* com = command;
  bool same_var;
  bool free_var;
  void* bind;
  
  DCONTROL13("Y1 in IU_MACH::CompleteSearch()");
  
  check_tag:
  
   CHECK13((com > code) && (com < end_of_code),"B1 in IU_MACH::CompleteSearch()");
   CHECK13((com->Tag() == IU_COMMAND::END) || (!(proc.CurrNode()->IsLeaf())),
          "B2 in IU_MACH::CompleteSearch()");
  
   
   switch (com->Tag())
    {
     #ifdef _ALL_ENUM_VALUES_IN_SWITCHES_
     case IU_COMMAND::START : 
      RuntimeError("R0 in IU_MACH::CompleteSearch()");
      return false;
     #endif 
     case IU_COMMAND::VAR :
      bind = com->Symbol().Var()->Unref(free_var);
      if (free_var)
       {
        if (proc.BindWithOccCheck((VAR*)bind,same_var))
         {
          PushBacktrackPoint(com);
          if (same_var) { PushBacktrackAction(ALTER_BINDING_WITH_OC_SAME_VAR); }
          else { PushBacktrackAction(ALTER_BINDING_WITH_OC); };
          com++;
          
          DCONTROL13("Y2 in IU_MACH::CompleteSearch()");
          goto check_tag;  
         }
        else { goto backtrack; };
       }
      else
       {
       
        CHECK13(PREFIX::CorrectComplexTerm((PREFIX_SYM*)bind),
               "C1 in IU_MACH::CompleteSearch()");
       
        if (proc.ExtractUnifiable((PREFIX_SYM*)bind))
         {
          PushBacktrackPoint(com);
          PushBacktrackAction(ALTER_UNIFIABLE);
          com++;
          
          DCONTROL13("Y3 in IU_MACH::CompleteSearch()");
          goto check_tag;
         }
        else 
         {
          goto backtrack; 
         };
       };
      
     case IU_COMMAND::VAR_FIRST_OCC :
      proc.Bind(com->Symbol().Var());
      PushBacktrackPoint(com);
      PushBacktrackAction(ALTER_BINDING);
      com++;
      
      DCONTROL13("Y4 in IU_MACH::CompleteSearch()");
      goto check_tag;  
   
     case IU_COMMAND::FUNC :
      func_check_node:
       switch (proc.CurrNode()->Tag())
        {
         case OR_INDEX_TREE_NODE::FUNC : 
          if (com->Symbol() == proc.CurrNode()->Symbol())
           {
            PushBacktrackPoint(com);
            PushBacktrackAction(TRY_ALTERNATIVE_NODE);
            com++;
            proc.Below();
            
            DCONTROL13("Y5 in IU_MACH::CompleteSearch()"); 
            goto check_tag;
           }
          else
           {
            func_try_alternative:
             proc.Alternative();
             if (proc.CurrNode()) { goto func_check_node;}
             else { goto backtrack; };
           };
         case OR_INDEX_TREE_NODE::CONST : goto func_try_alternative;
         case OR_INDEX_TREE_NODE::VAR :
          bind = (proc.Shift(proc.CurrNode()->Symbol().Var()))->Unref(free_var);
          if (free_var)
           {
            if (OCCUR_CHECK::Occurs((VAR*)bind,com->Term()))
             {
              goto func_try_alternative;
             }
            else // assignment possible
             {
              GlobalSubst.BindAlone((VAR*)bind,com->Term());
              PushBacktrackPoint(com);
              PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);
              com = com->NextCommand();
              proc.Below();
              
              DCONTROL13("Y6 in IU_MACH::CompleteSearch()");
              goto check_tag;
             };
           }
          else // complex term, unification needed
           {
            if (UNIFICATION::UnifyComplex((PREFIX_SYM*)bind,com->Term()))
             {
              PushBacktrackPoint(com);
              PushBacktrackAction(UNBIND_SECTION_TRY_ALTERNATIVE_NODE);
              com = com->NextCommand();
              proc.Below();
              
              DCONTROL13("Y7 in IU_MACH::CompleteSearch()");
              goto check_tag;
             }
            else { goto func_try_alternative; };
           };
         case OR_INDEX_TREE_NODE::VAR_FIRST_OCC :
          GlobalSubst.BindAlone(proc.Shift(proc.CurrNode()->Symbol().Var()),com->Term());
          PushBacktrackPoint(com);
          PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);
          com = com->NextCommand();
          proc.Below();
          
          DCONTROL13("Y8 in IU_MACH::CompleteSearch()");
          goto check_tag;
         #ifdef _ALL_ENUM_VALUES_IN_SWITCHES_
         case OR_INDEX_TREE_NODE::LEAF :
          RuntimeError("R1 in  IU_MACH::CompleteSearch()");
          return false;
         #endif
          
         #ifdef _DEBUG13_
         default : 
          DebugError("A1 in  IU_MACH::CompleteSearch()");
          return false;    
         #endif
        };
     
     case IU_COMMAND::FUNC_PLAIN :
      func_plain_check_node:
       switch (proc.CurrNode()->Tag())
        {
         case OR_INDEX_TREE_NODE::FUNC : 
          if (com->Symbol() == proc.CurrNode()->Symbol())
           {
            PushBacktrackPoint(com);
            PushBacktrackAction(TRY_ALTERNATIVE_NODE);
            com++;
            proc.Below();
            
            DCONTROL13("Y9 in IU_MACH::CompleteSearch()");
            goto check_tag;
           }
          else
           {
            func_plain_try_alternative:
             proc.Alternative();
             if (proc.CurrNode()) { goto func_plain_check_node;}
             else { goto backtrack; };
           };
         case OR_INDEX_TREE_NODE::CONST : goto func_plain_try_alternative;
         case OR_INDEX_TREE_NODE::VAR :
          bind = (proc.Shift(proc.CurrNode()->Symbol().Var()))->Unref(free_var);
          if (free_var)
           {
            GlobalSubst.BindAlone((VAR*)bind,com->Term());
            PushBacktrackPoint(com);
            PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);
            com = com->NextCommand();
            proc.Below();
            
            DCONTROL13("Y10 in IU_MACH::CompleteSearch()");
            goto check_tag;
           }
          else // complex term
           {
            if (UNIFICATION::UnifyPlain((PREFIX_SYM*)bind,com->Term()))
             {
              PushBacktrackPoint(com);
              PushBacktrackAction(UNBIND_SECTION_TRY_ALTERNATIVE_NODE);
              com = com->NextCommand();
              proc.Below();
              
              DCONTROL13("Y11 in IU_MACH::CompleteSearch()");
              goto check_tag;
             }
            else { goto func_plain_try_alternative; }; 
           };
         case OR_INDEX_TREE_NODE::VAR_FIRST_OCC : 
          GlobalSubst.BindAlone(proc.Shift(proc.CurrNode()->Symbol().Var()),com->Term());
          PushBacktrackPoint(com);
          PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);
          com = com->NextCommand();
          proc.Below();
          
          DCONTROL13("Y12 in IU_MACH::CompleteSearch()");
          goto check_tag;
         #ifdef _ALL_ENUM_VALUES_IN_SWITCHES_
         case OR_INDEX_TREE_NODE::LEAF :
          RuntimeError("S1 in  IU_MACH::CompleteSearch()");
          return false;
         #endif
         #ifdef _DEBUG13_
         default : 
          DebugError("L1 in  IU_MACH::CompleteSearch()");
          return false;    
         #endif
        }; 
      
      /************/
      
     case IU_COMMAND::FUNC_GROUND :
      func_ground_check_node:
       switch (proc.CurrNode()->Tag())
        {
         case OR_INDEX_TREE_NODE::FUNC : 
          if (com->Symbol() == proc.CurrNode()->Symbol())
           {
            PushBacktrackPoint(com);
            PushBacktrackAction(TRY_ALTERNATIVE_NODE);
            com++;
            proc.Below();
            
            DCONTROL13("Y13 in IU_MACH::CompleteSearch()");
            goto check_tag;
           }
          else
           {
            func_ground_try_alternative:
             proc.Alternative();
             if (proc.CurrNode()) { goto func_ground_check_node;}
             else { goto backtrack; };
           };
         case OR_INDEX_TREE_NODE::CONST : goto func_ground_try_alternative;
         case OR_INDEX_TREE_NODE::VAR :
          bind = (proc.Shift(proc.CurrNode()->Symbol().Var()))->Unref(free_var);
          if (free_var)
           {
            GlobalSubst.BindAlone((VAR*)bind,com->Term());
            PushBacktrackPoint(com);
            PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);
            com = com->NextCommand();
            proc.Below();
            
            DCONTROL13("Y14 in IU_MACH::CompleteSearch()");
            goto check_tag;
           }
          else // complex term
           {
            //if (UNIFICATION::Unify((PREFIX_SYM*)bind,com->Term()))
            if (UNIFICATION::MatchGroundPlain((PREFIX_SYM*)bind,com->Term()))
             {
              PushBacktrackPoint(com);
              PushBacktrackAction(UNBIND_SECTION_TRY_ALTERNATIVE_NODE);
              com = com->NextCommand();
              proc.Below();
              
              DCONTROL13("Y15 in IU_MACH::CompleteSearch()");
              goto check_tag;
             }
            else { goto func_ground_try_alternative; }; 
           };
         case OR_INDEX_TREE_NODE::VAR_FIRST_OCC : 
          GlobalSubst.BindAlone(proc.Shift(proc.CurrNode()->Symbol().Var()),com->Term());
          PushBacktrackPoint(com);
          PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);
          com = com->NextCommand();
          proc.Below();
          
          DCONTROL13("Y16 in IU_MACH::CompleteSearch()");
          goto check_tag;
         #ifdef _ALL_ENUM_VALUES_IN_SWITCHES_
         case OR_INDEX_TREE_NODE::LEAF :
          RuntimeError("R2 in  IU_MACH::CompleteSearch()");
          return false;
         #endif
         #ifdef _DEBUG13_
         default : 
          DebugError("A2 in  IU_MACH::CompleteSearch()");
          return false;    
         #endif
        };
     
     case IU_COMMAND::CONST :
      const_check_node:
       switch (proc.CurrNode()->Tag())
        {
         case OR_INDEX_TREE_NODE::FUNC :
          const_try_alternative:
           proc.Alternative();
           if (proc.CurrNode()) { goto const_check_node;}
           else { goto backtrack; };
         case OR_INDEX_TREE_NODE::CONST : 
          if (com->Symbol() == proc.CurrNode()->Symbol())
           {
            PushBacktrackPoint(com);
            PushBacktrackAction(TRY_ALTERNATIVE_NODE);
            com++;
            proc.Below();
            
            DCONTROL13("Y17 in IU_MACH::CompleteSearch()");
            goto check_tag;
           }
          else { goto const_try_alternative; };
         case OR_INDEX_TREE_NODE::VAR : 
          bind = (proc.Shift(proc.CurrNode()->Symbol().Var()))->Unref(free_var);
          if (free_var)
           {
            GlobalSubst.BindAlone((VAR*)bind,com->Term());
            PushBacktrackPoint(com);
            PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);
            com++;
            proc.Below();
            
            DCONTROL13("Y18 in IU_MACH::CompleteSearch()");
            goto check_tag;
           }
          else // complex term
           {
            if (com->Symbol() == ((PREFIX_SYM*)bind)->Head())
             {
              PushBacktrackPoint(com);
              PushBacktrackAction(TRY_ALTERNATIVE_NODE);
              com++;
              proc.Below();
              
              DCONTROL13("Y19 in IU_MACH::CompleteSearch()");
              goto check_tag;
             }
            else { goto const_try_alternative; };
           };
         case OR_INDEX_TREE_NODE::VAR_FIRST_OCC : 
          GlobalSubst.BindAlone(proc.Shift(proc.CurrNode()->Symbol().Var()),com->Term());
          PushBacktrackPoint(com);
          PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);
          com++;
          proc.Below();
          
          DCONTROL13("Y20 in IU_MACH::CompleteSearch()");
          goto check_tag;
         #ifdef _ALL_ENUM_VALUES_IN_SWITCHES_
         case OR_INDEX_TREE_NODE::LEAF :
          RuntimeError("R3 in  IU_MACH::CompleteSearch()");
          return false;
         #endif
         #ifdef _DEBUG13_
         default : 
          DebugError("A3 in  IU_MACH::CompleteSearch()");
          return false;    
         #endif
        };
     case IU_COMMAND::END :
      CHECK13(proc.CurrNode()->IsLeaf(),"C1 in IU_MACH::CompleteSearch()");
      
      return true;
     #ifdef _DEBUG13_
     default :
      DebugError("A3 in IU_MACH::CompleteSearch()");
      return false;   
     #endif
    };
  backtrack:
   command = com;
   if (Backtrack()) { com = command; goto check_tag; }
   else { return false; };
 }; // bool IU_MACH::CompleteSearch() 


inline bool IU_MACH::FindFirstLeaf(OR_INDEX_TREE_NODE* tree)
 {
  command = code+1;
  if (tree)
   {
    PushBacktrackPoint(code);
    PushBacktrackAction(FAILURE);
    proc.Start(tree);
    return CompleteSearch();
   }
  else 
   {
    return false; 
   };
 }; // bool IU_MACH::FindFirstLeaf()

inline bool IU_MACH::FindNextLeaf()
 {
  return ((Backtrack()) && (CompleteSearch()));
 }; // inline bool IU_MACH::FindNextLeaf()


inline bool IU_MACH::Backtrack()
 {
  DCONTROL13("Z1 in IU_MACH::Backtrack()");
 
  IU_COMMAND* com;
  bool same_var;
  VAR* var;
  
  DCONTROL13("Y1 in IU_MACH::Backtrack()");
  
  rollback:
   com = PopBacktrackPoint();
   
   DCONTROL13("C1 in IU_MACH::Backtrack()");
   
   BACKTRACK_ACTION act = PopBacktrackAction(); // debug
   
   DCONTROL13("C2 in IU_MACH::Backtrack()");
   
  // switch (PopBacktrackAction())
   switch (act) // debug
    {
     case FAILURE : return false;
     case ALTER_BINDING_WITH_OC : 
      DCONTROL13("D1 in IU_MACH::Backtrack()"); 
     
      CHECK13(com->Symbol().Var()->Binding(),"B2 in IU_MACH::Backtrack()");
      if (proc.AlterBindingWithOccCheck(GlobalSubst.LastBound(),same_var))
       {
        PushBacktrackPoint(com);
        if (same_var)
         {
          PushBacktrackAction(ALTER_BINDING_WITH_OC_SAME_VAR);
         }
        else
         {
          CHECK13(com->Symbol().Var()->Binding(),"B4 in IU_MACH::Backtrack()");
          PushBacktrackAction(ALTER_BINDING_WITH_OC);
         };
        command = com + 1;
        return true;
       }
      else 
       {
        DCONTROL13("Y2 in IU_MACH::Backtrack()");
        goto rollback; 
       };
     case ALTER_BINDING_WITH_OC_SAME_VAR :
      DCONTROL13("D2 in IU_MACH::Backtrack()"); 
      
      var = (VAR*)(com->Symbol().Var()->Unref());
      CHECK13(VAR::IsVar((void*)var),"B6 in IU_MACH::Backtrack()");
      if (proc.AlterBindingWithOccCheck(var,same_var))
       {
        PushBacktrackPoint(com);
        if (same_var)
         {
          PushBacktrackAction(ALTER_BINDING_WITH_OC_SAME_VAR);
         }
        else
         {
          PushBacktrackAction(ALTER_BINDING_WITH_OC);
         };
        command = com + 1;
        return true;
       }
      else 
       {
        DCONTROL13("Y3 in IU_MACH::Backtrack()");
        goto rollback; 
       };
      
     case ALTER_BINDING : 
      DCONTROL13("D3 in IU_MACH::Backtrack()"); 
     
      CHECK13(com->Tag() == IU_COMMAND::VAR_FIRST_OCC,"C1 in IU_MACH::Backtrack()");
      CHECK13(com->Symbol().Var()->Binding(),"C2 in IU_MACH::Backtrack()");
      
      if (proc.AlterBinding())
       {
        CHECK13(com->Symbol().Var()->Binding(),"C3 in IU_MACH::Backtrack()");
        PushBacktrackPoint(com);
        PushBacktrackAction(ALTER_BINDING);
        command = com + 1;
        return true;
       }
      else 
       {
        DCONTROL13("Y4 in IU_MACH::Backtrack()");
        goto rollback; 
       };
      
     case ALTER_UNIFIABLE :
      DCONTROL13("D4 in IU_MACH::Backtrack()"); 
     
      if (proc.AlterUnifiable())
       {
        DCONTROL13("D4.1 in IU_MACH::Backtrack()"); 
        
        PushBacktrackPoint(com);
        PushBacktrackAction(ALTER_UNIFIABLE);
        command = com + 1;
        return true;
       }
      else 
       {
        DCONTROL13("Y5 in IU_MACH::Backtrack()");
        goto rollback; 
       };
      
     case TRY_ALTERNATIVE_NODE :
      DCONTROL13("D5 in IU_MACH::Backtrack()"); 
      
      proc.RestoreCurrent();
      proc.Alternative();
      if (proc.CurrNode()) { command = com; return true; }
      else 
       {
        DCONTROL13("Y6 in IU_MACH::Backtrack()");
        goto rollback; 
       };
      
     case UNBIND_LAST_TRY_ALTERNATIVE_NODE :
      DCONTROL13("D6 in IU_MACH::Backtrack()"); 
     
      GlobalSubst.UnbindLast();
      proc.RestoreCurrent();
      proc.Alternative();
      if (proc.CurrNode()) { command = com; return true; }
      else 
       {
        DCONTROL13("Y7 in IU_MACH::Backtrack()");
        goto rollback; 
       };
      
     case UNBIND_SECTION_TRY_ALTERNATIVE_NODE :
      DCONTROL13("D7 in IU_MACH::Backtrack()"); 
     
      GlobalSubst.UnbindLastSection();
      proc.RestoreCurrent();
      proc.Alternative();
      if (proc.CurrNode()) { command = com; return true; }
      else 
       {
        DCONTROL13("Y8 in IU_MACH::Backtrack()");
        goto rollback; 
       };
    };   
  RuntimeError("R1 in IU_MACH::Backtrack()");
  return false;
 };

//****************************************************************************
