#define __KEPT_H
#ifndef __CLAUSE_H
 #include "clause.h"
#endif
#ifndef __OUTPUT_H
 #include "output.h"
#endif
#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __CLAUSE_SET_H
 #include "clause_set.h"
#endif
//*************************************************************************

const int NUMBER_OF_KEPT_CATEGORIES = MAX_CLAUSE_WEIGHT+1;

class KEPT
 {
  private:
   bool max_priority_flag;
   int pick_given_ratio; // <=0 if no weighting is used (default)
   float magic_coefficient; // must be = 1 + (1/pick_given_ratio);
   int picked_by_weight;
   Clause category[NUMBER_OF_KEPT_CATEGORIES];
   int minCat;
   int maxCat;
   int max_accepted_cat;
  private: // statistics
   long currently_kept[NUMBER_OF_KEPT_CATEGORIES];
   long used[NUMBER_OF_KEPT_CATEGORIES];
   long removed[NUMBER_OF_KEPT_CATEGORIES];
   long num_of_currently_kept;
   long num_of_used;
   long num_of_removed;
  private:
   long numOfReachable;

  private: // iterator
   Clause* next_clause;
   int iter_cat; 
  public:
   inline void ResetIterator()
    {
     //while ((EmptyCat(minCat)) && (minCat < maxCat)) { minCat++; };
     if (EmptyCat(minCat)) { next_clause = (Clause*)0; return; };
     iter_cat = minCat;
     next_clause = category[iter_cat].Next();
    };
   inline Clause* IterNext() 
    {
     Clause* res = next_clause;
     
     if (res)
      {
       next_clause = res->Next();
       if (next_clause == &(category[iter_cat]))
        {
         // try another category
         iter_cat++;
         while ((iter_cat <= maxCat) && (EmptyCat(iter_cat))) { iter_cat++; };
         if (iter_cat > maxCat) { next_clause = (Clause*)0; }
         else { next_clause = category[iter_cat].Next(); };
        };
      };
     return res;
    };
   
  public:
   inline void Reset()
    {
     max_priority_flag = false;
     num_of_currently_kept = 0L;
     num_of_used = 0L;
     num_of_removed = 0L;

     numOfReachable = MAXLONG;
     for (int cat = 0; cat < NUMBER_OF_KEPT_CATEGORIES; cat++)
      {
       StatReset(cat);
      };
     minCat = 0;
     maxCat = 0;
     max_accepted_cat = NUMBER_OF_KEPT_CATEGORIES-1;
     picked_by_weight = 0;
    };
   KEPT() : pick_given_ratio(0), magic_coefficient(0.0)
    {
     Reset();
    };
  
  private: // statistics
  
   inline void NormalizeMinMaxCat(int lb,int rb)
    {
     minCat = lb;
     maxCat = rb;
     while ((EmptyCat(minCat)) && (minCat < maxCat))
      {
       minCat++;
      };
     while ((EmptyCat(maxCat)) && (minCat < maxCat))
      {
       maxCat--;
      }; 
    };
  
   inline void StatReset(const int cat) 
    {
     currently_kept[cat] = 0L;
     used[cat] = 0L;
     removed[cat] = 0L;
    };
  
   inline void StatInserted(const int cat)
    {
     currently_kept[cat]++;
     num_of_currently_kept++;    
     if (cat > maxCat) 
      {
       NormalizeMinMaxCat(minCat,cat);
      }
     else
      {
       if (cat < minCat) 
        {
         NormalizeMinMaxCat(cat,maxCat);
        };
      };
     //cout << "INSERTED " << cat << "\n"; CheckStatistics(); // debug
     //OutputMinMaxCat(); // debug
    };
   
   inline void StatUsed(const int cat)
    {
     used[cat]++;
     currently_kept[cat]--;
     num_of_used++;     
     num_of_currently_kept--;
     numOfReachable--; // experimental
     if (EmptyCat(cat))
      {
       if ((cat == minCat) || (cat == maxCat))
        {
         NormalizeMinMaxCat(minCat,maxCat);
        };
      };
     
     //cout << "USED " << cat << "\n"; CheckStatistics(); // debug
     //OutputMinMaxCat(); // debug
    };
   
   inline void StatRemoved(const int cat)
    {
     //cout << "     before " << currently_kept[cat] << "\n";
     
     removed[cat]++;
     currently_kept[cat]--;
     num_of_removed++;
     num_of_currently_kept--;
     numOfReachable--; // experimental
     if (EmptyCat(cat))
      {
       if ((cat == minCat) || (cat == maxCat))
        {
         NormalizeMinMaxCat(minCat,maxCat);
        };
      };
     
    

    };

   inline long Total(const int cat) const
    {
     return currently_kept[cat] + used[cat] + removed[cat];
    };
  
  public:
   inline long Total() const 
    {
     return num_of_currently_kept + num_of_used + num_of_removed;
    };
 
   inline long UsedAndRemoved() const 
    {
     return (num_of_used + num_of_removed);
    };
  
  private:
   inline long NumOfLighter(const int cat)
    {
     long res = 0L;
     for (int c = 0; c < cat; c++)
      {
       res += currently_kept[c]; 
      };
     return res;
    };
   
   inline long NumOfHeavier(const int cat)
    {
     long res = 0L;
     for (int c = cat + 1; c <= maxCat; c++)
      {
       res += currently_kept[c];
      };
     return res;
    };
 
  private: // reachability
 
   inline bool Reachable(const int cat)
    {
     
     // here we assume that num_of_currently_kept > numOfReachable
     long num_of_lighter_or_same_weight = NumOfLighter(cat) + currently_kept[cat];
     
     if (!num_of_lighter_or_same_weight) return true; // experimental
     
     if (numOfReachable <= num_of_lighter_or_same_weight) return false;
     long magic_number = (long)(num_of_lighter_or_same_weight * magic_coefficient);
     
     return (numOfReachable - (num_of_currently_kept - numOfReachable) > magic_number);
    };
 
   inline void CorrectMaxAcceptedCat()
    {
     if (num_of_currently_kept < numOfReachable) 
      {
       max_accepted_cat = NUMBER_OF_KEPT_CATEGORIES-1;
       return;
      };
     if (Reachable(max_accepted_cat))
      {
       while (Reachable(max_accepted_cat + 1))
        {
         max_accepted_cat++;
         if (max_accepted_cat > maxCat) 
          {
           max_accepted_cat = NUMBER_OF_KEPT_CATEGORIES-1;
           return;
          };
        };
      }
     else
      {
       do
        {
         max_accepted_cat--;
        }
       while ((max_accepted_cat > 0) && (!(Reachable(max_accepted_cat))));
      };     
    };

  public: // reachability
   inline void SetNumOfReachable(const long n) 
    {
     CHECK3(CheckStatistics(),"corrupt statistics in KEPT");
    
     numOfReachable = n; 
     CorrectMaxAcceptedCat();
    };
   inline long MaxAcceptedWeight() const
    {
     if (num_of_currently_kept < numOfReachable) { return MAX_CLAUSE_WEIGHT; };
     if (pick_given_ratio > 0)
      {
       return Weight(max_accepted_cat);
      }
     else
      {
       return 0L;
      };
    };
   
  private:
   inline int Category(int weight)
    {
     if (pick_given_ratio > 0)
      {
       return weight;
      }
     else
      {
       return 0;
      };
    };
    
   inline long Weight(const int cat) const 
    {
     return cat; 
    };
   
   inline int Category(Clause* c) 
    {
     if (c->MaxPriority()) { return 0; }   
     else { return Category(c->Weight()); };
    };
   inline void InsertAs(Clause* c,int cat)
    {
     Clause::InsertBefore(c,&(category[cat]));
     StatInserted(cat);
    };
   inline bool EmptyCat(int cat) 
    { return (category[cat].Next() == &(category[cat])); };

  public: 
   inline bool MaxPriorityEmpty() 
    {
     return EmptyCat(0);
    }; 
    
  public:
   inline void CollectUnusable(CLAUSE_SET& unusable,long max_res_weight,long max_num_of_unusable)   
    {
     CHECK3(max_res_weight > 0,"A0 in KEPT::CollectUnusable(..)");
     if (max_num_of_unusable <= 0) return;    
     int min_cand_cat = Category(max_res_weight) + 1;
     long collected = 0;
     for (int cat = maxCat; cat >= min_cand_cat; cat--)
      if (!EmptyCat(cat))
       {
        Clause* curr_clause = category[cat].Previous();
        Clause* next_clause = curr_clause->Previous(); 
        try_curr_clause:
         if (curr_clause->MinResWeight() > max_res_weight)
          {
           Remove(curr_clause);
           unusable.Insert(curr_clause);
           collected++;
           if (collected >= max_num_of_unusable) return; 
          };
         
         if (next_clause != &(category[cat]))
          {
           curr_clause = next_clause;
           next_clause = curr_clause->Previous();
           goto try_curr_clause;
          };
       };
    }; // void CollectUnusable(CLAUSE_SET& unusable,long max_res_weight,long max_num_of_unusable)   

   inline long UnreachableKept() const
    {
     return (num_of_currently_kept > numOfReachable)? 
             (num_of_currently_kept - numOfReachable) : 0; 
    };
   
   inline void DestroyAllClauses()
    {
     for (int cat = minCat; cat <= maxCat; cat++)
      {
       while (!(EmptyCat(cat)))
        {
         Clause* c = category[cat].Next();
         Clause::Remove(c);
         c->Destroy();
        };
      };
     Reset();  
    };
   inline void SetPickGivenRatio(const int pgr) 
    {
     pick_given_ratio = pgr; 
     if (pgr > 0)
      {
       magic_coefficient = 1.0 + ((1.0)/((float)pgr));
      };
    };
   
   inline void Insert(Clause* c) 
    {
     if (max_priority_flag) c->AssignMaxPriority(); 
     InsertAs(c,Category(c));
    };
   
   inline Clause* TheWorstClause()
    {
     for (int cat = maxCat; cat >= minCat; cat--)
      {
       if (!(EmptyCat(cat)))
        {
         return category[cat].Previous();
        };
      };
     return (Clause*)0;
    };
   
   inline void InsertInputClause(Clause* const c) { InsertAs(c,0); };

   inline void MaxPriorityOn() { max_priority_flag = true; };
   inline void MaxPriorityOff() { max_priority_flag = false; };   
   //inline void InsertWithMaximalPriority(Clause* const c) { InsertAs(c,1); };
   
   inline void Remove(Clause* c)
    {         
     Clause::Remove(c);
     StatRemoved(Category(c));
    };
   inline long Size() { return num_of_currently_kept; };
   inline Clause* Next() 
    {
     if (!(EmptyCat(0)))
      {
       Clause* res = category[0].Next();
       Clause::Remove(res);
       StatUsed(0); 
       return res;
      };
      
     if (pick_given_ratio > 0)
      {
       if (EmptyCat(minCat))
        {
         return (Clause*)0; 
        };
       if (picked_by_weight < pick_given_ratio)
        // pick by weight
        {
         picked_by_weight++;
         Clause* res = category[minCat].Next();
         Clause::Remove(res);
         StatUsed(minCat); 
         return res;
        }
       else // pick by history
        {
         long oldestNum = category[minCat].Next()->Number();
         int oldestCat = minCat;
         for (int cat = minCat + 1; cat<=maxCat; cat++)
          {
           if ((!(EmptyCat(cat))) && (category[cat].Next()->Number() < oldestNum))
            {
             oldestNum = category[cat].Next()->Number();
             oldestCat = cat;
            };
          };
         picked_by_weight = 0;
         Clause* res = category[oldestCat].Next();
         Clause::Remove(res);
         StatUsed(oldestCat); 
         return res;
        };
      }
     else
      {
       if (EmptyCat(0)) { return (Clause*)0; };
       Clause* res = category[0].Next();
       Clause::Remove(res);
       StatUsed(0);
       return res;
      };
    }; 
     
  public: // output statistics
   inline void OutputMinMaxCat()
    {
     Output("minCat = "); Output(minCat); Output("  ");
     Output("maxCat = "); Output(maxCat); Output("\n");
    };
   inline void OutputStatistics(int cat)
    {
     if (EmptyCat(cat))
      {
       Output(' ');
      }
     else
      {
       Output('*');
      };
     if (cat == 0) 
      {
       Output("(input)"); 
      }
     else
      {
       Output("(weight = "); Output(cat); Output(")");
      };
     Output(" kept("); Output(currently_kept[cat]);
     Output(") = total("); Output(Total(cat));
     Output(") - (used("); Output(used[cat]);
     Output(") + removed("); Output(removed[cat]);
     Output("))  ");
     Output("lighter = "); Output(NumOfLighter(cat)); 
    };
   inline void OutputStatistics(bool extended)
    {
     #ifdef _DEBUG_
      if (!(CheckStatistics())) Output("WARNING: CORRUPT STATISTICS ON KEPT!\n");
     #endif 
    
     Output("total          "); Output(Total()); Output("\n");
     Output("used           "); Output(num_of_used); Output("\n");
     Output("removed        "); Output(num_of_removed); Output("\n");
     Output("currently kept "); Output(num_of_currently_kept); Output("\n");
     Output("reachable      "); Output(numOfReachable); Output("\n");
     
     Output("min.nonempty cat.    "); Output(minCat); Output("\n");
     Output("max.nonempty cat.    "); Output(maxCat); Output("\n");
     
     if (extended)
      {
       for (int cat = 0; cat < NUMBER_OF_KEPT_CATEGORIES; cat++)
        {
         if (Total(cat)) 
          {
           OutputStatistics(cat);
           Output("\n");
          };
        };
      };
    };
  public: // for debugging
   inline long CountCurrentlyKept(const int cat)
    {
     Clause* cl = category[cat].Next();
     long res = 0L;
     while (cl != &(category[cat]))
      {
       res++;
       cl = cl->Next();
      };
     return res;
    };
   
   inline void OutputClauses(int cat)
    {
     ::Output("Kept Clauses "); ::Output(cat); ::Output("\n");
     Clause* cl = category[cat].Next();
     while (cl != &(category[cat]))
      {
       ::Output(cl); ::Output("\n");
       cl = cl->Next();
      };
    };
    
   inline bool CheckStatistics()
    {
     long sum_used = 0L;
     long sum_removed = 0L;
     long sum_currently_kept = 0L;
     for (int c = 0; c < NUMBER_OF_KEPT_CATEGORIES; c++)
      {
       sum_used += used[c];
       sum_removed += removed[c];
       sum_currently_kept += currently_kept[c];
       
       long real_curr_kept = CountCurrentlyKept(c);
       if (currently_kept[c] != real_curr_kept)
        {
         Output("WARNING! ");Output("currently_kept["); Output(c); 
         Output("] CORRUPT!\n");
         Output("stat = "); Output(currently_kept[c]); 
         Output(", real = ");
         Output(real_curr_kept); cout << "\n";
         return false;
        };
      };  
     if (sum_used != num_of_used) { return false; };
     if (sum_removed != num_of_removed) { return false; };
     if (sum_currently_kept != num_of_currently_kept) { return false; };
     if (minCat > maxCat) { return false; };
     if (minCat < maxCat) 
      {
       if (EmptyCat(minCat)) 
        {
         Output("WARNING! ");Output("minCat("); Output(minCat); 
         Output(") < maxCat(");
         Output(maxCat); Output(") but EMPTY!\n");
         return false;
        };
       if (EmptyCat(maxCat))
        {
         Output("WARNING! ");Output("maxCat("); Output(maxCat); 
         Output(") > minCat(");
         Output(minCat); Output(") but EMPTY!\n");
         return false;
        };
      };
     return true; 
    };
 }; // class KEPT
 
//**************************************************************************** 
