#      /************* Index for matching. *************/

#define __M_INDEX_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif
#ifndef __MEMORY_H
 #include "memory.h"
#endif
#ifndef __STACK_H
 #include "stack.h"
#endif
#ifndef __TERM_H
 #include "term.h"
#endif
#ifndef __CLAUSE_H
 #include "clause.h"
#endif 
#ifndef __VARIABLES_H
 #include "variables.h"
#endif 
#ifndef __FLAT_TERM_H
 #include "flat_term.h"
#endif 
#ifndef __MI_TREE_H
 #include "mi_tree.h"
#endif

//*************************************************************************

class M_INDEX
 {
  private:
   MI_TREE::NODE* tree[MAX_NUMBER_OF_FUNCTORS];
   const FLAT_TERM* subst[MAX_NUMBER_OF_VARS];
   MI_TREE::INTEGRATOR integrator;
   MI_TREE::REMOVAL removal;
   MI_TREE::RETRIEVAL retrieval;
   MI_TREE::OUTPUT output;
  private:
   inline MI_TREE::NODE*& Tree(const TERM& func) { return tree[func.Functor()]; };
  public:
   M_INDEX() : integrator(subst), removal(subst), retrieval(), output()
    {
     output.SetSubst(subst);
     for (int f = 0;f<MAX_NUMBER_OF_FUNCTORS;f++) tree[f] = (MI_TREE::NODE*)0;     
    };

   inline MI_TREE::NODE* Integrate(TERM* t);     
   inline VAR_MULTISET& IntegrVars() { return integrator.Vars(); };
   inline const FLAT_TERM** IntegrMapVar(int v) { return integrator.MapVar(v); };
   inline bool Remove(TERM* t,MI_TREE::NODE*& leaf,MI_TREE::NODE**& branch_to_destroy); 

   inline MI_TREE::NODE* Match(const FLAT_TERM* query);

   inline void SetLeafInfo(void (*f)(MI_TREE::NODE*)) { output.SetLeafInfo(f); };
   inline void Output();   
 }; // class M_INDEX


//****************** Definitions for M_INDEX: ****************************

inline MI_TREE::NODE* M_INDEX::Match(const FLAT_TERM* query)
 {
  CHECK20(query->Symbol().IsComplex(),"A0 in M_INDEX::Match(..)");
  return retrieval.Match(query + 1,Tree(query->Symbol()));    
 }; // MI_TREE::NODE* M_INDEX::Match(const FLAT_TERM* query) 

inline MI_TREE::NODE* M_INDEX::Integrate(TERM* t)
 { 

  //cout << "INTEGR "; ::Output(t); cout << "\n"; // debug
  

  DCONTROL20("BEGIN in M_INDEX::Integrate(TERM* t)"); 
  CHECK20(t->IsComplex(),"A0 in M_INDEX::Integrate(TERM* t)");  
  integrator.Reset(&(Tree(*t)));
  TERM::ITERATOR iter;
  iter.Reset(t);  
  while (iter.Next())  
   {
    DCONTROL20("X50 in M_INDEX::Integrate(TERM* t)"); 
    integrator.Push(iter.SymbolRef());
    DCONTROL20("X60 in M_INDEX::Integrate(TERM* t)"); 
   };
  return integrator.Leaf();
 }; // MI_TREE::NODE* M_INDEX::Integrate(TERM* t)   


inline bool M_INDEX::Remove(TERM* t,MI_TREE::NODE*& leaf,MI_TREE::NODE**& branch_to_destroy)
 {
  CHECK20(!t->IsVariable(),"A0 in M_INDEX::Remove(TERM* t)"); 
  MI_TREE::NODE** tree_addr = &(Tree(*t));  
  if (!(*tree_addr)) return false;
  removal.Reset(tree_addr);
  TERM::ITERATOR iter;
  iter.Reset(t);  
  while (iter.Next())  
   {    
    if (!removal.Skip(iter.SymbolRef())) return false;
   };
  if (!removal.SkipComps()) return false;
  leaf = removal.Leaf();
 
  CHECK20(leaf->IsLeaf(),"B10 in M_INDEX::Remove(TERM* t)");    
    
  branch_to_destroy = removal.BranchToDestroy();
  return true;
 }; // bool M_INDEX::Remove(TERM* t,MI_TREE::NODE*& leaf,MI_TREE::NODE**& branch_to_destroy)  

inline void M_INDEX::Output()
 {
  for (int f = 0;f<MAX_NUMBER_OF_FUNCTORS;f++)
   {
    MI_TREE::NODE* t = tree[f];
    if (t)
     {
      ::Output("HEADER ");
      ::OutputFunctor(f);
      ::Output('\n');
      output.Tree(t);
      ::Output("\nEND OF TREE FOR ");
      ::OutputFunctor(f);
      ::Output("\n\n");
     };  
   };  
 }; // void M_INDEX::Output()
  


//*************************************************************************






















