#define __MEMORY_H

#include <stdlib.h>

#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif

#ifndef __OUTPUT_H
 #include "output.h"
#endif

#ifndef __DEBUG_H
 #include "debug.h"
#endif

extern void SpellPointer(void* ptr); // debug

#define CHUNK_SIZE_FOR_32_BITS 16000 

#ifdef _DEBUG_MEMORY_
 #undef CHUNK_SIZE_FOR_32_BITS
 #define CHUNK_SIZE_FOR_32_BITS 16000000 
#endif 

class Memory
{
  private: 
   static void* memories[MAX_NUMBER_OF_32_BITS_STRUCTURES+1];
   static void** memory = 0;
   static int free_memory = 0;
   static long occupied;
   static long requested;
 
  public:
    static inline void* Alloc(int number);	     // allocates number+1 32-bits
    static inline void Dealloc(void* str,int number);  // deallocates number+1 32-bits
    static inline long Occupied() { return occupied; };
    static inline long Requested() { return requested; };
    static inline void* Allocate(int number);
    static inline void Deallocate(void* str,int number);
  private:
    Memory();  // just initializes the memory routines
    static Memory dummy; // just to initialize the memory routines
  
  #ifdef _DEBUG_MEMORY_
  public:
    static inline bool Free(void* obj)  
     {
      return (*(((long*)obj) - 1) == 12345678);
     };
    static inline void MarkFree(void* obj)  
     {
      *(((long*)obj) - 1) = 12345678;
     };
    static inline bool Captured(void* obj)
     {
      return (*(((long*)obj) - 1) == 87654321);
     };
    static inline void MarkCaptured(void* obj)  
     {
      *(((long*)obj) - 1) = 87654321;
     };
  #endif  
};  // class Memory

inline void* Memory::Alloc(int number)
 {
  return Allocate(number+1);
 };  // Alloc(int number)

inline void Memory::Dealloc(void* obj,int number)
 {
  Deallocate(obj,number+1);
 }; // Dealloc()

inline void* Memory::Allocate(int number)
 {
  
  #ifdef _DEBUG_MEMORY_
   number++;
  #endif
 
  void *result;
  void** mem = memories + number;
  
  if (*mem)
   {
    result = *mem;               
    *mem = *((void**)*mem); 
    occupied += number;
    
    #ifdef _DEBUG_MEMORY_
     result = (void*)(((void**)result) + 1);
     MarkCaptured(result);
    #endif
    
    
    /*** debug ****
    
     if (((long)result) == 42983464)
      {
       cout << "PRECISE ";
       REPORT_CONTROL_POINT();
       cout << "\n";
      };
     
     if ((((long)result) > 42983464) && (((long)result) - 9  < 42983464))
      {
       cout << "APPROX ";
       REPORT_CONTROL_POINT();
       cout << "\n";
      };
    
    **************/
    
    return result;
   };
  
  if (free_memory < number)  // not enough memory requested from the system
   {
    memory = new (void *)[CHUNK_SIZE_FOR_32_BITS];
    
    free_memory = CHUNK_SIZE_FOR_32_BITS;
    requested += CHUNK_SIZE_FOR_32_BITS;
    
    #ifdef _DEBUG_MEMORY_
     DEBUG_OUTPUT("MEMORY : [");
     DEBUG_OUTPUT((long)memory);
     DEBUG_OUTPUT(" - ");
     DEBUG_OUTPUT((long)((void**)memory + CHUNK_SIZE_FOR_32_BITS - 1));
     DEBUG_OUTPUT("]\n");
    #endif
    
   };
  
  free_memory = free_memory - number;
  result = memory;
  memory = memory + number;
  occupied += number;
  
  #ifdef _DEBUG_MEMORY_
   result = (void*)(((void**)result) + 1);
   MarkCaptured(result);
  #endif      
  
  /*** debug ****
    
     if (((long)result) == 42983464)
      {
       cout << "PRECISE ";
       REPORT_CONTROL_POINT();
       cout << "\n";
      };
     
     if ((((long)result) > 42983464) && (((long)result) - 9  < 42983464))
      {
       cout << "APPROX ";
       REPORT_CONTROL_POINT();
       cout << "\n";
      };
    
    **************/
  
  return result;
 }; // inline void* Memory::Allocate(const int number) 

inline void Memory::Deallocate(void* obj,int number)
 {  
  
  #ifdef _DEBUG_MEMORY_
   if (!(Captured(obj))) 
    {
     if (Free(obj))
      {
       
       DebugError("Releasing memory which was released");
      }
     else
      {
       /***** debug *****/
       cout << "DIST\n";
       long* o = (long*)obj;
       o--;
       int dist = 1;
       check_o:   
        if ((Free((void*)o)) || (Captured((void*)o)))
         {
          cout << dist << "\n";
          if (Free((void*)o)) { cout << "FREE\n"; }
          else { cout << "CAPT\n"; };
          cout << (long)o << "\n";
         }
        else
         {      
          dist++;
          o--;
          goto check_o; 
         }; 
       /*****************/
       
      
       DEBUG_OUTPUT("OBJECT: "); 
       DEBUG_OUTPUT((long)obj);
       DEBUG_OUTPUT(" MARK: "); 
       DEBUG_OUTPUT((long)(*((void**)obj - 1)));
      
       /******* 
       cout << "AGA !!\n";
       
       SpellPointer((void*)(*((void**)obj - 1))); 
       cout << "\n";
       ********/
       
       DEBUG_OUTPUT("   NEXT NONCAPTURED: "); 
       DEBUG_OUTPUT((long)memory);
       DEBUG_OUTPUT("\n");
       DebugError("Releasing memory which was not captured yet");
      };
    };
   /********
   for (int i = 0; i < number; i++)
    {
     long* ptr = ((long*)obj) + i;
     if (((*ptr) == 12345678) || ((*ptr) == 87654321))
      {
       DebugError("Releasing memory from another object");
      };
    };
   *********/
    
   MarkFree(obj); 
   number++;
   obj = (void*)(((void**)obj) - 1);
  #endif
 
  void** mem = memories + number;
  *((void**)obj) = *mem;
  *mem = obj;
  occupied -= number;
 }; // void Memory::Deallocate(void* const obj,int number)


//****************************************************************************