#include "mi_tree.h"

//*******************************************************

#ifdef _DEBUG19_ 
long MI_TREE::NODE::next_id = 0L; 
#endif


//****************** Definitions for MI_TREE::INTEGRATOR: *********************

void MI_TREE::INTEGRATOR::PushVar(const TERM& sym)
 {
  DCONTROL19("BEGIN in MI_TREE::INTEGRATOR::PushVar(const TERM& sym)"); 
  int var = sym.Var();
  if (read_mode)
   {
    read_mode = ReadVar();
    if (read_mode)
     {
      if (vars.Occurences(var)) { ReadCompare(first_occ[var],mi_var); }
      else // first occurence of the variable
       {
        vars.Register(var);
        first_occ[var] = mi_var;
       };        
     } 
    else // variable must be inserted 
     {
      InsertVar(mi_var);
      if (vars.Occurences(var))
       {
        EnterWriteMode();
        WriteCompare(first_occ[var],mi_var);
       } 
      else // first occurence of the variable
       {
        vars.Register(var);
        first_occ[var] = mi_var;
        EnterWriteMode();
       };     
     };
    mi_var++;
   }
  else // write mode
   WriteVar(var);          
 }; // void MI_TREE::INTEGRATOR::PushVar(const TERM& sym)

void MI_TREE::INTEGRATOR::PushFunc(const TERM& sym)
 {
  DCONTROL19("BEGIN in MI_TREE::INTEGRATOR::PushFunc(const TERM& sym)");
  if (read_mode)
   {
    DCONTROL19("X10 in MI_TREE::INTEGRATOR::PushFunc(const TERM& sym)");  
    read_mode = ReadFunc(sym);
    DCONTROL19("X20 in MI_TREE::INTEGRATOR::PushFunc(const TERM& sym)"); 
    if (!read_mode) 
     {
      InsertFunc(sym);
      EnterWriteMode(); 
     }; 
   }
  else // write mode
   {
    DCONTROL19("Y10 in MI_TREE::INTEGRATOR::PushFunc(const TERM& sym)"); 
    WriteFunc(sym);
   };   
  DCONTROL19("END in MI_TREE::INTEGRATOR::PushFunc(const TERM& sym)"); 
 }; // void MI_TREE::INTEGRATOR::PushFunc(const TERM& sym)

bool MI_TREE::INTEGRATOR::ReadFunc(const TERM& sym)
 { 
  DCONTROL19("BEGIN in MI_TREE::INTEGRATOR::ReadFunc(const TERM& sym)");
  int var_num1,var_num2;
  const FLAT_TERM** var1;
  const FLAT_TERM** var2;
  MI_TREE::NODE* curr_node;  
  check_tag:   
   curr_node = *node_addr; 
   CHECK19(curr_node,"B10 in MI_TREE::INTEGRATOR::ReadFunc(const TERM& sym)");  
   DCONTROL19("X20 in MI_TREE::INTEGRATOR::ReadFunc(const TERM& sym)");  
   switch (curr_node->Tag())
    {
     case MI_TREE::NODE::FUNC:
      DCONTROL19("FUNC in MI_TREE::INTEGRATOR::ReadFunc(const TERM& sym)");   
      if (curr_node->Func() == sym)
       {
        if (curr_node->Fork()) CopyComps(constr_to_shift,curr_node->ForkAddr());
        MoveDown();
        return true;
       };
      // different functor
      if (curr_node->Func() > sym)
       {
        InsertComps(constr_to_shift,node_addr);
        return false;
       };
      // curr_node->Func() < sym 
      // no break here! go to the next case

     case MI_TREE::NODE::VAR:
      DCONTROL19("VAR in MI_TREE::INTEGRATOR::ReadFunc(const TERM& sym)");   
      // mismatch occured, try fork
      if (curr_node->Fork())      
       {
        CopyComps(constr_to_shift,curr_node->NextAddr());
        MoveRight();
        goto check_tag;
       };
      // no more forks
      InsertComps(constr_to_shift,curr_node->NextAddr());
      MoveRight();
      return false; 

     case MI_TREE::NODE::COMPARE:
      DCONTROL19("COMPARE in MI_TREE::INTEGRATOR::ReadFunc(const TERM& sym)"); 
      CHECK19(!curr_node->Fork(),"A10 in MI_TREE::INTEGRATOR::ReadFunc(const TERM& sym)"); 
      curr_node->BothVars(var1,var2);      
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      CHECK19(!constr_to_shift.Equivalent(var_num1,var_num2),"A20 in MI_TREE::INTEGRATOR::ReadFunc(const TERM& sym)"); 
      if (constr_to_insert.Equivalent(var_num1,var_num2))
       {
        constr_passed.MakeEquivalent(var_num1,var_num2);

        /******* debug *******
        if (!curr_node->Next())
	 {
	   ::Output("BAD NODE "); curr_node->Output(mi_subst); ::Output("\n");
         };

	  **************/


        MoveDown();
       }
      else // COMPARE <var1,var2> must be shifted
       {
        constr_to_shift.MakeEquivalent(var_num1,var_num2);
        *node_addr = curr_node->Next();   
        delete curr_node; 
       };
      goto check_tag;

     #ifdef _DEBUG19_
      default: 
       DebugError("X0 in MI_TREE::INTEGRATOR::ReadFunc(const TERM& sym)");
       return false;
     #else 
      #ifdef _SUPPRESS_WARNINGS_
       default: return false;
      #endif
     #endif
    };
 }; // bool MI_TREE::INTEGRATOR::ReadFunc(const TERM& sym)

bool MI_TREE::INTEGRATOR::ReadVar()
 {
  int var_num1,var_num2;
  const FLAT_TERM** var1;
  const FLAT_TERM** var2;
  MI_TREE::NODE* curr_node;
  check_tag:
   curr_node = *node_addr;        
   switch (curr_node->Tag())
    {
     case MI_TREE::NODE::FUNC:
      // no functor can preceed variable in a tree
      InsertComps(constr_to_shift,node_addr);      
      return false; 

     case MI_TREE::NODE::VAR:
      if (curr_node->Fork()) CopyComps(constr_to_shift,curr_node->ForkAddr());               
      MoveDown();
      return true;

     case MI_TREE::NODE::COMPARE:
      CHECK19(!curr_node->Fork(),"A10 in MI_TREE::INTEGRATOR::ReadVar()"); 
      curr_node->BothVars(var1,var2);      
      var_num1 = VarNum(var1);
      var_num2 = VarNum(var2);
      CHECK19(!constr_to_shift.Equivalent(var_num1,var_num2),"A20 in MI_TREE::INTEGRATOR::ReadVar()"); 
      if (constr_to_insert.Equivalent(var_num1,var_num2))
       {
        constr_passed.MakeEquivalent(var_num1,var_num2);
        MoveDown();
       }
      else // COMPARE <var1,var2> must be shifted
       {
        constr_to_shift.MakeEquivalent(var_num1,var_num2);
        *node_addr = curr_node->Next();
        delete curr_node; 
       };
      goto check_tag;

     #ifdef _DEBUG19_
      default: 
       DebugError("X0 in MI_TREE::INTEGRATOR::ReadVar()");
       return false;
     #else 
      #ifdef _SUPPRESS_WARNINGS_
       default: return false;
      #endif
     #endif
    };  
 }; // bool MI_TREE::INTEGRATOR::ReadVar() 


void MI_TREE::INTEGRATOR::ReadCompare(const FLAT_TERM** v1,const FLAT_TERM** v2)
 {
  DCONTROL19("BEGIN in MI_TREE::INTEGRATOR::ReadCompare(..)");
  CHECK19(v1<v2,"A0 in MI_TREE::INTEGRATOR::ReadCompare(..)");
  int var_num1 = VarNum(v1);
  int var_num2 = VarNum(v2);
  if (constr_to_shift.SplitOn(var_num1,var_num2))
   {
    constr_passed.MakeEquivalent(var_num1,var_num2);
    MI_TREE::NODE* curr_node = MI_TREE::NODE::NewCOMPARE(v1,v2);
    curr_node->SetNext(*node_addr);
    *node_addr = curr_node;
    MoveDown();
   }
  else constr_to_insert.MakeEquivalent(var_num1,var_num2);     
 }; // void MI_TREE::INTEGRATOR::ReadCompare(const FLAT_TERM** v1,const FLAT_TERM** v2) 


bool MI_TREE::INTEGRATOR::ReadCollectedComps()
 {
  CHECK19(read_mode,"A0 in MI_TREE::INTEGRATOR::ReadCollectedComps()");
  const FLAT_TERM** var1;
  const FLAT_TERM** var2;  
  int var_num1;
  int var_num2;
  MI_TREE::NODE* curr_node;
  if (constr_to_insert.Identity())
   {
    if (constr_to_shift.Identity())
     {
      both_empty_skip_comps:
       curr_node = *node_addr;
       if (curr_node->IsCompare())
	{
         MoveRight();
         if (curr_node->Fork()) goto both_empty_skip_comps;
         // no more forks
         return false;
        };
       // !curr_node->IsCompare()
       return true;
     }
    else // constr_to_insert.Identity() but !constr_to_shift.Identity()
     {
      InsertComps(constr_to_shift,node_addr); 
      MoveRight();
      return false;
     };
   }
  else // !constr_to_insert.Identity() 
   {
    if (constr_to_shift.Identity())
     {
      insert_nonempty_check_tag: 
       curr_node = *node_addr;
       if (curr_node->IsCompare())
	{
         curr_node->BothVars(var1,var2);
         var_num1 = VarNum(var1);
         var_num2 = VarNum(var2);
         if (constr_to_insert.Equivalent(var_num1,var_num2))
	  {
           constr_passed.MakeEquivalent(var_num1,var_num2);
           MoveDown();
           goto insert_nonempty_check_tag;  
          }
         else // try fork
	  {
           MoveRight();
           if (curr_node->Fork()) goto insert_nonempty_check_tag; 
	   // no fork 
           constr_to_insert.Subtract(constr_passed);
           MoveComps(constr_to_insert,node_addr);
           return false;
          };
        }
       else // !curr_node->IsCompare()
	{
         constr_to_insert.Subtract(constr_passed);
         if (constr_to_insert.Identity()) return true;

	 MI_TREE::NODE** node_addr_backup = node_addr;
         *node_addr = (MI_TREE::NODE*)0;
         MoveComps(constr_to_insert,node_addr);
         (*node_addr_backup)->SetFork(curr_node);
         return false;

        };
     }    
    else // !constr_to_insert.Identity() and !constr_to_shift.Identity()
     {
      both_nonempty_check_tag: 
       curr_node = *node_addr;
       if (curr_node->IsCompare()) 
	{
         curr_node->BothVars(var1,var2);
         var_num1 = VarNum(var1);
         var_num2 = VarNum(var2);
         if (constr_to_insert.Equivalent(var_num1,var_num2))         
	  {
           constr_passed.MakeEquivalent(var_num1,var_num2);
           if (curr_node->Fork()) CopyComps(constr_to_shift,curr_node->ForkAddr());          
           MoveDown();
           goto both_nonempty_check_tag;
          }
         else // try fork
	  {
           CopyComps(constr_to_shift,curr_node->NextAddr());
           MoveRight();           
           if (curr_node->Fork()) goto both_nonempty_check_tag;
           // no fork            
           constr_to_insert.Subtract(constr_passed);           
           MoveComps(constr_to_insert,node_addr);
           constr_to_shift.Reset();
           return false; 
          };
        }
       else // !curr_node->IsCompare()
	{
         constr_to_insert.Subtract(constr_passed);
         InsertComps(constr_to_shift,node_addr);
         MoveRight();
         MoveComps(constr_to_insert,node_addr);        
         return false;
        };
     };
   };  
 }; // bool MI_TREE::INTEGRATOR::ReadCollectedComps()


void MI_TREE::INTEGRATOR::InsertVar(const FLAT_TERM** v)
 {
  MI_TREE::NODE* tmp = *node_addr;
  *node_addr = MI_TREE::NODE::NewVAR(v);
  (*node_addr)->SetFork(tmp);
  MoveDown(); 
 }; // void MI_TREE::INTEGRATOR::InsertVar(const FLAT_TERM** v)

void MI_TREE::INTEGRATOR::InsertFunc(const TERM& func)
 {
  MI_TREE::NODE* tmp = *node_addr;
  *node_addr = MI_TREE::NODE::NewFUNC(func);
  (*node_addr)->SetFork(tmp);
  MoveDown();
 }; // void MI_TREE::INTEGRATOR::InsertFunc(const TERM& func)  

void MI_TREE::INTEGRATOR::EnterWriteMode()
 {
  // must be used if ReadVar or ReadFunc failed
  CHECK19(constr_to_shift.Identity(),"A0 in MI_TREE::INTEGRATOR::EnterWriteMode()");
  constr_to_insert.Subtract(constr_passed);
  MoveComps(constr_to_insert,node_addr);
 }; // void MI_TREE::INTEGRATOR::EnterWriteMode() 

void MI_TREE::INTEGRATOR::WriteVar(int v)
 {
  *node_addr = MI_TREE::NODE::NewVAR(mi_var);  
  MoveDown();
  if (vars.Occurences(v)) { WriteCompare(first_occ[v],mi_var); }
  else // first occurence
   {
    vars.Register(v);
    first_occ[v] = mi_var;
   };   
  mi_var++;
 }; // void MI_TREE::INTEGRATOR::WriteVar(int v) 

void MI_TREE::INTEGRATOR::WriteFunc(const TERM& func)
 {
  *node_addr = MI_TREE::NODE::NewFUNC(func);
  MoveDown();
 }; // void MI_TREE::INTEGRATOR::WriteFunc(const TERM& func) 


void MI_TREE::INTEGRATOR::WriteCompare(const FLAT_TERM** v1,const FLAT_TERM** v2)
 {
  DCONTROL19("BEGIN in MI_TREE::INTEGRATOR::WriteCompare(..)");
  *node_addr = MI_TREE::NODE::NewCOMPARE(v1,v2); 
  MoveDown();
 }; // void MI_TREE::INTEGRATOR::WriteCompare(const FLAT_TERM** v1,const FLAT_TERM** v2)


void MI_TREE::INTEGRATOR::CopyComps(const MI_EQ_CONSTR& constr,MI_TREE::NODE** addr)
 {
  DCONTROL19("BEGIN in MI_TREE::INTEGRATOR::CopyComps(..)");
  int vn1,vn2;
  MI_EQ_CONSTR::ITER iter(constr);
  MI_TREE::NODE* old_node = *addr;
  MI_TREE::NODE* new_node; 
  while (iter.Next(vn2,vn1))
   {
    CHECK19(vn1<vn2,"A0 in MI_TREE::INTEGRATOR::CopyComps()");
    new_node = MI_TREE::NODE::NewCOMPARE(Var(vn1),Var(vn2));
    *addr = new_node;
    addr = new_node->NextAddr();      
   };
  *addr = old_node;
 }; // void MI_TREE::INTEGRATOR::CopyComps(const MI_EQ_CONSTR& constr,MI_TREE::NODE** addr)

void MI_TREE::INTEGRATOR::InsertComps(MI_EQ_CONSTR& constr,MI_TREE::NODE** addr)
 {
  DCONTROL19("BEGIN in MI_TREE::INTEGRATOR::InsertComps(..)");
  int vn1,vn2;
  MI_TREE::NODE* old_node = *addr;
  MI_TREE::NODE* new_node; 
  while (constr.Split(vn2,vn1))
   {
    CHECK19(vn1<vn2,"A0 in MI_TREE::INTEGRATOR::InsertComps()");
    new_node = MI_TREE::NODE::NewCOMPARE(Var(vn1),Var(vn2));
    *addr = new_node;
    addr = new_node->NextAddr();      
   };
  CHECK19(constr.Identity(),"A10 in MI_TREE::INTEGRATOR::InsertComps()");
  *addr = old_node;
 }; // void MI_TREE::INTEGRATOR::InsertComps(MI_EQ_CONSTR& constr,MI_TREE::NODE** addr)

void MI_TREE::INTEGRATOR::MoveComps(MI_EQ_CONSTR& constr,MI_TREE::NODE**& addr)
 {
  DCONTROL19("BEGIN in MI_TREE::INTEGRATOR::MoveComps(..)");
  int vn1,vn2;
  MI_TREE::NODE* old_node = *addr;
  MI_TREE::NODE* new_node;
  while (constr.Split(vn2,vn1))
   {
    CHECK19(vn1<vn2,"A0 in MI_TREE::INTEGRATOR::MoveComps()");
    new_node = MI_TREE::NODE::NewCOMPARE(Var(vn1),Var(vn2));
    *addr = new_node;
    addr = new_node->NextAddr();      
   };
  CHECK19(constr.Identity(),"A10 in MI_TREE::INTEGRATOR::MoveComps()");
  *addr = old_node;
 }; // void MI_TREE::INTEGRATOR::MoveComps(MI_EQ_CONSTR& constr,MI_TREE::NODE**& addr)


//***************** Definitions for MI_TREE::RETRIEVAL: *****************

MI_TREE::NODE* MI_TREE::RETRIEVAL::Match(const FLAT_TERM* query,MI_TREE::NODE* tree)
 {
  if (!tree) return (MI_TREE::NODE*)0;
  STACK<MI_TREE::NODE*,MAX_TERM_WEIGHT> backtrack_nodes("MAX_TERM_WEIGHT in MI_TREE::RETRIEVAL::Match(..):backtrack_nodes"); 
  STACK<const FLAT_TERM*,MAX_TERM_WEIGHT> backtrack_cursors("MAX_TERM_WEIGHT in MI_TREE::RETRIEVAL::Match(..):backtrack_cursors");  
    
  const FLAT_TERM* cursor = query;
  MI_TREE::NODE* curr_node = tree;
  TERM curr_sym;

  check_node:
   switch (curr_node->Tag())
    {
     case MI_TREE::NODE::FUNC:
      curr_sym = cursor->Symbol();                 
      if (curr_sym.IsVariable()) goto backtrack; 
   
      compare_func:
       if (curr_sym == curr_node->Func())
	{
         cursor++;
         curr_node = curr_node->Next();
         goto check_node;  
        };      
       // different symbols
       if (curr_sym < curr_node->Func()) goto backtrack;
       // here curr_sym > curr_node->Func()     
       curr_node = curr_node->Fork();
       if (!curr_node) goto backtrack;
       func_check_tag:
        if (curr_node->IsFunc()) goto compare_func;
        // here curr_node must point to a COMPARE without fork
        CHECK19(curr_node->IsCompare(),"A100 in MI_TREE::RETRIEVAL::Match(..)");  
        CHECK19(!curr_node->Fork(),"A110 in MI_TREE::RETRIEVAL::Match(..)");  
        if ((*(curr_node->Var1()))->Equal(*(curr_node->Var2())))
	 {
          curr_node = curr_node->Next();
          goto func_check_tag;
         }
	else // COMPARE failed
         goto backtrack;
        
     case MI_TREE::NODE::VAR:
      *(curr_node->Var()) = cursor;        
      if (curr_node->Fork() && cursor->Symbol().IsNotVariable())
       {
        backtrack_nodes.Push(curr_node->Fork());
        backtrack_cursors.Push(cursor);
       };
      cursor = cursor->NextTerm();
      curr_node = curr_node->Next();
      goto check_node;     

     case MI_TREE::NODE::COMPARE:
      if ((*(curr_node->Var1()))->Equal(*(curr_node->Var2())))
       {
        if (curr_node->Fork())
         {
          backtrack_nodes.Push(curr_node->Fork());
          backtrack_cursors.Push(cursor);
         };
        curr_node = curr_node->Next();
        goto check_node;
       }
      else // COMPARE failed, try fork
       {
        curr_node = curr_node->Fork();
        if (curr_node) goto check_node;
        goto backtrack;
       };

     case MI_TREE::NODE::LEAF: return curr_node; // success

     #ifdef _DEBUG19_
      default: DebugError("X10 in MI_TREE::RETRIEVAL::Match(..)"); return (MI_TREE::NODE*)0;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default: return (MI_TREE::NODE*)0;  
      #endif
     #endif
    };
 
      //*************** Backtrack section: **********
   backtrack:
    if (backtrack_nodes.Empty())
     {
      // failure
      return (MI_TREE::NODE*)0;
     };
    curr_node = backtrack_nodes.Pop(); 
    cursor = backtrack_cursors.Pop();
    goto check_node;
 }; // MI_TREE::NODE* MI_TREE::RETRIEVAL::Match(const FLAT_TERM* query,MI_TREE::NODE* tree) 


//***************** Definitions for MI_TREE::OUTPUT: ******************** 


void MI_TREE::OUTPUT::LeafInfoDefault(MI_TREE::NODE*) {}; 

//******************************************************  
