#include "options.h"

int OPTION::next_option = 0;  
OPTION::DESCRIPTION OPTION::descriptions[MAX_NUMBER_OF_OPTIONS];
void* OPTION::values[MAX_NUMBER_OF_OPTIONS];
bool OPTION::wasSet[MAX_NUMBER_OF_OPTIONS];
bool OPTION::main_option[MAX_NUMBER_OF_OPTIONS];

//SPELL OPTION::SwitchEnum[] = { {0,"off"}, {1,"on"} };
//OPTION::DATATYPE OPTION::SWITCH = 
 //OPTION::DATATYPE("SWITCH",&(OPTION::SWITCHSpellValue),&(OPTION::StringToSWITCH));

OPTION::DATATYPE OPTION::SWITCH("SWITCH",&(OPTION::SWITCHSpellValue),&(OPTION::StringToSWITCH));
OPTION::DATATYPE OPTION::LONG_INT("LONG_INT",&(OPTION::LONG_INTSpellValue),&(OPTION::StringToLONG_INT));
OPTION::DATATYPE OPTION::DOUBLE("DOUBLE",&(OPTION::DOUBLESpellValue),&(OPTION::StringToDOUBLE));
OPTION::DATATYPE OPTION::STRING("STRING",&(OPTION::STRINGSpellValue),&(OPTION::StringToSTRING));


//*********************************************************************

int OPTION::Number(char* name)
 {
  for (int i = 0; i < next_option; i++)
   {
    if (!(strcmp(descriptions[i].Name(),name))) { return i; };
   };
  return -1;
 }; // int OPTION::Number(char* name)


bool OPTION::Described(char* name)
 {
  return (Number(name) != -1);
 }; // bool OPTION::Described(char* name)

void OPTION::Describe(bool main_opt,char* name,DATATYPE& dt,void* dfl,char* un)
 {
  int num = Number(name);
  if (num != -1) { RuntimeError("Option ",name," described two times.\n"); };
  if (next_option >= MAX_NUMBER_OF_OPTIONS) 
   { RuntimeError("Compilation parameter MAX_NUMBER_OF_OPTIONS exceeded.\n"); };
  descriptions[next_option].DESCRIPTION(name,dt,dfl,un);
  values[next_option] = dfl;
  wasSet[next_option] = false;
  main_option[next_option] = main_opt;
  next_option++;
 }; // void OPTION::Describe(bool main_opt,char* name,DATATYPE dt,void* dfl,char* un)

OPTION::DESCRIPTION* OPTION::Description(char* name)
 {
  int num = Number(name);
  if (num == -1)
   {
    return (DESCRIPTION*)0;
   }
  else
   {
    return descriptions + num;
   };
 }; // DESCRIPTION* OPTION::Description(char* name)

bool OPTION::Set(char* name,char* value,char*& expl)
 {
  int num = Number(name);
  if (num == -1) 
   { 
    char* s1 = "unknown option \"";
    char* s2 = "\"";
    expl = new char[strlen(s1)+strlen(name)+strlen(s2)+1];
    char* s = expl;
    strcpy(s,s1);
    s += strlen(s1);
    strcpy(s,name);
    s += strlen(name);
    strcpy(s,s2);
    s += strlen(s2);
    *s = (char)0;
    return false;
   };
  bool transformed = false;
  void* val = descriptions[num].Datatype().StringTo(value,transformed);
  if (transformed)
   {
    values[num] = val;
    expl = (char*)0;
    wasSet[num] = true;
    return true;
   };
  char* s1 = "value ";
  char* s2 = " can not be interpreted as ";
  expl = new char[strlen(s1)
                  +strlen(value)
                  +strlen(s2)
                  +strlen(descriptions[num].Datatype().Name())];
  char* s = expl;
  strcpy(s,s1);
  s += strlen(s1);
  strcpy(s,value);
  s += strlen(value);
  strcpy(s,s2);
  s += strlen(s2);
  strcpy(s,descriptions[num].Datatype().Name());
  s += strlen(descriptions[num].Datatype().Name());
  *s = (char)0;
  return false;
 }; // bool OPTION::InitOption(char* name,char* value,char*& expl)

void OPTION::Set(char* name,void* value)
 {
  int num = Number(name);
  if (num == -1) 
   { RuntimeError("Assigning value to undescribed option ",name,".\n"); };
  values[num] = value;
  wasSet[num] = true;
 }; // void OPTION::Set(char* name,void* value)

void* OPTION::Value(char* name)
 {
  int num = Number(name);
  if (num == -1) 
   { RuntimeError("Trying to read value of undescribed option ",name,".\n"); };
  return values[num];
 }; // void* OPTION::Value(char* name)

void OPTION::OutputOption(int num)
 {
  Output(descriptions[num].Datatype().Name());
  Output(" ");
  Output(descriptions[num].Name());
  Output(" = ");
  Output(descriptions[num].Datatype().SpellValue(values[num]));
  Output(" ");
  Output(descriptions[num].Units());
  if (!(wasSet[num]))
   {
    Output(" (default)");
   };    
 }; // void OPTION::OutputOption(int num)

void OPTION::OutputOption(char* name)
 {
  int num = Number(name);
  if (num == -1) 
   {
    Output("Option ");
    Output(name);
    Output(" is not described.");
   }
  else
   {
    OutputOption(num);
   };
 }; // void OPTION::OutputOption(char* name)

void OPTION::OutputOptions(bool main_only,bool output_default)
 {
  for (int i = 0; i < next_option; i++)
   {
    if (((!(main_only)) || (main_option[i])) 
        && ((output_default) || (wasSet[i])))
     {
      OutputOption(i);
      Output("\n");
     };
   };
 }; // void OPTION::OutputOptions()


char* OPTION::DATATYPE::EnumSpellValue(int en_size,SPELL* en,void* value)
 {
  for (int i = 0; i < en_size; i++)
   {
    if (en[i].Val() == (int)value) { return en[i].Spell(); };
   };
  return "<<inacceptable value>>";
 }; // char* OPTION::DATATYPE::EnumSpellValue(int en_size,SPELL* en,void* value)

void* OPTION::DATATYPE::EnumStringTo(int en_size,SPELL* en,char* s,bool& successful)
 {
  for (int i = 0; i < en_size; i++)
   {
    if (!(strcmp(en[i].Spell(),s)))
     { 
      successful = true;
      return (void*)(en[i].Val()); 
     };
   };
  successful = false;
  return (void*)0;
 }; // void* OPTION::DATATYPE::EnumStringTo(int en_size,SPELL* en,char* s,bool& successful) 

void* OPTION::StringToSWITCH(char* s,bool& successful)
 {
  static char* on = "on";
  static char* off = "off";
  if (!(strcmp(s,on))) { successful = true; return (void*)1; };
  if (!(strcmp(s,off))) { successful = true; return (void*)0; };
  successful = false;
  return (void*)0;
 }; // void* OPTION::StringToSWITCH(char* s,bool& successful)

char* OPTION::SWITCHSpellValue(void* value)
 {
  static char* on = "on";
  static char* off = "off";
  if (value)
   {
    return on;
   }
  else
   {
    return off;
   };
 }; // char* OPTION::SWITCHSpellValue(void* value) 

void* OPTION::StringToLONG_INT(char* s,bool& successful)
 {
  char* tail = (char*)0;
  errno = 0;
  long val = strtol(s,&tail,0);
  successful = (!((errno) || (tail == s)));
  return (void*)val;
 }; // void* OPTION::StringToLONG_INT(char* s,bool& successful)

char* OPTION::LONG_INTSpellValue(void* value)
 {
  ostrstream str;
  if ((long)value == MAXLONG) { return "MAXLONG"; };
  if ((long)value == -MAXLONG) { return "-MAXLONG"; };
  str << (long)value << (char)0;
  str.freeze(1);
  return str.str();
 }; // char* OPTION::LONG_INTSpellValue(void* value)

void* OPTION::StringToDOUBLE(char* s,bool& successful)
 {
  char* tail = (char*)0;
  errno = 0;
  double* val = new double;
  *val = strtod(s,&tail);
  successful = (!((errno) || (tail == s)));
  return (void*)val;
 }; // void* OPTION::StringToDOUBLE(char* s,bool& successful)

char* OPTION::DOUBLESpellValue(void* value)
 {
  ostrstream str;
  str << *((double*)value) << (char)0;
  str.freeze(1);
  return str.str();
 }; // char* OPTION::DOUBLESpellValue(void* value)

void* OPTION::StringToSTRING(char* s,bool& successful)
 {
  char* res = new char[strlen(s)];
  strcpy(res,s);
  successful = true;
  return (void*)res;
 }; // void* OPTION::StringToDOUBLE(char* s,bool& successful)

char* OPTION::STRINGSpellValue(void* value)
 {
  return (char*)value;
 }; // char* OPTION::DOUBLESpellValue(void* value)



//*****************************************************************************












