#define __OPTIONS_H

#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif

#ifndef __OUTPUT_H
 #include "output.h"
#endif

#ifndef __READCMD_H
 #include "readcmd.h"
#endif


#include <string.h>
#include <stdlib.h>
#include <strstream.h>
#include <errno.h>

struct SPELL 
 {
  int val;
  char* spell;
  inline int Val() { return val; };
  inline char* Spell() { return spell; };
 };
 
class OPTION
 {
  public:
   class DATATYPE
    {
     private:
      char* name;
      char* (*spell_value)(void* value);
      void* (*string_to)(char* s,bool& successful);
      int enumeration_size;
      SPELL* enumeration;
     private:
      static char* EnumSpellValue(int en_size,SPELL* en,void* value);
      static void* EnumStringTo(int en_size,SPELL* en,char* s,bool& successful);
     public:
      DATATYPE() {};
      DATATYPE(char* nm,
               char* (*sv)(void* value),
               void* (*st)(char* s,bool& successful)) 
       : name(nm),spell_value(sv),string_to(st),enumeration((SPELL*)0) 
       {};             
      DATATYPE(char* nm,
               int enum_size,
               SPELL* en)
        : name(nm),
          spell_value(0),
          string_to(0),
          enumeration_size(enum_size),
          enumeration(en)
       {};           
      inline char* Name() { return name; };
      inline char* SpellValue(void* value) 
       {
        if (enumeration)
         {
          return EnumSpellValue(enumeration_size,enumeration,value);
         }
        else
         {
          return spell_value(value); 
         };
       };
      inline void* StringTo(char* s,bool& successful) 
       {
        if (enumeration)
         {
          return EnumStringTo(enumeration_size,enumeration,s,successful);
         }
        else
         {
          return string_to(s,successful); 
         };
       };                
    }; 
   class DESCRIPTION
    {
     private:
      char* name;
      DATATYPE dtype;
      void* default_value;
      char* units;
     public: 
      DESCRIPTION() {};
      DESCRIPTION(char* nm,DATATYPE dt,void* def_val,char* un)
       : name(nm), dtype(dt), default_value(def_val), units(un) 
       {};
      inline char* Name() { return name; };
      inline DATATYPE& Datatype() { return dtype; };
      inline void* DefaultValue() { return default_value; };
      inline char* Units() { return units; };
    };
  private: // data
   static int next_option;  
   static DESCRIPTION descriptions[MAX_NUMBER_OF_OPTIONS];
   static void* values[MAX_NUMBER_OF_OPTIONS];
   static bool wasSet[MAX_NUMBER_OF_OPTIONS];
   static bool main_option[MAX_NUMBER_OF_OPTIONS];
  private:
   static int Number(char* name); // returns -1 if the option is not described yet 
   static void OutputOption(int num);
  public:
   static void Reset() { next_option = 0; };
   static inline void Describe(bool main_opt,char* name,DATATYPE& dt,void* dfl)
    { Describe(main_opt,name,dt,dfl,""); }; 
   static void Describe(bool main_opt,char* name,DATATYPE& dt,void* dfl,char* un); 
   static bool Set(char* name,char* value,char*& expl);
   static void Set(char* name,void* value);
   static bool Described(char* name);
   static DESCRIPTION* Description(char* name);
   static void* Value(char* name);
   static void OutputOption(char* name);
   static void OutputOptions(bool main_only,bool output_default);
   static void DescribeAllOptions(); // see option_list.cc
  private: // commonly used datatypes
   static void* StringToSWITCH(char* s,bool& successful);
   static char* SWITCHSpellValue(void* value);
   static void* StringToLONG_INT(char* s,bool& successful);
   static char* LONG_INTSpellValue(void* value);
   static void* StringToDOUBLE(char* s,bool& successful);
   static char* DOUBLESpellValue(void* value);
   static void* StringToSTRING(char* s,bool& successful);
   static char* STRINGSpellValue(void* value);
  public: // commonly used datatypes
   static DATATYPE SWITCH;
   static DATATYPE LONG_INT;
   static DATATYPE DOUBLE;
   static DATATYPE STRING;
 }; // class OPTION

//**********************************************************************













