#define __OR_INDEX_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __TERM_H
 #include "term.h"
#endif
#ifndef __CLAUSE_H
 #include "clause.h"
#endif
#ifndef __PREFIX_H
 #include "prefix.h"
#endif 
#ifndef __VARIABLES_H
 #include "variables.h"
#endif 
#ifndef __UNIFICATION_H
 #include "unification.h"
#endif


class OR_INDEX_TREE_NODE;

class OR_INDEX
 {
  private: // auxilliary data
   TERM::ITERATOR_WITH_DEPTH l_iter;
   VAR_MULTISET var_counter;
   TERM pseudo_literal[2];
  private: // structure
   OR_INDEX_TREE_NODE* header_index[MAX_NUMBER_OF_FUNCTORS * 2];
   OR_INDEX_TREE_NODE* forward_para_from_var_tree;  
   OR_INDEX_TREE_NODE* backward_para_tree; 
   bool paramodulation;
  public:
   OR_INDEX() : 
    l_iter(), 
    var_counter(), 
    forward_para_from_var_tree((OR_INDEX_TREE_NODE*)0),
    backward_para_tree((OR_INDEX_TREE_NODE*)0)
     {
      for (int i = 0; i < MAX_NUMBER_OF_FUNCTORS * 2; i++)
       header_index[i] = (OR_INDEX_TREE_NODE*)0;
      pseudo_literal[0] = TERM((ARITY)1,(FUNCTOR)0,POL_POS);
     };
   inline void SetParamodulation(bool fl) { paramodulation = fl; }; 
  private:
   inline OR_INDEX_TREE_NODE* CreateBranch(OR_INDEX_TREE_NODE*& leaf);
   inline void Integrate(TERM* lit,Clause* cl,unsigned lit_num);
   inline OR_INDEX_TREE_NODE* IntegratePropLit(TERM lit); // returns the leaf
   inline OR_INDEX_TREE_NODE* OR_INDEX::IntegrateArgs(OR_INDEX_TREE_NODE** tree); // returns the leaf
   inline void IntegrateForBackwardPara(TERM* lit,Clause* cl,unsigned lit_num);
   inline void IntegrateBPRedex(TERM* redex,Clause* cl,unsigned lit_num);
   inline void IntegrateForForwardPara(TERM* lit,Clause* cl,unsigned lit_num);
   inline OR_INDEX_TREE_NODE* IntegrateNonpropLit(TERM* lit); // returns the leaf
   
   OR_INDEX_TREE_NODE* LeafAndBranch(TERM* t,OR_INDEX_TREE_NODE** tree,OR_INDEX_TREE_NODE**& branch);
   void RemovePropLit(TERM* lit,Clause* cl);
   void Remove(TERM* t,Clause* cl,OR_INDEX_TREE_NODE** tree);
   void RemoveBPRedex(TERM* t,Clause* cl);
   void RemoveFromBPI(TERM* lit,Clause* cl);
   void RemoveFromFPI(TERM* lit,Clause* cl);
   void Remove(TERM* lit,Clause* cl);
  
  public:
   inline OR_INDEX_TREE_NODE** Tree(TERM hd)
    { 
     return header_index + (hd.Functor()*2) + hd.Polarity();
    };
   inline OR_INDEX_TREE_NODE** FParaTree(TERM fun)
    { 
     return header_index + (2*fun.Functor());
    }; 
   inline OR_INDEX_TREE_NODE** FParaFromVarTree() { return &forward_para_from_var_tree; }; 
   inline OR_INDEX_TREE_NODE** BParaTree() { return &backward_para_tree; }; 
   inline void Integrate(Clause* const c);
   void Remove(Clause* c);
  public:
   void Destroy();
 }; // class OR_INDEX 

class RES_PAIR_LIST
 {
  private: // sructure
   unsigned lit_num;
   Clause* clause;
   TERM term; // for forward and backward paramodulation
   RES_PAIR_LIST* next_pair;
  public: // constructors and memory management
   RES_PAIR_LIST(unsigned ln,Clause* cl,RES_PAIR_LIST* np)
    : lit_num(ln), clause(cl), next_pair(np) 
    {};
   RES_PAIR_LIST(unsigned ln,Clause* cl,TERM t,RES_PAIR_LIST* np) 
    // for forward paramodulation
    : lit_num(ln), 
      clause(cl), 
      term(t),
      next_pair(np) 
    {}; 
   RES_PAIR_LIST(unsigned ln,Clause* cl,TERM* t,RES_PAIR_LIST* np) 
    // for backward paramodulation
    : lit_num(ln), 
      clause(cl), 
      next_pair(np) 
    {
     term.SetFirst(t);
    }; 
   inline void* operator new(size_t)
    {
     return Memory::Allocate(sizeof(RES_PAIR_LIST)/4);
    };
   inline void operator delete(void* obj)
    {
     Memory::Deallocate(obj,sizeof(RES_PAIR_LIST)/4);
    };
  public:
   inline unsigned LitNum() const { return lit_num; };
   inline Clause* TheClause() const { return clause; };
   inline TERM ReplacingTerm() const { return term; }; // for forward paramodulation
   inline TERM* Redex() const { return term.First(); }; // for backward paramodulation
   inline RES_PAIR_LIST* Next() { return next_pair; };
   inline RES_PAIR_LIST** Tail() { return &next_pair; };
   
   
   static inline void Remove(Clause* c,RES_PAIR_LIST*& lst)
    {
     RES_PAIR_LIST* pair_to_delete;
     while ((lst) && (lst->TheClause() == c))
      {
       pair_to_delete = lst;
       lst = lst->Next();
       delete pair_to_delete;
      };
     if (!lst) return;
     RES_PAIR_LIST** tail = lst->Tail();
     while (*tail) 
      {
       if ((*tail)->TheClause() == c)
        {
         pair_to_delete = *tail;
         *tail = (*tail)->Next();
         delete pair_to_delete;  
        }
       else tail = (*tail)->Tail(); 
      }; 
    }; // void Remove(Clause* const c,RES_PAIR_LIST* lst)
    
  public: // for debugging
   inline void Output()
    {
     int max_len = 5;
     RES_PAIR_LIST* rest = this;
     ::Output("[");
     while ((rest) && (max_len))
      {
       ::Output("<");
       ::Output((short unsigned int)(rest->lit_num));
       ::Output(",");
       ::Output(rest->clause->Number());
       ::Output(">");
       max_len--;
       rest = rest->Next();
      };
     if (rest) { ::Output("..."); };
     ::Output("]");
    }; // void Output()
      
 }; // class RES_PAIR_LIST
 
class OR_INDEX_TREE_NODE
 {
  public:
   enum TAG
    {
     FUNC,
     CONST,
     VAR,
     VAR_FIRST_OCC,
     LEAF
    };
  private:
   struct INTERMEDIATE
    {
     TAG tag;
     unsigned depth;
     PREF_SYM symbol;
     OR_INDEX_TREE_NODE* below;
     OR_INDEX_TREE_NODE* alternative;
    };
   struct LEAF_NODE
    {
     TAG tag;
     unsigned depth;
     TERM term;
     RES_PAIR_LIST* clauses;
    };
  public: 
   static inline OR_INDEX_TREE_NODE* NewNode(TAG t,PREF_SYM s,unsigned d)
    {
     INTERMEDIATE* res = (INTERMEDIATE*)(Memory::Allocate(sizeof(INTERMEDIATE)/4));
     res->tag = t;
     res->depth = d;
     res->symbol = s;
     res->below = 0;
     res->alternative = 0;
     return (OR_INDEX_TREE_NODE*)res;
    };
   inline void DeleteNode()
    {
     Memory::Deallocate((void*)this,sizeof(INTERMEDIATE)/4);
    };
   static inline OR_INDEX_TREE_NODE* NewLeaf(TERM* lit,unsigned ln,Clause* cl)
    {
     LEAF_NODE* res = (LEAF_NODE*)(Memory::Allocate(sizeof(LEAF_NODE)/4));
     res->tag = LEAF;
     res->depth = 0;
     res->term.SetFirst(lit);
     res->clauses = new RES_PAIR_LIST(ln,cl,(RES_PAIR_LIST*)0);
     return (OR_INDEX_TREE_NODE*)res;
    };
   static inline OR_INDEX_TREE_NODE* NewLeaf()
    {
     LEAF_NODE* res = (LEAF_NODE*)(Memory::Allocate(sizeof(LEAF_NODE)/4));
     res->tag = LEAF;
     res->depth = 0;
     res->term.SetFirst((TERM*)0);
     res->clauses = (RES_PAIR_LIST*)0; 
     return (OR_INDEX_TREE_NODE*)res;
    }; 
   inline void DeleteLeaf()
    {
     Memory::Deallocate((void*)this,sizeof(LEAF_NODE)/4);
    };
   inline TAG Tag() { return ((INTERMEDIATE*)this)->tag; }; // applicable to leaves
   inline bool IsLeaf() { return (Tag() == LEAF); };
   inline bool IsVar() { return ((Tag() == VAR) || (Tag() == VAR_FIRST_OCC)); };
   inline PREF_SYM Symbol() { return ((INTERMEDIATE*)this)->symbol; };
   inline OR_INDEX_TREE_NODE* Below() { return ((INTERMEDIATE*)this)->below; };
   inline OR_INDEX_TREE_NODE** BelowAddr() { return &(((INTERMEDIATE*)this)->below); };
   inline void SetBelow(OR_INDEX_TREE_NODE* b) { ((INTERMEDIATE*)this)->below = b; };
   inline OR_INDEX_TREE_NODE* Alternative() { return ((INTERMEDIATE*)this)->alternative; };
   inline OR_INDEX_TREE_NODE** AlternativeAddr() { return &(((INTERMEDIATE*)this)->alternative); };
   inline void SetAlternative(OR_INDEX_TREE_NODE* a) { ((INTERMEDIATE*)this)->alternative = a; };
   inline unsigned Depth() { return ((INTERMEDIATE*)this)->depth; }; // applicable to leaves
   
   inline bool SameSymbol(const TERM s)
    {
     if (s.IsVariable())
      {
       return (Symbol().Var() == VAR::Base(s.Var()));
      }
     else // here s must be a functor
      { 
       return (Symbol().Func() == s);
      };
    }; // bool SameSymbol(const TERM s)

   // for leaves:
   inline TERM Term() { return ((LEAF_NODE*)this)->term; };
   inline void SetTerm(TERM* t) { ((LEAF_NODE*)this)->term.SetFirst(t); };
   inline void SetTerm(TERM t) { ((LEAF_NODE*)this)->term = t; };
   inline RES_PAIR_LIST* Clauses() { return ((LEAF_NODE*)this)->clauses; };
   inline void AddPair(unsigned ln,Clause* cl)
    {
     ((LEAF_NODE*)this)->clauses = new RES_PAIR_LIST(ln,cl,((LEAF_NODE*)this)->clauses);
    };
   inline void AddFParaTripple(unsigned ln,Clause* cl,TERM t)
    {
     ((LEAF_NODE*)this)->clauses = new RES_PAIR_LIST(ln,cl,t,((LEAF_NODE*)this)->clauses);
    }; 
   inline void AddBParaTripple(unsigned ln,Clause* cl,TERM* t)
    {
     ((LEAF_NODE*)this)->clauses = new RES_PAIR_LIST(ln,cl,t,((LEAF_NODE*)this)->clauses);
    }; 
   inline void Remove(Clause* const c)
    {
     RES_PAIR_LIST::Remove(c,((LEAF_NODE*)this)->clauses);
    };
  public: 
   static void Destroy(OR_INDEX_TREE_NODE* tree);
  public: // methods for debugging
   static inline void OutputTag(TAG t);
   inline void Output(); 
   inline void Output(unsigned var_offset); 
 }; // class OR_INDEX_TREE_NODE

 
//********************** Definitions: **************************************** 

inline void OR_INDEX::Integrate(Clause* const cl)
 {
  TERM* lit = cl->Literals();
  unsigned num_of_max = lit->MaxLiterals();
  lit = lit->Args();
  for (unsigned lit_num = 0; lit_num < num_of_max; lit_num++)
   {
    Integrate(lit->First(),cl,lit_num);
    lit = lit->Next();
   };
 }; // void OR_INDEX::Integrate(Clause* const cl)

#undef DO

//#define DO(x) { if (cl->Number() == 6) { x; }; }
#define DO(x) {}

inline void OR_INDEX::Integrate(TERM* lit,Clause* cl,unsigned lit_num)
 {
  if (lit->Arity())
   {
    if (paramodulation) IntegrateForBackwardPara(lit,cl,lit_num);
    if (lit->IsEquality())
     {      
      if (lit->Positive()) IntegrateForForwardPara(lit,cl,lit_num);
     }
    else
     {
      OR_INDEX_TREE_NODE* leaf = IntegrateNonpropLit(lit);
      CHECK15(leaf->Term().IsReference(),"A9 in OR_INDEX::Integrate(TERM* lit,Clause* cl,unsigned lit_num)");
      CHECK15((!leaf->Term().First()) || (leaf->Term().First() == lit),"A10 in OR_INDEX::Integrate(TERM* lit,Clause* cl,unsigned lit_num)");
      leaf->SetTerm(lit);
      leaf->AddPair(lit_num,cl); 
     }; 
   }
  else
   {
    OR_INDEX_TREE_NODE* leaf = IntegratePropLit(*lit);    
    CHECK15(leaf->Term().IsReference(),"A19 in OR_INDEX::Integrate(TERM* lit,Clause* cl,unsigned lit_num)");
    CHECK15((!leaf->Term().First()) || (leaf->Term().First() == lit),"A20 in OR_INDEX::Integrate(TERM* lit,Clause* cl,unsigned lit_num)");
    leaf->SetTerm(lit);
    leaf->AddPair(lit_num,cl); 
   };  
 }; // void OR_INDEX::Integrate(TERM* lit,Clause* cl,unsigned lit_num)

inline void OR_INDEX::IntegrateBPRedex(TERM* redex,Clause* cl,unsigned lit_num)
 {
  CHECK15(redex->IsComplex(),"A0 in OR_INDEX::IntegrateBPRedex(..)");
  OR_INDEX_TREE_NODE** tree = &backward_para_tree;
  OR_INDEX_TREE_NODE* leaf;
  pseudo_literal->Args()->SetFirst(redex);
  l_iter.Reset(pseudo_literal,0);
  l_iter.Next();
  leaf = IntegrateArgs(tree);
  leaf->SetTerm(redex);
  leaf->AddBParaTripple(lit_num,cl,redex);  
 }; // void OR_INDEX::IntegrateBPRedex(TERM* redex,Clause* cl,unsigned lit_num)

inline void OR_INDEX::IntegrateForBackwardPara(TERM* lit,Clause* cl,unsigned lit_num)
 {
  CHECK15(lit->Arity(),"A0 in OR_INDEX::IntegrateForBackwardPara(..)");
  TERM::ITERATOR iter;
  if (lit->IsEquality() && lit->IsOrderedEq())
   {
    // only the first argument can be rewritten
    TERM* arg = lit->Args()->First();
    if (!arg->IsVariable())
     {
      iter.Reset(arg);
      do
       {
        if (!iter.CurrentSymIsVar()) IntegrateBPRedex(iter.CurrPos(),cl,lit_num); 
       }
      while (iter.Next());
     };
   }
  else // ordinary literal or unordered equality
   {
    iter.Reset(lit);
    while (iter.Next())
     if (!iter.CurrentSymIsVar()) IntegrateBPRedex(iter.CurrPos(),cl,lit_num);
   };
 }; // void OR_INDEX::IntegrateForBackwardPara(TERM* lit,Clause* cl,unsigned lit_num)

inline void OR_INDEX::IntegrateForForwardPara(TERM* lit,Clause* cl,unsigned lit_num)
 {
  CHECK15(lit->IsEquality(),"ARG1 in OR_INDEX::IntegrateForForwardPara(TERM* lit,Clause* cl,unsigned lit_num)");
  CHECK15(lit->Positive(),"ARG2 in OR_INDEX::IntegrateForForwardPara(TERM* lit,Clause* cl,unsigned lit_num)");
  
  OR_INDEX_TREE_NODE* leaf;
  // integrate the first argument
  TERM* arg = lit->Args();
  TERM replacing_term = *(arg->Next());
  bool stop = false;
  integrate_arg: 
   if (arg->IsVariable())
    {
     pseudo_literal[1] = *arg;
     l_iter.Reset(pseudo_literal,0);
     l_iter.Next();
     leaf = IntegrateArgs(&forward_para_from_var_tree);
     leaf->SetTerm(*arg);
     leaf->AddFParaTripple(lit_num,cl,replacing_term); 
    }
   else // complex term, integrate as ordinary literal
    {
     if (arg->First()->Arity())
      {
       l_iter.Reset(arg->First(),0);
       l_iter.Next();
       leaf = IntegrateArgs(FParaTree(*(arg->First())));
      }
     else
      {
       leaf = *(FParaTree(*(arg->First())));
       if (!leaf) 
        {
         leaf = OR_INDEX_TREE_NODE::NewLeaf();
         *(FParaTree(*(arg->First()))) = leaf;
        };
      }; 
     leaf->SetTerm(arg->First());
     leaf->AddFParaTripple(lit_num,cl,replacing_term);
    };    
   
  if (lit->IsOrderedEq() || stop) return; 
  // integrate the second argument 
  replacing_term = *arg;
  arg = arg->Next();
  stop = true;
  goto integrate_arg;
 }; // void OR_INDEX::IntegrateForForwardPara(TERM* lit,Clause* cl,unsigned lit_num)


inline OR_INDEX_TREE_NODE* OR_INDEX::IntegratePropLit(TERM lit)
 {
  OR_INDEX_TREE_NODE** tree = Tree(lit);
  if (!(*tree))
   {
    // no tree for this header yet, create it here
    *tree = OR_INDEX_TREE_NODE::NewLeaf();
   };
  return *tree;
 }; // OR_INDEX_TREE_NODE* OR_INDEX::IntegratePropLit(TERM lit)

inline OR_INDEX_TREE_NODE* OR_INDEX::IntegrateNonpropLit(TERM* lit)
 {
  OR_INDEX_TREE_NODE** tree = Tree(*lit);
  l_iter.Reset(lit,0);
  l_iter.Next();
  return IntegrateArgs(tree);
 }; // OR_INDEX_TREE_NODE* OR_INDEX::IntegrateNonpropLit(TERM* lit)
 
inline OR_INDEX_TREE_NODE* OR_INDEX::IntegrateArgs(OR_INDEX_TREE_NODE** tree)
 {
  var_counter.Reset();
  if (*tree)
   {
    OR_INDEX_TREE_NODE* node = *tree;
    PREF_SYM sym;
    take_next_sym:
     if (l_iter.CurrentSymIsVar())
      {
       int var_num = l_iter.Symbol().Var();
       sym.MkVar(VAR::Base(var_num));
       var_counter.Register(var_num);
      }
     else
      {
       sym.MkFunc(l_iter.Symbol());
      };
     check_node:
      if (node->Symbol() == sym)
       {
        node = node->Below();
        if (l_iter.Next())
         {
          goto take_next_sym;
         }
        else // node must be a leaf
         {
          CHECK15(node->IsLeaf(),"A20 in OR_INDEX::Integrate(TERM* lit)");
          return node;
         };
       }
      else 
       {
        if (node->Alternative())
         {
          node = node->Alternative();
          goto check_node;
         }
        else // no alternative, new branch must be constructed
         {
          OR_INDEX_TREE_NODE* leaf;
          node->SetAlternative(CreateBranch(leaf));
          return leaf;
         };
       };      
   }
  else // empty tree, create branch
   {
    OR_INDEX_TREE_NODE* leaf;
    *tree = CreateBranch(leaf);
    return leaf;
   };
 }; // OR_INDEX_TREE_NODE* OR_INDEX::IntegrateArgs(OR_INDEX_TREE_NODE** tree)





//**************************************************************************

inline OR_INDEX_TREE_NODE* OR_INDEX::CreateBranch(OR_INDEX_TREE_NODE*& leaf)
 {
  PREF_SYM sym;
  OR_INDEX_TREE_NODE::TAG tag;
  if (l_iter.CurrentSymIsVar())
   {
    if (var_counter.Occurences(l_iter.Symbol().Var()))
     {
      tag = OR_INDEX_TREE_NODE::VAR;
     }
    else // first occurence of the variable
     {
      var_counter.Register(l_iter.Symbol().Var());
      tag = OR_INDEX_TREE_NODE::VAR_FIRST_OCC;
     };
    sym.MkVar(VAR::Base(l_iter.Symbol().Var()));
   }
  else
   {
    if (l_iter.Symbol().Arity())
     {
      tag = OR_INDEX_TREE_NODE::FUNC;
     }
    else
     {
      tag = OR_INDEX_TREE_NODE::CONST;
     };
    sym.MkFunc(l_iter.Symbol());
   };
  OR_INDEX_TREE_NODE* result = OR_INDEX_TREE_NODE::NewNode(tag,sym,l_iter.Depth());
  OR_INDEX_TREE_NODE* node = result;
  while (l_iter.Next())
   {
    if (l_iter.CurrentSymIsVar())
     {
      if (var_counter.Occurences(l_iter.Symbol().Var()))
       {
        tag = OR_INDEX_TREE_NODE::VAR;
       }
      else // first occurence of the variable
       {
        var_counter.Register(l_iter.Symbol().Var());
        tag = OR_INDEX_TREE_NODE::VAR_FIRST_OCC;
       };
      sym.MkVar(VAR::Base(l_iter.Symbol().Var()));
     }
    else
     {
      if (l_iter.Symbol().Arity())
       {
        tag = OR_INDEX_TREE_NODE::FUNC;
       }
      else
       {
        tag = OR_INDEX_TREE_NODE::CONST;
       };
      sym.MkFunc(l_iter.Symbol());
     };
    node->SetBelow(OR_INDEX_TREE_NODE::NewNode(tag,sym,l_iter.Depth()));
    node = node->Below();
   };
  
  leaf = OR_INDEX_TREE_NODE::NewLeaf();
  node->SetBelow(leaf);
  return result;
 }; // OR_INDEX_TREE_NODE* OR_INDEX::CreateBranch(OR_INDEX_TREE_NODE*& leaf)






//********************** Auxilliary class declarations: ****************
 
class OR_INDEX_PATH
 {
  private:
   OR_INDEX_TREE_NODE** bookmark[MAX_TERM_WEIGHT];
   OR_INDEX_TREE_NODE*** next_bookmark;
   OR_INDEX_TREE_NODE* path[MAX_TERM_WEIGHT];
   OR_INDEX_TREE_NODE** next;
   OR_INDEX_TREE_NODE** changed_suffix;
   OR_INDEX_TREE_NODE* curr_node;
   unsigned var_offset;
   VAR* var_extracted;
  public:
   OR_INDEX_PATH() {};
   inline void SetIndex(const int index) 
    {
     var_offset = VAR::Offset(index); 
    };
   inline int GetIndex() const { return VAR::Index(var_offset); };
   inline VAR* Shift(VAR* const v) { return (v + var_offset); }; 
   inline void Reset(OR_INDEX_TREE_NODE* const tree)
    {
     CHECK15(tree,"null argument in OR_INDEX_PATH::Reset()");
     next = path;
     curr_node = tree;
     next_bookmark = bookmark;
    };
   static inline void ResetStatic() { VAR::Reset(); };
   inline OR_INDEX_TREE_NODE* CurrNode() const { return curr_node; };
   inline OR_INDEX_TREE_NODE** Position() const { return next; };
   inline OR_INDEX_TREE_NODE** ChangedSuffix() const { return changed_suffix; };
   inline void MakeBookmark() { *next_bookmark = Position(); next_bookmark++; };
   inline void ForgetBookmark() { next_bookmark--; };
   inline OR_INDEX_TREE_NODE** LastBookmark() const { return *(next_bookmark-1); };
   inline VAR* VarExtracted() const { return var_extracted; };
   inline void MemorizeCurrent() { *next = curr_node; next++; };
   inline void RestoreCurrent() { next--; curr_node = *next; };
   inline void Below() 
    {
     CHECK15(!(curr_node->IsLeaf()),"in OR_INDEX_PATH::Below()");
     MemorizeCurrent();
     curr_node = curr_node->Below();
    };
   inline void Alternative() 
    { 
     CHECK15(!(curr_node->IsLeaf()),"in OR_INDEX_PATH::Alternative()");
     curr_node = curr_node->Alternative();
    };
   inline void ExtractTerm()
    {
     MakeBookmark();
     if (curr_node->IsVar())
      {
       var_extracted = Shift(curr_node->Symbol().Var());
       Below();
      }
     else // complex term to be extracted
      {
       var_extracted = (VAR*)0;
       unsigned init_depth = curr_node->Depth();
       do 
        {
         Below();
        }
       while (init_depth < curr_node->Depth());
      };
    }; // void ExtractTerm()
   inline bool ModifyTerm();
   inline bool ExtractTermWithout(VAR* const var,bool& same_var);
   inline bool ModifyTermWithout(VAR* const var,
                                 bool& same_var);
   inline void ToPrefix(OR_INDEX_TREE_NODE** pos,PREFIX_SYM*& memory);
  public: // methods for debugging
   inline void Output() 
    {
     for (OR_INDEX_TREE_NODE** p = path; p<next; p++)
      {
       (*p)->Output(var_offset);
       ::Output("\n");
      };
     ::Output("CURRENT NODE = ");
     if (curr_node) { curr_node->Output(var_offset); } 
     else { ::Output("["); ::Output((long)0); ::Output("]"); };
     ::Output("\n");
    };
 }; // class OR_INDEX_PATH 

class OR_PROCESSOR : private OR_INDEX_PATH
 {
  private:
   PREFIX_SYM memory[MAX_TERM_WEIGHT];
   PREFIX_SYM* free_memory;
   PREFIX::ITER_WITH_BACKTRACK* term[MAX_TERM_WEIGHT];
   PREFIX::ITER_WITH_BACKTRACK** next_term;
   PREFIX::ITER_WITH_BACKTRACK* curr_iter;
   enum SITUATION
    {
     START_OF_EXTRACTION,
     VAR_SAME_VAR,
     VAR_ANYTHING,
     FUNCTOR_FUNCTOR,
     FUNCTOR_FREE_VAR,
     FUNCTOR_BOUND_VAR
    };
   SITUATION backtrack_situation[MAX_TERM_WEIGHT*2];
   SITUATION* curr_sit; 
  private:
   inline void NextIter() 
    {
     next_term++;
     curr_iter = *next_term;
     if (!curr_iter) 
      {
       curr_iter = new PREFIX::ITER_WITH_BACKTRACK(); 
       *next_term = curr_iter;
      };
     CHECK15(next_term - term <= MAX_TERM_WEIGHT,"in OR_PROCESSOR::NextIter()");
    };
   inline void ForgetIter() 
    {
     next_term--; 
     CHECK15(next_term >= term,"in OR_PROCESSOR::ForgetIter()");
     curr_iter = *next_term;
    };
   inline PREFIX::ITER_WITH_BACKTRACK* LastIter() { return *(next_term-1); };
  public:
   OR_PROCESSOR() : OR_INDEX_PATH() 
    {
     for (int i = 0; i < MAX_TERM_WEIGHT; i++)
      {
       term[i] = (PREFIX::ITER_WITH_BACKTRACK*)0;
      };
    };
   static inline void ResetStatic() { OR_INDEX_PATH::ResetStatic(); };
   inline void SetIndex(int index) { OR_INDEX_PATH::SetIndex(index); };
   inline int GetIndex() const { return OR_INDEX_PATH::GetIndex(); };
   inline OR_INDEX_TREE_NODE* CurrNode() const { return OR_INDEX_PATH::CurrNode(); };
   inline void Below() { OR_INDEX_PATH::Below(); };
   inline void Alternative() { OR_INDEX_PATH::Alternative(); };
   inline void RestoreCurrent() { OR_INDEX_PATH::RestoreCurrent(); };
   inline VAR* Shift(VAR* v) { return OR_INDEX_PATH::Shift(v); };
  public: // commands
   inline void Start(OR_INDEX_TREE_NODE* tree) 
    {
     free_memory = memory;
     curr_sit = backtrack_situation;
     next_term = term;
     OR_INDEX_PATH::Reset(tree); 
    };
   inline void Bind(VAR* v);
   inline bool AlterBinding();
   inline bool BindWithOccCheck(VAR* v,bool& same_var);
   inline bool AlterBindingWithOccCheck(VAR* v,bool& same_var);
  private:
   inline bool BacktrackExtraction();
   inline bool CompleteExtraction();
  public:
   inline bool ExtractUnifiable(PREFIX_SYM* word);
   inline bool AlterUnifiable();
  public: // methods for debugging
   inline void OutputPath() { OR_INDEX_PATH::Output(); };
   inline void OutputSit(SITUATION sit)
    {
     switch (sit)
      {
       case START_OF_EXTRACTION : ::Output("START_OF_EXTRACTION"); break;
       case VAR_SAME_VAR : ::Output("VAR_SAME_VAR"); break;
       case VAR_ANYTHING : ::Output("VAR_ANYTHING"); break;
       case FUNCTOR_FUNCTOR : ::Output("FUNCTOR_FUNCTOR"); break;
       case FUNCTOR_FREE_VAR : ::Output("FUNCTOR_FREE_VAR"); break;
       case FUNCTOR_BOUND_VAR : ::Output("FUNCTOR_BOUND_VAR"); break;
      };
    };
 }; // class OR_PROCESSOR : private OR_INDEX_PATH

//****************** Definitions: ********************************************

extern void OutputPrefix(void* word); // debug

inline bool OR_PROCESSOR::CompleteExtraction()
 {
  //static long call_num = 0; // debug

  //call_num++; // debug
  
  check_iter:
  
   if (curr_iter->Empty()) 
    {
     return true; 
    };
   
   check_sym:
   
   
    VAR* var = curr_iter->CurrVar();
    if (var) // free variable
     {
     
      CHECK15(!(var->Binding()),"C1 OR_PROCESSOR::CompleteExtraction()");
      bool same_var;
      
      if (BindWithOccCheck(var,same_var))
       {
       
        CHECK15(!(same_var && var->Binding()),"C2 OR_PROCESSOR::CompleteExtraction()"); 
        if (same_var) { *curr_sit = VAR_SAME_VAR; }
        else { *curr_sit = VAR_ANYTHING; };
        curr_sit++;
        curr_iter->SkipSym();
        goto check_iter;
       }
      else
       {
        if (BacktrackExtraction()) { goto check_sym; }
        else { return false; };
       };
     }
    else // functor
     {
     
      check_node:
       switch (OR_INDEX_PATH::CurrNode()->Tag())
        {
         case OR_INDEX_TREE_NODE::FUNC : 
          // currently the same as for OR_INDEX_TREE_NODE::CONST,
          // but can be optimized by jumping to check_sym instead of check_iter
          
         case OR_INDEX_TREE_NODE::CONST : 
         
          if (OR_INDEX_PATH::CurrNode()->Symbol() == curr_iter->CurrSubterm()->Head())
           { // same functor
            
            curr_iter->SkipSym();
            OR_INDEX_PATH::Below();
            *curr_sit = FUNCTOR_FUNCTOR;
            curr_sit++;
            goto check_iter;
           }
          else // different functors, try alternative
           {
            OR_INDEX_PATH::Alternative();
            if (OR_INDEX_PATH::CurrNode())
             {
              goto check_node;
             }
            else // backtrack needed
             {
              if (BacktrackExtraction()) { goto check_sym; }
              else { return false; };
             };
           };
         
         case OR_INDEX_TREE_NODE::VAR :
          {
           bool free_var;
           void* bind = OR_INDEX_PATH::Shift(OR_INDEX_PATH::CurrNode()->Symbol().Var())->Unref(free_var);
           PREFIX::CorrectBracket(curr_iter->CurrSubterm());
           
           CHECK15(PREFIX::CorrectComplexTerm(curr_iter->CurrSubterm()),
                  "A1 in OR_PROCESSOR::CompleteExtraction()");
           
           if (free_var)
            {
             if (OCCUR_CHECK::Occurs((VAR*)bind,curr_iter->CurrSubterm()))
              {
               // try alternative
               OR_INDEX_PATH::Alternative();
               if (OR_INDEX_PATH::CurrNode())
                {
                 goto check_node;
                }
               else // backtrack needed
                {
                 if (BacktrackExtraction()) { goto check_sym; }
                 else { return false; };
                };
              }
             else // the variable doesn't occur in the subterm, make binding
              {
               GlobalSubst.BindAlone((VAR*)bind,curr_iter->CurrSubterm());
               curr_iter->SkipTerm();
               OR_INDEX_PATH::Below();
               *curr_sit = FUNCTOR_FREE_VAR;
               curr_sit++;
               goto check_iter;
              };
            }
           else // variable bound by a complex term, try to unify
            {
             if (UNIFICATION::UnifyComplex((PREFIX_SYM*)bind,curr_iter->CurrSubterm()))
              {
               curr_iter->SkipTerm();
               OR_INDEX_PATH::Below();
               *curr_sit = FUNCTOR_BOUND_VAR;
               curr_sit++;
               goto check_iter;
              }
             else // try alternative
              {
               OR_INDEX_PATH::Alternative();
               if (OR_INDEX_PATH::CurrNode())
                {
                 goto check_node;
                }
               else // backtrack needed
                {
                 if (BacktrackExtraction()) { goto check_sym; }
                 else { return false; };
                };
              };
            };
          };
         case OR_INDEX_TREE_NODE::VAR_FIRST_OCC :
          PREFIX::CorrectBracket(curr_iter->CurrSubterm());
          
          CHECK15(PREFIX::CorrectComplexTerm(curr_iter->CurrSubterm()),
                 "A1 in in OR_PROCESSOR::CompleteExtraction()");
          
          GlobalSubst.BindAlone(OR_INDEX_PATH::Shift(OR_INDEX_PATH::CurrNode()->Symbol().Var()),
                           curr_iter->CurrSubterm());
          
          curr_iter->SkipTerm();
          OR_INDEX_PATH::Below();
          *curr_sit = FUNCTOR_FREE_VAR;
          curr_sit++;
          goto check_iter;         
         
         #ifdef _ALL_ENUM_VALUES_IN_SWITCHES_
         case OR_INDEX_TREE_NODE::LEAF : 
          RuntimeError("1 in OR_PROCESSOR::CompleteExtraction()");
          return false; 
         #endif
         #ifdef _DEBUG15_
         default : 
          DebugError("2 in OR_PROCESSOR::CompleteExtraction()");
          return false; 
         #endif 
        }; // switch
       RuntimeError("3 in OR_PROCESSOR::CompleteExtraction()"); return false;
     };  
 }; // bool OR_PROCESSOR::CompleteExtraction()

inline bool OR_PROCESSOR::BacktrackExtraction()
 {
  DEBUG15_OP(static long call_num = 0);
  DEBUG15_OP(call_num++);
  
  DCONTROL15("Z1 in OR_PROCESSOR::BacktrackExtraction()");
  
  rollback:
   CHECK15((!(OR_INDEX_PATH::CurrNode())) || OR_INDEX_PATH::CurrNode()->IsLeaf(),
          "D1 in OR_PROCESSOR::BacktrackExtraction()");
  
   curr_sit--;
   
   switch (*curr_sit)
    {
     case START_OF_EXTRACTION : 
      DCONTROL15("X1 in OR_PROCESSOR::BacktrackExtraction()");
      ForgetIter(); 
      return false;
     case VAR_SAME_VAR :
      DCONTROL15("X2 in OR_PROCESSOR::BacktrackExtraction()");
      {
       bool same_var;
       curr_iter->Backtrack();

       //cout << "SYMBOL "; curr_iter->CurrSubterm()->Head().Output(); cout << "\n"; // debug

       CHECK15(curr_iter->CurrVar(),"A1 in OR_PROCESSOR::BacktrackExtraction()");
       CHECK15(!(curr_iter->CurrVar()->Binding()),"B1 in OR_PROCESSOR::BacktrackExtraction()");
       
       if (AlterBindingWithOccCheck(curr_iter->CurrVar(),same_var)) 
        {
         //CHECK15(!(same_var),
         //       "two occurences of the same var, in OR_PROCESSOR::BacktrackExtraction()");
         if (!(same_var)) { *curr_sit = VAR_ANYTHING; };
         curr_sit++;
         curr_iter->SkipSym();
         return true; 
        }
       else 
        {
         DCONTROL15("Y1 in OR_PROCESSOR::BacktrackExtraction()");
         goto rollback; 
        };
      };
     case VAR_ANYTHING : 
      DCONTROL15("X3 in OR_PROCESSOR::BacktrackExtraction()");
      {
       bool same_var;
       curr_iter->Backtrack();
       
       //cout << "SYMBOL "; curr_iter->CurrSubterm()->Head().Output(); cout << "\n"; // debug
       
       if (AlterBindingWithOccCheck(curr_iter->CurrVar(),same_var)) 
        {
         if (same_var) 
          {
           *curr_sit = VAR_SAME_VAR; 
          };
         curr_sit++; 
         curr_iter->SkipSym();
         return true; 
        }
       else 
        {
         DCONTROL15("Y2 in OR_PROCESSOR::BacktrackExtraction()");
         goto rollback; 
        };
      };

     case FUNCTOR_FUNCTOR : 
      DCONTROL15("X4 in OR_PROCESSOR::BacktrackExtraction()");
      curr_iter->Backtrack();
      
      //cout << "SYMBOL "; curr_iter->CurrSubterm()->Head().Output(); cout << "\n"; // debug
      
      CHECK15(curr_iter->CurrSubterm()->Head().IsFunctor(),"A3 in OR_PROCESSOR::BacktrackExtraction()");

      OR_INDEX_PATH::RestoreCurrent();
      OR_INDEX_PATH::Alternative();
      if (OR_INDEX_PATH::CurrNode()) { return true; }
      else 
       {
        DCONTROL15("Y3 in OR_PROCESSOR::BacktrackExtraction()");
        goto rollback; 
       };

     case FUNCTOR_FREE_VAR :
      DCONTROL15("X5 in OR_PROCESSOR::BacktrackExtraction()");
      GlobalSubst.UnbindLast();
      curr_iter->Backtrack();

      CHECK15(curr_iter->CurrSubterm()->Head().IsFunctor(),"A4 in OR_PROCESSOR::BacktrackExtraction()");
      
      OR_INDEX_PATH::RestoreCurrent();
      OR_INDEX_PATH::Alternative();
      if (OR_INDEX_PATH::CurrNode()) { return true; }
      else 
       {
        DCONTROL15("Y4 in OR_PROCESSOR::BacktrackExtraction()");
        goto rollback; 
       };

     case FUNCTOR_BOUND_VAR :
      DCONTROL15("X6 in OR_PROCESSOR::BacktrackExtraction()");
      
      //if (call_num == 2300) // debug
       //{
        //GlobalSubst.Output(OutputPrefix); // debug
       //};
      
      GlobalSubst.UnbindLastSection();
      
      DCONTROL15("X6.1 in OR_PROCESSOR::BacktrackExtraction()");
      
      curr_iter->Backtrack();
      
      DCONTROL15("X6.2 in OR_PROCESSOR::BacktrackExtraction()");
      
      CHECK15(curr_iter->CurrSubterm()->Head().IsFunctor(),"A5 in OR_PROCESSOR::BacktrackExtraction()");
      
      OR_INDEX_PATH::RestoreCurrent();
      
      DCONTROL15("X6.3 in OR_PROCESSOR::BacktrackExtraction()");
      
      OR_INDEX_PATH::Alternative();
      
      DCONTROL15("X6.4 in OR_PROCESSOR::BacktrackExtraction()");
      
      if (OR_INDEX_PATH::CurrNode()) 
       {
        DCONTROL15("X6.5 in OR_PROCESSOR::BacktrackExtraction()");
        return true; 
       }
      else 
       {
        DCONTROL15("Y5 in OR_PROCESSOR::BacktrackExtraction()");
        goto rollback; 
       };
     #ifdef _DEBUG15_
     default : DebugError("1 in OR_PROCESSOR::BacktrackExtraction()"); return false;
     #endif
    };  
  RuntimeError("2 in OR_PROCESSOR::BacktrackExtraction()"); return false; 
 }; // bool OR_PROCESSOR::BacktrackExtraction()

inline bool OR_PROCESSOR::ExtractUnifiable(PREFIX_SYM* word) 
  // word must represent a complex term
 {
  NextIter();
  curr_iter->Reset(word);
  *curr_sit = START_OF_EXTRACTION;
  curr_sit++;
  return CompleteExtraction();
 }; // bool OR_PROCESSOR::ExtractUnifiable(PREFIX_SYM* word)
 
inline bool OR_PROCESSOR::AlterUnifiable()
 {
  DCONTROL15("A1 in OR_PROCESSOR::AlterUnifiable()");
 
  if (BacktrackExtraction()) 
   {
    DCONTROL15("A2 in OR_PROCESSOR::AlterUnifiable()");
    return CompleteExtraction(); 
   }
  else 
   {
    DCONTROL15("A3 in OR_PROCESSOR::AlterUnifiable()");
    return false; 
   };
 }; // bool OR_PROCESSOR::AlterUnifiable()

//****************************************************************************

inline bool OR_INDEX_PATH::ModifyTerm()
 {
  OR_INDEX_TREE_NODE** begin = LastBookmark();
   
  CHECK15(begin < Position(),"A1 in OR_INDEX_PATH::ModifyTerm()"); 
  
  rollback:
   RestoreCurrent();
   changed_suffix = Position();
   if (Position() == begin)  
    {
     Alternative();
     if (curr_node) { ForgetBookmark(); ExtractTerm(); return true; }
     else { ForgetBookmark(); return false; };
    };
   Alternative();
   if (curr_node)
    {
     unsigned init_depth = (*begin)->Depth();
     do 
      {
       Below();
      }
     while (init_depth < curr_node->Depth());
     var_extracted = (VAR*)0;
     return true;
    }
   else { goto rollback; };
 }; // bool OR_INDEX_PATH::ModifyTerm()

inline bool OR_INDEX_PATH::ExtractTermWithout(VAR* const var,bool& same_var)
 {
  MakeBookmark();
  same_var = false;
  check_node:
   switch (curr_node->Tag())
    {
     case OR_INDEX_TREE_NODE::FUNC:
      {
       var_extracted = (VAR*)0;
       unsigned init_depth = curr_node->Depth();
       OR_INDEX_TREE_NODE** init_pos = Position();
       Below();
       check_depth:
        if (curr_node->Depth() <= init_depth) { return true; };
        intern_check_node:
         switch (curr_node->Tag())
          {
           case OR_INDEX_TREE_NODE::FUNC: Below(); goto intern_check_node;
           case OR_INDEX_TREE_NODE::CONST: Below(); goto check_depth;
           case OR_INDEX_TREE_NODE::VAR: 
            {
             bool free_var;
             void* bind = Shift(curr_node->Symbol().Var())->Unref(free_var);
             if (free_var)
              {
               if ((VAR*)bind == var) { goto try_alternative; } 
               else { Below(); goto check_depth; };
              }
             else // variable bound by a complex term in the node
              {
               if (OCCUR_CHECK::Occurs(var,(PREFIX_SYM*)bind))
                { goto try_alternative; }
               else { Below(); goto check_depth; };
              };
            };
           case OR_INDEX_TREE_NODE::VAR_FIRST_OCC: Below(); goto check_depth;
           
           #ifdef _ALL_ENUM_VALUES_IN_SWITCHES_
           case OR_INDEX_TREE_NODE::LEAF: 
            RuntimeError("R1 in OR_INDEX_PATH::ExtractTermWithout()"); return false;
           #endif
           
           #ifdef _DEBUG15_
           default: DebugError("A1 in OR_INDEX_PATH::ExtractTermWithout()"); return false;
           #endif
          };
       try_alternative:
        Alternative();
        if (curr_node) { goto intern_check_node; }
        else // backtrack needed
         {
          RestoreCurrent();
          if (Position() == init_pos) 
           {
            Alternative();
            if (curr_node) { goto check_node; }
            else { ForgetBookmark(); return false; };
           }
          else { goto try_alternative; };
         };
      }; // case OR_INDEX_TREE_NODE::FUNC
     case OR_INDEX_TREE_NODE::CONST: 
      Below(); 
      var_extracted = (VAR*)0;
      return true;
     case OR_INDEX_TREE_NODE::VAR: 
      {
       bool free_var;
       void* bind = Shift(curr_node->Symbol().Var())->Unref(free_var);
       if (free_var)
        {
         var_extracted = Shift(curr_node->Symbol().Var());
         Below();
         same_var = (((VAR*)bind) == var);
         return true;
        }
       else
        {
         if (OCCUR_CHECK::Occurs(var,(PREFIX_SYM*)bind))
          {
           Alternative();
           if (curr_node) { goto check_node; }
           else { ForgetBookmark(); return false; };
          }
         else
          {
           var_extracted = Shift(curr_node->Symbol().Var());
           Below();
           return true;
          };
        };
      };
     case OR_INDEX_TREE_NODE::VAR_FIRST_OCC:
      var_extracted = Shift(curr_node->Symbol().Var());
      Below(); 
      return true;
     #ifdef _ALL_ENUM_VALUES_IN_SWITCHES_
     case OR_INDEX_TREE_NODE::LEAF :  
      RuntimeError("R2 in OR_INDEX_PATH::ExtractTermWithout()"); return false;
     #endif
      
     #ifdef _DEBUG15_
     default: DebugError("A2 in OR_INDEX_PATH::ExtractTermWithout()"); return false;
     #endif
    };
  RuntimeError("in OR_INDEX_PATH::ExtractTermWithout()"); return false;  
 }; // bool OR_INDEX_PATH::ExtractTermWithout(VAR* var,bool& same_var)

inline bool OR_INDEX_PATH::ModifyTermWithout(VAR* const var,bool& same_var)
 {
  OR_INDEX_TREE_NODE** begin = LastBookmark();
  
  CHECK15(begin < Position(),"A1 in OR_INDEX_PATH::ModifyTermWithout()");
  
  unsigned init_depth = (*begin)->Depth();
  
  same_var = false; 
  changed_suffix = Position()-1;
  rollback:
   RestoreCurrent();
   if (Position() < changed_suffix) { changed_suffix = Position(); };
   if (Position() == begin)  
    {
     Alternative();
     ForgetBookmark();
     return ((curr_node) && (ExtractTermWithout(var,same_var)));
    };
   try_alternative:
    Alternative();
    if (curr_node)
     {
      check_node:
       switch (curr_node->Tag())
        {
         case OR_INDEX_TREE_NODE::FUNC: Below(); goto check_node;
         case OR_INDEX_TREE_NODE::CONST: Below(); goto check_depth;
         case OR_INDEX_TREE_NODE::VAR:
          {
           bool free_var;
           void* bind = Shift(curr_node->Symbol().Var())->Unref(free_var);
           if (free_var)
            {
             if ((VAR*)bind == var) { goto try_alternative; } 
             else { Below(); goto check_depth; };
            }
           else // variable bound by a complex term in the node
            {
             if (OCCUR_CHECK::Occurs(var,(PREFIX_SYM*)bind))
              { goto try_alternative; }
             else { Below(); goto check_depth; };
            };
          };
         case OR_INDEX_TREE_NODE::VAR_FIRST_OCC: Below(); goto check_depth;
         #ifdef _ALL_ENUM_VALUES_IN_SWITCHES_
         case OR_INDEX_TREE_NODE::LEAF :  
          RuntimeError("R3 in OR_INDEX_PATH::ModifyTermWithout()"); return false;
         #endif
         
         #ifdef _DEBUG15_
         default: DebugError("in OR_INDEX_PATH::ModifyTermWithout()"); return false;
         #endif
        };
       RuntimeError("in OR_INDEX_PATH::ModifyTermWithout()"); return false;
      check_depth:
       if (curr_node->Depth() <= init_depth) { var_extracted = (VAR*)0; return true; }
       else 
        {
         goto check_node; 
        };
     }
    else { goto rollback; };
 }; // bool OR_INDEX_PATH::ModifyTermWithout(VAR* var,.......

inline void OR_INDEX_PATH::ToPrefix(OR_INDEX_TREE_NODE** pos,PREFIX_SYM*& memory)
 {
  OR_INDEX_TREE_NODE* node;
  while (pos < Position())
   {
    node = *pos;
    switch (node->Tag())
     {
      case OR_INDEX_TREE_NODE::FUNC:
       memory->Head() = node->Symbol();
       memory->ResetBracket();
       break;
      case OR_INDEX_TREE_NODE::CONST:
       memory->Head() = node->Symbol();
       memory->SetBracket(memory+1);
       break; 
      case OR_INDEX_TREE_NODE::VAR: 
      case OR_INDEX_TREE_NODE::VAR_FIRST_OCC: 
       memory->Head().MkVar(Shift(node->Symbol().Var()));
       memory->SetBracket(memory+1); 
       break;
      #ifdef _ALL_ENUM_VALUES_IN_SWITCHES_
       case OR_INDEX_TREE_NODE::LEAF :
        RuntimeError("in OR_INDEX_PATH::ToPrefix()"); return;
      #endif
      #ifdef _DEBUG15_
      default: DebugError("in OR_INDEX_PATH::ToPrefix()"); return; 
      #endif
     };
    pos++;
    memory++;   
   }; 
 }; // void OR_INDEX_PATH::ToPrefix(OR_INDEX_TREE_NODE** pos,PREFIX_SYM*& memory)

inline void OR_PROCESSOR::Bind(VAR* v)
 {
  OR_INDEX_PATH::ExtractTerm();
  if (OR_INDEX_PATH::VarExtracted())
   {
    GlobalSubst.BindAlone(v,(void*)(OR_INDEX_PATH::VarExtracted()));
   }
  else // complex term extracted
   {
    PREFIX_SYM* word = free_memory;
    OR_INDEX_PATH::ToPrefix(OR_INDEX_PATH::LastBookmark(),free_memory);
    word->SetBracket(free_memory);
    GlobalSubst.BindAlone(v,(void*)word);
   };
 }; // void OR_PROCESSOR::Bind(VAR* v)
    
inline bool OR_PROCESSOR::AlterBinding()
 {
  DCONTROL15("Z1 in OR_PROCESSOR::AlterBinding()");
  VAR* v = GlobalSubst.LastBound();
  void* bind = v->Binding();
  
  CHECK15(bind,"A1 in OR_PROCESSOR::AlterBinding()");

  DCONTROL15("Z2 in OR_PROCESSOR::AlterBinding()");
  
  GlobalSubst.UnbindLast();
  
  DCONTROL15("Z3 in OR_PROCESSOR::AlterBinding()");
  
  if (OR_INDEX_PATH::ModifyTerm())
   {
    if (OR_INDEX_PATH::VarExtracted())
     {
      GlobalSubst.BindAlone(v,(void*)(OR_INDEX_PATH::VarExtracted()));
      if (!(VAR::IsVar(bind))) 
       {
        free_memory = (PREFIX_SYM*)bind; 
       };
     }
    else // complex term extracted
     {
      if (VAR::IsVar(bind))
       {
        PREFIX_SYM* word = free_memory;
        OR_INDEX_PATH::ToPrefix(OR_INDEX_PATH::LastBookmark(),free_memory);
        word->SetBracket(free_memory);
        GlobalSubst.BindAlone(v,(void*)word);
       }
      else
       {
        PREFIX_SYM* changed_suffix = (PREFIX_SYM*)bind + (OR_INDEX_PATH::ChangedSuffix() - OR_INDEX_PATH::LastBookmark());
        free_memory = changed_suffix;
        CHECK15(free_memory >= ((PREFIX_SYM*)bind),"A2 in OR_PROCESSOR::AlterBinding()");
        
        PREFIX_SYM* old_end_of_word = ((PREFIX_SYM*)bind)->Bracket();
        OR_INDEX_PATH::ToPrefix(OR_INDEX_PATH::ChangedSuffix(),free_memory);
        
        CHECK15(free_memory > ((PREFIX_SYM*)bind),"A3 in OR_PROCESSOR::AlterBinding()");
        
        ((PREFIX_SYM*)bind)->SetBracket(free_memory);
        PREFIX::ResetCorruptedBrackets(((PREFIX_SYM*)bind)+1,changed_suffix,old_end_of_word,free_memory);
        GlobalSubst.BindAlone(v,bind);
       };
     }
    return true;
   }
  else 
   {
    if (!(VAR::IsVar(bind))) { free_memory = (PREFIX_SYM*)bind; };
    return false; 
   };
 }; // bool OR_PROCESSOR::AlterBinding()
    
inline bool OR_PROCESSOR::BindWithOccCheck(VAR* v,bool& same_var)
 {
  if (OR_INDEX_PATH::ExtractTermWithout(v,same_var))
   {
    if (same_var) { return true; };
    if (OR_INDEX_PATH::VarExtracted())
     {
      GlobalSubst.BindAlone(v,(void*)(OR_INDEX_PATH::VarExtracted())); 
     }
    else // complex term extracted
     {
      PREFIX_SYM* word = free_memory;
      CHECK15(free_memory,"null free_memory in OR_PROCESSOR::BindWithOccCheck()");
      
      OR_INDEX_PATH::ToPrefix(OR_INDEX_PATH::LastBookmark(),free_memory);
      word->SetBracket(free_memory);
      
      CHECK15(PREFIX::CorrectComplexTerm(word),"in OR_PROCESSOR::BindWithOccCheck()");
      
      if (OCCUR_CHECK::Occurs(v,word)) // debug
       {
        v->Output(); cout << " ----- ";
        PREFIX::OutputTerm(word); cout << "\n"; 
        OutputPath(); 
       };
      
      CHECK15(!(OCCUR_CHECK::Occurs(v,word)),"cycle in OR_PROCESSOR::BindWithOccCheck()");
      
      GlobalSubst.BindAlone(v,(void*)word);
     };
    return true;
   }
  else 
   {
    return false;
   };
 }; // bool OR_PROCESSOR::BindWithOccCheck(VAR* v,bool& same_var)
 
inline bool OR_PROCESSOR::AlterBindingWithOccCheck(VAR* v,bool& same_var)
 {
  void* bind = v->Binding();
  bool was_var = true;
  if (bind) 
   {
    // here we suppose that v is the last bound variable
    CHECK15(v == GlobalSubst.LastBound(),"B1 in OR_PROCESSOR::AlterBindingWithOccCheck()");
    GlobalSubst.UnbindLast(); 
    if (!(VAR::IsVar(bind))) { was_var = false; free_memory = (PREFIX_SYM*)bind; };
   };
  if (OR_INDEX_PATH::ModifyTermWithout(v,same_var))
   {
    if (OR_INDEX_PATH::VarExtracted())
     {
      if (!(same_var)) { GlobalSubst.BindAlone(v,(void*)(OR_INDEX_PATH::VarExtracted())); };
     }
    else // complex term extracted
     {
      if (was_var)
       {
        PREFIX_SYM* word = free_memory;
        OR_INDEX_PATH::ToPrefix(OR_INDEX_PATH::LastBookmark(),free_memory);
        CHECK15(free_memory,"A2 in OR_PROCESSOR::AlterBindingWithOccCheck()");
        
        word->SetBracket(free_memory);

        CHECK15(PREFIX::CorrectComplexTerm(word),"1 in OR_PROCESSOR::AlterBindingWithOccCheck()");
        CHECK15(!(OCCUR_CHECK::Occurs(v,word)),"cycle1 in OR_PROCESSOR::AlterBindingWithOccCheck()");
        
        GlobalSubst.BindAlone(v,(void*)word);
       }
      else
       {
        PREFIX_SYM* changed_suffix = (PREFIX_SYM*)bind + (OR_INDEX_PATH::ChangedSuffix() - OR_INDEX_PATH::LastBookmark());
        free_memory = changed_suffix;
        
        CHECK15(free_memory,"A3 in OR_PROCESSOR::AlterBindingWithOccCheck()");
        
        PREFIX_SYM* old_end_of_word = ((PREFIX_SYM*)bind)->Bracket();
        OR_INDEX_PATH::ToPrefix(OR_INDEX_PATH::ChangedSuffix(),free_memory);
        CHECK15(free_memory,"A4 in OR_PROCESSOR::AlterBindingWithOccCheck()");
        
        ((PREFIX_SYM*)bind)->SetBracket(free_memory);
        PREFIX::ResetCorruptedBrackets(((PREFIX_SYM*)bind)+1,changed_suffix,old_end_of_word,free_memory);
       
        CHECK15(PREFIX::CorrectComplexTerm((PREFIX_SYM*)bind),"2 in OR_PROCESSOR::AlterBindingWithOccCheck()");
        CHECK15(!(OCCUR_CHECK::Occurs(v,(PREFIX_SYM*)bind)),"cycle2 in OR_PROCESSOR::AlterBindingWithOccCheck()");

        GlobalSubst.BindAlone(v,bind);
       };
     }
    return true;
   }
  else 
   {
    return false; 
   };
 }; // bool OR_PROCESSOR::AlterBindingWithOccCheck(VAR* v,bool& same_var)

//*****************  Methods for debugging ****************

inline void OR_INDEX_TREE_NODE::OutputTag(TAG t)
 {
  switch (t)
   {
    case LEAF: ::Output("LEAF"); break;
    case FUNC: ::Output("FUNC"); break;
    case CONST: ::Output("CONST"); break;
    case VAR: ::Output("VAR"); break;
    case VAR_FIRST_OCC: ::Output("VAR_FIRST_OCC"); break;
    default: ::Output("UNCKNOWN TAG"); 
   };
 };

inline void OR_INDEX_TREE_NODE::Output()
 {
  ::Output("["); ::Output((long)this); ::Output("] ");
  OutputTag(Tag());
  ::Output(" ");
  switch (Tag())
   {
    case LEAF: 
     ::Output(" ");
     ::Output(Term());
     //::Output(" {");
     //::Output((long)(Clauses()));
     //::Output("}");
     ::Output(" ");
     Clauses()->Output();
     break;
    case FUNC: 
    case CONST:
    case VAR: 
    case VAR_FIRST_OCC: 
     Symbol().Output();
     ::Output(" [");
     ::Output((long)(Below()));
     ::Output("]");
     ::Output(" [");
     ::Output((long)(Alternative()));
     ::Output("] ");
     ::Output((short unsigned int)Depth());
     break;
    default: break; 
   };
  
 }; // void OR_INDEX_TREE_NODE::Output()

inline void OR_INDEX_TREE_NODE::Output(unsigned var_offset)
 {
  ::Output("["); ::Output((long)this); ::Output("] ");
  OutputTag(Tag());
  ::Output(" ");
  switch (Tag())
   {
    case LEAF: 
     ::Output(" ");
     ::Output(Term());
     ::Output(" {");
     ::Output((long)(Clauses()));
     ::Output("}");
     break;
    case FUNC: 
    case CONST:
    case VAR: 
    case VAR_FIRST_OCC: 
     Symbol().Output(var_offset);
     ::Output(" [");
     ::Output((long)(Below()));
     ::Output("]");
     ::Output(" [");
     ::Output((long)(Alternative()));
     ::Output("] ");
     ::Output((short unsigned int)Depth());
     break;
    default: break; 
   };
 }; // void OR_INDEX_TREE_NODE::Output()



//********************************************************************************
