#define __OR_PREMISE_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif 
#ifndef __CLAUSE_H
 #include "clause.h"
#endif
#ifndef __PREFIX_H
 #include "prefix.h" 
#endif
#ifndef __OR_INDEX_H
 #include "or_index.h"
#endif
#ifndef __RESOLVENT_H
 #include "resolvent.h"
#endif

extern RESOLVENT Resolvent; 

//***************************************************************************

class OR_PREMISE
 {
  protected: // structure
   Clause* clause;
   int index;
   
   PREFIX_SYM memory[MAX_TERM_WEIGHT*MAX_NUMBER_OF_LITERALS];
   PREFIX_SYM* free_memory;
   
   PREFIX_SYM* max_literal[MAX_NUMBER_OF_LITERALS];
   int last_max;
   bool mark[MAX_NUMBER_OF_LITERALS];

   TERM* nonmax_literals;
   
   int cut_lit;
   bool nonmarkedExist;

  public: // methods
   OR_PREMISE() {};
   inline PREFIX_SYM** MaxLiterals() { return max_literal; };
   inline TERM* NonmaxLiterals() { return nonmax_literals; };
   inline Clause* TheClause() { return clause; };
   inline bool* Marks() { return mark; };
   inline int LastMax() { return last_max; };
   inline int Index() { return index; };
   inline int CutLit() { return cut_lit; };
   inline void Load(int ind,Clause* c);
   inline void LoadOptimized(int ind,Clause* c);
   inline void Unload() { free_memory = memory; };
   inline void Cut(int lit) 
    {
     cut_lit = lit;
     mark[cut_lit] = true;
    };
   inline void Load(int ind,RES_PAIR_LIST* res_pair)
    {
     Load(ind,res_pair->TheClause());
     Cut((int)(res_pair->LitNum()));
    };
   inline void LoadOptimized(int ind,RES_PAIR_LIST* res_pair)
    {
     LoadOptimized(ind,res_pair->TheClause());
     Cut((int)(res_pair->LitNum()));
    };
   inline void Cut(TERM* lit) { Cut(clause->NumberOfLiteral(lit)); };
   inline void UnmarkCut() { mark[cut_lit] = false; };
   inline bool CollectNonmarkedLiterals();
   inline bool NonmarkedExist() { return nonmarkedExist; };
  public: // output for debugging
   inline void Output();
 }; // class OR_PREMISE
 
//******************* Definitions: *************************************

inline void OR_PREMISE::Load(int ind,Clause* c)
 {
  free_memory = memory;
  clause = c;
  index = ind;

  TERM* lit = clause->Literals();
  int max = lit->MaxLiterals();
  last_max = max - 1;
  lit = lit->Args();
  int max_lit = 0;
  while (max)
   {
    max_literal[max_lit] = free_memory;
    mark[max_lit] = false;
    StandardTermToPrefix(ind,lit->First(),free_memory);
    lit = lit->Next();
    max--;
    max_lit++;
   };
  nonmax_literals = lit;
 }; // void OR_PREMISE::Load(int ind,Clause* c) 

inline void OR_PREMISE::LoadOptimized(int ind,Clause* c)
 {
  // if there is only one maximal literal we do not transform it
  // into prefix representation
  free_memory = memory;
  clause = c;
  index = ind;

  TERM* lit = clause->Literals();
  int max = lit->MaxLiterals();
  last_max = max - 1;
  if (last_max)
   {
    lit = lit->Args();
    int max_lit = 0;
    while (max)
     {
      max_literal[max_lit] = free_memory;
      mark[max_lit] = false;
      StandardTermToPrefix(ind,lit->First(),free_memory);
      lit = lit->Next();
      max--;
      max_lit++;
     };
    nonmax_literals = lit;
   }
  else
   {
    max_literal[0] = (PREFIX_SYM*)0; // just in case
    mark[0] = false; // just in case
    nonmax_literals = lit->Args()->Next();
   };
 }; // void OR_PREMISE::LoadOptimized(int ind,Clause* c) 

inline bool OR_PREMISE::CollectNonmarkedLiterals()
 {
  int currMax = 0;
  nonmarkedExist = !(nonmax_literals->IsNull());
  
  DCONTROL11("A3 in OR_PREMISE::CollectNonmarkedLiterals()");
  
  while (currMax <= last_max)
   {
    DCONTROL11("A4 in OR_PREMISE::CollectNonmarkedLiterals()");
    
    if (!(mark[currMax]))
     {
      nonmarkedExist = true;
      
      DCONTROL11("A10 in OR_PREMISE::CollectNonmarkedLiterals()");
      if (!(Resolvent.PushPrefixLit(max_literal[currMax]))) 
       {
        DCONTROL11("END1 in OR_PREMISE::CollectNonmarkedLiterals()");
        return false; 
       };
      DCONTROL11("A11 in OR_PREMISE::CollectNonmarkedLiterals()");
     };
    currMax++;
   };
  
  DCONTROL11("A17 in OR_PREMISE::CollectNonmarkedLiterals()");
   
  TERM* nonMax = nonmax_literals;
  while (!(nonMax->IsNull()))
   {
    DCONTROL11("A20 in OR_PREMISE::CollectNonmarkedLiterals()");
    if (!(Resolvent.PushStandardLit(index,nonMax->First()))) 
     {
      DCONTROL11("END2 in OR_PREMISE::CollectNonmarkedLiterals()");
      return false; 
     }; 
    DCONTROL11("A21 in OR_PREMISE::CollectNonmarkedLiterals()"); 
    nonMax = nonMax->Next();
   };
  DCONTROL11("END3 in OR_PREMISE::CollectNonmarkedLiterals()"); 
  return true;
 }; // inline bool PREMISE::CollectNonmarkedLiterals()

inline void OR_PREMISE::Output()
 {
  ::Output("<<<<<< PREMISE >>>>>>>\n");
  ::Output(" Clause : ");
  ::Output(clause);
  ::Output("\n Index = ");
  ::Output(index);
  ::Output("\n Maximal literals: \n");
  int ml = 0;
  while (ml <= last_max)
   {
    ::Output("  [");
    ::Output((int)(mark[ml]));
    ::Output("] ");
    PREFIX::OutputTerm(max_literal[ml]);
    ::Output("\n");
    ml++;
   };
 }; // void OR_PREMISE::Output() 
 
 
//*********************************************************************** 