
#ifndef __SIGNATURE_H
 #include "signature.h"
#endif 

extern SIGNATURE Signature;

#include <fstream.h>
#include <string.h>



//#include "formula.h"

#ifndef __CLAUSE_H
 #include "clause.h"
#endif

ofstream file;
ofstream tab_file;

bool output_to_file_allowed = false;
bool output_to_stdout_allowed = true; 

void AdjustOutput()
 {
  output_to_file_allowed = output_to_file;
  if ((tabulate_to_file) && (!(strcmp(output_file_name,tabulation_file_name))))
   {
    output_to_file_allowed = false;
   };
  output_to_stdout_allowed = true; // temporary
  if ((output_to_file_allowed) && (file))
   {
    file.open(output_file_name,ios::app);
   };
  if (tabulate_to_file) 
   {
    tab_file.open(tabulation_file_name,ios::app);
   };
 };

void CloseOutput()
 {
  if (!file) { file.close(); };
  if (!tab_file) 
   {
    tab_file.close(); 
   };
 };

// for printing variables
const char var_symbol = 'X';

void OutputArgs(const TERM* ts);  // to output arguments of a term
void OutputArg(const TERM* ts,int pr);  // to output one argument of a term


void Output(const char* s)
 {
  if (output_to_stdout_allowed) { cout << s; };
  if (output_to_file_allowed) { file << s; };
 };

void Tabulate(const char* s,int len)
 {
  tab_file.width(len);
  if (tabulate_to_file) { tab_file << s; }; 
 };

void Output(char s)
{
  if (output_to_stdout_allowed) { cout << s; };
  if (output_to_file_allowed) { file << s; };
}

void Tabulate(char s,int len)
 {
  tab_file.width(len);
  if (tabulate_to_file) { tab_file << s; }; 
 };

void Output(void* p)
{
  if (output_to_stdout_allowed) { cout << p; };
  if (output_to_file_allowed) { file << p; };
}


void Output(int s)
{
  if (output_to_stdout_allowed) { cout << s; };
  if (output_to_file_allowed) { file << s; };
}

void Tabulate(int s,int len)
 {
  tab_file.width(len);
  if (tabulate_to_file) { tab_file << s; }; 
 };

void Output(long s)
{
  if (output_to_stdout_allowed) { cout << s; };
  if (output_to_file_allowed) { file << s; };
}

void Tabulate(long s,int len)
 {
  tab_file.width(len);
  if (tabulate_to_file) { tab_file << s; }; 
 };

void Output(unsigned short s)
{
  if (output_to_stdout_allowed) { cout << s; };
  if (output_to_file_allowed) { file << s; };
}

void Tabulate(unsigned short s,int len)
 {
  tab_file.width(len);
  if (tabulate_to_file) { tab_file << s; }; 
 };

void Output(unsigned long s)
{
  if (output_to_stdout_allowed) { cout << s; };
  if (output_to_file_allowed) { file << s; };
}

void Tabulate(unsigned long s,int len)
 {
  tab_file.width(len);
  if (tabulate_to_file) { tab_file << s; }; 
 };

void Output(unsigned s)
{
  if (output_to_stdout_allowed) { cout << s; };
  if (output_to_file_allowed) { file << s; };
}

void Tabulate(unsigned s,int len)
 {
  tab_file.width(len);
  if (tabulate_to_file) { tab_file << s; }; 
 };

void Output(float s)
 {
  if (output_to_stdout_allowed) { cout << s; };
  if (output_to_file_allowed) { file << s; };
 };

void Tabulate(float s,int len)
 {
  tab_file.width(len);
  if (tabulate_to_file) { tab_file << s; }; 
 };

void Output(double s)
 {
  if (output_to_stdout_allowed) { cout << s; };
  if (output_to_file_allowed) { file << s; };
 };
 
void Tabulate(double s,int len)
 {
  tab_file.width(len);
  if (tabulate_to_file) { tab_file << s; }; 
 };

void TabulateFileName(char* fn)
 {
  int len = strlen(fn);
  char* begin = fn+len-1;
  while ((begin > fn) && ((*begin) != '/')) { begin--; };
  if (*begin == '/') { begin++; };
  char s[tabulated_file_name_width+1];
  if (strlen(begin) > (unsigned long)tabulated_file_name_width)
   {
    int i = 0;
    while (i < tabulated_file_name_width-2)
     {
      s[i] = begin[i];
      i++;
     };
    s[i] = '.';
    s[i+1] = '.';
    s[i+2] = (char)0;
    begin = s;
   };
  tab_file.setf(ios::left);
  Tabulate(begin,tabulated_file_name_width+1);
 };

void TabulateTime(float t)
 {
  tab_file.setf(ios::right);
  Tabulate(t,10);
  tab_file.setf(ios::left);
  Tabulate("s",3);
 };

void TabulateMemory(long mem)
 {
  tab_file.setf(ios::right);
  Tabulate(mem,8);
  tab_file.setf(ios::left);
  Tabulate("Kb",4);
 };

void TabulateChar(char c)
 {
  tab_file.setf(ios::left);
  Tabulate(c,2);
 };
 
void TabulateString(const char* const str)
 {
  tab_file.setf(ios::left);
  Tabulate(str,strlen(str)+1);
 };
 
void TabulateLong(long s)
 {
  tab_file.setf(ios::right);
  if (s == MAXLONG)
   {
    Tabulate("       MAXLONG",14);
   }
  else
   {
    Tabulate(s,14);
   };
 }; 

void OutputVar(int var)
 {
  Output("X");
  Output(var); 
 };

void OutputFunctor(int func)
 {
  Output(Signature.FunName(func));
  //Output((TABLE::Symbol(func)).name);
 }; 

void Output(const TERM& t)
 {
  if (t.IsReference()) { Output(t.First()); }
  else
   {
    if (t.IsVariable()) { OutputVar(t.Var()); }
    else 
     {
      if (t.Polarity() == POL_NEG) 
       {
        if (IsEquality(t.Functor()))
         {
          Output("!");
         }
        else
         {  
          Output("~");
         }; 
       };
      OutputFunctor(t.Functor()); 
     };
   };
 };  

void OutputEqLit(const TERM* lit)
 {
  Output(*(lit->Args()));
  Output(" ");
  Output(*lit); 
  Output(" ");
  Output(*(lit->Args()->Next()));
 };

void Output(const TERM* term,int pr)	// to output atoms, not terms
 {
  pr++; // to surpress warnings
 
  if (IsEquality(term->Functor()))
   {
    OutputEqLit(term);
    return;
   };
   
  const char* fun_name = Signature.FunName(term->Functor());
  Output(fun_name);

  /************ debug **********
  Output("["); 
  //Output(term->Functor()); 
  Output(term->Content());
  Output("]"); 
  *****************************/
  
  if (term->Arity())
   {
    //Output("(");
    OutputArgs(term->Args());
    //Output(")");
   };
  
  /*****************************************************************
  SYMBOL& symb = TABLE::Symbol(term->Functor());
  char* name = symb.name;
  int priority = symb.priority;
  
  // flag showing whether parentheses should be printed
  int parentheses = pr < priority ? 1 : 0;

  if (parentheses)
    Output('(');

  switch (symb.mode)
    {
    case OP_CONS:
    case OP_NO:
      Output(name);
      OutputArgs(term->Args());
      break;

    case OP_XFX:
      OutputArg(term->Args(),priority-1);
      Output(' ');
      Output(name);
      Output(' ');
      OutputArg(term->Args()->Next(),priority-1);
      break;

    case OP_XFY:
      OutputArg(term->Args(),priority-1);
      Output(' ');
      Output(name);
      Output(' ');
      OutputArg(term->Args()->Next(),priority);
      break;

    case OP_YFX:
      OutputArg(term->Args(),priority);
      Output(' ');
      Output(name);
      Output(' ');
      OutputArg(term->Args()->Next(),priority-1);
      break;

    case OP_XF:
      OutputArg(term->Args(),priority-1);
      Output(' ');
      Output(name);
      break;

    case OP_YF:
      OutputArg(term->Args(),priority);
      Output(' ');
      Output(name);
      break;

    case OP_FX:
      Output(name);
      Output(' ');
      OutputArg(term->Args(),priority-1);
      break;

    case OP_FY:
      Output(name);
      Output(' ');
      OutputArg(term->Args(),priority);
      break;

    }

  if (parentheses)
    Output(')');
 ****************************************************************/
}  // Output(const TERM*,PRIORITY)


void Output(const TERM* t)	// to output literals, not terms
{

 if (IsEquality(t->Functor()))
   {
    OutputEqLit(t);
    return;
   };
   
  if (t->Polarity() == POL_NEG)
    Output('~');
  Output(t,PR_HIGH);
}


void OutputTerm(const TERM* t,int priority)  // to output terms, temporarily the same as for literals
{
  Output(t,priority);
}


void OutputTerm(const TERM* t)  
{
 if (t->IsVariable())
  {
   OutputVar(t->Var());
  }
 else
  {
   if (t->IsReference())
    {
     Output(t->First());
    }
   else
    {
     Output(t);
    };
  };
}

int PriorityOf(int connective)
{
  switch (connective)
    {
    case AND:
      return PR_AND;
    case OR:
      return PR_OR;
    case IMP:
      return PR_IMP;
    case IFF:
      return PR_IFF;
    case NOT:
      return PR_NOT;
    case ALL:
      return PR_ALL;
    case SOME:
      return PR_SOME;
    case BOX:
      return PR_BOX;
    case DIAM:
      return PR_DIAM;
    case ATOMIC:
      return PR_MIN;
    }

  // to supress warnings
  return 0;
}  // PriorityOf()


/***********
void Output(const FORMULA* f,int priority)
{
  int connective = f->Connective();
  int new_priority = PriorityOf(connective);
  int print_par = new_priority > priority;

  if (print_par)
    Output('(');

  switch (connective)
    {
    case AND:
      Output(f->LeftArg(),new_priority);
      Output(" & ");
      Output(f->RightArg(),new_priority-1);
      break;

    case OR:
      Output(f->LeftArg(),new_priority);
      Output(" \\/ ");
      Output(f->RightArg(),new_priority-1);
      break;

    case IMP:
      Output(f->LeftArg(),new_priority-1);
      Output(" => ");
      Output(f->RightArg(),new_priority);
      break;

    case IFF:
      Output(f->LeftArg(),new_priority);
      Output(" <=> ");
      Output(f->RightArg(),new_priority-1);
      break;

    case NOT:
      Output('~');
      Output(f->Arg(),new_priority);
      break;

    case BOX:
      Output("[] ");
      Output(f->Arg(),new_priority);
      break;

    case DIAM:
      Output("<> ");
      Output(f->Arg(),new_priority);
      break;

    case ALL:
      Output("(! x");
      Output(f->QuantVar());
      Output(')');
      Output(f->QuantArg(),new_priority);
      break;

    case SOME:
      Output("(? x");
      Output(f->QuantVar());
      Output(')');
      Output(f->QuantArg(),new_priority);
      break;

    case ATOMIC:
      Output(f->Term());
      break;
    }

  if (print_par)
    Output(')');

}  // Output(const FORMULA*,PRIORITY)


void Output(const FORMULA* f)
{
  Output(f,PR_HIGH);
}

********************/

void Output(ANCESTORS* p) 
 {
  Output("[");
  Output(ANCESTORS::BackgroundStr(p->Background())); 
  for (int i = 0; i < p->Number(); i++)
   {
    Output(((p->Clauses())[i])->Number());
    if (p->Number() - i > 1) Output(",");
   };
  Output("]");
 }; // void Output(ANCESTORS* p) 

void Output(const Clause* c)
 {
  if (!c) { Output("<< null >>"); return; };

  //Output((int)c->Store()); // debug
  //Output("& "); // debug


  Output(c->Number());
  Output(". ");

  TERM* lits = c->Literals();

  //Output((long)lits); // debug

  if (lits->EmptyLiterals())  // empty clause
    Output('#');
  else  // not empty clause
    {
    int maxl = lits->MaxLiterals();
    lits = lits->Args();
    
    Output(lits->First());
    
    lits = lits->Next();
    if (!lits->IsNull())  // more than 1 literal
      do
	{
	Output(maxl == 1 ? " | " : " \\/ ");
	maxl--;
	Output(lits->First());
	lits = lits->Next();
	}
      while (!lits->IsNull());
    else ;
    }

  Output(' ');
  Output(c->Weight());
  Output('(');
  Output(c->MinResWeight());
  Output(") ");
  Output(" ");
  Output(c->NumOfPosEq());
  Output("pe ");
  
  Output(c->Ancestors());
}  // OUTPUT(Clause*)

void OutputArgs(const TERM* ts)  // to output arguments of a term
{
  if (ts->IsNull()) // no argument, the term was a constant
    return;

  Output('(');
  for (;;)
    {
    OutputArg(ts,PR_HIGH);
    ts = ts->Next();
    if (ts->IsNull())  // last argument
      {
      Output(')');
      return;
      }
    Output(',');
    }
}  // OutputArgs()


void OutputArg(const TERM* ts,int pr)  // to output one argument of a term
{
  if (ts->IsVariable())  // variable
    {
    Output('X');
    Output(ts->Var());
    return;
    }

  // compound term
  OutputTerm(ts->First(),pr);
}  // OutputArg()

void OutputInference(Clause* root)
 {
  Output("----------------- PROOF ------------------\n");
  INFERENCE::CollectInference(root);
  INFERENCE::ResetIterator();
  Clause** c = INFERENCE::Next();
  while (c)
   {
    Output(*c);
    Output("\n");
    c = INFERENCE::Next();
   };
  Output("------------ END OF PROOF ----------------\n");
 };


//*************************************************************************
















