#include "path_index.h"

//***************** Definitions for PI_TREE: ******************************

void PI_TREE::Integrate(const Clause* cl,int lit,const TERM* t_ref,PI_LEVEL* lev)
 {
  DCONTROL4("A1 in PI_TREE::Integrate(..)");
  
  if (t_ref->IsVariable())
   {
    DCONTROL4("A2 in PI_TREE::Integrate(..)");
    lev->AddOccurence(cl,lit,*t_ref);
    DCONTROL4("A3 in PI_TREE::Integrate(..)");
   }
  else // t_ref->First() is a complex term
   {
    DCONTROL4("A4 in PI_TREE::Integrate(..)");
    
    lev->AddOccurence(cl,lit,*t_ref);
    
    DCONTROL4("A5 in PI_TREE::Integrate(..)");
    
    TERM* trm = t_ref->First();
    TERM header = *trm;
    int arity = header.Arity();
    
    DCONTROL4("A6 in PI_TREE::Integrate(..)");
    
    PI_NODE* node = lev->ComplexSubterms();
    while ((node) && (node->Header() != header)) node = node->Alternative();
    if (node)
     {
      if (arity) // not a constant, subterms must be integrated
       {
        for (int n = 0; n < arity; n++)
         {
          Integrate(cl,lit,trm->NthArg(n),node->Subterms() + n);
         };
       }
      else // constant
       {
        PI_CL_LIST::Insert(&(node->ConstOccurences()),(Clause*)cl,lit); 
       };
       
     }
    else // no branch for this header yet, create a new branch 
     {
      PI_NODE* new_branch = CreateBranch(cl,lit,trm);
      new_branch->Alternative() = lev->ComplexSubterms();
      lev->ComplexSubterms() = new_branch;
     };
   };
  DCONTROL4("END in PI_TREE::Integrate(..)");
 }; // void PI_TREE::Integrate(const Clause* cl,int lit,const TERM* t_ref,PI_LEVEL* lev)

void PI_TREE::Remove(const Clause* cl,const TERM* t_ref,PI_LEVEL* lev)
 {
  DCONTROL4("BEGIN in PI_TREE::Remove(const Clause* cl,const TERM* t_ref,PI_LEVEL* lev)");      
  //CHECK2(lev->Occurences(),"A1 in PI_TREE::Remove(..,PI_LEVEL* lev)");
  
  PI_CLT_LIST::Remove(&(lev->Occurences()),cl);
  if (lev->Occurences())
   {
    DCONTROL4("A0 in PI_TREE::Remove(const Clause* cl,const TERM* t_ref,PI_LEVEL* lev)");
    if (t_ref->IsReference()) // not a variable
     {      
      Remove(cl,t_ref->First(),&(lev->ComplexSubterms())); 
     };
   }
  else 
   { 
    DCONTROL4("A1 in PI_TREE::Remove(const Clause* cl,const TERM* t_ref,PI_LEVEL* lev)");
    Destroy(lev->ComplexSubterms());
    
    DCONTROL4("A2 in PI_TREE::Remove(const Clause* cl,const TERM* t_ref,PI_LEVEL* lev)");
    lev->ComplexSubterms() = (PI_NODE*)0;
   };
  DCONTROL4("END in PI_TREE::Remove(const Clause* cl,const TERM* t_ref,PI_LEVEL* lev)"); 
 }; // void PI_TREE::Remove(const Clause* cl,const TERM* t_ref,PI_LEVEL* lev)

void PI_TREE::Remove(const Clause* cl,const TERM* t,PI_NODE** node)
 { 
  DCONTROL4("BEGIN in PI_TREE::Remove(const Clause* cl,const TERM* t,PI_NODE** node)");
  if (!(*node)) return;
  TERM header = *t;
  int arity = header.Arity();
    
  DCONTROL4("X0 in PI_TREE::Remove(const Clause* cl,const TERM* t,PI_NODE** node)");
 
  
  while ((*node)->Header() != header)
   {
    DCONTROL4("X1 in PI_TREE::Remove(const Clause* cl,const TERM* t,PI_NODE** node)");
 
    
    node = &((*node)->Alternative());
    
    DCONTROL4("X2 in PI_TREE::Remove(const Clause* cl,const TERM* t,PI_NODE** node)");
 
    //cout << "BRRRRR\n";  // debug
    //::Output((*node)->Header()); // debug 
    //::Output((long)(*node));
    //cout << "\n"; // debug
 
    if (!(*node)) return;
   };

  DCONTROL4("X3 in PI_TREE::Remove(const Clause* cl,const TERM* t,PI_NODE** node)");
 
  if (arity)
   {
    
    Remove(cl,t->Args(),(*node)->Subterms());
    if ((*node)->Subterms()->Occurences())
     {
      for (int arg = 1 /* not 0! */; arg < arity; arg++)
       Remove(cl,t->NthArg(arg),(*node)->Subterms() + arg);
     }
    else
     {
      PI_NODE* tmp = *node;
      *node = tmp->Alternative();

      tmp->Alternative() = (PI_NODE*)0;
      Destroy(tmp);
     };
   }
  else // constant
   {
    PI_CL_LIST::Remove(&((*node)->ConstOccurences()),(Clause*)cl);
    
    DCONTROL4("X6 in PI_TREE::Remove(const Clause* cl,const TERM* t,PI_NODE** node)");
     
    if (!((*node)->ConstOccurences()))
     {
      PI_NODE* tmp = *node;
      *node = tmp->Alternative();

      tmp->Alternative() = (PI_NODE*)0;
      
      DCONTROL4("X7 in PI_TREE::Remove(const Clause* cl,const TERM* t,PI_NODE** node)");
    
      Destroy(tmp);
      
      DCONTROL4("X8 in PI_TREE::Remove(const Clause* cl,const TERM* t,PI_NODE** node)");
    
     };   
   };
  DCONTROL4("END in PI_TREE::Remove(const Clause* cl,const TERM* t,PI_NODE** node)");  
 }; // void PI_TREE::Remove(const Clause* cl,const TERM* t,PI_NODE** node)


void PI_TREE::Destroy(PI_NODE* node)
 {
  if (node)
   {
    Destroy(node->Alternative());
    int arity = node->Header().Arity();
    if (arity)
     {
      for (int arg = 0; arg < arity; arg++)
       {
        PI_LEVEL* lev = node->Subterms() + arg;
        PI_CLT_LIST::Destroy(lev->Occurences()); 
        Destroy(lev->ComplexSubterms());
       };
      PI_NODE::Delete(node);
     }
    else 
     {
      PI_CL_LIST::Destroy(node->ConstOccurences());
      PI_NODE::DeleteConst(node);
     };  
   };
 }; // void PI_TREE::Destroy(PI_NODE* node) 

void PI_TREE::DestroyFromTop(PI_NODE* node)
 {
  if (node)
   {
    DestroyFromTop(node->Alternative());
    int arity = node->Header().Arity();
    if (arity)
     {
      for (int arg = 0; arg < arity; arg++)
       {
        PI_LEVEL* lev = node->Subterms() + arg;
        PI_CLT_LIST::Destroy(lev->Occurences()); 
        Destroy(lev->ComplexSubterms());
       };
      PI_NODE::Delete(node);
     }
    else // propositional
     {
      PI_C_LIST::Destroy(node->PropOccurences());
      PI_NODE::DeleteProp(node);
     };  
   };
 }; // void PI_TREE::DestroyFromTop(PI_NODE* node) 


PI_NODE* PI_TREE::CreateBranch(const Clause* cl,int lit,const TERM* t)
 {
  // t must be a complex term 
  CHECK2(t,"A1 in PI_TREE::CreateBranch(..)");
  CHECK2(t->IsComplex(),"A2 in PI_TREE::CreateBranch(..)");
  int arity = t->Arity();
  PI_NODE* result;
  if (arity)
   { 
    result = PI_NODE::NewNode(*t);
    const TERM* curr_arg;
    PI_LEVEL* subterms = result->Subterms();
    for (int arg = 0; arg < arity; arg++)
     {
      curr_arg = t->NthArg(arg);
      subterms[arg].AddOccurence(cl,lit,*curr_arg);
      if (curr_arg->IsReference()) 
       {
        subterms[arg].ComplexSubterms() = CreateBranch(cl,lit,curr_arg->First());
       };
     }; 
   }
  else // constant
   {
    result = PI_NODE::NewConstNode(*t);
    PI_CL_LIST::Insert(&(result->ConstOccurences()),(Clause*)cl,lit);
   };
  return result;   
 }; // PI_NODE* PI_TREE::CreateBranch(Clause* cl,int lit,TERM* t)


//**************** Definitions for SMALL_PATH_INDEX: ******************************

void SMALL_PATH_INDEX::OutputHeaders()
 {
  PI_NODE* node = tree;
  while (node)
   {
    node->Output(); ::Output(" ");
    node = node->Alternative();
   };
 }; // void SMALL_PATH_INDEX::OutputHeaders() 



//***************************************************************************