#define __PI_CLAUSE_LIST_H


#ifndef __DEBUG_H
 #include "debug.h"
#endif

#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif

#ifndef __TERM_H
 #include "term.h"
#endif

#ifndef __CLAUSE_H
 #include "clause.h"
#endif

#ifndef __OUTPUT_H
 #include "output.h"
#endif

//***************************************************************************


class PI_C_LIST
 {
  private: 
   long clause_number;
   const Clause* clause;
   PI_C_LIST* next;
  public:
   PI_C_LIST(const Clause* cl,PI_C_LIST* lst)
    : clause_number(cl->Number()), 
      clause(cl), 
      next(lst) 
     {}; 
   inline void* operator new(size_t)
    {      
     return Memory::Allocate(sizeof(PI_C_LIST)/4);
    };
   inline void operator delete(void* obj)
    {
     Memory::Deallocate(obj,sizeof(PI_C_LIST)/4);
    };
   inline long ClauseNumber() const { return clause_number; }; 
   inline const Clause* TheClause() const { return clause; };
   inline PI_C_LIST*& Next() { return next; };
   inline const PI_C_LIST* Next() const { return next; };
   static inline void SkipYounger(long cl_num,PI_C_LIST*& lst)
    {
     while ((lst) && (lst->clause_number > cl_num)) lst = lst->Next();
    }; 
   static inline void SkipYoungerOrEqual(long cl_num,PI_C_LIST*& lst)
    {
     while ((lst) && (lst->clause_number >= cl_num)) lst = lst->Next();
    };
   static inline void Insert(PI_C_LIST** lst,const Clause* cl);
   static inline void Remove(PI_C_LIST** lst,const Clause* cl);
   static inline void Destroy(PI_C_LIST* lst);
  public: // for debugging
   inline void Output() const;
   static inline void Output(const PI_C_LIST* lst);
   static inline bool OccursIn(const Clause* cl,const PI_C_LIST* lst);
   static inline int Length(const PI_C_LIST* lst)
    {
     if (lst) return (Length(lst->Next()) + 1);
     return 0; 
    };
 }; // class PI_C_LIST

inline void PI_C_LIST::Insert(PI_C_LIST** lst,const Clause* cl)
 {  
 
  DEBUG2_OP(PI_C_LIST** lst_init_val = lst); 
  DEBUG2_OP(int lst_init_length = Length(*lst)); 
  // here we assume that a list doesn't contain duplicates
  long cl_num = cl->Number();
  while ((*lst) && ((*lst)->TheClause()->Number() > cl_num)) 
   {
    lst = &((*lst)->Next());
   };  
   
  CHECK2((!(*lst)) || ((*lst)->TheClause()->Number() < cl_num),"A1 in PI_C_LIST::Insert(..)");
 
  *lst = new PI_C_LIST(cl,*lst); 
  
  CHECK2(OccursIn(cl,*lst_init_val),"Wrong result(1) in PI_C_LIST::Insert(..)"); 
  CHECK2(Length(*lst_init_val) == lst_init_length + 1,"Wrong result(2) in PI_C_LIST::Insert(..)"); 
 }; // void PI_C_LIST::Insert(PI_C_LIST** lst,const Clause* cl)

inline void PI_C_LIST::Remove(PI_C_LIST** lst,const Clause* cl)
 {
  // here we assume that the list (*lst) does not contain duplicates
  long cl_num = cl->Number();
  while ((*lst) && ((*lst)->TheClause()->Number() > cl_num)) lst = &((*lst)->Next());
  
  if ((*lst) && ((*lst)->TheClause()->Number() == cl_num))
   {
    PI_C_LIST* tmp = *lst;
    *lst = tmp->Next();
    delete tmp;
   };
 }; // void PI_C_LIST::Remove(PI_C_LIST** lst,Clause* cl)


inline void PI_C_LIST::Destroy(PI_C_LIST* lst)
 {
  PI_C_LIST* tmp;
  while (lst)
   {
    tmp = lst;
    lst = lst->Next();
    delete tmp;
   };
 }; // void PI_C_LIST::Destroy(PI_C_LIST* lst)

inline void PI_C_LIST::Output() const
 {
  ::Output((long)clause); // debug
  ::Output("[");
  //::Output(clause->Number());
  ::Output(clause_number);
  ::Output("]");
 }; // void PI_C_LIST::Output() const

inline void PI_C_LIST::Output(const PI_C_LIST* lst)
 {
  ::Output("{");
  while (lst)
   {
    lst->Output();
    lst = ((PI_C_LIST*)lst)->Next();
    if (lst) ::Output(",");
   };  
  ::Output("}"); 
 }; // void PI_C_LIST::Output(const PI_C_LIST* lst)

inline bool PI_C_LIST::OccursIn(const Clause* cl,const PI_C_LIST* lst)
 {
  return (lst) && ((lst->TheClause() == cl) || OccursIn(cl,lst->Next()));
 }; // bool PI_C_LIST::OccursIn(const Clause* cl,const PI_C_LIST* lst) 

//***********************************************************************

class PI_CL_LIST : public PI_C_LIST
 {
  private: 
   int lit_num;
  public:
   PI_CL_LIST(const Clause* cl,int ln,PI_CL_LIST* lst) 
    : PI_C_LIST(cl,lst), lit_num(ln) {}; 
   inline void* operator new(size_t)
    {
     return Memory::Allocate(sizeof(PI_CL_LIST)/4);
    };
   inline void operator delete(void* obj)
    {
     Memory::Deallocate(obj,sizeof(PI_CL_LIST)/4);
    };
   inline int LitNum() const { return lit_num; };
   inline PI_CL_LIST*& Next() { return (PI_CL_LIST*&)PI_C_LIST::Next(); };
   inline const PI_CL_LIST* Next() const { return (const PI_CL_LIST*)PI_C_LIST::Next(); };
   static inline void Insert(PI_CL_LIST** lst,Clause* cl,int ln);
   static inline void Remove(PI_CL_LIST** lst,Clause* cl);
   static inline void Destroy(PI_CL_LIST* lst);
  public: // for debugging
   inline void Output() const;
   static inline void Output(const PI_CL_LIST* lst);
   static inline bool OccursIn(const Clause* cl,const PI_CL_LIST* lst);
   static inline int Length(const PI_CL_LIST* lst)
    {
     if (lst) return (Length(lst->Next()) + 1);
     return 0;
    };
 }; // class PI_CL_LIST

inline void PI_CL_LIST::Insert(PI_CL_LIST** lst,Clause* cl,int ln)
 {
  DEBUG2_OP(PI_CL_LIST** lst_init_val = lst); 
  DEBUG2_OP(int lst_init_length = Length(*lst)); 
  long cl_num = cl->Number();
  
  while ((*lst) && ((*lst)->TheClause()->Number() > cl_num)) 
   lst = &((*lst)->Next()); 
  while ((*lst) 
         && ((*lst)->TheClause()->Number() == cl_num) 
         && ((*lst)->LitNum() > ln))
   lst = &((*lst)->Next()); 
  *lst = new PI_CL_LIST(cl,ln,*lst);
  CHECK2(OccursIn(cl,*lst_init_val),"Wrong result(1) in PI_CL_LIST::Insert(..)"); 
  CHECK2(Length(*lst_init_val) == lst_init_length + 1,"Wrong result(2) in PI_CL_LIST::Insert(..)");

 }; // void PI_CL_LIST::Insert(PI_CL_LIST** lst,const Clause* cl)

inline void PI_CL_LIST::Remove(PI_CL_LIST** lst,Clause* cl)
 {
  // the clause cl must occur in the list (*lst) at least once
  long cl_num = cl->Number();
  while ((*lst) && ((*lst)->TheClause()->Number() > cl_num)) 
   lst = &((*lst)->Next());
  while ((*lst) && ((*lst)->TheClause()->Number() == cl_num))
   {
    PI_CL_LIST* tmp = *lst;
    *lst = tmp->Next();
    delete tmp;
   };
 }; // void PI_CL_LIST::Remove(PI_CL_LIST** lst,Clause* cl)


inline void PI_CL_LIST::Destroy(PI_CL_LIST* lst)
 {
  PI_CL_LIST* tmp;
  while (lst)
   {
    tmp = lst;
    lst = lst->Next();
    delete tmp;
   };
 }; // void PI_CL_LIST::Destroy(PI_CL_LIST* lst)


inline void PI_CL_LIST::Output() const
 {
  ::Output("[");
  ::Output(TheClause()->Number());
  ::Output(":");
  ::Output(lit_num);
  ::Output("]");
 }; // void PI_CL_LIST::Output() const

inline void PI_CL_LIST::Output(const PI_CL_LIST* lst)
 {
  while (lst)
   {
    lst->Output();
    lst = ((PI_CL_LIST*)lst)->Next();
    if (lst) ::Output(",");
   };  
 }; // void PI_CL_LIST::Output(const PI_CL_LIST* lst)

inline bool PI_CL_LIST::OccursIn(const Clause* cl,const PI_CL_LIST* lst)
 {
  return (lst) && ((lst->TheClause() == cl) || OccursIn(cl,lst->Next()));
 }; // bool PI_CL_LIST::OccursIn(const Clause* cl,const PI_CL_LIST* lst) 

//************************************************************************


class PI_CLT_LIST : public PI_CL_LIST
 {
  private:    
   TERM term;
  public:
   PI_CLT_LIST(const Clause* cl,int ln,TERM t,PI_CLT_LIST* lst) 
    : PI_CL_LIST(cl,ln,(PI_CL_LIST*)lst), term(t) {}; 
   inline void* operator new(size_t)
    {
     return Memory::Allocate(sizeof(PI_CLT_LIST)/4);
    };
   inline void operator delete(void* obj)
    {
     Memory::Deallocate(obj,sizeof(PI_CLT_LIST)/4);
    };
   inline TERM Term() const { return term; };
   inline PI_CLT_LIST*& Next() { return (PI_CLT_LIST*&)PI_CL_LIST::Next(); };
   inline const PI_CLT_LIST* Next() const { return (const PI_CLT_LIST*)PI_CL_LIST::Next(); };
   static inline void Insert(PI_CLT_LIST** lst,const Clause* cl,int ln,TERM t);
   static inline void Remove(PI_CLT_LIST** lst,const Clause* cl);
   static inline void Destroy(PI_CLT_LIST* lst);
  public: // for debugging
   inline void Output() const;
   static inline void Output(const PI_CLT_LIST* lst);
   static inline bool OccursIn(const Clause* cl,const PI_CLT_LIST* lst);
   static inline int Length(const PI_CLT_LIST* lst)
    {
     if (lst) return (Length(lst->Next()) + 1);
     return 0;
    };
   static inline bool CheckIntegrity(const PI_CLT_LIST* lst)
    {
     if (!lst) return true;
     #ifdef _DEBUG_MEMORY_
      if (!(Memory::Captured((void*)lst))) return false;
     #endif
     return true; 
    };
      
 }; // class PI_CLT_LIST

inline void PI_CLT_LIST::Insert(PI_CLT_LIST** lst,const Clause* cl,int ln,TERM t)
 {
 
  DCONTROL4("BEGIN in PI_CLT_LIST::Insert(..)");
 
 
  DEBUG2_OP(PI_CLT_LIST** lst_init_val = lst); 
  
  DCONTROL4("A0 in PI_CLT_LIST::Insert(..)");
  
  DEBUG2_OP(int lst_init_length = Length(*lst)); 
  
  long cl_num = cl->Number();
  
  DCONTROL4("A1 in PI_CLT_LIST::Insert(..)");
  
  while ((*lst) && ((*lst)->TheClause()->Number() > cl_num)) 
   {
    DCONTROL4("A2 in PI_CLT_LIST::Insert(..)");
    lst = &((*lst)->Next()); 
    DCONTROL4("A3 in PI_CLT_LIST::Insert(..)");
   };
  
  DCONTROL4("A4 in PI_CLT_LIST::Insert(..)"); 
   
  while ((*lst) 
         && ((*lst)->TheClause()->Number() == cl_num) 
         && ((*lst)->LitNum() > ln))
   lst = &((*lst)->Next()); 
  *lst = new PI_CLT_LIST(cl,ln,t,*lst);
  
  CHECK2(OccursIn(cl,*lst_init_val),"Wrong result(1) in PI_CLT_LIST::Insert(..)"); 
  CHECK2(Length(*lst_init_val) == lst_init_length + 1,"Wrong result(2) in PI_CLT_LIST::Insert(..)"); 
  DCONTROL4("END in PI_CLT_LIST::Insert(..)");
 }; // void PI_CLT_LIST::Insert(PI_CLT_LIST** lst,const Clause* cl)

inline void PI_CLT_LIST::Remove(PI_CLT_LIST** lst,const Clause* cl)
 {
  long cl_num = cl->Number();
  while ((*lst) && ((*lst)->TheClause()->Number() > cl_num)) 
   lst = &((*lst)->Next());
  while ((*lst) && ((*lst)->TheClause()->Number() == cl_num))
   {
    PI_CLT_LIST* tmp = *lst;
    *lst = tmp->Next();
    delete tmp;
   };
 }; // void PI_CLT_LIST::Remove(PI_CLT_LIST** lst,Clause* cl)


inline void PI_CLT_LIST::Destroy(PI_CLT_LIST* lst)
 {
  PI_CLT_LIST* tmp;
  while (lst)
   {
    tmp = lst;
    lst = lst->Next();
    delete tmp;
   };
 }; // void PI_CLT_LIST::Destroy(PI_CLT_LIST* lst)


inline void PI_CLT_LIST::Output() const
 {
  ::Output((long)this); // debug
  ::Output("/"); //debug
  
  ::Output("[");
  ::Output(TheClause()->Number());  
  ::Output(":");
  ::Output(LitNum());
  ::Output(":");
  ::Output(term);
  ::Output("]");  
 }; // void PI_CLT_LIST::Output() const

inline void PI_CLT_LIST::Output(const PI_CLT_LIST* lst)
 { 
  while (lst)
   {
    lst->Output();
    lst = ((PI_CLT_LIST*)lst)->Next();
    if (lst) ::Output(",");
    ::Output("\n");
   };  
 }; // void PI_CLT_LIST::Output(const PI_CLT_LIST* lst)


inline bool PI_CLT_LIST::OccursIn(const Clause* cl,const PI_CLT_LIST* lst)
 {
  return (lst) && ((lst->TheClause() == cl) || OccursIn(cl,lst->Next()));
 }; // bool PI_CLT_LIST::OccursIn(const Clause* cl,const PI_CLT_LIST* lst) 


//***************************************************************************