#define __PREMISE_FACTORING_H
#ifndef __TERM_H
 #include "term.h"
#endif
#ifndef __STACK_H
 #include "stack.h"
#endif
#ifndef __OR_PREMISE_H
 #include "or_premise.h"
#endif
#ifndef __PREFIX_H
 #include __PREFIX_H
#endif

class OR_PREMISE_WITH_FACTORING : public OR_PREMISE
 {
  private:
   STACK<int,MAX_NUMBER_OF_LITERALS> factored;
   
   PREFIX_SYM* cut_literal; // the literal itself in prefix representation
   
   int first_factor;
   int last_factor;
   bool was_factored;
  public:
   OR_PREMISE_WITH_FACTORING() : OR_PREMISE(),factored("MAX_NUMBER_OF_LITERALS") {};
   inline void ResetFactoring()
    {
     was_factored = false;
     if (FactoringPossible())
      {
       factored.Reset();
       cut_literal = max_literal[cut_lit];
      };
    };
   inline bool WasFactored() { return was_factored; };
   inline bool FactoringPossible() { return (bool)LastMax(); };
   bool GetFirstFactor(int first_candidate);
   bool GetNextFactor();
  private: // aux. methods
   bool FindFirstFactor(int& candidate);
 }; // class OR_PREMISE_WITH_FACTORING
 
template <int numOfPremises> 
 class OR_MULTIPREMISE  // for hyperresolution
  {
   private:
    OR_PREMISE_WITH_FACTORING premise[numOfPremises];
    STACK_WITH_ITERATOR<int,numOfPremises> loaded;
   private: // might be changed in the future
    int next;
   public:
    OR_MULTIPREMISE() : loaded("number of premises") { next = 0; };
    inline void Load(int ind,Clause* cl,TERM* cut_liter)
     {
      premise[next].LoadOptimized(ind,cl);
      premise[next].Cut(cut_liter);
      loaded.Push(next);
      next++;
     }; 
    inline void Load(int ind,RES_PAIR_LIST* res_pair)
     {
      premise[next].LoadOptimized(ind,res_pair);
      loaded.Push(next);
      next++;
     }; 
    inline void UnloadLast() 
     {
      next = loaded.Pop();
      premise[next].Unload();
     };
    inline void Reset() 
     {
      while (!loaded.Empty()) 
       {
        premise[loaded.Pop()].Unload();
       };
      next = 0;
     };
    
    inline bool CollectNonmarkedLiterals()
     {
      loaded.ResetIterator();
      int* i;
      while ((i = loaded.Next()))
       {
        if (!(premise[*i].CollectNonmarkedLiterals()))
         {
          return false; 
         };
       }; 
      return true; // all literals successfully collected
     };
    inline void ResetFactoring()
     {
      loaded.ResetIterator();
      int* i;
      while ((i = loaded.Next()))
       {
        premise[*i].ResetFactoring();
       };
     };
    inline bool NextFactor()
     {
      loaded.ResetIterator();
      int* i;
      OR_PREMISE_WITH_FACTORING* prem;
      while ((i = loaded.Next()))
       {
        prem = &(premise[*i]);
        if (prem->WasFactored())
         {
          if (prem->GetNextFactor())
           {
            return true;
           }
          else
           {
            prem->ResetFactoring();
           };
         }
        else 
         {
          if (prem->GetFirstFactor(0))
           {
            return true;
           };
         };
       };
      return false; // no more factors 
     };
   inline void WriteAllClauses(Clause** dest)
    {
     loaded.ResetIterator();
     int* i;
     while ((i = loaded.Next()))
      {
       *dest = premise[*i].TheClause();
       dest++;
      }; 
    };
  }; // class OR_MULTIPREMISE<int numOfPremises>
 
//*********************************************************************** 