#define __RESOLVENT_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif
#ifndef __TERM_H
 #include "term.h"
#endif
#ifndef __PREFIX_H
 #include "prefix.h"
#endif
#ifndef __VARIABLES_H
 #include "variables.h"
#endif
#ifndef __STATISTICS_H
 #include "statistics.h"
#endif
#ifndef __FLAT_TERM_H
 #include "flat_term.h"
#endif 
#ifndef __TMP_LITERAL_H
 #include "tmp_literal.h"
#endif 
#ifndef __NEW_CLAUSE_H
 #include "new_clause.h"
#endif 
#ifndef __ANCESTORS_H
 #include "ancestors.h"
#endif 
#ifndef __CLAUSE_H
 #include "clause.h"
#endif
#ifndef __FORWARD_SUBSUMPTION_H
 #include "forward_subsumption.h"
#endif 
#ifndef __FORW_UEQ_SIMPLIFICATION_H 
 #include "forw_ueq_simplification.h"
#endif
#ifndef __SPLITTING_H
 #include "splitting.h"
#endif

//extern CLAUSE_INDEX ClauseIndex; 
extern FS ForwardSubsumption;
extern FUEQ_SIMPLIFICATION FSimplificationByUEq;
extern SPLITTING_INDEX SplittingIndex;
extern void RegisterPreprocessedSupported(Clause* cl);
extern void RegisterPreprocessedUnsupported(Clause* cl);
extern void RegisterNewResolvent(Clause* cl);
extern void InternalTickEvent(const long ticks);
//*************************************************************************

class RESOLVENT
 {
  private: // structure
   NEW_CLAUSE clause;
   NEW_CLAUSE basic_component;
   TMP_LITERAL positive_name;
   TMP_ANCESTORS ancestors;
   VAR_TO_INT_RENAMING var_renaming;
   INT_VAR_RENAMING int_var_renaming;
   INT_VAR_RENAMING split_var_renaming;
   VAR_WEIGHTING var_weighting;
   VAR_WEIGHTING split_var_weighting; 
   //bool eliminated_pred[MAX_NUMBER_OF_FUNCTORS];  
   bool* eliminated_pred; 
  private: // internal clocks
   long next_tick;
   long ticks; 
  private: // options and parameters
   bool forward_subsumption; 
   bool simplification_by_fs;
   bool simplification_by_ueq;
   bool simplification_by_eq_res;
   bool static_splitting;
   bool negative_selection;
   bool equality_possible;
   bool eliminate_predicates;
   long max_weight;
   long previous_max_weight;
   long internal_tick_frequency;
  public:
   RESOLVENT() 
    : clause(), 
      basic_component(),
      positive_name(),
      ancestors(), 
      var_renaming(), 
      int_var_renaming(),
      split_var_renaming(),
      var_weighting(),
      split_var_weighting(),

      eliminated_pred((bool*)0), 
 
      next_tick(MAXLONG),
      ticks(0L),
      forward_subsumption(false), // default
      simplification_by_fs(false), // default
      simplification_by_ueq(false), // default
      simplification_by_eq_res(false), // default
      static_splitting(false), // default
      negative_selection(false), // default
    
      eliminate_predicates(false), // default  

      max_weight(MAX_CLAUSE_WEIGHT),
      previous_max_weight(MAX_CLAUSE_WEIGHT),
      internal_tick_frequency(MAXLONG)
    {
     //ResetEliminationTable();
    };
   
    /*****
   inline void ResetEliminationTable()
    {
     for (int p = 0; p < MAX_NUMBER_OF_FUNCTORS; p++) eliminated_pred[p] = false;
    };
   inline void Eliminate(int pred) { eliminated_pred[pred] = true; };
   *********/
   inline void SetEliminationTable(bool* table) { eliminated_pred = table; };
   inline void Tick()
    {
     if (next_tick) { next_tick--; }
     else { next_tick = internal_tick_frequency; ticks++; InternalTickEvent(ticks); };
    };
     
   inline void Reset()
    {
     clause.Reset();
     ancestors.Reset();
     STATISTICS_ON_GENERATED::AnotherGenerated();
     Tick();
    };
   inline void RegisterTooHeavy()
    {
     STATISTICS_ON_GENERATED::AnotherGenerated();
     Tick();
     STATISTICS_ON_GENERATED::AnotherTooHeavyClause();
    }; 
   inline void ResetInternalTick(const long freq) 
    {
     internal_tick_frequency = freq; 
     next_tick = freq;
     ticks = 0L;
    }; 
  public: // setting options and parameters
   inline void SetForwardSubsumption(bool fl) { forward_subsumption = fl; };
   inline void SetSimplificationByFS(bool fl) { simplification_by_fs = fl; };
   inline void SetSimplificationByUEq(bool fl) { simplification_by_ueq = fl; };
   inline void SetSimplificationByEqRes(bool fl) { simplification_by_eq_res = fl; };
   inline void SetStaticSplitting(bool fl) { static_splitting = fl; };
   inline void SetNegativeSelection(bool fl) { negative_selection = fl; };
   inline void SetEqualityPossible(bool fl) { equality_possible = fl; };
   inline void SetEliminatePredicates(bool fl) { eliminate_predicates = fl; };
   inline void SetMaxWeight(long mw) 
    {
     previous_max_weight = max_weight;
     max_weight = (mw > MAX_CLAUSE_WEIGHT) ? MAX_CLAUSE_WEIGHT : mw;
     if (max_weight < MIN_MAX_WEIGHT) max_weight = MIN_MAX_WEIGHT;
     clause.SetMaxWeight(max_weight);
    };
   inline long MaxWeight() const { return max_weight; };
   inline long PreviousMaxWeight() const { return previous_max_weight; };
   inline void SetMaxLength(long ml) { clause.SetMaxLength(ml); };
   inline void SetMaxDepth(long md) { clause.SetMaxDepth(md); };
   inline void SetMaxLitWeight(long mlw) { clause.SetMaxLitWeight(mlw); };
  public:
   //inline bool WasTautology() const { return clause.WasTautology(); }; 
   
   // collecting new clause
  private: 
   inline bool PushVar(const int v) { return clause.PushVar(v); };
   inline void PushVar(const TERM v) { clause.PushVar(v); };  
   inline void PushConst(const TERM c) { clause.PushConst(c); };
   inline bool PushNonconstFun(const TERM f) { return clause.PushNonconstFun(f); };

  public:
   inline bool PushSymbol(const TERM& sym)
    {
     if (sym.IsVariable()) { PushVar(sym); return true; };
     if (sym.Arity()) return PushNonconstFun(sym);
     PushConst(sym);
     return true;
    };   

   //inline void OpenLiteral() { clause.OpenLiteral(); };
 
  public:    
   inline bool OpenNonpropLiteral(TERM header) { return clause.OpenNonpropLiteral(header); };
   inline bool CloseLiteral() { return clause.CloseLiteral(); };
   inline bool PushPropLit(const TERM lit) { return clause.PushPropLit(lit); }; 

  public: 
   inline bool PushRewrittenLit(int index,TERM* lit,TERM* redex,PREFIX_SYM* replacing_term);
   bool PushRewrittenLit(PREFIX_SYM* lit,
                         PREFIX_SYM* redex,
                         TERM replacing_term,
                         int index);
   inline bool PushPrefixLit(PREFIX_SYM* word); 
   inline bool PushPrefixEqLit(TERM header,PREFIX_SYM* arg1,PREFIX_SYM* arg2);
   inline bool PushPrefixComplexTerm(PREFIX_SYM* word);
   inline bool PushPrefixTerm(PREFIX_SYM* t);
   inline bool PushStandardLit(int index,TERM* t); 
   inline bool PushStandardTerm(int index,TERM t);
   inline bool PushStandardComplexTerm(int index,TERM* t);
   bool PushLiteral(TERM* const t);   
   bool LoadClause(const Clause* const c);
   inline void SetBackground(ANCESTORS::BACKGR b) { ancestors.SetBackground(b); };
   inline void PushAncestor(Clause* cl) { ancestors.Push(cl); };
   
   // processing collected clause
  private:  
   inline void FSQueryPushLit(TMP_LITERAL* lit);
   inline Clause* FSubsume();
   inline int SimplifyByFS();
   Clause* RewriteByUEq(TMP_LITERAL* lit,TMP_LITERAL* new_lit);
   bool SimplifyByUEq(TMP_LITERAL*& lit,TMP_LITERAL*& spare_lit);
   bool SimplifyByUEq();
   bool SimplifyByEqRes();  

   Clause* MakeUnitClause(TMP_LITERAL* lit);
   inline Clause* MakeGroundSharedClause(TMP_LIT_LIST& lits)
    {
     return MakeSharedClause(lits); // temporary version
    };
   Clause* MakeSharedClause(TMP_LIT_LIST& lits);
   Clause* MakeUnsharedClause(TMP_LIT_LIST& lits);
   Clause* MakeInputClause(TMP_LIT_LIST& lits); 
  public: 
   Clause* MakeEmptyClause();
   inline Clause* FSubsume(Clause* cl);
   Clause* MakeInputClause(ANCESTORS::BACKGR backgr); 
   void ProcessCollected();
   void PreprocessSupported(Clause* cl); 
   void PreprocessUnsupported(Clause* cl); 
    
    // splitting
  private:
   int ComponentName(TMP_LIT_LIST& component,bool& new_component);
   bool StaticSplitting(bool supported);
   void StatSplitGroundComp(TMP_LIT_LIST& gr_comp,bool supported);
   
  public: // for debugging
   inline void Output() { clause.Output(); };
      
 }; // class RESOLVENT

//****************** Definitions for RESOLVENT: ***************************


inline bool RESOLVENT::PushRewrittenLit(int index,TERM* lit,TERM* redex,PREFIX_SYM* replacing_term)
 {
  CHECK8(lit->Arity(),"A0 in RESOLVENT::PushRewrittenLit(int index,TERM* lit,..)");
  TERM header = *lit;
  if (header.IsEquality()) header.MakeUnordered();
  if (!(OpenNonpropLiteral(header))) return false;
  
  ST_TERM_WITH_SUBST iter;
  iter.Reset(index,lit);
  iter.SkipSym();
  PREF_SYM sym;
  
  while (!(iter.Empty()))
   {
    if (iter.CurrPosInBase() == redex)
     {
      if (!PushPrefixTerm(replacing_term)) return false;
      iter.SkipTerm();
     }
    else
     { 
      sym = iter.CurrSym();
      if (sym.IsVariable()) // free variable, must be renamed
       {
        if (!(PushVar(var_renaming.Rename(sym.Var())))) { return false; };
       }
      else
       {
        if (sym.Func().Arity())
         {
          if (!(PushNonconstFun(sym.Func()))) return false;
         }
        else PushConst(sym.Func()); 
       };
      iter.SkipSym();
     };  
   };
   
  return CloseLiteral();
 }; // bool RESOLVENT::PushRewrittenLit(int index,TERM* lit,TERM* redex,PREFIX_SYM* replacing_term)


inline bool RESOLVENT::PushPrefixLit(PREFIX_SYM* word)
 {
  TERM header = word->Head().Func();
  if (!(header.Arity())) { return PushPropLit(header); };
  
  if (!(OpenNonpropLiteral(header))) return false;
  
  PREFIX::ITER iter;
  iter.Reset(word);
  TERM sym;
  iter.SkipSym(); // skipping the header
  do
   {
    if (iter.CurrVar()) // free variable
     { 
      if (!(PushVar(var_renaming.Rename(iter.CurrVar())))) { return false; };  
     }
    else // functor
     {
      sym = iter.CurrSubterm()->Head().Func();
      if (sym.Arity())
       {
        if (!(PushNonconstFun(sym))) return false;
       }
      else // constant
       PushConst(sym);
     };
    iter.SkipSym();
   }
  while (!(iter.Empty()));
  return CloseLiteral();
 }; // bool RESOLVENT::PushPrefixLit(PREFIX_SYM* word)



inline bool RESOLVENT::PushPrefixEqLit(TERM header,PREFIX_SYM* arg1,PREFIX_SYM* arg2)
 {
  CHECK8(header.IsEquality(),"A0 in RESOLVENT::PushPrefixEqLit(..)");    
  
  return OpenNonpropLiteral(header)
         && PushPrefixTerm(arg1)
         && PushPrefixTerm(arg2)
         && CloseLiteral();
 }; // bool RESOLVENT::PushPrefixEqLit(TERM header,PREFIX_SYM* arg1,PREFIX_SYM* arg2)


inline bool RESOLVENT::PushPrefixComplexTerm(PREFIX_SYM* word)
 {
  PREFIX::ITER iter;
  iter.Reset(word);
  TERM sym;
  do
   {
    if (iter.CurrVar()) // free variable
     { 
      if (!(PushVar(var_renaming.Rename(iter.CurrVar())))) return false; 
     }
    else // functor
     {
      sym = iter.CurrSubterm()->Head().Func();
      if (sym.Arity())
       {
        if (!(PushNonconstFun(sym))) return false;
       }
      else // constant
       PushConst(sym);
     };
    iter.SkipSym();
   }
  while (!(iter.Empty()));
  return true;
 }; // bool RESOLVENT::PushPrefixComplexTerm(PREFIX_SYM* word)

inline bool RESOLVENT::PushPrefixTerm(PREFIX_SYM* t)
 {
  if (t->Head().IsVariable())
   {
    bool free_var;
    void* bind = t->Head().Var()->Unref(free_var);
    if (free_var) return PushVar(var_renaming.Rename((VAR*)bind));
    return PushPrefixComplexTerm((PREFIX_SYM*)bind); 
   };
  return PushPrefixComplexTerm(t);
 }; // bool RESOLVENT::PushPrefixTerm(PREFIX_SYM* t)

inline bool RESOLVENT::PushStandardTerm(int index,TERM t)
 {
  // t is variable or reference 
  if (t.IsVariable())
   {
    bool free_var;
    void* bind = VAR::Var(t.Var(),index)->Unref(free_var);
    
    if (free_var) return PushVar(var_renaming.Rename((VAR*)bind));
    return PushPrefixComplexTerm((PREFIX_SYM*)bind);
   };
  // t is reference to a complex term
  return PushStandardComplexTerm(index,t.First());
 }; // bool RESOLVENT::PushStandardTerm(int index,TERM t)

inline bool RESOLVENT::PushStandardComplexTerm(int index,TERM* t)
 {
  ST_TERM_WITH_SUBST term;
  term.Reset(index,t);
  PREF_SYM sym;
  do
   {
    sym = term.CurrSym();
    if (sym.IsVariable()) // free variable, must be renamed
     {
      if (!(PushVar(var_renaming.Rename(sym.Var())))) return false; 
     }
    else
     {
      if (sym.Func().Arity())
       {
        if (!(PushNonconstFun(sym.Func()))) return false;
       }
      else PushConst(sym.Func()); 
     };
    term.SkipSym();
   }
  while (!(term.Empty()));
  return true; 
 }; // inline bool RESOLVENT::PushStandardComplexTerm(int index,TERM* t)

inline bool RESOLVENT::PushStandardLit(int index,TERM* t)
 {
  DCONTROL8("BEGIN in RESOLVENT::PushStandardLit(..)");
  // Note: this function uses the global substitution to substitute the variables!

  TERM header = *t;
  if (!(header.Arity())) { return PushPropLit(header); };
  
  DCONTROL8("A13 in RESOLVENT::PushStandardLit(..)");
  
  if (!(OpenNonpropLiteral(header))) return false;
  
  DCONTROL8("A15 in RESOLVENT::PushStandardLit(..)");
  
  ST_TERM_WITH_SUBST term;
  term.Reset(index,t);
  term.SkipSym();
  PREF_SYM sym;
  
  DCONTROL8("A20 in RESOLVENT::PushStandardLit(..)");
  
  while (!(term.Empty()))
   {
    DCONTROL8("A21 in RESOLVENT::PushStandardLit(..)");
    sym = term.CurrSym();
    if (sym.IsVariable()) // free variable, must be renamed
     {
      if (!(PushVar(var_renaming.Rename(sym.Var())))) { return false; };
     }
    else
     {
      if (sym.Func().Arity())
       {
        if (!(PushNonconstFun(sym.Func()))) return false;
       }
      else PushConst(sym.Func()); 
     };
    term.SkipSym();
    DCONTROL8("A30 in RESOLVENT::PushStandardLit(..)");
   };
   
  DCONTROL8("END5 in RESOLVENT::PushStandardLit(..)"); 
  return CloseLiteral();  
 }; // bool RESOLVENT::PushStandardLit(int index,TERM* t)


inline void RESOLVENT::FSQueryPushLit(TMP_LITERAL* lit)
 {
  if (lit->IsEquality())
   {
    
    //cout << "BEFORE "; lit->Output(); cout << "\n"; lit->FlatTerm()->OutputStructure(); cout << "\n"; // debug
   
    lit->LinkEqArgs();

    //cout << "AFTER "; lit->Output(); cout << "\n"; lit->FlatTerm()->OutputStructure(); cout << "\n"; // debug

    if (lit->IsOrderedEq())
     {
      if (lit->EqSwapArgs())
       {
        ForwardSubsumption.PushOrdEq(lit->Header(),lit->Arg2(),lit->Arg1());
       }
      else
       { 
        ForwardSubsumption.PushOrdEq(lit->Header(),lit->Arg1(),lit->Arg2());
       }; 
     }
    else ForwardSubsumption.PushUnordEq(lit->Header(),lit->Arg1(),lit->Arg2()); 
   }
  else ForwardSubsumption.PushLit(lit->FlatTerm()); 
  
 }; // void RESOLVENT::FSQueryPushLit(TMP_LITERAL* lit)


inline Clause* RESOLVENT::FSubsume() 
 {
  Clause* subsuming_cl;
  
  ForwardSubsumption.ResetQuery(); 
  for (TMP_LITERAL** lit = clause.AllLiterals().First(); lit <= clause.AllLiterals().Last(); lit++)
   FSQueryPushLit(*lit);
  
  //Output(); cout << "\n ------->>>\n";
  //ForwardSubsumption.OutputQuery(); cout << "\n\n"; // debug 
  
   
  subsuming_cl = ForwardSubsumption.Subsume();
  return subsuming_cl;  
 }; // Clause* RESOLVENT::FSubsume()

inline Clause* RESOLVENT::FSubsume(Clause* cl)
 {
  LoadClause(cl);
  clause.EndOfClause();
  return FSubsume();
 }; // Clause* RESOLVENT::FSubsume(Clause* cl)

inline int RESOLVENT::SimplifyByFS()
 {
  if (clause.AllLiterals().Length() <= 1) return 0;
  int res = 0;
  
  TMP_LITERAL** cand_lit = clause.AllLiterals().Last(); 
  try_lit: 
   if ((*cand_lit)->IsEquality())   
    {
     if (cand_lit == clause.AllLiterals().First()) return res;
     cand_lit--;
     goto try_lit;
    };

   // try to simplify by resolving cand_lit
   ((FLAT_TERM*)((*cand_lit)->FlatTerm()))->SymbolRef().InversePolarity();
   
   ForwardSubsumption.ResetQuery(); 
   for (TMP_LITERAL** lit = clause.AllLiterals().First(); lit <= clause.AllLiterals().Last(); lit++)
    FSQueryPushLit(*lit);
            
   Clause* simplifier = ForwardSubsumption.Subsume();
   if (simplifier)
    {
     res++;
     PushAncestor(simplifier);
     simplifier->MarkAsSimplifier();
     clause.AllLiterals().ReplaceByTop(cand_lit);
     if ((clause.AllLiterals().Length() <= 1) 
         || (cand_lit == clause.AllLiterals().First()))
      return res;
     cand_lit--;
     goto try_lit;  
    }
   else 
    {
     ((FLAT_TERM*)((*cand_lit)->FlatTerm()))->SymbolRef().InversePolarity();
     if (cand_lit == clause.AllLiterals().First()) return res;
     cand_lit--;
     goto try_lit;
    }; 
 }; // int RESOLVENT::SimplifyByFS()

 
//*************************************************************************
