#define __SHARING_H
#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif
#ifndef __TERM_H
 #include "term.h"
#endif
#ifndef __FLAT_TERM_H
 #include "flat_term.h"
#endif 
#ifndef __OUTPUT_H
 #include "output.h"
#endif

class SHARING_TREE_NODE;

class SHARING_INDEX
 {  
  private:
   typedef STACK<TERM*,MAX_TERM_WEIGHT> POS_STACK;
  private: // auxilliary data
   TERM::ITERATOR t_iter;
   POS_STACK subterms;
  private: // structure
   SHARING_TREE_NODE* nonconst_index[MAX_NUMBER_OF_FUNCTORS];
   TERM const_index[MAX_NUMBER_OF_FUNCTORS*2];
  public:
   class SURFER; 
  public:
   SHARING_INDEX() : t_iter(), subterms("MAX_TERM_WEIGHT in SHARING_INDEX::subterms")
    {
     for (int f = 0; f < MAX_NUMBER_OF_FUNCTORS; f++)
      {
       nonconst_index[f] = (SHARING_TREE_NODE*)0;
      };
     for (int c = 0; c < MAX_NUMBER_OF_FUNCTORS; c++)
      {
       //const_index[c].MakeComplex((ARITY)0,(FUNCTOR)c,POL_POS);
       const_index[2*c].MakeComplex((ARITY)0,(FUNCTOR)c,POL_POS);
       const_index[2*c+1].SetFirst((TERM*)0);

      };
    };
   inline void Destroy() 
    {
     for (int f = 0; f < MAX_NUMBER_OF_FUNCTORS; f++)
      {
       DestroyTree(nonconst_index[f]);
       nonconst_index[f] = (SHARING_TREE_NODE*)0;
      };
    }; 
  public:
   inline SHARING_TREE_NODE** Tree(const TERM& fun)
    {
     CHECK6(fun.Arity(),"wrong argument in SHARING_INDEX::Tree(..)");
     return nonconst_index + fun.Functor();
    };
  private:
   inline SHARING_TREE_NODE* CreateBranch(TERM* whole);
   inline TERM* IntegrateComplexNonconst(TERM* t);
   inline TERM* IntegrateFlatComplexNonconst(const FLAT_TERM* t);
   static inline void DestroyTree(SHARING_TREE_NODE* tree);
  public: // main methods
   /****/
   inline TERM* IntegrateConst(TERM* c)
    {
     return &(const_index[2*(c->Functor())]);
    }; // TERM* IntegrateConst(TERM* c)
   inline TERM* Const(const TERM& c)
    {
     return &(const_index[2*(c.Functor())]);
    };
    /*******/
   
    /******
   inline TERM* IntegrateConst(TERM* c)
    {
     return &(const_index[c->Functor()]);
    }; // TERM* IntegrateConst(TERM* c)
   inline TERM* Const(const TERM& c)
    {
     return &(const_index[c.Functor()]);
    };
    ***********/

   inline TERM* IntegrateComplex(TERM* t);
   inline TERM* IntegrateFlatComplex(const FLAT_TERM* t);
   inline TERM* CopyLiteral(TERM* source);
   inline TERM* CopyFlatLiteral(const FLAT_TERM* source);
   inline TERM* CopyFlatEq(const TERM& header,const FLAT_TERM* arg1,const FLAT_TERM* arg2);
   inline void DestroyLiteralCopy(TERM* t);
   inline TERM* CopyComplex(TERM* source);
   inline void DestroyCopy(TERM* t); 
   inline void RemoveComplex(TERM* t);
   inline void RemoveComplexNonconst(TERM* t);
 }; // class SHARING_INDEX
 
class SHARING_TREE_NODE
 {
  
  private: // structure
   TERM tag;
   union
    {
     SHARING_TREE_NODE* below;
     long num;
    };
   union
    {
     SHARING_TREE_NODE* alternative;
     long dummy;
    };
  public: // construction and destruction
   SHARING_TREE_NODE(const TERM& t) : tag(t) // for making intermediate nodes
    {
     below = (SHARING_TREE_NODE*)0;
     alternative = (SHARING_TREE_NODE*)0;
    };
   SHARING_TREE_NODE(TERM* t) : tag(t) // for making leaves
    {
     num = 1;
    };
   inline void* operator new(size_t)
    {
     return Memory::Allocate(sizeof(SHARING_TREE_NODE)/4);
    };
   inline void operator delete(void* obj)
    {
     Memory::Deallocate(obj,sizeof(SHARING_TREE_NODE)/4);
    };
  public:
   inline bool IsLeaf() { return tag.IsReference(); };
   inline TERM Tag() { return tag; };
    // methods for leaves
   inline TERM* Term() { return tag.First(); }; 
   inline long Num() { return num; };
   inline void IncNum() { if (num < MAXLONG) { num++; }; };
   inline void DecNum() { if (num < MAXLONG) { num--; }; };
    // methods for intermediate nodes
   inline SHARING_TREE_NODE* Below() { return below; };
   inline SHARING_TREE_NODE** BelowAddr() { return &below; };
   inline SHARING_TREE_NODE* Alternative() { return alternative; };
   inline SHARING_TREE_NODE** AlternativeAddr() { return &alternative; };
   inline void SetBelow(SHARING_TREE_NODE* n) { below = n; };  
   inline void SetAlternative(SHARING_TREE_NODE* n) { alternative = n; }; 
 }; // class SHARING_TREE_NODE  

class SHARING_INDEX::SURFER
 {
  private:
   SHARING_INDEX* index;
   SHARING_TREE_NODE** node_addr;
  public:
   SURFER(SHARING_INDEX* si) : index(si) {};
   inline bool Reset(const TERM& fun)
    {
     CHECK6(fun.Arity(),"wrong argument in SHARING_INDEX::SURFER::Reset(..)");
     node_addr = index->Tree(fun);
     return (bool)(*node_addr);
    };
   inline bool Read(const TERM& sym)
    {
     check_tag:
      if ((*node_addr)->Tag() == sym)
       {
        node_addr = (*node_addr)->BelowAddr();
        return true;
       }
      else 
       {
        node_addr = (*node_addr)->AlternativeAddr();
        if (*node_addr) { goto check_tag; }
        else return false;
       };
    }; // bool Read(const TERM& sym) 
   inline void Write(const TERM& sym)
    {
     *node_addr = new SHARING_TREE_NODE(sym);
     node_addr = (*node_addr)->BelowAddr();
    }; // void Write(const TERM& sym)
   inline void CreateLeaf(TERM* t) { *node_addr = new SHARING_TREE_NODE(t); };
   inline SHARING_TREE_NODE* Leaf() { return (*node_addr); }; 
 }; // class SHARING_INDEX::SURFER


 
//*********************** Definitions for SHARING_INDEX: *********************

inline SHARING_TREE_NODE* SHARING_INDEX::CreateBranch(TERM* whole)
 {
  SHARING_TREE_NODE* result = new SHARING_TREE_NODE(t_iter.Symbol());
  SHARING_TREE_NODE* node = result;
  while (t_iter.Next())
   {
    node->SetBelow(new SHARING_TREE_NODE(t_iter.Symbol()));
    node = node->Below();
   };
  node->SetBelow(new SHARING_TREE_NODE(whole));
  return result; 
 }; // SHARING_TREE_NODE* SHARING_INDEX::CreateBranch(TERM* whole)
 
 

inline TERM* SHARING_INDEX::IntegrateFlatComplexNonconst(const FLAT_TERM* t)
 {
  CHECK6(t->Symbol().Arity(),"whrong argument in SHARING_INDEX::IntegrateFlatComplexNonconst(..)");
  
  
  FLAT_TERM::ITER iter;
  iter.Reset(t);
  SURFER surf(this);
  SHARING_TREE_NODE* leaf;
  TERM sym;
  
  if (surf.Reset(t->Symbol()))
   {
    read_mode_next_sym:
     iter.NextSym();
     if (iter.EndOfTerm()) // the term is in the index already
      {
       leaf = surf.Leaf();
       
       CHECK6(leaf,"A3 in SHARING_INDEX::IntegrateFlatComplexNonconst(..)");
       CHECK6(leaf->IsLeaf(),"A4 in SHARING_INDEX::IntegrateFlatComplexNonconst(..)");
  
       leaf->IncNum();
       return leaf->Term();
      }; 
     sym = iter.CurrPos()->Symbol();
     if (surf.Read(sym)) { goto read_mode_next_sym; };
     surf.Write(sym);
     goto write_mode_next_sym;
   }
  else
   {
    write_mode_next_sym:
     iter.NextSym();
     if (iter.EndOfTerm()) // branch created, leaf with a copy of the term must be created also
      {
       TERM* copy = new (t->Symbol().Arity()) TERM(t->Symbol());

       copy->Info()->SetFirst((TERM*)0);

       surf.CreateLeaf(copy);
       TERM* targ = copy->Args();
       iter.Reset(t);
       iter.NextSym();
       copy_argument:
        if (iter.CurrPos()->Symbol().IsVariable())
         {
          *targ = iter.CurrPos()->Symbol();
         }
        else // complex term
         {
          if (iter.CurrPos()->Symbol().Arity())
           {
            targ->SetFirst(IntegrateFlatComplexNonconst(iter.CurrPos()));
           }
          else // constant
           {
            //targ->SetFirst(&(const_index[iter.CurrPos()->Symbol().Functor()]));
            targ->SetFirst(Const(iter.CurrPos()->Symbol())); 
           }; 
         }; 
        iter.NextTerm();
        if (iter.EndOfTerm())
         {
          return copy;
         };
        targ = targ->Next(); 
        goto copy_argument;
       
      };
     surf.Write(iter.CurrPos()->Symbol());
     goto write_mode_next_sym;
   };
 }; // TERM* SHARING_INDEX::IntegrateFlatComplexNonconst(const FLAT_TERM* t)

inline TERM* SHARING_INDEX::IntegrateComplexNonconst(TERM* t)
 {
  CHECK6(t->Arity(),"whrong argument in SHARING_INDEX::IntegrateComplexNonconst(TERM* t)");
  TERM::ITERATOR iter;
  iter.Reset(t);
  SURFER surf(this);
  SHARING_TREE_NODE* leaf;
  TERM sym; 
  
  if (surf.Reset(*t))
   {
    read_mode_next_sym:
     if (iter.Next())
      {
       sym = iter.Symbol();
       if (surf.Read(sym)) { goto read_mode_next_sym; }
       else
        {
         surf.Write(sym);
         goto write_mode_next_sym;
        };
      }
     else // the term is in the index already
      {
       leaf = surf.Leaf();
       
       CHECK6(leaf,"A3 in SHARING_INDEX::IntegrateComplexNonconst(TERM* t)");
       CHECK6(leaf->IsLeaf(),"A4 in SHARING_INDEX::IntegrateComplexNonconst(TERM* t)");
  
       leaf->IncNum();
       return leaf->Term();
      }; 
   }
  else
   {
    write_mode_next_sym:
     if (iter.Next())
      {
       surf.Write(iter.Symbol());
       goto write_mode_next_sym;
      }
     else // branch created, leaf with a copy of the term must be created also          
      {
       TERM* copy = t->CopyTop();
     

       copy->Info()->SetFirst((TERM*)0);



       TERM* src = t->Args();
       TERM* targ = copy->Args();
       check_arg:
        switch (src->Flag())
         {
          case TERM_VAR:
           *targ = *src;
           src = src->Next();
           targ = targ->Next();
           goto check_arg;
          case TERM_REF:
           if (src->First()->Arity())
            {
             *targ = *src;
             subterms.Push(targ);
            }
           else // constant
            {
             //targ->SetFirst(&(const_index[src->First()->Functor()]));
             targ->SetFirst(Const(*(src->First())));
            };
           src = src->Next();
           targ = targ->Next();
           goto check_arg;
          case TERM_COMPLEX:
           surf.CreateLeaf(copy); 
           return copy;
         };
        return (TERM*)0; // to surpress warnings 
      };
   };
 }; // TERM* SHARING_INDEX::IntegrateComplexNonconst(TERM* t) 


inline TERM* SHARING_INDEX::IntegrateComplex(TERM* t)
 {
  if (t->Arity())
   {
    subterms.Reset();
    TERM* result = IntegrateComplexNonconst(t);
    TERM* subt;
    while (!(subterms.Empty()))
     {
      subt = subterms.Pop();
      subt->SetFirst(IntegrateComplexNonconst(subt->First()));
     };
    return result;
   }
  else // constant
   {
    //return &(const_index[t->Functor()]);
    return Const(*t);
   };
 }; // TERM* SHARING_INDEX::IntegrateComplex(TERM* t)


inline TERM* SHARING_INDEX::IntegrateFlatComplex(const FLAT_TERM* t)
 {
  if (t->Symbol().Arity())
   {
    return IntegrateFlatComplexNonconst(t);
   }
  else // constant
   {
    //return &(const_index[t->Symbol().Functor()]);
    return Const(t->Symbol());
   };
 }; // TERM* SHARING_INDEX::IntegrateFlatComplex(const FLAT_TERM* t)


inline TERM* SHARING_INDEX::CopyLiteral(TERM* source)
 {
  TERM* target = source->CopyLitTop();
  source = source->Args();
  TERM* targ = target->Args();
  copy_arg:
   switch (source->Flag())
    {
     case TERM_VAR:
      *targ = *source;
      source = source->Next();
      targ = targ->Next();
      goto copy_arg;
     case TERM_REF:
      targ->SetFirst(IntegrateComplex(source->First()));
      source = source->Next();
      targ = targ->Next();
      goto copy_arg;
     case TERM_COMPLEX:
      return target;
    };
  return (TERM*)0; // to surpress warnings
 }; // TERM* SHARING_INDEX::CopyLiteral(TERM* source)
 
inline TERM* SHARING_INDEX::CopyFlatLiteral(const FLAT_TERM* source)
 {
  TERM* target = TERM::AllocLit(source->Symbol());
  TERM* targ = target->Args();
  FLAT_TERM::ITER iter;
  iter.Reset(source);
  iter.NextSym();
  copy_arg:
   if (iter.EndOfTerm()) { return target; };
   if (iter.CurrPos()->Symbol().IsVariable())
    {
     *targ = iter.CurrPos()->Symbol(); 
    }
   else // complex
    {
     targ->SetFirst(IntegrateFlatComplex(iter.CurrPos()));
    }; 
   iter.NextTerm();
   targ = targ->Next(); 
   goto copy_arg;
 }; // TERM* SHARING_INDEX::CopyFlatLiteral(const FLAT_TERM* source)

inline TERM* SHARING_INDEX::CopyFlatEq(const TERM& header,
                                       const FLAT_TERM* arg1,
                                       const FLAT_TERM* arg2)
 {
  TERM* result = TERM::AllocLit(header);
  TERM* rarg = result->Args();
  if (arg1->Symbol().IsVariable()) { *rarg = arg1->Symbol(); }
  else rarg->SetFirst(IntegrateFlatComplex(arg1));
  rarg = rarg->Next();
  if (arg2->Symbol().IsVariable()) { *rarg = arg2->Symbol(); }
  else rarg->SetFirst(IntegrateFlatComplex(arg2));
  return result;
 }; // TERM* SHARING_INDEX::CopyFlatEq(const TERM& header,..  
 
inline void SHARING_INDEX::DestroyLiteralCopy(TERM* t)
 {
  for (TERM* arg = t->Args(); ; arg = arg->Next())
   {
    switch (arg->Flag())
     {
      case TERM_VAR : break;
      case TERM_REF : 
       RemoveComplex(arg->First());
       break;
      case TERM_COMPLEX : 
       t->DeleteLit();
       //delete t;
       return;
     };
   };
 }; // void SHARING_INDEX::DestroyLiteralCopy(TERM* t)     
 
 
inline TERM* SHARING_INDEX::CopyComplex(TERM* source)
 {
  TERM* target = source->CopyTop();


  target->Info()->SetFirst((TERM*)0);


  source = source->Args();
  TERM* targ = target->Args();
  copy_arg:
   switch (source->Flag())
    {
     case TERM_VAR:
      *targ = *source;
      source = source->Next();
      targ = targ->Next();
      goto copy_arg;
     case TERM_REF:
      targ->SetFirst(IntegrateComplex(source->First()));
      source = source->Next();
      targ = targ->Next();
      goto copy_arg;
     case TERM_COMPLEX:
      return target;
    };
  return (TERM*)0; // to surpress warnings
 }; // TERM* SHARING_INDEX::CopyComplex(TERM* source)
  
inline void SHARING_INDEX::DestroyCopy(TERM* t)
 {
  for (TERM* arg = t->Args(); ; arg = arg->Next())
   {
    switch (arg->Flag())
     {
      case TERM_VAR : break;
      case TERM_REF : 
       RemoveComplex(arg->First());
       break;
      case TERM_COMPLEX : 
       delete t;
       return;
     };
   };
 }; // void SHARING_INDEX::DestroyCopy(TERM* t)     
 
  
inline void SHARING_INDEX::RemoveComplex(TERM* t)
 {
  if (t->Arity())
   {
    RemoveComplexNonconst(t);
   };
 }; // void SHARING_INDEX::Remove(TERM* t)
  
inline void SHARING_INDEX::RemoveComplexNonconst(TERM* t)
 {
  CHECK0((t->Flag() == TERM_COMPLEX) && (t->Arity()),"A1 in in SHARING_INDEX::RemoveComplexNonconst()");
  
  STACK<TERM*,MAX_TERM_WEIGHT> subterms("MAX_TERM_WEIGHT in SHARING_INDEX::RemoveComplexNonconst()");
  TERM* subterm;
  SHARING_TREE_NODE* node;     
  SHARING_TREE_NODE** branch;
  SHARING_TREE_NODE** below_addr; 
  TERM::ITERATOR iter;
  SHARING_TREE_NODE* branch_to_delete;
  SHARING_TREE_NODE* node_to_delete;
 
  process_next_term: 
  
   CHECK0((t->Flag() == TERM_COMPLEX) && (t->Arity()),"A2 in in SHARING_INDEX::RemoveComplexNonconst()");
  
   node = nonconst_index[t->Functor()];  
   
   CHECK0(node,"A3 in in SHARING_INDEX::RemoveComplexNonconst()");
      
   branch = &(nonconst_index[t->Functor()]);
   below_addr = branch;
    iter.Reset(t);
    iter.Next();
    do
     {
     
      CHECK0(!(node->IsLeaf()),"A4 in in SHARING_INDEX::RemoveComplexNonconst()");
       
      if (node->Alternative())
       {
        branch = below_addr;
        check_node:
         if (node->Tag() == iter.Symbol())
          {
           below_addr = node->BelowAddr();
           node = node->Below();
          }
         else // try alternative
          {
           branch = node->AlternativeAddr();
           node = node->Alternative();
           goto check_node;
          };
       }
      else
       {
        below_addr = node->BelowAddr();
        node = node->Below();
       };
     }
    while (iter.Next());
   
    // here node is a leaf
   
    node->DecNum();
    if (!(node->Num())) 
     {
      // last occurence of the term, top level of the term must be destroyed,
      // the branch must be deleted
     
      subterm = node->Term()->Args();
      while (!(subterm->IsNull()))
       {
        if ((subterm->IsReference()) && (subterm->First()->Arity()))
         {
          subterms.Push(subterm->First());
         };
        subterm = subterm->Next();
       }; 
     
      delete node->Term(); 
      
      branch_to_delete = *branch;
      *branch = (*branch)->Alternative(); 
      while (!(branch_to_delete->IsLeaf()))
       {
        node_to_delete = branch_to_delete;
        branch_to_delete = branch_to_delete->Below();
        delete node_to_delete;
       };
      delete branch_to_delete; // leaf
      if (!(subterms.Empty()))
       {
        t = subterms.Pop();
        goto process_next_term;
       };
     };     
}; // void SHARING_INDEX::RemoveComplexNonconst(TERM* const t)

inline void SHARING_INDEX::DestroyTree(SHARING_TREE_NODE* tree)
 {
  if (tree)
   {
    if (tree->IsLeaf()) { delete tree->Term(); delete tree; }
    else
     {
      if (tree->Alternative()) DestroyTree(tree->Alternative());
      if (tree->Below()) DestroyTree(tree->Below());
      delete tree;
     };
   };
 }; // void SHARING_INDEX::DestroyTree(SHARING_TREE_NODE* tree)

//*************************************************************************** 
 









