#define __SIGNATURE_H

#include <string.h>

#ifndef __ARRAY_H
 #include "array.h"
#endif

#ifndef __STACK_H
 #include "stack.h"
#endif

enum PRIORITY
 {
  PR_NOOP = -1,	// must be less than the minimal priority
  PR_CONS = -1,
  PR_MIN = 0,

  PR_EQUAL = 100,
  PR_UNEQUAL = 100,

  PR_MAX = 1000,

  PR_NOT = 1001,
  PR_ALL = 1001,
  PR_SOME = 1001,
  PR_DIAM = 1001,
  PR_BOX = 1001,
  PR_AND = 1002,
  PR_OR = 1003,
  PR_IMP = 1004,
  PR_IFF = 1005,

  PR_HIGH = 2000 // must be more than any priority
 };

class SYM_TAB
 {
  private:
   ARRAY<char*> symbol;
   int next_symbol;
  public:
   SYM_TAB() : symbol((char*)0), next_symbol(0) {};
   
   inline void Reset() { ResetDownto(0); };
   
   inline void ResetDownto(const int num)
    {
     for (int i = num; i < next_symbol; i++)
      {
       symbol[i] = (char*)0;
      };
     next_symbol = num;
    };
   
   inline int Number(const char* const name)
    {
     for (int i = 0; i < next_symbol; i++)
      {
       if (!(strcmp(name,symbol[i]))) { return i; };
      };
     //symbol[next_symbol] = name;
     
     symbol[next_symbol] = new char[strlen(name)+1];
     strcpy(symbol[next_symbol],name);
     
     next_symbol++;
     return next_symbol-1;
    }; // int Number(const char* const name)
   inline int MaxNumber() { return (next_symbol - 1); };
   inline const char* Name(const int num) { return symbol[num]; };
 }; // class SYM_TAB 

#include <iostream.h>

class SIGNATURE
 {
  private:
   SYM_TAB variable;
   SYM_TAB functor;
   //ARRAY<bool> skolem;
   const char* const skolem_const_prefix = "$c";
   const char* const skolem_func_prefix = "$f";
   const char* const skolem_pred_prefix = "$p";
   int max_input_fun;
   int skolem_next;
  public:
   SIGNATURE() : 
    variable(), 
    functor(), 
    //skolem(false), 
    max_input_fun(0),
    skolem_next(0) 
    {
     int eq_num = InputFunctorNum("equal");
     if (eq_num != EQ_NUM) RuntimeError("WRONG NUMBER FOR EQUALITY!");
     int ord_eq_num = InputFunctorNum("Equal");
     if (ord_eq_num != ORD_EQ_NUM) RuntimeError("WRONG NUMBER FOR ORDERED EQUALITY!");
    }; 
   inline void ResetSkolem()
    {
     functor.ResetDownto(max_input_fun+1);
     skolem_next = 0;
    };
   inline void ResetVariables() 
    {
     variable.Reset();
    };
   inline int VarNum(const char* const var_name)
    {
     return variable.Number(var_name);
    };
   inline int InputFunctorNum(const char* const fun_name)
    {
    
     //cout << "FUNCTOR " << fun_name << "\n"; // debug
     
     //int res = functor.Number(fun_name); 
    
     //cout << res << " " << FunName(res) << "\n";
     
     //return res;
    
     return functor.Number(fun_name);
    };
   inline void MemoLastInputFun()
    {
     max_input_fun = functor.MaxNumber();
    };
   inline const char* FunName(const int num) 
    {
     switch (num)
      {
       case EQ_NUM : return "=";
       case ORD_EQ_NUM : return "==";
       default : return functor.Name(num); 
      }; 
    };
   inline int SkolemConst();
   inline int SkolemFunc();
   inline int SkolemPred();   
 }; // class SIGNATURE

//*********** Auxilliary functions: ******************************************

inline char DigitToChar(int digit)
 {
  switch (digit)
   {
    case 0 : return '0';
    case 1 : return '1';
    case 2 : return '2';
    case 3 : return '3';
    case 4 : return '4';
    case 5 : return '5';
    case 6 : return '6';
    case 7 : return '7';
    case 8 : return '8';
    case 9 : return '9';
    default : return ' ';
   };
 };

inline char* IntToString(int num)
 {
  STACK<char,32> characters("32 in IntToString:characters");
  bool negative = false;
  if (num < 0)
   {
    num = -1 * num;
    negative = true;
   }; 
  do
   {
    characters.Push(DigitToChar(num % 10));
    num = num / 10;
   }
  while (num);
  if (negative) { characters.Push('-'); };
  int len = characters.NumberOfElements();
  char* res = new char[len+1];
  res[len] = 0;
  for (int i = 0; i < len; i++)
   {
    res[i] = characters.Pop();
   };
  return res;
 };

//********************* Skolem symbols: *************************************

inline int SIGNATURE::SkolemConst()
 {
  char* suffix = IntToString(skolem_next);
  skolem_next++;
  int len = strlen(skolem_const_prefix) + strlen(suffix);
  char* name = new char[len+1];
  name[len] = 0;
  strcpy(name,skolem_const_prefix);
  strcpy(name+strlen(skolem_const_prefix),suffix);
  delete [] suffix;
  int num = functor.Number(name);
  //skolem[num] = true;
  return num;
 }; // int SIGNATURE::SkolemConst()

inline int SIGNATURE::SkolemFunc()
 {
  char* suffix = IntToString(skolem_next);
  skolem_next++;
  int len = strlen(skolem_func_prefix) + strlen(suffix);
  char* name = new char[len+1];
  name[len] = 0;
  strcpy(name,skolem_func_prefix);
  strcpy(name+strlen(skolem_func_prefix),suffix);
  delete [] suffix;
  int num = functor.Number(name);
  //skolem[num] = true;
  return num;
 }; // int SIGNATURE::SkolemFunc()

inline int SIGNATURE::SkolemPred()
 {
  char* suffix = IntToString(skolem_next);
  skolem_next++;
  int len = strlen(skolem_pred_prefix) + strlen(suffix);
  char* name = new char[len+1];
  name[len] = 0;
  strcpy(name,skolem_pred_prefix);
  strcpy(name+strlen(skolem_pred_prefix),suffix);
  delete [] suffix;
  int num = functor.Number(name);
  //skolem[num] = true;
  return num;
 }; // int SIGNATURE::SkolemPred() 

//********************************************************************************