#define __SPLITTING_H

#ifndef __NEW_CLAUSE_H
 #include "new_clause.h"
#endif

#ifndef __SIGNATURE_H
 #include "signature.h"; 
#endif

extern SIGNATURE Signature;
 
//**************************************************************************

class SPLITTING_INDEX_TREE_NODE
 {
  private: // structure
   TERM symbol;
   SPLITTING_INDEX_TREE_NODE* less;
   SPLITTING_INDEX_TREE_NODE* greater;
   SPLITTING_INDEX_TREE_NODE* next;
  public:
   SPLITTING_INDEX_TREE_NODE()
    : less((SPLITTING_INDEX_TREE_NODE*)0),
      greater((SPLITTING_INDEX_TREE_NODE*)0),
      next((SPLITTING_INDEX_TREE_NODE*)0)
    {};
    
   SPLITTING_INDEX_TREE_NODE(const SPLITTING_INDEX_TREE_NODE& node) // copy constructor
    : symbol(node.symbol),
      less(node.less),
      greater(node.greater),
      next(node.next)
    {};
   inline void* operator new(size_t)
    {
     return Memory::Allocate(sizeof(SPLITTING_INDEX_TREE_NODE)/4);
    };
   inline void operator delete(void* obj)
    {
     Memory::Deallocate(obj,sizeof(SPLITTING_INDEX_TREE_NODE)/4);
    };
   
   inline SPLITTING_INDEX_TREE_NODE& operator= (const SPLITTING_INDEX_TREE_NODE& node)
    {
     symbol = node.symbol;
     less = node.less;
     greater = node.greater;
     next = node.next;
     return *this;
    };
    
   inline TERM& Symbol() { return symbol; };
   inline bool IsLeaf() const { return symbol.IsReference(); };
   inline void MakeLeaf() { symbol.SetFirst((TERM*)0); };
   inline SPLITTING_INDEX_TREE_NODE*& Less() { return less; };
   inline SPLITTING_INDEX_TREE_NODE*& Greater() { return greater; };
   inline SPLITTING_INDEX_TREE_NODE*& Next() { return next; }; 
   
   // for leaves only:
   inline int Name() const { return (int)less; };
   inline void SetName(const int n) { less = (SPLITTING_INDEX_TREE_NODE*)n; };

   // inserting a leaf
   inline void InsertLeaf()
    {
     SPLITTING_INDEX_TREE_NODE* copy_of_this = new SPLITTING_INDEX_TREE_NODE(*this);
     MakeLeaf();
     next = copy_of_this;
    };
   
 }; // class SPLITTING_INDEX_TREE_NODE 
 
class SPLITTING_INDEX_TREE
 {
  private:
   static inline bool Track(const TERM sym,SPLITTING_INDEX_TREE_NODE*& node)
    {
     compare_sym:
      if (sym == node->Symbol()) 
       {
        node = node->Next();
        return true; 
       };
      if ( sym.Content() > node->Symbol().Content())
       {
        if (node->Greater()) 
         {
          node = node->Greater();
          goto compare_sym;
         }
        else { return false; };
       }
      else
       {
        if (node->Less())
         {
          node = node->Less();
          goto compare_sym;
         }
        else { return false; };
       };
    }; // bool Track(const TERM sym,SPLITTING_INDEX_TREE_NODE*& node)
    
 public:
  static inline bool Track(TERM::ITERATOR& iter,SPLITTING_INDEX_TREE_NODE*& node)
   {
    track_symbol:
     if (Track(iter.Symbol(),node)) 
      {
       if (iter.Next()) { goto track_symbol; }
       else { return true; };
      }
     else { return false; }; 
   }; // bool Track(TERM::ITERATOR& iter,SPLITTING_INDEX_TREE_NODE*& node)
   
  static inline bool Track(TMP_LIT_LIST::CONCAT& iter,SPLITTING_INDEX_TREE_NODE*& node)
   {
    track_symbol:
     if (Track(iter.Symbol(),node)) 
      {
       iter.NextSym();
       if (iter.Empty()) return true;
       goto track_symbol;
      }
     else return false;  
   }; // bool Track(TMP_LIT_LIST::CONCAT& iter,SPLITTING_INDEX_TREE_NODE*& node) 
 
  static inline SPLITTING_INDEX_TREE_NODE* NewBranch(TERM::ITERATOR& iter, 
                                                     SPLITTING_INDEX_TREE_NODE*& leaf)
   {
    SPLITTING_INDEX_TREE_NODE* result = new SPLITTING_INDEX_TREE_NODE();
    leaf = result;
    do 
     {
      leaf->Symbol() = iter.Symbol();
      leaf->Next() = new SPLITTING_INDEX_TREE_NODE();
      leaf = leaf->Next();
     }
    while (iter.Next());
    leaf->MakeLeaf();
    return result;
   }; // SPLITTING_INDEX_TREE_NODE* NewBranch(TERM::ITERATOR& iter,SPLITTING_INDEX_TREE_NODE*& leaf)
  
  static inline SPLITTING_INDEX_TREE_NODE* NewBranch(TMP_LIT_LIST::CONCAT& iter, 
                                                     SPLITTING_INDEX_TREE_NODE*& leaf)
   {
    SPLITTING_INDEX_TREE_NODE* result = new SPLITTING_INDEX_TREE_NODE();
    leaf = result;
    do 
     {
      leaf->Symbol() = iter.Symbol();
      leaf->Next() = new SPLITTING_INDEX_TREE_NODE();
      leaf = leaf->Next();
      iter.NextSym();
     }
    while (iter.Nonempty());
    leaf->MakeLeaf();
    return result;
   }; // SPLITTING_INDEX_TREE_NODE* NewBranch(TMP_LIT_LIST::CONCAT& iter,SPLITTING_INDEX_TREE_NODE*& leaf)
  
  
  static inline void Continue(TERM::ITERATOR& iter,SPLITTING_INDEX_TREE_NODE*& node)
   {
    // After applying this function node will point to the newly created leaf.
    // Here iter.Symbol().Content() != node->Symbol().Content().
    
    if (iter.Symbol().Content() > node->Symbol().Content())
     {
      node->Greater() = new SPLITTING_INDEX_TREE_NODE();
      node = node->Greater();
     }
    else // iter.Symbol().Content() > node->Symbol().Content()
     {
      node->Less() = new SPLITTING_INDEX_TREE_NODE();
      node = node->Less();
     };
    node->Symbol() = iter.Symbol(); 
       
    if (iter.Next())
     {
      SPLITTING_INDEX_TREE_NODE* leaf;
      node->Next() = NewBranch(iter,leaf);
      node = leaf;
     }
    else
     {
      node->Next() = new SPLITTING_INDEX_TREE_NODE();
      node = node->Next();
      node->MakeLeaf();      
     };
   }; // void Continue(TERM::ITERATOR& iter,SPLITTING_INDEX_TREE_NODE*& node)
  
  static inline void Continue(TMP_LIT_LIST::CONCAT& iter,SPLITTING_INDEX_TREE_NODE*& node)
   {
    // After applying this function node will point to the newly created leaf.
    // Here iter.Symbol().Content() != node->Symbol().Content().
    
    if (iter.Symbol().Content() > node->Symbol().Content())
     {
      node->Greater() = new SPLITTING_INDEX_TREE_NODE();
      node = node->Greater();
     }
    else // iter.Symbol().Content() > node->Symbol().Content()
     {
      node->Less() = new SPLITTING_INDEX_TREE_NODE();
      node = node->Less();
     };
    node->Symbol() = iter.Symbol(); 
    
    iter.NextSym();
       
    if (iter.Nonempty())
     {
      SPLITTING_INDEX_TREE_NODE* leaf;
      node->Next() = NewBranch(iter,leaf);
      node = leaf;
     }
    else
     {
      node->Next() = new SPLITTING_INDEX_TREE_NODE();
      node = node->Next();
      node->MakeLeaf();      
     };
   }; // void Continue(TMP_LIT_LIST::CONCAT& iter,SPLITTING_INDEX_TREE_NODE*& node) 
   
   
  static void Destroy(SPLITTING_INDEX_TREE_NODE* tree)
   {
    if (tree)
     {
      if (!(tree->IsLeaf()))
       {
        Destroy(tree->Less());
        Destroy(tree->Greater());
        Destroy(tree->Next()); 
       };
      delete tree; 
     };
   }; // void Destroy(SPLITTING_INDEX_TREE_NODE* tree)
   
 }; // class SPLITTING_INDEX_TREE
 
 
//************************************************************************** 

class SPLITTING_INDEX
 {
  private:
   SPLITTING_INDEX_TREE_NODE* index_tree;
  public:
   SPLITTING_INDEX() : index_tree((SPLITTING_INDEX_TREE_NODE*)0) {};
   inline int ComponentName(TERM* lit_list,bool& new_component);
   inline int ComponentName(TMP_LIT_LIST& lit_list,bool& new_component); 
   void Destroy()
    {
     SPLITTING_INDEX_TREE::Destroy(index_tree);
     index_tree = (SPLITTING_INDEX_TREE_NODE*)0;
    };
   void Reset() { Destroy(); Signature.ResetSkolem(); }; 
 }; // class SPLITTING_INDEX
 
//*************** Definitions for SPLITTING_INDEX: *************************

inline int SPLITTING_INDEX::ComponentName(TERM* lit_list,bool& new_component)
 {
  TERM::ITERATOR iter;
  iter.Reset(lit_list);
  iter.Next();
  SPLITTING_INDEX_TREE_NODE* node = index_tree;
  if (node)
   {
    if (SPLITTING_INDEX_TREE::Track(iter,node))
     {
      if (node->IsLeaf())
       {
        new_component = false;
       }
      else
       {
        new_component = true;
        node->InsertLeaf();
       };
     }
    else
     {
      new_component = true;
      SPLITTING_INDEX_TREE::Continue(iter,node);
     };
   }
  else
   {
    new_component = true;
    index_tree = SPLITTING_INDEX_TREE::NewBranch(iter,node);
   };
  
  // node is a leaf here
  if (new_component)
   {
    node->SetName(Signature.SkolemPred());
   };
  return node->Name();
 }; // int SPLITTING_INDEX::ComponentName(TERM* lit_list,bool& new_component)

inline int SPLITTING_INDEX::ComponentName(TMP_LIT_LIST& lit_list,bool& new_component)
 {

  //int res =  Signature.SkolemPred(); // debug
  //cout << res << "\n"; // debug
  //return res; // debug


  TMP_LIT_LIST::CONCAT iter(lit_list);
  SPLITTING_INDEX_TREE_NODE* node = index_tree;
  if (node)
   {
    if (SPLITTING_INDEX_TREE::Track(iter,node))
     {
      if (node->IsLeaf())
       {
        new_component = false;
       }
      else
       {
        new_component = true;
        node->InsertLeaf();
       };
     }
    else
     {
      new_component = true;
      SPLITTING_INDEX_TREE::Continue(iter,node);
     };
   }
  else
   {
    new_component = true;
    index_tree = SPLITTING_INDEX_TREE::NewBranch(iter,node);
   };
  
  // node is a leaf here
  if (new_component)
   {
    node->SetName(Signature.SkolemPred());
   };
  return node->Name();
 }; // int SPLITTING_INDEX::ComponentName(TMP_LIT_LIST& lit_list,bool& new_component)
 
//**************************************************************************
