#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif

#ifndef __TERM_H
 #include "term.h" 
#endif

#ifndef __WORD_H
 #include "word.h"
#endif

#ifndef __OUTPUT_H
 #include "output.h"
#endif

#define __SUBST_H
   
//const int MAX_NUMBER_OF_INDICES = MAX_NUMBER_OF_LITERALS + 1;

class SUBST
 {
  private: // structure
   TERM* subst[MAX_NUMBER_OF_VARS*MAX_NUMBER_OF_INDICES*2];
  public:
   static inline void Bind(TERM** bind,int index,TERM* word)
    {
     *bind = word;
     *(bind+1) = (TERM*)index;
    };
   static inline void Unbind(TERM** bind) { *bind = (TERM*)0; };
  public:
   inline TERM** Binding(int index,int var)
    {
     return subst + ((index * MAX_NUMBER_OF_VARS) + var) * 2;
    };
   inline bool Bound(int index,int var)
    {
     return (bool)(*(Binding(index,var)));
    };
   inline void VarAndIndex(TERM** bind,int& ind,int& var) // for debugging
    {
     int tmp = ((int)(bind - subst))/2;
     ind = tmp/MAX_NUMBER_OF_VARS;
     var = tmp%MAX_NUMBER_OF_VARS;
    };
   static inline int Index(TERM** bind) { return (int)(*(bind+1)); };
   inline void UnrefUpToLastVar(int& index,int& var);
   inline void Unref(int& index,int var,TERM*& word);
   inline void UnrefCompletely(TERM**& bind);
   inline void UnrefCompletely(int& index,int var,TERM*& word);
   inline void OutputBinding(TERM** bind); // for debugging
   SUBST() {};
 }; // class SUBST

class BINDINGS : public SUBST
 {
  private: // structure
   STACK<TERM**,MAX_NUMBER_OF_VARS*MAX_NUMBER_OF_INDICES> bindings;
  public: 
   inline void MarkSection() 
    { bindings.Push((TERM**)0); };
   inline void AddBinding(TERM** bind,int index,TERM* word)
    {
     Bind(bind,index,word);
     bindings.Push(bind);
    };
   inline void AddBinding(int var_index,int var,int index,TERM* word)
    {
     AddBinding(Binding(var_index,var),index,word);
    };
   inline void UnbindSection()
    {
     TERM** bind = bindings.Pop();
     while (bind)
      {
       Unbind(bind);
       bind = bindings.Pop();
      };
    };
   inline void UndoSingleAssignment() { Unbind(bindings.Pop()); }; 
   inline void Reset()
    {
     while (!(bindings.Empty()))
      {
       UnbindSection();
      };
     MarkSection();
    };
   BINDINGS() : bindings("MAX_NUMBER_OF_VARS*MAX_NUMBER_OF_INDICES") 
    { MarkSection(); };
   inline void* State() { return (void*)(bindings.Mark()); };
   inline void RestoreState(void* state)
    {
     TERM** bind;
     TERM*** mark = (TERM***)state;
     while (!(bindings.OnMark(mark)))
      {
       bind = bindings.Pop();
       if (bind) { Unbind(bind); };
      };
    };
   
   inline void OutputBindings(); // for debugging
 }; // class BINDINGS

class VAR_SET
 {
  protected: // structure
   bool member[MAX_NUMBER_OF_VARS][MAX_NUMBER_OF_INDICES];
   STACK<bool*,MAX_NUMBER_OF_VARS> all_members;
  public: 
   VAR_SET() : all_members("MAX_NUMBER_OF_VARS") 
    {
     for (int i = 0; i < MAX_NUMBER_OF_INDICES; i++)
      for (int v = 0; v < MAX_NUMBER_OF_VARS; v++)
       { member[v][i] = false; };
    };
   inline void Reset()
    {
     while (!(all_members.Empty()))
      {
       *(all_members.Pop()) = false;
      };
    };  
   inline bool Member(int index,int var) { return member[var][index]; };
   inline void NewMember(int index,int var)
    {
     bool* tmp = &(member[var][index]);
     *tmp = true;
     all_members.Push(tmp);
    }; 
   inline void ForgetLast()
    {
     *(all_members.Pop()) = false;
    };
 }; // class VAR_SET 
 
class VAR_SET_WITH_BACKTRACK : public VAR_SET
 {
  private:
   int last_section;
   STACK<int,MAX_NUMBER_OF_VARS> sections;
  public:
   VAR_SET_WITH_BACKTRACK() 
    : VAR_SET(),
      last_section(0),
      sections("MAX_NUMBER_OF_VARS") 
    { };
   inline void Reset() 
    {
     VAR_SET::Reset(); 
     sections.Reset();
     last_section = 0;
    };
   inline void NewMember(int index,int var)
    {
     last_section++;
     VAR_SET::NewMember(index,var);
    };
   inline void Mark() 
    {
     sections.Push(last_section);
     last_section = 0;
    };
   inline void Backtrack() 
    {
     while (last_section)
      {
       ForgetLast();
       last_section--;
      };
     last_section = sections.Pop();
    };
 }; // VAR_SET_WITH_BACKTRACK 

class BINDINGS_WITH_OCCUR_CHECK : public BINDINGS
 {
  private: 
   VAR_SET VarSet;
   STACK<TERM**,MAX_NUMBER_OF_VARS> links_to_check;  
  public:
   BINDINGS_WITH_OCCUR_CHECK() 
    : VarSet(),links_to_check("MAX_NUMBER_OF_VARS")
    {};
   bool Occurs(int var_ind,int var,int index,TERM* word);
 }; // BINDINGS_WITH_OCCUR_CHECK

class NEW_VARS
 {
  private:
   int new_var[MAX_NUMBER_OF_VARS][MAX_NUMBER_OF_INDICES];
   STACK<int*,MAX_NUMBER_OF_VARS> all_new_vars;
   int next_var;
  public:
   NEW_VARS() : all_new_vars("MAX_NUMBER_OF_VARS") 
    {
     for (int i = 0; i < MAX_NUMBER_OF_INDICES; i++)
      for (int v = 0; v < MAX_NUMBER_OF_VARS; v++)
       { new_var[v][i] = 0; };
    };
   inline void Reset()
    {
     next_var = 1; // this is not a bug. See int NewVar(int index,int var).
     while (!(all_new_vars.Empty()))
      {
       *(all_new_vars.Pop()) = 0;
      };
    };
   inline int NewVar(int index,int var)
    {
     int v = new_var[var][index];
     if (!v)
      { 
       new_var[var][index] = next_var;
       all_new_vars.Push(&(new_var[var][index]));
       v = next_var;
       next_var++;
      };
     return v-1;
    };
 }; // class NEW_VARS

//********** Methods of SUBST: ********************************** 

inline void SUBST::UnrefUpToLastVar(int& index,int& var)
 {
  TERM** bind;
  next:
   bind = Binding(index,var);
   if ((*bind) && ((*bind)->IsVariable()))
    {
     index = Index(bind);
     var = (*bind)->Var();
     goto next;
    };
 }; // void SUBST::UnrefUpToLastVar(int& index,int& var)
 
inline void SUBST::Unref(int& index,int var,TERM*& word)
 {
  TERM** bind = Binding(index,var);
  word = *bind;
  if (word)
   {
    index = Index(bind);
   };
 }; // void SUBST::Unref(int& index,int var,TERM*& word)

inline void SUBST::UnrefCompletely(TERM**& bind)
 {
  int var;
  TERM** tmp_bind;
  next:
   if ((*bind)->IsVariable())
    {
     var = (*bind)->Var();
     tmp_bind = bind;
     bind = Binding(Index(bind),var);
     if (*bind) // bound variable
      {
       goto next;
      }
     else // free variable
      {
       bind = tmp_bind;
       return;
      };
    }
   else
    {
     return;
    };
 }; // void SUBST::UnrefCompletely(TERM**& bind)
 
inline void SUBST::UnrefCompletely(int& index,int var,TERM*& word)
 {
  Unref(index,var,word);
  if ((word) && (word->IsVariable()))
   {
    TERM* tmp_word;  
    unref_next:     
     Unref(index,word->Var(),tmp_word);
     if (tmp_word)
      {
       word = tmp_word;
       if (word->IsVariable())
        {
         goto unref_next;
        };
      };
   };
 }; // void SUBST::UnrefCompletely(int& index,int var,TERM*& word)
 
inline void SUBST::OutputBinding(TERM** bind) // for debugging
 {
  int ind;
  int var;
  VarAndIndex(bind,ind,var);
  OutputVar(var);
  Output("{");
  Output(ind);
  Output("} -> <");
  WORD::OutputWord(*bind);
  Output(">");
  Output("{");
  Output(Index(bind));
  Output("}");
 }; // void SUBST::OutputBinding(TERM** bind) // for debugging 

//********** Methods of BINDINGS: ********************************** 

inline void BINDINGS::OutputBindings() // for debugging
 {
  STACK<TERM**,MAX_NUMBER_OF_VARS> rev_bindings("MAX_NUMBER_OF_VARS");
  TERM** bind;
  while (!(bindings.Empty()))
   {
    rev_bindings.Push(bindings.Pop());
   };
  Output("[");
  while (!(rev_bindings.Empty()))
   { 
    bind = rev_bindings.Pop();
    bindings.Push(bind);
    if (bind)
     {
      OutputBinding(bind);
      if (!(rev_bindings.Empty()))
       {
        Output(";");
       };
      }
     else
      {
       Output("|");
      };
   };
  Output("]");
 }; // void BINDINGS::OutputBindings() // for debugging 

//*******************************************************************













