#define __SUPERPOSITION_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif 
#ifndef __PARA_MACHINE_H
 #include "para_machine.h"
#endif 
#ifndef __RESOLVENT_H
 #include "resolvent.h"
#endif
#ifndef __FORWARD_SUPERPOSITION_H
 #include "forward_superposition.h" 
#endif


//*************************************************************

class SUPERPOSITION;
extern SUPERPOSITION Superposition;

extern void AnotherBackwardSimplified(Clause* simplified_cl,Clause* simplifier);

//*************************************************************

extern RESOLVENT Resolvent;


class BPARA_MACH
 {
  private:
   PARA_MACH para_mach;
   bool init_state;
   ORD_CONSTR* ord_constr; // experimental
   Clause* query; // mostly for debugging
   bool unit_query;
   bool simplification_possible;
  public:
   BPARA_MACH(ORD_CONSTR* constr) : para_mach(), ord_constr(constr) {};
   inline void SetIndex(int ind) { para_mach.SetIndex(ind); };
   inline int GetIndex() const { return para_mach.GetIndex(); };
   inline void SetUnitQuery(bool fl) { unit_query = fl; };
   inline void SetQuery(Clause* cl) { query = cl; };
   inline bool VariablesFromIndexAffected() { return para_mach.VariablesFromIndexAffected(); };
   inline bool SimplificationPossible() const { return simplification_possible; }; 
   inline void LoadLHS(PREFIX_SYM* t)
    {
     para_mach.CompileComplexTerm(t);     
     init_state = true;
    };
   inline RES_PAIR_LIST* NextCandidate() 
    {
     RES_PAIR_LIST* res;
     if (init_state)
      {
       init_state = false;
       if (para_mach.FindFirstLeaf(*(ORIndex.BParaTree())))
        {
	         
         if (ord_constr->Violated()) // experimental
          {           
           goto try_next_leaf;
          };     
	
         simplification_possible = unit_query && ord_constr->Trivial() && (!VariablesFromIndexAffected()); 
                    
         para_mach.ResetCandidatesFromLeaf();
         return para_mach.NextCandidateFromTheLeaf();
        }
       else return (RES_PAIR_LIST*)0;
      }
     else
      {
       res = para_mach.NextCandidateFromTheLeaf();
       if (res) 
        {
         return res;
        };
 
       try_next_leaf: 
        if (para_mach.FindNextLeaf())
         {
	 
          if (ord_constr->Violated()) // experimental
           {
            //ord_constr->Output(); cout << "\n\n"; // debug

            goto try_next_leaf;
           };
         
          simplification_possible = unit_query && ord_constr->Trivial() && (!VariablesFromIndexAffected());          

          para_mach.ResetCandidatesFromLeaf();
          return para_mach.NextCandidateFromTheLeaf();
         }
        else return (RES_PAIR_LIST*)0;  
      }; 
    }; 
 }; // class BPARA_MACH
 
class BPARA_MACHINE 
 {
  private:
   BPARA_MACH bpara_mach;
   PREFIX_SYM fst_arg_mem[MAX_TERM_WEIGHT];
   PREFIX_SYM snd_arg_mem[MAX_TERM_WEIGHT];
   PREFIX_SYM* rhs;
   int index;
   int eq_index;
   bool unordered_eq;
   bool wrap_args;
   RES_PAIR_LIST* candidate;

   ORD_CONSTR ordering_constraint; // experimental    

   Clause* query;
  public: 
   BPARA_MACHINE() 
    : bpara_mach(&ordering_constraint),
     
      ordering_constraint() // experimental 
   {
    fst_arg_mem[0].Head().MkFunc(TERM((ARITY)1,(FUNCTOR)0,POL_POS));
    snd_arg_mem[0].Head().MkFunc(TERM((ARITY)1,(FUNCTOR)0,POL_POS));
   };
   inline void SetIndex(int ind) { index = ind; bpara_mach.SetIndex(ind); };
   inline int GetIndex() const { return index; };    
   inline bool VariablesFromIndexAffected() { return bpara_mach.VariablesFromIndexAffected(); };
   inline void SetQuery(Clause* cl) 
    {
     query = cl;
     bpara_mach.SetQuery(cl);
     bpara_mach.SetUnitQuery(cl->Unit());
    };
   inline RES_PAIR_LIST* Candidate() const { return candidate; };
   inline void LoadEq(const TERM* eq_lit,int eq_ind)
    {
     CHECK16(eq_lit->IsEquality(),"A0 in BPARA_MACHINE::LoadEq(..)");
     CHECK16(eq_lit->Positive(),"A1 in BPARA_MACHINE::LoadEq(..)");
     
     eq_index = eq_ind; 
     const TERM* arg_ref = eq_lit->Args();
     PREFIX_SYM* mem = fst_arg_mem + 1;
     StandardTermToPrefix(eq_index,*arg_ref,mem);
     fst_arg_mem[0].SetBracket(mem);
     
     // another argument
     arg_ref = arg_ref->Next();
     mem = snd_arg_mem + 1;
     StandardTermToPrefix(eq_index,*arg_ref,mem);
     snd_arg_mem[0].SetBracket(mem);
     
     bpara_mach.LoadLHS(fst_arg_mem);
     rhs = snd_arg_mem + 1;
     unordered_eq = eq_lit->IsUnorderedEq();
     wrap_args = unordered_eq;
   
     /****** experimental ******/
     if (unordered_eq)
      {
       ordering_constraint.Compile(fst_arg_mem,snd_arg_mem);
      }
     else ordering_constraint.MakeAlwaysValid();

     /************************/
    };
   
   inline bool Paramodulate()
    {
     try_next_cand:
      candidate = bpara_mach.NextCandidate();
      if (candidate) 
       {  
        return true;
       };
      if (wrap_args)
       {
        wrap_args = false;
        bpara_mach.LoadLHS(snd_arg_mem);
        rhs = fst_arg_mem + 1;    

        /******* experimental ********/ 
         ordering_constraint.Inverse();
        /*****************************/
        goto try_next_cand;  
       };
      return false; 
    };  
  
  inline bool CollectLiterals() 
   {
    Clause* the_clause = candidate->TheClause();
    if (!the_clause->Active()) return false; 

    /********* experimental  *********/

    if (bpara_mach.SimplificationPossible() && (query != the_clause))
     {           
      //::Output("CLAUSE "); ::Output(query); ::Output("\n\n");
      //::Output("SIMPLIFIES "); ::Output(candidate->TheClause()); ::Output("\n\n\n\n");

      AnotherBackwardSimplified(the_clause,query);

     };
       
    /************************/

    int rewritten_lit_num = candidate->LitNum();
    int curr_lit_num = 0;
    for (TERM* lit_ref = the_clause->Literals()->Args();
         !(lit_ref->IsNull());
         lit_ref = lit_ref->Next())
     {
      if (curr_lit_num == rewritten_lit_num)
       {
        
        if (!Resolvent.PushRewrittenLit(index,lit_ref->First(),candidate->Redex(),rhs))
         return false;
       }
      else
       {
        if (!Resolvent.PushStandardLit(index,lit_ref->First())) return false;  
       }; 
      curr_lit_num++;     
     };    
    return true;
   };
    
 }; // class BPARA_MACHINE 




class BPARA_PREMISE
 { 
  private:
   int index;
   Clause* clause;
   const TERM* first_max_lit_ref;
   const TERM* first_nonmax_lit_ref;
   const TERM* next_max_lit_ref;
   WEIGHT_FORMULA nonmax_weight_form;
   WEIGHT_FORMULA max_weight_form;
  public:
   BPARA_PREMISE() : nonmax_weight_form(), max_weight_form() {}; 
   inline void SetIndex(int ind) 
    {
     index = ind; 
     nonmax_weight_form.SetIndex(ind);
     max_weight_form.SetIndex(ind);
    };
   inline int GetIndex() const { return index; };
   inline Clause* TheClause() const { return clause; };
   inline void CollectNonmaxWeightForm()
    {
     nonmax_weight_form.Reset();
     for (const TERM* lit_ref = first_nonmax_lit_ref; !lit_ref->IsNull(); lit_ref = lit_ref->Next())
      nonmax_weight_form.Compile(lit_ref->First());
    };
   inline void RecollectMaxWeightForm()
    {
     max_weight_form.Reset();
     for (const TERM* lit_ref = first_max_lit_ref;lit_ref != first_nonmax_lit_ref; lit_ref = lit_ref->Next())
      {
       if (lit_ref->Next() == next_max_lit_ref)
        {
         CHECK17(lit_ref->First()->IsEquality(),"A0 in BPARA_PREMISE::RecollectMaxWeightForm()");
         CHECK17(lit_ref->First()->Positive(),"A1 in BPARA_PREMISE::RecollectMaxWeightForm()");
         if (lit_ref->First()->IsOrderedEq())
          max_weight_form.Compile(*(lit_ref->First()->Arg2()));
        }
       else max_weight_form.Compile(lit_ref->First());
      }; 
    }; 
   
   inline void Load(Clause* cl) 
    {
     clause = cl;
     TERM* lits = cl->Literals(); 
     first_max_lit_ref = lits->Args();
     first_nonmax_lit_ref = lits->NthArg(lits->MaxLiterals());
     next_max_lit_ref = first_max_lit_ref;
    };
   inline const TERM* NextEq()
    {
     const TERM* lit;
     check_lit:
      if (next_max_lit_ref == first_nonmax_lit_ref) return (TERM*)0;
      lit = next_max_lit_ref->First();
      next_max_lit_ref = next_max_lit_ref->Next();
      if (lit->Positive() && lit->IsEquality()) 
       {
        RecollectMaxWeightForm();
        return lit;
       }; 
      goto check_lit;
    }; 
    
   inline long MinResultWeight()
    {
     return nonmax_weight_form.Compute() + max_weight_form.Compute();
    }; 
   inline bool CollectLiterals() 
    {
     // collect max. literals except the used equality
     for (const TERM* mlr = first_max_lit_ref; mlr != first_nonmax_lit_ref; mlr = mlr->Next())
      if (mlr != next_max_lit_ref->Previous())
       if (!Resolvent.PushStandardLit(index,mlr->First())) return false;
      
     // collect nonmax. literals
     for (const TERM* nmlr = first_nonmax_lit_ref; !nmlr->IsNull(); nmlr = nmlr->Next())
      if (!Resolvent.PushStandardLit(index,nmlr->First())) return false;
      
     return true;
    }; 
 }; // class BPARA_PREMISE

class BACKWARD_SUPERPOSITION
 {
  private:
   BPARA_PREMISE bpara_premise;
   BPARA_MACHINE bpara_machine;
  public:
   BACKWARD_SUPERPOSITION()
    : bpara_premise(),
      bpara_machine()
    {
     bpara_premise.SetIndex(0);
     bpara_machine.SetIndex(1);
    }; 
   inline void SetQuery(Clause* cl) { bpara_machine.SetQuery(cl); };
   inline bool CollectResult()
    {
     if (bpara_premise.MinResultWeight()
         + bpara_machine.Candidate()->TheClause()->MinResWeight()
         > Resolvent.MaxWeight())
      {
       Resolvent.RegisterTooHeavy();
       return false;
      };   
         
     
     Resolvent.Reset();
     if (bpara_machine.CollectLiterals() && bpara_premise.CollectLiterals())
      {
       Resolvent.SetBackground(ANCESTORS::RESOL);
       Resolvent.PushAncestor(bpara_premise.TheClause());
       Resolvent.PushAncestor(bpara_machine.Candidate()->TheClause());
       return true;
      }
     else return false;
    }; 
   void Process(Clause* cl);
     
 }; // class BACKWARD_SUPERPOSITION

class EQ_FACTOR
 {
  private:
   bool st_unordered;
   bool wrap_st;  
   PREFIX_SYM* s;
   PREFIX_SYM* t;
   bool wrap_s_t_;
   PREFIX_SYM* s_;
   PREFIX_SYM* t_;
   bool init_state;
   PREFIX_SYM* s_t_;
  public:
   EQ_FACTOR() {};
   inline void LoadST(PREFIX_SYM* eq_lit) 
    {
     st_unordered = eq_lit->Head().Func().IsUnorderedEq();
     s = eq_lit + 1;
     t = s->Bracket();
    };
   inline void LoadS_T_(PREFIX_SYM* eq_lit) 
    {
     s_t_ = eq_lit;
     s_ = eq_lit + 1;
     t_ = s_->Bracket();
     init_state = true;
    };
  private: 
   inline bool UnifySS_()
    {
     return UNIFICATION::Unify(s,s_);
    };
   inline void WrapST()
    {
     PREFIX_SYM* tmp = s;
     s = t;
     t = tmp;
    }; 
   inline void WrapS_T_()
    {
     PREFIX_SYM* tmp = s_;
     s_ = t_;
     t_ = tmp;
    };
   
   inline bool FirstInf()
    {
     wrap_st = st_unordered;
     wrap_s_t_ = true;
     
     if (UnifySS_()) return true;
     
     wrap_s_t_ = false;
     WrapS_T_();
     if (UnifySS_()) return true;
     
     if (!wrap_st) return false;
     
     wrap_st = false;     
     WrapST();
     if (UnifySS_()) { wrap_s_t_ = true; return true; };
    
     // last attempt
     WrapS_T_();
     CHECK17(!wrap_s_t_,"A10 in EQ_FACTOR::FirstInf()");
     CHECK17(!wrap_st,"A11 in EQ_FACTOR::FirstInf()");
     return UnifySS_();
    };
   
   inline bool NextInf()
    {
     GlobalSubst.UnbindLastSection();
     if (wrap_s_t_)
      {
       wrap_s_t_ = false;
       WrapS_T_();
       if (UnifySS_()) return true;
      };
     if (wrap_st)
      {
       wrap_st = false; 
       WrapST();
       if (UnifySS_()) { wrap_s_t_ = true; return true; };
       
       // last attempt
       WrapS_T_();
       CHECK17(!wrap_s_t_,"A10 in EQ_FACTOR::NextInf()");
       CHECK17(!wrap_st,"A11 in EQ_FACTOR::NextInf()");
       return UnifySS_(); 
      }
     else return false;
    };
  
  public:
   inline bool NextInference()
    {
     if (init_state)
      {
       init_state = false;
       return FirstInf();
      };
     return NextInf(); 
    };   
    
   inline bool CollectResult()
    {
     if (!Resolvent.PushPrefixLit(s_t_)) return false;
     return Resolvent.PushPrefixEqLit(NEG_UNORD_EQ_HEADER,t,t_);
    };   
  
  public: // for debugging 
   inline void Output()
    {
     ::Output("S = "); PREFIX::OutputTerm(s); ::Output("\n");
     ::Output("T = "); PREFIX::OutputTerm(t); ::Output("\n"); 
     ::Output("S_ = "); PREFIX::OutputTerm(s_); ::Output("\n"); 
     ::Output("T_ = "); PREFIX::OutputTerm(t_); ::Output("\n"); 
     ::Output("SUBSTITUTION:\n");
     GlobalSubst.Output(OutputPrefix);
     ::Output("\n"); 
    };
 }; // class EQ_FACTOR


class EQ_FACTORING
 {
  private:
   EQ_FACTOR eq_factor;   
   PREFIX_SYM pos_eq[MAX_TERM_WEIGHT][MAX_NUMBER_OF_LITERALS];
   int last_pos_eq;
   int last_max_pos_eq;
   
   STACK_WITH_ITERATOR<TERM*,MAX_NUMBER_OF_LITERALS> other_literals;
   
   Clause* clause;
   int index;
   bool init_state;
   int curr_st;
   int curr_s_t_;
  public:
   EQ_FACTORING()
    : eq_factor(),
      other_literals("MAX_NUMBER_OF_LITERALS in EQ_FACTORING::other_maximal_lits")
     {};
   inline void SetIndex(int ind) { index = ind; };  
  private:   
   inline void Reset()
    {
     other_literals.Reset();
     last_pos_eq = -1;
     last_max_pos_eq = -1;
     init_state = true;
    };  
   inline Clause* TheClause() const { return clause; };
   inline void LoadMax(TERM* lit)
    {
     if (lit->Positive() && lit->IsEquality())
      {
       last_pos_eq++;
       last_max_pos_eq++;
       PREFIX_SYM* mem = pos_eq[last_pos_eq];
       StandardTermToPrefix(index,lit,mem);
      }
     else other_literals.Push(lit); 
    };
   inline void LoadNonmax(TERM* lit)
    {
     if (lit->Positive() && lit->IsEquality())
      {
       last_pos_eq++;
       PREFIX_SYM* mem = pos_eq[last_pos_eq];
       StandardTermToPrefix(index,lit,mem);
      }
     else other_literals.Push(lit); 
    };
   
   inline bool Load(Clause* cl) 
    {     
     Reset();
     clause = cl;
     TERM* lits = cl->Literals(); 
     int num_of_max = lits->MaxLiterals();
     if (num_of_max == 1) return false;
     TERM* lit_ref = lits->Args();
     TERM* lit;
     do
      {
       lit = lit_ref->First();
       LoadMax(lit);
       
       lit_ref = lit_ref->Next(); 
       num_of_max--; 
      }
     while (num_of_max);  
     if (last_pos_eq == -1) return false;
     
     while (!lit_ref->IsNull())
      {
       lit = lit_ref->First();
       LoadNonmax(lit);
       lit_ref = lit_ref->Next(); 
      };
     
     return (last_pos_eq > 0);
    };
   
   inline bool FirstInf()
    {
     curr_st = 0;
     curr_s_t_ = 1;
     eq_factor.LoadST(pos_eq[curr_st]);
     try_current_s_t_:
      eq_factor.LoadS_T_(pos_eq[curr_s_t_]); 
      
      if (eq_factor.NextInference()) return true;
      
      curr_s_t_++;
      if (curr_s_t_ == curr_st) curr_s_t_++;
      if (curr_s_t_ <= last_pos_eq) goto try_current_s_t_;
      
      curr_st++;
      if (curr_st > last_max_pos_eq) return false;  
      eq_factor.LoadST(pos_eq[curr_st]);
      curr_s_t_ = 0; 
      goto try_current_s_t_;
    }; 
   
   inline bool NextInf()
    {
     if (eq_factor.NextInference()) return true;
     try_next_s_t_:
      curr_s_t_++;
      if (curr_s_t_ == curr_st) curr_s_t_++;
      if (curr_s_t_ <= last_pos_eq)
       {
        try_curr_s_t_:
         eq_factor.LoadS_T_(pos_eq[curr_s_t_]); 
         if (eq_factor.NextInference()) return true;
         goto try_next_s_t_;
       };
       
      curr_st++;
      if (curr_st > last_max_pos_eq) return false;  
      eq_factor.LoadST(pos_eq[curr_st]);
      curr_s_t_ = 0;
      goto try_curr_s_t_; 
    }; 
    
   inline bool NextInference()
    {
     if (init_state)
      {
       init_state = false;
       return FirstInf();
      };
     return NextInf();  
    }; 
   
   inline bool CollectOtherPosEq()
    {
     for (int ln = 0; ln <= last_pos_eq; ln++)
      if ((ln != curr_st) && (ln != curr_s_t_))
       if (!Resolvent.PushPrefixLit(pos_eq[ln])) return false;
     return true; 
    };
   
   inline bool CollectOtherLiterals()
    {
     other_literals.ResetIterator();
     for (TERM** lit = other_literals.Next(); lit; lit = other_literals.Next())
      if (!Resolvent.PushStandardLit(index,*lit)) return false;
     return true; 
    };
    
   inline bool CollectResult()
    {
     return CollectOtherPosEq()  
            && eq_factor.CollectResult() 
            && CollectOtherLiterals();
    };
     
  public:   
   void Process(Clause* cl);
 }; // class EQ_FACTORING



class SUPERPOSITION
 {
  private:
   FORWARD_SUPERPOSITION forward;
   BACKWARD_SUPERPOSITION backward;
   EQ_FACTORING eq_factoring;
  public:
   SUPERPOSITION()
    : forward(),
      backward(),
      eq_factoring()
     {
      eq_factoring.SetIndex(0);
     };
   inline void Process(Clause* cl)
    {
     forward.Process(cl);
     backward.Process(cl);
     eq_factoring.Process(cl);
    };  
 }; // class SUPERPOSITION
 
 
 
 
 
//***********************************************************************










 
