#define  __TERM_H
#ifndef __STDDEF_H
#  include <stddef.h>
#endif
#ifndef __DEBUG_H
 #include "debug.h"
#endif 
#ifndef __VAMPIRE_H
#  include "vampire.h"
#endif
#ifndef __MEMORY_H
#  include "memory.h"
#endif
#ifndef __OUTPUT_H
 #include "output.h"
#endif
#ifndef __STACK_H
 #include "stack.h"
#endif

class TERM;


// Has to be #defined, GnuC++ does not like it to be typedef'ed
#define TERMLIST TERM


// I have no idea why this should be a special type including both
// connectives and special symbols
enum FUNCTOR
  {
  ATOMIC = 0,
  ALL = 1,
  SOME = 2,
  IMP = 3,
  NOT = 4,
  AND = 5,
  OR = 6,
  IFF = 7,
  BOX = 8,
  DIAM = 9,

  LITERALS = 1020,
  // all lists end on this symbol
  NIL = 1021,
  TERM_LIST = 1022,  // lists of literals (in a clause)
  FORMULA_LIST = 1023 // 2*10-1 for lists of formulas
  };


enum TERM_FLAG
  {
   // These values are fixed! This is used in substitution trees! (A.R.)
  TERM_REF = 0x0000, // must be zero!!!!
  TERM_VAR = 0x0001,
  TERM_COMPLEX = 0x0002
  };


enum POLARITY
 {
  POL_POS = 0,  // the values are fixed! Used in indexing. See also func. Positive();
  POL_NEG = 1
 };

inline POLARITY Inverse(POLARITY pol) { return (POLARITY)(!pol); };

// introduced just to have a separate type
enum ARITY
  {
  };


enum COMPARISON
  {
  LESS = -1,
  EQUAL = 0,
  GREATER = 1,
  INCOMPARABLE = 2
  };

inline COMPARISON Inverse(COMPARISON cmp)
 {
  switch (cmp)
   {
    case LESS: return GREATER;
    case GREATER: return LESS;
    default: return cmp;
   };
 };

class TERM
{
  public:
   class ITERATOR;
   class ITERATOR_WITH_DEPTH;
  protected:
    union
    {
      struct
      {
        unsigned arity : 5; // Must be always equal to 0! (A.R.)
	int var : 11;
	int dummy : 14;
	unsigned int flag : 2;
      } var;

      // WARNING: changing of field sizes in this structure must be
      // followed by the check of appropriate enum definitions in "number.h"
      struct
      {
	// same order as in STREE!
	unsigned arity : 5;
	unsigned functor : 10;
	unsigned polarity : 1;
	unsigned reserved : 14;
	unsigned flag : 2;
      } complex;

      struct
      {
	// same order as in complex
	// used in clauses, max_literals is used to denote
	// the number of maximal literals
	unsigned arity : 5;
	unsigned functor : 10;
	unsigned polarity : 1;
	unsigned max_literals : 5;
	unsigned reserved : 9;
	unsigned flag : 2;
      } clause;

      struct
      {
	unsigned short first;
	unsigned short second;
      } halves;

      TERM* ref;
      long content;  // just for safe passing by value

      struct
      {
	int dummy1 : 16;
	int dummy2 : 14;
	unsigned int flag : 2;
      } term;

    };

  public:  // constructors
    TERM() {}
   
    TERM(const TERM& t) : // copy constructor, redefined for efficiency
      content(t.content)
      {}
    

    // for making variables, the first dummy argument
    // is just to avoid casts from int to TERM
    TERM(TERM_FLAG,int v)
    {
      var.arity = 0;        // New! (A.R.)
      var.flag = TERM_VAR;
      var.dummy = 0;
      var.var = v;
    }


    // for the empty list, both arguments are dummy
    inline TERMLIST(int dummy1,int dummy2);
    // for term lists
    inline TERMLIST(TERM* first,TERMLIST* next);


    TERM(ARITY ar,FUNCTOR f,POLARITY p)
     {
      complex.arity = ar;
      complex.functor = f;
      complex.polarity = p;
      complex.reserved = 0;
      complex.flag = TERM_COMPLEX;
     };
    
    inline void MakeComplex(ARITY ar,FUNCTOR f,POLARITY p)
     {
      complex.arity = ar;
      complex.functor = f;
      complex.polarity = p;
      complex.reserved = 0;
      complex.flag = TERM_COMPLEX;
     };
    
    TERM(ARITY ar,FUNCTOR f,POLARITY p,unsigned res,TERM_FLAG fl)
    {
      complex.arity = ar;
      complex.functor = f;
      complex.polarity = p;
      complex.reserved = res;
      complex.flag = fl;
    }
    
    
    //TERM(long& sym) : content(sym) { complex.reserved = 0; }; // New!
    TERM(long sym) : content(sym) {}; 
   
    
    TERM(TERM*& t) : ref(t) {};
  public:  // assignment, redefined for efficiency

    inline TERM& operator= (const TERM& t)
    {
      content = t.content;
      return *this;
    } 

  public: // memory
    inline void* operator new(size_t,int arity)
      { return Memory::Alloc(arity + 1); }
    inline void operator delete(void* obj);
    //inline void Delete()  // for deleting vars
    //  { Memory::Dealloc(this,0); }
    
    
    void Destroy();

  public: // structure
    inline unsigned Reserved() const { return complex.reserved; }; 
    inline void SetReserved(const unsigned r) { complex.reserved = r; };
    inline long Content() const { return content; };
    inline TERM* Next()
      { return this - 1; }
    inline TERM* Previous()  // WARNING: does not check that the prev. arg exists!
      { return this + 1; }
    inline TERM* LastArg()  // WARNING: does not check if there is any arg!
      { return this+1; }
    inline int Arity() const
      { return complex.arity; }
      
    inline TERM* Args() 
      { return this + Arity(); }

    inline TERM* Info() { return this + Arity() + 1; }  
    
    inline const TERM* NthArg(const int n) const
     { 
      return Args() - n;
     };
    inline const TERM* Arg1() const 
     {
      return Args();
     }; 
    inline const TERM* Arg2() const 
     {
      return Args()->Next();
     };
     
    inline TERM* First() const
      { return ref; }
    inline void SetFirst(TERM* t)
      { ref = t; }
    
    inline void RenameVar(const int v) { var.var = v; };
      
    inline void MakeVar(int v);
    // this func uses that references have even addresses
    inline void MakeReference(TERM* t)
      { ref = t; }
    /******** old version:
    inline int IsReference() const
      { return term.flag == TERM_REF; }
    ****************************/ 
    
    inline int IsReference() const
      { return !term.flag; }  
    
    inline TERM* Reference() const
      { return ref; };
    inline POLARITY Polarity() const
      { return (POLARITY)complex.polarity; }
    
    inline bool Positive() const { return !(Polarity()); };  
    inline bool Negative() const { return (bool)(Polarity()); };
      
    inline void SetPolarity(POLARITY p)
      { complex.polarity = p; }
    inline void MakeNegative()
      { complex.polarity = POL_NEG; }
    inline void MakePositive()
      { complex.polarity = POL_POS; }
    inline void InversePolarity()
     {
      if (Polarity() == POL_POS)
       {
        MakeNegative();
       }
      else
       {
        MakePositive();   
       };
     }; 
     
    inline int IsVariable() const
      { return term.flag == TERM_VAR; }
    inline int IsNotVariable() const
      { return term.flag != TERM_VAR; }
    inline int Var() const
      { return var.var; }
    inline int Flag() const
      { return var.flag; }
    inline void SetVar(int v)	// does not set flag to VAR
      { var.var = v; }
    inline int IsComplex() const
      { return var.flag == TERM_COMPLEX; }
    inline int IsNull() const	// term ref or var considered as term list 
      { return var.flag == TERM_COMPLEX; };
    
    inline bool IsEquality() const { return ::IsEquality(Functor()); };
    inline bool IsOrderedEq() const 
     {
      CHECK0(IsEquality(),"A0 in TERM::IsOrderedEq()"); 
      return ::IsOrderedEq(Functor());
     };
    inline bool IsUnorderedEq() const 
     {
      CHECK0(IsEquality(),"A0 in TERM::IsUnorderedEq()"); 
      return ::IsUnorderedEq(Functor());
     };
    inline void MakeUnordered() { SetFunctor(EQ_NUM); };
    
    inline int Functor() const
      { return complex.functor; }
    inline int MaxLiterals() const
      { return clause.max_literals; }
    inline unsigned short ComplexFunctor() const
      { return halves.first; }
    inline void SetComplexFunctor(unsigned short fun)
      { halves.first = fun; }
    inline void SetFunctor(int fun)
      { complex.functor = fun; }
    inline void SetMaxLiterals(int maxl)
      { clause.max_literals = maxl; }
    inline void SetArity(int arity)
      { complex.arity = arity; }
    inline int IsConstant() const
      { return Arity() == 0; }
    inline const TERM* Next() const
      { return this - 1; }
    inline const TERM* Previous() const
      { return this + 1; }  
    inline const TERM* Args() const
      { return this + Arity(); }
    inline const TERM* UnaryArg() const
      { return this[1].First(); }
    inline TERM* UnaryArg()
      { return this[1].First(); }
    inline const TERM* BinaryFirstArg() const
      { return this[2].First(); }
    inline TERM* BinaryFirstArg()
      { return this[2].First(); }
    inline const TERM* BinarySecondArg() const
      { return this[1].First(); }
    inline TERM* BinarySecondArg()
      { return this[1].First(); }
    // for lists of terms
    inline int IsEmpty() const;
    inline TERM* Head() const;
    inline TERM* Tail() const;
    inline void SetHead(TERM* head)
      { this[2].SetFirst(head); }
    inline void SetTail(TERM* tail)
      { this[1].SetFirst(tail); }

  public:
    inline int SameFunctor(const TERM& t) const
      {
       //cout << "SameFunctor\n"; // debug
       //cout << "this = " << (long)this << "\n";
       //cout << "t = " << (long)&t << "\n";
        
       return content == t.content; 
      };
    inline int operator==(const TERM& t) const 
     { return content == t.content; };
    inline int operator!=(const TERM& t) const 
     { return content != t.content; };
    inline int operator>(const TERM& t) const  
     { return content > t.content; };
    inline int operator<(const TERM& t) const  
     { return content < t.content; };
      
    inline int GreaterFunctor(const TERM& t) const
      { return content > t.content; }

    inline COMPARISON CompareFunctor(const int f) const
     {
      // don't change the order, it's used in splitting
      if (Functor() < f) { return GREATER; };
      if (Functor() > f) { return LESS; };
      return EQUAL;
     };
    inline COMPARISON CompareFunctor(const TERM f) const
     {
      return CompareFunctor(f.Functor());
     };
   private:
    static bool positive_is_greater; // = false by default, see "term.cc"
   public:
    static inline void SetPositiveIsGreater(bool flag) 
     { positive_is_greater = flag; };
    static inline bool PositiveIsGreater() { return positive_is_greater; };
    inline COMPARISON ComparePolarity(int p) const
     {
      if (Polarity() == p) { return EQUAL; };
      if (Polarity() == POL_POS)
       {
        if (positive_is_greater)
         {
          return GREATER;
         }
        else
         {
          return LESS;
         };
       }
      else // Polarity() == POL_NEG
       {
        if (positive_is_greater)
         {
          return LESS;
         }
        else
         {
          return GREATER;
         };
       };
     };  
    static inline POLARITY MinimalPolarity() 
     {
      return (positive_is_greater) ? POL_NEG : POL_POS;
     };
    
    static inline POLARITY MaximalPolarity() 
     {
      return (positive_is_greater) ? POL_POS : POL_NEG;
     };
     
    inline COMPARISON CompareHeader(const TERM hd) const
     {
      if (Polarity() == hd.Polarity()) 
       {
        // experimental:
        /*******
        if (Polarity() == POL_POS)
         {
          return CompareFunctor(hd.Functor());
         }
        else
         {
          return Inverse(CompareFunctor(hd.Functor()));
         };
        *************/
        
        return CompareFunctor(hd.Functor()); 
       }
      else
       { 
        return ComparePolarity(hd.Polarity()); 
       }; 
     };

   inline bool ComplementaryHeader(TERM hd) const 
    {
     return (Functor() == hd.Functor()) && (Polarity() != hd.Polarity());
    };
   
   /***
   inline bool SameHeader(TERM hd)
    {
     return SameFunctor(hd);
    };
   ***/
   
   inline int HeaderNum() const { return (Functor()*2) + Polarity(); };
   
   inline void MakeHeader(int header_num) //  without arity!
    {
     int fun = header_num/2;
     int pol = header_num%2;
     MakeComplex((ARITY)0,(FUNCTOR)fun,(POLARITY)pol);
    }; 
    
// *************************************************
    inline long ComplementaryFunctor(const TERM& t) const;
    inline int SameVariable(const TERM& t) const
      { return content == t.content; }

  public: // literals
    inline static TERM* NewLiterals(int arity,int max_lits);
    inline int EmptyLiterals() const
      { return Arity() == 0; }

  private:
    inline COMPARISON CompareArgs(const TERM*) const;
    //COMPARISON KBCompareArgs(TERM*);

  public:
    //void Rename();
    //TERM* Copy() const;
    inline TERM* CopyTop() const // for complex term only (including literals)
     {
      return new(Arity()) TERM(*this);
     };
  
    inline TERM* CopyTop(int info_field_size) const // for complex term only (including literals)
     {
      TERM* result = ((TERM*)(Memory::Allocate(1 + Arity() + info_field_size))) + info_field_size;
      *result = *this;
      return result;
     };
    
    inline TERM* CopyLitTop() const // for literals only
     {
      return AllocLit(*this);
     
      /*****
      TERM* result = ((TERM*)(Memory::Allocate(1 + Arity() + 1))) + 1;
      *result = *this;
      return result;
      ******/
     };
    static inline TERM* AllocLit(const TERM& header)
     {
      TERM* result = ((TERM*)(Memory::Allocate(1 + header.Arity() + 1))) + 1;
      *result = header;
      return result;
     };
     
    
    inline void DeleteLit() // for literals only
     {
      Memory::Deallocate(this - 1,1 + Arity() + 1);  
     }; 
     
    inline void** InfoField() const { return (void**)(this - 1); }; 
    
     
    inline TERM* CopyTopLevel() const // for complex term only (including literals)
     {
      TERM* res = CopyTop();
      TERM* new_arg = res->Args();
      
      for (TERM* arg = (TERM*)Args(); !(arg->IsNull()); arg = arg->Next())
       {
        *new_arg = *arg;
        new_arg = new_arg->Next();
       };
      return res;
     };
    /******* 
    inline TERM* CopyRecursively()
     {
      STACK<TERM*,MAX_TERM_DEPTH> terms("MAX_TERM_DEPTH in TERM::CopyRecursively()");
      
      //terms.Reset();
      TERM* res = CopyTopLevel();
      TERM* args = res->Args();
      copy_args:
       if (args->IsNull())
        {
         if (!(terms.Empty()))
          {
           args = terms.Pop();
           goto copy_args;
          }
         else
          {
           return res;
          };
        }
       else
        {
         if (args->IsReference())
          {
           terms.Push(args->Next());
           args->SetFirst(args->First()->CopyTopLevel());
           args = args->First()->Args();
           goto copy_args;
          }
         else // avriable
          {
           args = args->Next();
           goto copy_args;
          };
        };
     };
    ***********/
    static inline bool Ground(TERM* t); // t must be complex
    inline bool Plain(); // this must be complex 
    
    //int Occurs(int var) const;
    //void Replace(int var1,int var2);
    inline int Equal(const TERM*) const;
    static inline bool EqualComplexTerms(TERM* const t1,TERM* const t2);
    inline int Weight() const;
    //int Length() const;  // for lists
    inline int ComplementaryLiteral(const TERM* t) const;
    static inline int EqualArgs(const TERM*,const TERM*);
    inline int CompareVarOrReference(const TERM*) const;
    inline int CompareVarOrReferenceIntegrated(const TERM*) const;
    //inline COMPARISON Compare(const TERM*) const;
    //inline COMPARISON KBCompare(TERM*);

  public:
   static inline int Depth(TERM* t);    
};  // class TERM


inline bool TERM::Plain() // all arguments are pairwise distinct variables
 {
  for (TERM* arg = Args(); !arg->IsNull(); arg = arg->Next())
   { 
    if (arg->IsVariable())
     {
      for (TERM* prev_var = Args(); prev_var != arg; prev_var = prev_var->Next())
       if ((*prev_var) == (*arg)) return false;
     }
    else return false;
   };  
  return true;
 }; // bool TERM::Plain()






// this function does not check whether the object to be deleted
// is actually a structure. Its call with a non-structure will
// result in a disaster
inline void TERM::operator delete(void* obj)
{
  Memory::Dealloc(obj,((TERM*)obj)->Arity() + 1);
}


// just functors

const TERM cons(ARITY(2),TERM_LIST,POL_POS);
const TERM complementary_mask(ARITY(0),FUNCTOR(0),POLARITY(1),0,(TERM_FLAG)0);
const TERM nil(ARITY(0),NIL,POL_POS);
const TERM literal(ARITY(0),TERM_LIST,POL_POS);
const TERM dummy_var(TERM_VAR,0);

const TERM POS_ORD_EQ_HEADER((ARITY)2,(FUNCTOR)ORD_EQ_NUM,POL_POS);
const TERM NEG_ORD_EQ_HEADER((ARITY)2,(FUNCTOR)ORD_EQ_NUM,POL_NEG);
const TERM POS_UNORD_EQ_HEADER((ARITY)2,(FUNCTOR)EQ_NUM,POL_POS);
const TERM NEG_UNORD_EQ_HEADER((ARITY)2,(FUNCTOR)EQ_NUM,POL_NEG);

// empty term list, dummy arguments
inline TERMLIST::TERMLIST(int,int)
{
  *this = nil;
}


inline TERMLIST::TERMLIST(TERM* first,TERMLIST* rest)
{
  *this = cons;
  this[2].SetFirst(first);
  this[1].SetFirst(rest);
}


// for lists of terms
inline int TERM::IsEmpty() const
{
  return SameFunctor(nil);
}


// for lists of terms
inline TERM* TERM::Head() const
{
  return (TERM*)BinaryFirstArg();
}


// for lists of terms
inline TERMLIST* TERM::Tail() const
{
  return (TERMLIST*)BinarySecondArg();
}


inline int TERM::Equal(const TERM* t) const
{
  return SameFunctor(*t) ?
	 EqualArgs(Args(),t->Args()) :
	 0;
}  // TERM::Equal()

inline long TERM::ComplementaryFunctor(const TERM& t) const
{
  return (content ^ t.content) == complementary_mask.content;
}  // TERM::ComplementaryFunctor()


inline int TERM::ComplementaryLiteral(const TERM* t) const
{
  return ComplementaryFunctor(*t) &&
	 EqualArgs(Args(),t->Args());
}


inline TERM* TERM::NewLiterals(int arity,int max_lits)
{
  TERM* result = new(arity) TERM(literal);
  result->SetArity(arity);
  result->SetMaxLiterals(max_lits);

  return result;
}  // TERM::NewLiterals()


inline void TERM::MakeVar(int v)
{
  *this = dummy_var;
  var.var = v;
}  // TERM::MakeVar()

inline int TERM::CompareVarOrReference(const TERM* t) const
 {
  if (IsVariable())  // first is variable
   {
    return SameVariable(*t);
   }
  else 
   {
    if (t->IsVariable())  // second is a variable, the first is not
     {
      return 0;
     }
    else // both are references
     {
      //cout << "this = " << (long)this << "\n"; // debug
      //cout << "t = " << (long)t << "\n"; // debug
     
      const TERM* s = First();
      t = t->First();
      
      //cout << "now\n"; // debug
      //cout << "s = " << (long)s << "\n"; // debug
      //cout << "t = " << (long)t << "\n"; // debug
      
      //cout << "same functor = \n   "; // debug
      //cout << s->SameFunctor(*t) << "\n"; // debug
      
      return ((s->SameFunctor(*t)) && (EqualArgs(s->Args(),t->Args())));
     };
   };
 }  // TERM::CompareVarOrReference()

inline int TERM::CompareVarOrReferenceIntegrated(const TERM* t) const
{
  return IsVariable() ?  // first is variable
	 SameVariable(*t) :
	 t->IsVariable() ?  // second is a variable, the first is not
	 0 :
	 First() == t->First();
}  // TERM::CompareVarOrReferenceIntegrated()

/*****
inline COMPARISON TERM::Compare(const TERM* t) const
{
 COMPARISON cmpHd = CompareHeader(*t);
 if (cmpHd == EQUAL)
  {
   return Args()->CompareArgs(t->Args());
  }
 else
  {
   return cmpHd;
  };
}  // TERM::Compare()
******/

inline void OutputSym(TERM sym)
 {
  if (sym.IsVariable())
   {
    OutputVar(sym.Var());
   }
  else
   {
    if (sym.Polarity() == POL_NEG)
     {
      Output("~");
     };
    OutputFunctor(sym.Functor());
   };
 };

inline int TERM::Depth(TERM* t) // t must be a complex term
 {
  STACK<TERM*,MAX_TERM_DEPTH*2> subterms("MAX_TERM_DEPTH*2 in TERM::Depth():subterms");
  STACK<int,MAX_TERM_DEPTH*2> depths("MAX_TERM_DEPTH*2 in TERM::Depth():depths");
  //subterms.Reset();
  //depths.Reset();
  
  int max_depth = 1;
  int depth = 1;
  t = t->Args();
  if (t->IsNull()) { return 0; }; // t is constant
  next:
   if (t->IsNull())
    {
     if (subterms.Empty())
      {
       return max_depth;
      }
     else
      {
       t = subterms.Pop();
       depth = depths.Pop();
       goto next;
      };
    }
   else
    {
     if (t->IsVariable()) 
      {
       t = t->Next();
       goto next;
      }
     else // reference
      {
       if (t->First()->Arity()) // not a constant
        {
         subterms.Push(t->Next());
         depths.Push(depth);
         t = t->First()->Args();
         depth++;
         if (depth > max_depth) { max_depth = depth; };
         goto next; 
        }
       else // constant
        {
         t = t->Next();
         goto next;
        };       
      };    
    };
 }; // int TERM::Depth(TERM* t)

inline int TERM::Weight() const
 {
  int w = 1;

  const TERM* term = Args();

  STACK<const TERM*,MAX_TERM_DEPTH> stack("MAX_TERM_DEPTH");
  //stack.Reset();

  for (;;)
    switch (term->Flag())
      {
      case TERM_REF:
	stack.Push(term->First());
	// then do as for variables

      case TERM_VAR:
	term = term->Next();
	break;

      case TERM_COMPLEX:
	w += term->Arity();
	if (stack.Empty())
	  return w;
	else
	  term = stack.Pop();
	break;
      }
 };  // TERM::Weight()


inline int TERM::EqualArgs(const TERM* first,const TERM* second)
 {
  //cout << "EqualArgs(..)\n"; // debug
  
  STACK<const TERM*,MAX_TERM_DEPTH> first_stack("MAX_TERM_DEPTH");
  STACK<const TERM*,MAX_TERM_DEPTH> second_stack("MAX_TERM_DEPTH");
  //first_stack.Reset();
  //second_stack.Reset();

  for (;;)
    {
    switch (first->Flag())
      {
      case TERM_REF:
        //cout << "TERM_REF:\n"; // debug
      
	if (second->IsVariable() ||
	    !first->First()->SameFunctor(*second->First()))
	  return 0;
	first_stack.Push(first->First());
	second_stack.Push(second->First());
	first = first->Next();
	second = second->Next();
	break;

      case TERM_VAR:
        //cout << "TERM_VAR:\n"; // debug
      
	if (!first->SameVariable(*second))
	  return 0;
	first = first->Next();
	second = second->Next();
	break;

      case TERM_COMPLEX:
        //cout << "TERM_COMPLEX:\n"; // debug
      
	if (first_stack.Empty())
	  return 1;
	first = first_stack.Pop()->Args();
	second = second_stack.Pop()->Args();
	break;
      }
    }
}  // TERM::EqualArgs()

inline COMPARISON TERM::CompareArgs(const TERM* t2) const
 {
  const TERM* t1 = this;

  STACK<const TERM*,MAX_TERM_DEPTH> ts1("MAX_TERM_DEPTH");
  STACK<const TERM*,MAX_TERM_DEPTH> ts2("MAX_TERM_DEPTH");
  //ts1.Reset();
  //ts2.Reset();

  for (;;)
    switch (t1->Flag())
      {
      case TERM_REF:
	if (t2->IsVariable())
	  // variable is always less then a compound term
	  return GREATER;
	ts1.Push(t1->Next());
	t1 = t1->First();
	ts2.Push(t2->Next());
	t2 = t2->First();
	// comparing functors
	if (t1->GreaterFunctor(*t2))
	  return GREATER;
	else if (t1->SameFunctor(*t2))
	  {
          t1 = t1->Args();
	  t2 = t2->Args();
	  }
	else  // smaller functor
	  return LESS;
	break;

      case TERM_COMPLEX:  // end of args
	if (ts1.Empty())
	  return EQUAL;
	t1 = ts1.Pop();
	t2 = ts2.Pop();
	break;

      case TERM_VAR:
	if (t2->IsReference())
	  // variable is always less then a compound term
	  return LESS;
	t1 = t1->Next();
	t2 = t2->Next();
	break;

      }  // for (;;), switch
 }  // TERM::Compare()
//***************************************************************************

class TERM::ITERATOR // for complex terms only!
 {
  private:
   STACK<TERM*,MAX_TERM_WEIGHT> subterms;   
   TERM* sym;
   bool isVar;  
  public:
   ITERATOR() : subterms("MAX_TERM_WEIGHT in TERM::ITERATOR.subterms") {};
   inline void Reset(TERM* t) 
    { 
     subterms.Reset(); 
     sym = t;
     isVar = false;
    };
    
   inline TERM Symbol() const { return *sym; };
   inline TERM& SymbolRef() { return *sym; };
   inline TERM* CurrPos() const { return sym; };
   inline bool CurrentSymIsVar() const { return isVar; };
  public:
   inline bool Next()
    {
     if (isVar)
      {
       sym = sym->Next();
      }
     else
      {
       sym = sym->Args();  
      };
     if (sym->IsNull()) 
      {
       if (subterms.Empty()) { return false; };
       sym = subterms.Pop();
      };
     if (!((isVar = sym->IsVariable()))) // reference
      {
       if (!(sym->Next()->IsNull())) { subterms.Push(sym->Next()); };
       sym = sym->First(); 
      };
     return true;
    };
   inline bool NextVar()
    {
     while (Next())
      {
       if (isVar) { return true; };
      };
     return false;
    };
   inline bool SkipTerm()
    {
     if (isVar)
      {
       sym = sym->Next();
       if (sym->IsNull()) 
        {
         if (subterms.Empty()) { return false; };
         sym = subterms.Pop();
        };
      }
     else
      {
       if (subterms.Empty()) { return false; };
       sym = subterms.Pop();
      }; 
     if (!((isVar = sym->IsVariable()))) // reference
      {
       if (!(sym->Next()->IsNull())) { subterms.Push(sym->Next()); };
       sym = sym->First(); 
      };
     return true; 
    }; 
 }; // class TERM::ITERATOR 

inline bool TERM::Ground(TERM* t) // t must be complex
 {
  if (!(t->Arity())) return true;
  ITERATOR iter;
  iter.Reset(t);
  while (iter.Next()) { if (iter.CurrentSymIsVar()) return false; }; 
  return true;
 }; // bool TERM::Ground(TERM* t)



class BRACKET_COUNTER
 {
  private:
   int holes[MAX_TERM_WEIGHT];
   int* next_holes;
   int curr_holes;  
  public:
   BRACKET_COUNTER() : curr_holes(0)
    {
     next_holes = holes;
    };
   inline void Reset() 
    {
     curr_holes = 0;
     next_holes = holes;
    };
   inline bool IsOpen() const 
    {
     return curr_holes;
    };
   inline bool AllClosed() const
    {
     return !(curr_holes);
    };
   inline void Open(const int arity) // arity != 0 here!
    {
     *next_holes = curr_holes-1;
     next_holes++;
     curr_holes = arity;
    };
   inline unsigned Close()
    {
     curr_holes--;
     if (curr_holes) // can be -1 
      {
       return 0;
      }
     else
      {
       unsigned closed = 1;
       pop_holes:
        if (next_holes == (holes + 1))
         {
          return closed;
         }
        else
         {
          next_holes--;
          curr_holes = (*next_holes);
          if (curr_holes)
           {
            return closed;
           }
          else
           {
            closed++;
            goto pop_holes;
           };
         };
      };
    };
 }; // class BRACKET_COUNTER

class DEPTH_COUNTER
 {
  private:
   unsigned depth;
  public: 
   DEPTH_COUNTER() {};
   inline void Reset(const unsigned d) { depth = d; };
   inline unsigned Depth() const { return depth; };
   inline void Down() { depth++; };
   inline void Up(const unsigned n) { depth -= n; };
 };

class TERM::ITERATOR_WITH_DEPTH  // for complex terms only!
 : private TERM::ITERATOR,
   private BRACKET_COUNTER,
   private DEPTH_COUNTER 
 {
  public:
   ITERATOR_WITH_DEPTH() {};
   inline void Reset(TERM* const t,const unsigned d) 
    {
     TERM::ITERATOR::Reset(t);
     BRACKET_COUNTER::Reset();
     DEPTH_COUNTER::Reset(d);
    };
   inline TERM Symbol() const { return TERM::ITERATOR::Symbol(); };
   inline bool CurrentSymIsVar() const { return TERM::ITERATOR::CurrentSymIsVar(); };
   inline bool Next()
    {
     if (Symbol().Arity())
      {
       BRACKET_COUNTER::Open(Symbol().Arity());
       DEPTH_COUNTER::Down();
      }
     else
      {
       DEPTH_COUNTER::Up(BRACKET_COUNTER::Close());
      };
     return TERM::ITERATOR::Next();
    };
  inline unsigned Depth() const { return DEPTH_COUNTER::Depth(); }; 
 }; // class TERM::ITERATOR_WITH_DEPTH

class SMART_DEPTH_COUNTER 
 : private BRACKET_COUNTER, private DEPTH_COUNTER
 {
  public:  
   SMART_DEPTH_COUNTER() : BRACKET_COUNTER(), DEPTH_COUNTER() {};
   inline void Reset(const unsigned d)
    {
     BRACKET_COUNTER::Reset();
     DEPTH_COUNTER::Reset(d);
    };
   inline void NonconstantFunctor(const int arity) 
    {
     BRACKET_COUNTER::Open(arity);
     DEPTH_COUNTER::Down();
    };
   inline void VarOrConst()
    {
     DEPTH_COUNTER::Up(BRACKET_COUNTER::Close());
    };
   inline void NextSym(const TERM& sym)
    {
     if (sym.Arity()) { NonconstantFunctor(sym.Arity()); }
     else { VarOrConst(); };
      
    }; 
   inline unsigned Depth() { return DEPTH_COUNTER::Depth(); };
 }; // class SMART_DEPTH_COUNTER 


inline bool TERM::EqualComplexTerms(TERM* const t1,TERM* const t2)
 {
  if (*t1 == *t2)
   {
    ITERATOR iter1;
    ITERATOR iter2;
    iter1.Reset(t1);
    iter2.Reset(t2);
    while (iter1.Next())
     {
      iter2.Next();
      if (iter1.Symbol() != iter2.Symbol()) { return false; };
     };
    return true;
   }
  else { return false; };
 }; // bool TERM::EqualComplexTerms(TERM* const t1,TERM* const t2)



//*****************************************************************************















