#define __TMP_CLAUSE_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif

#ifndef __VARIABLES_H
 #include "variables.h"
#endif

#ifndef __PREFIX_H
 #include "prefix.h"
#endif

#ifndef __TMP_TERM_H
 #include "tmp_term.h"
#endif

#ifndef __COMPONENT_H
 #include "component.h"
#endif

#ifndef __CLAUSE_H
 #include "clause.h"
#endif

#ifndef __FSUBSUME_H
 #include "fsubsume.h"
#endif

#ifndef __CLAUSE_SET_H
 #include "clause_set.h"
#endif

#ifndef __SPLITTING_H
 #include "splitting.h"
#endif

#ifndef __STATISTICS_H
 #include "statistics.h"
#endif

#ifndef __WEIGHTING_H
 #include "weighting.h"
#endif

#ifndef __FORWARD_SUBSUMPTION_H
 #include "forward_subsumption.h"
#endif 

/******* debug *********
#ifndef __RESOLVENT_H
 #include "resolvent.h"
#endif 
***********************/

typedef STACK_WITH_ITERATOR<Clause*,MAX_NUMBER_OF_LITERALS> CLSTACK;

extern void InternalTickEvent(const long ticks);
extern void IntegrateNewResolvents(CLSTACK& clauses);
extern void IntegrateNewResolvent(Clause* const cl);
extern void TestForUnitConflict(Clause* const cl);
extern KEPT Kept;
extern SPLITTING_INDEX SplittingIndex;

extern HEADER_WEIGHTING HeaderWeighting;

extern FS ForwardSubsumption;

class TMP_CLAUSE
 {
  private: // main data 
   PARENTS* parents;
   PARENTS* spare_parents;
   TMP_TERMLIST literals; 
   CLSTACK created_clauses;
   bool forward_subsumption_flag;
   bool subsume_shorter_flag;
   bool static_splitting_flag;
   bool dynamic_splitting_flag;
   
   bool simplify_resolvents;
   
   bool bin_res_negative_selection;  
   
   long internal_tick_frequency;
   long next_tick;
   long ticks;
   
  private: // auxilliary data 
   LARGE_VAR_RENAMING renaming;
   TERM* tmp_unit_clause_lits;
   Clause* tmp_unit_clause;
   
  public:  
   TMP_CLAUSE() 
    : parents((PARENTS*)0), 
      literals(), 
      created_clauses("MAX_NUMBER_OF_LITERALS in TMP_CLAUSE::created_clauses"),
      
      simplify_resolvents(false), // default
        
      bin_res_negative_selection(false), // default
      
      internal_tick_frequency(MAXLONG),
      next_tick(MAXLONG),
      ticks(0L),
       
      renaming(),
      tmp_unit_clause_lits(TERM::NewLiterals(1,1)),
      tmp_unit_clause(new Clause(tmp_unit_clause_lits))
     {};
   inline void Reset() 
    {
     renaming.Reset();
     literals.Reset(); 
     STATISTICS_ON_GENERATED::AnotherGenerated();
     
     if (next_tick) { next_tick--; }
     else { next_tick = internal_tick_frequency; ticks++; InternalTickEvent(ticks); };
    };

   inline void ResetInternalTick(const long freq) 
    {
     internal_tick_frequency = freq; 
     next_tick = freq;
     ticks = 0L;
    };
    
   inline long MaxWeight() const { return literals.MaxWeight(); };
   inline void SetMaxWeight(const long w) 
    {
    
     //cout << "MAX W = " << w << "\n"; // debug
    
     if (w < MIN_MAX_WEIGHT)
      { 
       literals.SetMaxWeight(MIN_MAX_WEIGHT);
      }
     else
      {
       literals.SetMaxWeight(w); 
      };
    };
   inline void SetMaxTermWeight(const long w) { literals.SetMaxTermWeight(w); };
   inline void SetMaxDepth(const long w) { literals.SetMaxDepth(w); };
   inline void SetMaxLength(const long w) { literals.SetMaxLength(w); };
   
   inline void SetForwardSubsumption(const bool fl) { forward_subsumption_flag = fl; };
   inline void SetSubsumeShorter(const bool fl) { subsume_shorter_flag = fl; };
    
   inline void SetStaticSplitting(const bool fl) { static_splitting_flag = fl; };
   inline void SetDynamicSplitting(const bool fl) { dynamic_splitting_flag = fl; };
   
   inline void SetSimplifyResolvents(const bool fl) { simplify_resolvents = fl; }; 
   inline void SetNegativeSelection(const bool fl) 
    {
     bin_res_negative_selection = fl; 
     literals.SetNegativeSelection(fl);
    }; 
    
   inline Clause* FSubsume()
    {
     /*******/
     Clause* subsuming_cl1;
     int len = Length();
     
     if (len != 1)
      {
       ForwardSubsumption.ResetQuery();
       for (int lit_n = 0; lit_n < len; lit_n++)
        {
         ForwardSubsumption.PushLit(literals.FlatTerm(lit_n));
        };
       subsuming_cl1 = ForwardSubsumption.Subsume();
      }
     else 
      {
       /***
       static long u = 0; // debug
       u++; // debug
       cout << "UNIT " << u << "\n";
       ***/
       subsuming_cl1 = ForwardSubsumption.SubsumeUnit(literals.FlatTerm(0));  
      };
     
     Clause* subsuming_cl2 = FSTREE::Subsume(subsume_shorter_flag);
     
     if ((subsuming_cl1) && (!subsuming_cl2))
      {
       cout << "NEW SUBSUMPTION\n";
       cout << "QUERY "; Output(); cout << "\n\n";
       ForwardSubsumption.OutputQuery(); cout << "\n\n";
       cout << "SUBSUMED BY "; ::Output(subsuming_cl1); cout << "\n\n\n";
      
       
       //ForwardSubsumption.OutputCode();
       
       RuntimeError("");
      };
     
     /*****/
     if ((subsuming_cl2) && (!subsuming_cl1))
      {
       cout << "OLD SUBSUMPTION\n";
       cout << "QUERY "; Output(); cout << "\n\n";
       cout << "SUBSUMED BY "; ::Output(subsuming_cl2); cout << "\n\n\n";
       RuntimeError("");
      };
     /*****/
     
     /**********/ 
     return subsuming_cl2;
     //return FSTREE::Subsume(subsume_shorter_flag);
    }; 
   
   inline Clause* FSubsume(Clause* const c)
    {
     Reset();
     LoadClause(c); 
     return FSubsume();
    };
    
   inline Clause* MakeInputClause(const CL_BACKGROUND backgr)
    {
     SetParents(PARENTS::NewInputParents(backgr));
     return MakeInputClause(); 
    }; // void MakeInputClause(const CL_BACKGROUND backgr) 
   
   inline int ComponentName(LIT_NUM_STACK& component,bool& new_component)
    {
     TERM lit_list[component.NumberOfElements()+1];
     *lit_list = TERM(literal);
     lit_list->SetArity(component.NumberOfElements());
     TERM* lit_ref = lit_list->Args();

     component.ResetBackIterator();
     int* lit_num = component.Previous();
     while (lit_num)
      {
       lit_ref->SetFirst(Literal(*lit_num));
       lit_ref = lit_ref->Next();
       lit_num = component.Previous();
      };
     
     return SplittingIndex.ComponentName(lit_list,new_component);
    }; // int ComponentName(LIT_NUM_STACK& component,bool& new_component)
   
   inline Clause* ComponentToClause(LIT_NUM_STACK& component,int& name)
    {
     static VAR_WEIGHTING var_weighting;
     
     literals.CreateVarWeighting(component,var_weighting);
     literals.Order(component,var_weighting);
     literals.ResetVarRenaming();
     literals.RenameVariables(component);
     bool new_component;
     
     name = ComponentName(component,new_component);
     
     if (new_component)
      {
       TERM header((ARITY)0,(FUNCTOR)name,POL_POS); // compatible with negative selection
       HeaderWeighting.MakeMinimal(header); 
       component.Push(literals.PushRenamingLit(header));
       return MakeClause(component);
      }
     else { return (Clause*)0; };
    }; // Clause* ComponentToClause(LIT_NUM_STACK& component,int& name)
   
   inline void Split(CLSTACK& result)
    {
     if (Length() == 1) { result.Push(MakeUnitClause()); return; };
     
     literals.FindLinkedComponents();
     if ((literals.NumOfNongroundComponents()) 
         && ((literals.NumOfNongroundComponents() > 1)
             || (literals.GroundLiterals().NumberOfElements() > 1)))
      { 
       // splitting makes sense
       
       LIT_NUM_STACK basic_component(" MAX_NUMBER_OF_LITERALS in TMP_CLAUSE::Split():basic_component");
       STACK_AS_SET<int,MAX_NUMBER_OF_LITERALS> all_names("MAX_NUMBER_OF_LITERALS in TMP_CLAUSE::Split():all_names"); 
       Clause* new_clause;
       int name; 
       
       if (literals.GroundLiterals().NumberOfElements())
        {
         if (literals.GroundLiterals().NumberOfElements() == 1)
          {
           basic_component.Push(literals.GroundLiterals().Top());
          }
         else
          {
           new_clause = ComponentToClause(literals.GroundLiterals(),name);
         
           if (!(all_names.Contains(name)))
            {
             TERM header((ARITY)0,(FUNCTOR)name,POL_NEG); // compatible with negative selection
             //HeaderWeighting.MakeMaximal(header);
             
             HeaderWeighting.MakeMinimal(header); // experimental
             
             basic_component.Push(literals.PushRenamingLit(header));
             all_names.Push(name);
            };
         
           if (new_clause)
            {
             result.Push(new_clause);
            };
          };
        };
       
       for (int comp = 0; comp < literals.NumOfNongroundComponents(); comp++)
        {
         new_clause = ComponentToClause(literals.NongroundComponent(comp),name);
         
         if (!(all_names.Contains(name)))
          {
           TERM header((ARITY)0,(FUNCTOR)name,POL_NEG); // compatible with negative selection
           
           //HeaderWeighting.MakeMaximal(header); 
           HeaderWeighting.MakeMinimal(header); // experimental
           
           basic_component.Push(literals.PushRenamingLit(header));
           all_names.Push(name);
          };
         
         if (new_clause)
          {
           result.Push(new_clause);
          };
        };
       
       // basic component forms another clause
       result.Push(MakeClause(basic_component));
      }
     else // nothing to split
      {
       result.Push(MakeClause());
      };
    }; // void Split(CLSTACK& result)

   inline void PreprocessSupportedInputClause(Clause* c,CLSTACK& result)
    {
     Reset();
     LoadClause(c); 
     SetParents(PARENTS::NewPreprocessedParents(c));
     result.Reset();

     if (static_splitting_flag)
      {
       Split(result);
      }
     else
      {
       result.Push(MakeClause());
      };
    }; // void PreprocessSupportedInputClause(Clause* c,CLSTACK& result)
   
   inline void PreprocessUnsupportedInputClause(Clause* c,CLSTACK& result)
    {
     Reset();
     LoadClause(c);
     SetParents(PARENTS::NewPreprocessedParents(c));
     result.Reset();
     
     // temporary version
     result.Push(MakeInputClause());
    }; // void void PreprocessUnsupportedInputClause(Clause* c,CLSTACK& result)
   
   inline void SimplifyResolvent()
    {
     if (simplify_resolvents)
      {
       STACK<Clause*,MAX_NUMBER_OF_LITERALS> simplifiers("MAX_NUMBER_OF_LITERALS in TMP_CLAUSE::SimplifyResolvent():simplifiers");
       Clause* curr_simplifier;
       int lit = Length() - 1; // the order is essential!
       while (lit >= 0)
        {
         Literal(lit)->InversePolarity();
         
         if ((curr_simplifier = FSubsume())) // a subsuming resolvent found
          {
           simplifiers.Push(curr_simplifier);
           literals.RemoveTerm(lit);
          }
         else
          {
           Literal(lit)->InversePolarity();
           
          };
         lit--; 
        };
       int num_of_simplifiers = simplifiers.NumberOfElements();
       if (num_of_simplifiers)
        {
         STATISTICS_ON_GENERATED::AnotherSimplifiedResolvent();
         int prev_num_of_parents = parents->NumberOfParents();
      
         parents = parents->ExpandBy(num_of_simplifiers);
         
         Clause** next_parent = parents->Clauses() + prev_num_of_parents;
         while (!(simplifiers.Empty()))
          {
           *next_parent = simplifiers.Pop();
           next_parent++;
          };
        }; 
      };
    }; // void SimplifyResolvent()
    
   inline void ProcessCollectedResolvent()
    {
     if (!(Kept.AcceptedWeight(literals.Weight())))
      {
      
       //cout << "REJECTED " << literals.Weight() << "\n"; // debug
      
       if (Length() == 1) // unit clause
        {
         literals.ResetVarRenaming();
         literals.RenameVariables();
         tmp_unit_clause_lits->Args()->SetFirst(Literal(0));
         tmp_unit_clause->SetParents(parents);
         tmp_unit_clause->SetWeight(literals.Weight());
         TestForUnitConflict(tmp_unit_clause);
         parents->Delete(); // otherwise memory leak
        };
       
       STATISTICS_ON_GENERATED::AnotherRejectedByKept();
       return;
      };
     if ((Length()) && (forward_subsumption_flag))
      {
       if (FSubsume())
        {
         STATISTICS_ON_GENERATED::AnotherForwardSubsumed();
         return;
        };
      };
     
     BackupParents();
     SimplifyResolvent();
      
     if (Length() == 1) // unit clause
      { IntegrateNewResolvent(MakeUnitClause()); }
     else
      {
       // temporary version
       created_clauses.Reset();
       created_clauses.Push(MakeClause());
       IntegrateNewResolvents(created_clauses);
      };
      
     RestoreParents();
    };
  
  inline bool WasTautology() const { return literals.WasTautology(); };
     
  private:
   inline void OpenLiteral()
    {
     literals.OpenTerm();
    };
   inline bool OpenLiteral(TERM head)
    {
     literals.OpenTerm();
     return literals.PushFunc(head); 
    };
   inline bool CloseLiteral() { return literals.CloseTerm(); };
   inline bool PushPropositional(const TERM fun)
    {
     return literals.PushConst(fun);
    };
   
   inline bool PushVar(VAR* v)
    {
     int index;
     int var_num;
     v->NumAndIndex(var_num,index);
     if (index) 
      {
       STATISTICS_ON_GENERATED::AnotherClauseWithTooManyVariables();
       return false; 
      }
     else
      { 
       literals.PushVar(var_num); 
       return true;
      };
    };
   
   inline bool PushFunc(PREF_SYM func)
    {
     return literals.PushFunc(func.Func()); 
    };
  public:
   inline int Length() { return literals.Length(); };
   inline TERM* Literal(const int n) { return literals.Term(n); };
   inline const TERM* RefToLiteral(const int n) { return literals.RefToTerm(n); }; 
   inline bool PushPrefixLit(PREFIX_SYM* word);
   inline bool PushStandardLit(int index,TERM* t);
   inline bool PushLiteral(TERM* const t);
   inline bool LoadClause(const Clause* const c);
  public:
   inline PARENTS* Parents() { return parents; };
   inline void SetParents(PARENTS* const p) 
    {
     if ((parents) && (parents->NumberOfChildren() == 1))
      {
       
       // parents with no children, memory must be released
       parents->UnconditionalDelete();
      };
     
     parents = p; 
     if (parents)
      {
       parents->RegisterNewChild();
      };
    };
   inline void BackupParents() { spare_parents = parents; };
   inline void RestoreParents() { parents = spare_parents; }; 
    
  private: 
   inline Clause* MakeClause(LIT_NUM_STACK& literals);
   inline Clause* MakeGroundClause(LIT_NUM_STACK& literals);
   inline Clause* MakeUnitClause();
   inline Clause* MakeInputClause();
   inline Clause* MakeClause() 
    {
     LIT_NUM_STACK all_lits("MAX_NUMBER_OF_LITERALS in TMP_CLAUSE::MakeClause():all_lits");
     bool ground;
     literals.PushAllLitNums(all_lits,ground);
     if (ground)
      {
       return MakeGroundClause(all_lits);
      }
     else
      {
       return MakeClause(all_lits);
      };
    };
  public: // for debugging
   inline void Output() { literals.Output(); };
   inline void OutputRenaming() { renaming.Output(); };
 }; // class TMP_CLAUSE

inline bool TMP_CLAUSE::PushPrefixLit(PREFIX_SYM* word)
 {
  TERM header = word->Head().Func();
  if (!(header.Arity())) { return PushPropositional(header); };

  OpenLiteral();
  if (!(PushFunc(header))) { return false; };

  PREFIX::ITER iter;
  iter.Reset(word);
  iter.SkipSym(); // skipping the header
  do
   {
    if (iter.CurrVar()) // free variable
     { 
      if (!(PushVar(renaming.Rename(iter.CurrVar())))) { return false; };  
     }
    else // functor
     {
      if (!(PushFunc(iter.CurrSubterm()->Head()))) { return false; };   
     };
    iter.SkipSym();
   }
  while (!(iter.Empty()));
  return CloseLiteral();
 }; // bool TMP_CLAUSE::PushPrefixLit(PREFIX_SYM* word)


inline bool TMP_CLAUSE::PushStandardLit(int index,TERM* t)
 {
  // Note: this function uses the global substitution to substitute the variables!
  TERM header = *t;
  if (!(header.Arity())) { return PushPropositional(header); };
  if (!(OpenLiteral(header))) { return false; };
  ST_TERM_WITH_SUBST term;
  term.Reset(index,t);
  term.SkipSym();
  PREF_SYM sym;
  while (!(term.Empty()))
   {
    sym = term.CurrSym();
    if (sym.IsVariable()) // free variable, must be renamed
     {
      if (!(PushVar(renaming.Rename(sym.Var())))) { return false; };
     }
    else
     {
      if (!(PushFunc(sym))) { return false; };  
     };
    term.SkipSym();
   };
  return CloseLiteral();  
 }; // bool TMP_CLAUSE::PushStandardLit(int index,TERM* t)

inline bool TMP_CLAUSE::PushLiteral(TERM* const t)
 {
  TERM header = *t;
  if (!(header.Arity())) { return PushPropositional(header); };
  if (!(OpenLiteral(header))) 
   {
    return false; 
   };
  TERM::ITERATOR iter;
  iter.Reset(t);
  while (iter.Next()) 
   {
    if (iter.CurrentSymIsVar())
     {
      literals.PushVar(iter.Symbol().Var());
     }
    else
     {
      if (!(literals.PushFunc(iter.Symbol()))) 
       {
        return false; 
       };
     };
   };
  return CloseLiteral();
 }; // bool TMP_CLAUSE::PushLiteral(TERM* const t)

inline bool TMP_CLAUSE::LoadClause(const Clause* const c)
 {
  TERM* lit = c->Literals()->Args();
  do
   {
    if (!(PushLiteral(lit->First()))) { return false; };
    lit = lit->Next();
   }
  while (!(lit->IsNull()));
  return true;
 }; // void TMP_CLAUSE::LoadClause(const int index,Clause* const c)

inline Clause* TMP_CLAUSE::MakeClause(LIT_NUM_STACK& lits)
 {
  static VAR_WEIGHTING var_weighting;
  
  if (lits.Empty())
   {
    Clause* empty_clause = Clause::MakeEmptyClause();
    empty_clause->SetParents(parents); 
    return empty_clause;
   };
  
  long weight = literals.Weight(lits);
  
  LIT_NUM_STACK max_lits("MAX_NUMBER_OF_LITERALS in TMP_CLAUSE::MakeClause(..):max_lits");
  LIT_NUM_STACK nonmax_lits("MAX_NUMBER_OF_LITERALS in TMP_CLAUSE::MakeClause(..):nonmax_lits");
  
  literals.KBOrder(lits,max_lits,nonmax_lits);
  
  int num_of_max = max_lits.NumberOfElements();
  int num_of_nonmax = nonmax_lits.NumberOfElements();
  
  if (num_of_max > 1)
   {
    literals.CreateVarWeighting(lits,var_weighting);
    literals.Order(max_lits,var_weighting);
    if (num_of_nonmax > 1)
     {
      literals.Order(nonmax_lits,var_weighting);
     };
   }
  else
   {
    if (num_of_nonmax > 1)
     {
      literals.CreateVarWeighting(lits,var_weighting);
      literals.Order(nonmax_lits,var_weighting);
     };
   };
  
  literals.ResetVarRenaming();
  
  
  literals.RenameVariables(max_lits);
  literals.RenameVariables(nonmax_lits);
  
  TERM* lit_list = TERM::NewLiterals(num_of_max + num_of_nonmax,num_of_max);
  TERM* lit_ref = lit_list->Args();
  while (!(max_lits.Empty())) 
   {
    lit_ref->SetFirst(Literal(max_lits.Pop())); 
    lit_ref = lit_ref->Next();
   };
  while (!(nonmax_lits.Empty())) 
   {
    lit_ref->SetFirst(Literal(nonmax_lits.Pop())); 
    lit_ref = lit_ref->Next();
   }; 
   
  Clause* result = new Clause(lit_list);
  result->SetParents(parents);
  result->SetWeight(weight);
  return result;
 }; // Clause* TMP_CLAUSE::MakeClause(LIT_NUM_STACK& literals)

inline Clause* TMP_CLAUSE::MakeGroundClause(LIT_NUM_STACK& lits)
 {
  static VAR_WEIGHTING var_weighting; // not used
 
  if (lits.Empty())
   {
    Clause* empty_clause = Clause::MakeEmptyClause();
    empty_clause->SetParents(parents); 
    return empty_clause;
   };
  
  long weight = literals.Weight(lits);
  
  LIT_NUM_STACK max_lits("MAX_NUMBER_OF_LITERALS in TMP_CLAUSE::MakeClause(..):max_lits");
  LIT_NUM_STACK nonmax_lits("MAX_NUMBER_OF_LITERALS in TMP_CLAUSE::MakeClause(..):nonmax_lits");
  
  literals.KBOrder(lits,max_lits,nonmax_lits);
  
  int num_of_max = max_lits.NumberOfElements();
  int num_of_nonmax = nonmax_lits.NumberOfElements();
  
  if (num_of_max > 1)
   {
    literals.CreateVarWeighting(lits,var_weighting);
    literals.Order(max_lits,var_weighting);
    if (num_of_nonmax > 1)
     {
      literals.Order(nonmax_lits,var_weighting);
     };
   }
  else
   {
    if (num_of_nonmax > 1)
     {
      literals.CreateVarWeighting(lits,var_weighting);
      literals.Order(nonmax_lits,var_weighting);
     };
   };
  
  TERM* lit_list = TERM::NewLiterals(num_of_max + num_of_nonmax,num_of_max);
  TERM* lit_ref = lit_list->Args();
  while (!(max_lits.Empty())) 
   {
    lit_ref->SetFirst(Literal(max_lits.Pop())); 
    lit_ref = lit_ref->Next();
   };
  while (!(nonmax_lits.Empty())) 
   {
    lit_ref->SetFirst(Literal(nonmax_lits.Pop())); 
    lit_ref = lit_ref->Next();
   }; 
   
  Clause* result = new Clause(lit_list);
  result->SetParents(parents);
  result->SetWeight(weight);
  return result;
 }; // Clause* TMP_CLAUSE::MakeGroundClause(LIT_NUM_STACK& literals)

inline Clause* TMP_CLAUSE::MakeUnitClause()
 {
  literals.ResetVarRenaming();
  literals.RenameVariables();        
  
  TERM* lit_list = TERM::NewLiterals(1,1);
  lit_list->Args()->SetFirst(Literal(0));
  Clause* result = new Clause(lit_list);
  result->SetParents(parents); 
  result->SetWeight(literals.Weight());
  return result;
 }; // Clause* TMP_CLAUSE::MakeUnitClause()

inline Clause* TMP_CLAUSE::MakeInputClause()
 {
  static VAR_WEIGHTING var_weighting;
  if (literals.Length() > 1) // nonunit clause
   {
    LIT_NUM_STACK all_lits("MAX_NUMBER_OF_LITERALS in TMP_CLAUSE::MakeClause():all_lits");
    bool ground;
    literals.PushAllLitNums(all_lits,ground);
    literals.CreateVarWeighting(all_lits,var_weighting);
    literals.Order(all_lits,var_weighting);
    literals.ResetVarRenaming();
    literals.RenameVariables(all_lits);
    TERM* lit_list = TERM::NewLiterals(literals.Length(),literals.Length());
    TERM* lit_ref = lit_list->Args();
    while (!(all_lits.Empty()))
     {
      lit_ref->SetFirst(Literal(all_lits.Pop())); 
      lit_ref = lit_ref->Next();
     };
    Clause* result = new Clause(lit_list);
    result->SetParents(parents);
    result->SetWeight(literals.Weight());
    return result;
   }
  else // unit clause
   {
    return MakeUnitClause();
   };
 }; // Clause* TMP_CLAUSE::MakeInputClause()

//***************************************************************************