#define __TMP_LITERAL_H

 
#ifndef __DEBUG_H
 #include "debug.h"
#endif
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif
#ifndef __TERM_H
 #include "term.h"
#endif
#ifndef __STACK_H
 #include "stack.h"
#endif
#ifndef __VARIABLES_H
 #include "variables.h"
#endif
#ifndef __WEIGHTING_H
 #include "weighting.h"
#endif
#ifndef __STATISTICS_H
 #include "statistics.h"
#endif
#ifndef __FLAT_TERM_H
 #include "flat_term.h"
#endif 
#ifndef __SIGNATURE_H
 #include "signature.h"
#endif
#ifndef __CLAUSE_INDEX_H
 #include "clause_index.h"
#endif

extern CLAUSE_INDEX ClauseIndex; 

extern HEADER_WEIGHTING HeaderWeighting;

class TMP_LITERAL
 {
  private:
   TERM header;
   FLAT_TERM::OPEN_WITH_MEMORY flat_term;
   
   //long bullshit[10];
   
   long weight;
   
   VAR_MULTISET variables;
   VAR_MULTISET eq_vars1;
   VAR_MULTISET eq_vars2;
   bool eq_different_order;
   long max_depth;
  public:
   TMP_LITERAL() 
    : flat_term(), 
      variables(), 
      max_depth(MAX_TERM_DEPTH) 
     {};
   inline void Reset()
    {
     flat_term.Reset();
     weight = 0;
     variables.Reset();
    };
   inline void SetMaxDepth(long md) 
    {
     if (md > MAX_TERM_DEPTH) { max_depth = MAX_TERM_DEPTH; }
     else max_depth = md;
    };
   inline const FLAT_TERM* FlatTerm() const { return flat_term.Term(); }; 
   inline long CurrWeight() const { return flat_term.CurrWeight(); }; 
   //inline TERM Header() const { return flat_term.Term()->Symbol(); };
   inline TERM Header() const { return header; };
   //inline long Weight() const { return FlatTerm()->Weight(); };
   inline unsigned Depth() { return ((FLAT_TERM*)FlatTerm())->Depth(); };
   inline long Weight() const { return weight; };
   inline long NumOfRedexes(); 
   inline VAR_MULTISET& Variables() { return variables; };
   inline bool Ground() const { return !(variables.Size()); };
   inline bool Nonground() const { return !(variables.Size()); };
   inline bool IsEquality() const { return header.IsEquality(); };
   inline bool IsOrderedEq() const { return header.IsOrderedEq(); };
   inline bool IsUnorderedEq() const { return header.IsUnorderedEq(); };
   inline bool Positive() const { return header.Positive(); };
   inline bool Negative() const { return header.Negative(); }; 
   inline POLARITY Polarity() const { return header.Polarity(); };
   inline bool NotPropositional() const { return (bool)header.Arity(); };
   inline bool Defined();  
     

   inline const FLAT_TERM* Arg1() const { return FlatTerm() + 1; };
   inline const FLAT_TERM* Arg2() const { return Arg1()->NextTerm(); };
   inline VAR_MULTISET& Arg1Vars() { return eq_vars1; };
   inline VAR_MULTISET& Arg2Vars() { return eq_vars2; };
   inline void LinkEqArgs() 
    { 
     CHECK8(FlatTerm() < (FlatTerm()->NextTerm()),"A0 in TMP_LITERAL::SetBackJump()");
     ((FLAT_TERM*)(FlatTerm()->NextTerm()))->SetBackJump(FlatTerm() + 1); 
    };
   inline bool EqSwapArgs() const { return eq_different_order; };
   inline void EqCollectVars();
   inline void EqMakeOrdered()
    {
     header.SetFunctor(ORD_EQ_NUM);
     ((FLAT_TERM*)FlatTerm())->SymbolRef() = header;
    };
   
   inline long PosEqMinLHSWeight() const
    {
     CHECK8(IsEquality() && Positive(),"A0 in TMP_LITERAL::PosEqMinLHSWeight()");
     if (IsOrderedEq())
      {
       if (EqSwapArgs()) return Arg1()->Weight();
       return Arg2()->Weight();
      }
     else // unordered 
      {
       long w1 = Arg1()->Weight();
       long w2 = Arg2()->Weight();
       return (w1 > w2) ? w2 : w1;
      }; 
    };
   
   inline void PushVar(const int v)
    {
     variables.Register(v);
     flat_term.PushVar(TERM(TERM_VAR,v));
    };
   inline void PushVar(const TERM v) 
    {
     variables.Register(v.Var());
     flat_term.PushVar(v); 
    };
   inline void PushConst(const TERM c) { flat_term.PushConst(c); };
   inline bool PushNonconstFun(const TERM f) 
    {
     if (((long)(flat_term.Depth())) < max_depth)
      { 
       flat_term.PushNonconstFun(f); 
       weight += f.Arity();
       return true;
      };
     return false;   
    };
   
   inline bool PushSym(const TERM& sym)
    {
     if (sym.IsVariable()) { PushVar(sym); return true; };
     if (sym.Arity()) return PushNonconstFun(sym);
     PushConst(sym);
     return true;   
    };

   inline void PushNonpropHeader(TERM hd)
    {
     weight = hd.Arity() + 1;
     flat_term.PushNonconstFun(hd);
     header = hd;
    };
   
   inline void PushPropLit(TERM lit)
    {
     weight = 1;
     flat_term.PushConst(lit);
     header = lit;
    };
   
   inline bool EqEqualArgs()
    {
     CHECK8(IsEquality(),"A0 in TMP_LITERAL::EqEqualArgs()");
     return Arg1()->Equal(Arg2());
    };
   inline bool Equal(TMP_LITERAL& lit) const { return FlatTerm()->Equal(lit.FlatTerm()); };
   inline bool Complementary(TMP_LITERAL& lit) const { return FlatTerm()->Complementary(lit.FlatTerm()); };
   
   inline bool EqualEqLit(TMP_LITERAL& lit) const;
   inline bool ComplementaryEqLit(TMP_LITERAL& lit) const;
  
   inline bool SubstPossible(int var,unsigned subst_depth,long subst_weight)
    {
     int occs = variables.Occurences(var);
     if (occs)
      {     
       long new_weight = Weight() + (subst_weight * occs) - occs;
       return (new_weight <= MAX_TERM_WEIGHT) && (Depth() + subst_depth <= MAX_TERM_DEPTH);
      }
     else return true;
    };

   inline bool Subst(TMP_LITERAL& new_lit,int var,FLAT_TERM* subst_term)
    {
     TERM new_header = Header();
     CHECK8(new_header.Arity(),"A0 in TMP_LITERAL::Subst(TMP_LITERAL& new_lit,int var,FLAT_TERM* subst_term)");             if(new_header.IsEquality()) new_header.MakeUnordered();
     new_lit.Reset();
     new_lit.PushNonpropHeader(new_header);
     FLAT_TERM::ITER iter(FlatTerm());
     FLAT_TERM::ITER subst_iter;
     iter.NextSym(); // skipping the header
     while (iter.NotEndOfTerm())
      {
       TERM sym = iter.CurrPos()->Symbol();  
       if (sym.IsVariable() && sym.Var() == var) 
	{
         subst_iter.Reset(subst_term);             
         do
          {
           if (!new_lit.PushSym(((FLAT_TERM*)(subst_iter.CurrPos()))->SymbolRef())) return false;             
           subst_iter.NextSym();
          }
         while (subst_iter.NotEndOfTerm());
        }
       else if (!new_lit.PushSym(sym)) return false;
       iter.NextSym();
      };
     return true;
    };

   inline void RenameVariables(INT_VAR_RENAMING& var_renaming)
    {
     if ((IsEquality()) && eq_different_order)
      {
       ((FLAT_TERM*)Arg2())->RenameVariables(var_renaming);
       ((FLAT_TERM*)Arg1())->RenameVariables(var_renaming);
      }
     else ((FLAT_TERM*)FlatTerm())->RenameVariables(var_renaming);
    };
   inline COMPARISON KBComparePos(TMP_LITERAL& lit);
   inline COMPARISON KBComparePosEq(TMP_LITERAL& lit);
   
   inline COMPARISON KBCompareEqOO(TMP_LITERAL& lit);
   inline COMPARISON KBCompareEqOU(TMP_LITERAL& lit);
   inline COMPARISON KBCompareEqUU(TMP_LITERAL& lit);
   
   inline COMPARISON KBCompareEqArgT1S1(TMP_LITERAL& lit);
   inline COMPARISON KBCompareEqArgT1S2(TMP_LITERAL& lit);
   inline COMPARISON KBCompareEqArgT2S1(TMP_LITERAL& lit);
   inline COMPARISON KBCompareEqArgT2S2(TMP_LITERAL& lit);
   
   inline COMPARISON KBCompareArgs(TMP_LITERAL& lit);
   inline COMPARISON Compare(TMP_LITERAL& lit,const VAR_WEIGHTING& var_weighting);
   inline void EqOrderArgs();
   inline TERM StandardRepCopy() // for input clauses
    {
     if ((IsEquality()) && eq_different_order)
      {
       TERM copy;
       copy.SetFirst(new(2) TERM(header));
       *(copy.First()->Args()) = Arg2()->StandardRepCopy(); 
       *(copy.First()->Args()->Next()) = Arg1()->StandardRepCopy(); 
       return copy;
      }
     else return FlatTerm()->StandardRepCopy(); 
    };  
   inline TERM* SharedCopy(Clause* cl)
    {
     if ((IsEquality()) && eq_different_order)
      {
       return ClauseIndex.IntegrateEq(header,Arg2(),Arg1(),cl); // temporary
      }
     else 
      {
       //cout << "FLAT:\n";
       //FlatTerm()->OutputTerm(); cout << "\n"; // debug
       
       return ClauseIndex.Integrate(FlatTerm(),cl); 
      }; 
    }; 
  public: // for debugging
   inline void Output() 
    {
     if (IsEquality())
      {
       if (IsOrderedEq() && EqSwapArgs())
	{
         ::Output("(");
         Arg2()->OutputTerm();
         ::Output(")");
         ::Output(header);
         ::Output("(");
         Arg1()->OutputTerm();
         ::Output(")");
        }
       else
	{
         ::Output("(");
         Arg1()->OutputTerm();
         ::Output(")");
         ::Output(header);
         ::Output("(");
         Arg2()->OutputTerm();
         ::Output(")");
	};  
      }
     else
      { 
       FlatTerm()->OutputTerm(); 
      }; 
     //::Output("{");
     //variables.Output();
     //::Output("}"); 
    };
 }; // class TMP_LITERAL 
 
//******************* Definitions for TMP_LITERAL: ************************** 

inline long TMP_LITERAL::NumOfRedexes()
 {
  FLAT_TERM* term_to_scan;
  if (IsEquality() && IsOrderedEq())
   {
    if (EqSwapArgs()) { term_to_scan = (FLAT_TERM*)(Arg2()); }  
    else term_to_scan = (FLAT_TERM*)(Arg1()); 
   }
  else term_to_scan = (FLAT_TERM*)(FlatTerm());
  long res = 0;
  if (!term_to_scan->SymbolRef().IsVariable()) 
   {
    FLAT_TERM::ITER iter(term_to_scan);
    do
     {
      TERM sym = ((FLAT_TERM*)(iter.CurrPos()))->SymbolRef();     
      if (sym.IsComplex() && sym.Arity()) res++;
      iter.NextSym();
     }
    while (iter.NotEndOfTerm()); 
   };
  return res;
 }; // long TMP_LITERAL::NumOfRedexes()


inline bool TMP_LITERAL::Defined() 
 {
  

  bool res =  ((!IsEquality()) 
                && (header.Arity() == variables.NumOfDifferent()));    
 
  //Output(); ::Output("\n"); // debug
  //::Output((int)res); ::Output("\n\n\n"); // debug 

  return res;           
 }; // bool TMP_LITERAL::Defined()


inline void TMP_LITERAL::EqCollectVars()
 {
  eq_vars1.Reset();
  eq_vars2.Reset();
  if (Ground()) return;
  const FLAT_TERM* end;
  if (Arg1()->Weight() < Arg2()->Weight())
   {
    end = Arg1()->NextTerm();
    for (const FLAT_TERM* pos = Arg1(); pos < end; pos++)
     {
      TERM sym = pos->Symbol();
      if (sym.IsVariable()) eq_vars1.Register(sym.Var());
     };
    eq_vars2.CollectDifference(variables,eq_vars1); 
   }
  else
   {
    end = Arg2()->NextTerm();
    for (const FLAT_TERM* pos = Arg2(); pos < end; pos++)
     {
      TERM sym = pos->Symbol();
      if (sym.IsVariable()) eq_vars2.Register(sym.Var());
     };
    eq_vars1.CollectDifference(variables,eq_vars2); 
   };
 }; // void TMP_LITERAL::EqCollectVars()

inline void TMP_LITERAL::EqOrderArgs()
 {
  // don't apply this to already ordered equations!
  // this must be an equational literal
  EqCollectVars();
  if (IsOrderedEq()) { eq_different_order = false; return; }; // already ordered  
  if (Arg1()->Weight() == Arg2()->Weight())
   {
    if (eq_vars1.Equal(eq_vars2))
     {
      // compare lexicographically
      switch (Arg1()->CompareLex(Arg2()))
       {
        case LESS :
         eq_different_order = true;
         EqMakeOrdered();
         break;
        case GREATER :
         eq_different_order = false; 
         EqMakeOrdered();
         break;
        case EQUAL : // same as for INCOMPARABLE
        case INCOMPARABLE :
         
         //Output(); cout << "\n"; // debug
         
         eq_different_order = false;
       };
     }
    else // incomparable
     {
      //Output(); cout << "\n"; // debug
     
      eq_different_order = false;
     };
   }
  else 
   if (Arg1()->Weight() > Arg2()->Weight())
    {
     eq_different_order = false;
     if (eq_vars1.GreaterOrEqual(eq_vars2)) 
      {
       EqMakeOrdered();
      }
     else 
      {
       //cout << "(1) "; Output(); cout << "\n"; // debug
      };  
    }
   else // Arg1()->Weight() < Arg2()->Weight()
    {
     eq_different_order = true;
     if (eq_vars2.GreaterOrEqual(eq_vars1)) 
      {
       EqMakeOrdered();
      }
     else
      {
       //cout << "(2) "; Output(); cout << "\n"; // debug
       //cout << "ALL VARS:\n"; variables.Output(); cout << "\n\n"; // debug
       //cout << "VARS1:\n"; eq_vars1.Output(); cout << "\n\n"; // debug
       //cout << "VARS2:\n"; eq_vars2.Output(); cout << "\n\n"; // debug
       
      };  
    };
 }; // void TMP_LITERAL::EqOrderArgs()

inline bool TMP_LITERAL::EqualEqLit(TMP_LITERAL& lit) const 
 {
  // both this and lit must be equational literals
  if (Polarity() != lit.Polarity()) return false;
  if (Arg1()->Equal(lit.Arg1()))
   {
    
    return Arg2()->Equal(lit.Arg2());
   }
  else
   {
    if ((IsUnorderedEq()) || (lit.IsUnorderedEq()))
     {
      return (Arg1()->Equal(lit.Arg2())) && (Arg2()->Equal(lit.Arg1()));
     }
    else return false; 
   }; 
 }; // bool TMP_LITERAL::EqualEqLit(TMP_LITERAL& lit) const 
    
inline bool TMP_LITERAL::ComplementaryEqLit(TMP_LITERAL& lit) const 
 {
  // both this and lit must be equational literals
  if (Polarity() == lit.Polarity()) return false;
  if (Arg1()->Equal(lit.Arg1()))
   {
    return Arg2()->Equal(lit.Arg2());
   }
  else
   {
    if ((IsOrderedEq()) && (lit.IsOrderedEq()))
     {
      return false;
     }
    else return (Arg1()->Equal(lit.Arg2())) && (Arg2()->Equal(lit.Arg1()));
   };
 }; // bool TMP_LITERAL::ComplementaryEqLit(TMP_LITERAL& lit) const 
 
inline COMPARISON TMP_LITERAL::KBCompareArgs(TMP_LITERAL& lit)
 {
  //Output(); ::Output("   VS  "); lit.Output(); ::Output("\n"); // debug

  DCONTROL8("BEGIN in TMP_LITERAL::KBCompareArgs(..)"); 
  if (Weight() == lit.Weight())
   {
    //::Output("EQ. WEIGHT\n"); // debug   

    if (variables.Equal(lit.variables))
     {
      //::Output("EQ. VARS\n"); // debug   
      
      // compare lexicographically
      const FLAT_TERM* fst = FlatTerm() + 1;
      const FLAT_TERM* const end_of_fst = FlatTerm()->NextTerm();
      const FLAT_TERM* snd = lit.FlatTerm();
      check_end:
       if (fst == end_of_fst) return EQUAL;
       snd++;
       if (fst->Symbol() != snd->Symbol())
        {
         if ((fst->Symbol().IsVariable()) || (snd->Symbol().IsVariable()))
          return INCOMPARABLE;
         return fst->Symbol().CompareFunctor(snd->Symbol()); 
        };
       fst++;
       goto check_end;
     };
 
   // ::Output("INCOMP!!!\n"); // debug

    return INCOMPARABLE;   
   } 
  else
   {
    if (Weight() > lit.Weight())
     return (variables.GreaterOrEqual(lit.variables)) ? GREATER : INCOMPARABLE;
    // Weight() < lit.Weight()
    return (lit.variables.GreaterOrEqual(variables)) ? LESS : INCOMPARABLE;
   }; 
 }; // COMPARISON TMP_LITERAL::KBCompareArgs(TMP_LITERAL& lit) 
 
inline COMPARISON TMP_LITERAL::KBComparePosEq(TMP_LITERAL& lit)
 {
  CHECK8(Positive(),"A0 in TMP_LITERAL::KBComparePosEq(..)");
  CHECK8(lit.Positive(),"A1 in TMP_LITERAL::KBComparePosEq(..)");
  
  if (IsOrderedEq())
   {
    if (lit.IsOrderedEq()) return KBCompareEqOO(lit); 
    return KBCompareEqOU(lit);
   }; 
  // this is unordered  
  if (lit.IsOrderedEq()) return lit.KBCompareEqOU(*this); 
  // both are unordered
  return KBCompareEqUU(lit);
  
 }; // COMPARISON TMP_LITERAL::KBComparePosEq(TMP_LITERAL& lit)

inline COMPARISON TMP_LITERAL::KBComparePos(TMP_LITERAL& lit)
 {
  CHECK8(Positive(),"A0 in TMP_LITERAL::KBComparePos(..)");
  CHECK8(lit.Positive(),"A1 in TMP_LITERAL::KBComparePos(..)");
  
  if (IsEquality())
   {
    if (lit.IsEquality())  
     {
      // both are equalities
      return KBComparePosEq(lit);
     }
    else return LESS; 
   }
  else // this is not an equality
   {
    if (lit.IsEquality()) { return GREATER; };
    // both are not equalities
    switch (HeaderWeighting.Compare(Header(),lit.Header()))
     {
      case GREATER : return GREATER;
      case LESS : return LESS;
      case EQUAL : return KBCompareArgs(lit); 
      #ifdef _DEBUG8_
       default: DebugError("E0 in TMP_LITERAL::KBComparePos(..)"); return INCOMPARABLE; 
      #else
        #ifdef _SUPPRESS_WARNINGS_
         default: return INCOMPARABLE; 
        #endif 
      #endif 
     };
   };  
 }; // COMPARISON TMP_LITERAL::KBComparePos(TMP_LITERAL& lit) 


inline COMPARISON TMP_LITERAL::KBCompareEqOO(TMP_LITERAL& lit)
 {
  CHECK8(IsOrderedEq(),"A0 in TMP_LITERAL::KBCompareEqOO(..)");
  CHECK8(lit.IsOrderedEq(),"A1 in TMP_LITERAL::KBCompareEqOO(..)");
  CHECK8(Positive(),"A2 in TMP_LITERAL::KBCompareEqOO(..)");
  CHECK8(lit.Positive(),"A3 in TMP_LITERAL::KBCompareEqOO(..)");
  
  const FLAT_TERM* max;
  const FLAT_TERM* lit_max;
  VAR_MULTISET* max_vars;
  VAR_MULTISET* lit_max_vars;
  
  if (EqSwapArgs()) { max = Arg2(); max_vars = &eq_vars2; }
  else { max = Arg1(); max_vars = &eq_vars1; };
  
  if (lit.EqSwapArgs()) { lit_max = lit.Arg2(); lit_max_vars = &(lit.eq_vars2); }
  else { lit_max = lit.Arg1(); lit_max_vars = &(lit.eq_vars1); };  
  
  COMPARISON cmp = FLAT_TERM::KBCompare(max,max_vars,lit_max,lit_max_vars);
  if (cmp != EQUAL) return cmp;
  return INCOMPARABLE;
 }; // COMPARISON TMP_LITERAL::KBCompareEqOO(TMP_LITERAL& lit)

inline COMPARISON TMP_LITERAL::KBCompareEqOU(TMP_LITERAL& lit)
 {
  CHECK8(IsOrderedEq(),"A0 in TMP_LITERAL::KBCompareEqOU(..)");
  CHECK8(lit.IsUnorderedEq(),"A1 in TMP_LITERAL::KBCompareEqOU(..)");
  CHECK8(Positive(),"A2 in TMP_LITERAL::KBCompareEqOU(..)");
  CHECK8(lit.Positive(),"A3 in TMP_LITERAL::KBCompareEqOU(..)");
  
  const FLAT_TERM* max;
  const FLAT_TERM* lit_max;
  VAR_MULTISET* max_vars;
  VAR_MULTISET* lit_max_vars; 
  
  if (EqSwapArgs()) { max = Arg2(); max_vars = &eq_vars2; }
  else { max = Arg1(); max_vars = &eq_vars1; };
  
  lit_max = lit.Arg1(); 
  lit_max_vars = &(lit.eq_vars1);  
  
  COMPARISON cmp = FLAT_TERM::KBCompare(max,max_vars,lit_max,lit_max_vars);

  switch (cmp)
   {
    case GREATER: 
     // t1 > s1
     lit_max = lit.Arg2(); 
     lit_max_vars = &(lit.eq_vars2);
     // compare t1 and s2
     cmp = FLAT_TERM::KBCompare(max,max_vars,lit_max,lit_max_vars);
     if (cmp == GREATER) return GREATER;
     return INCOMPARABLE;
    case LESS: 
     // t1 < s1
     return LESS;
    case INCOMPARABLE: 
     return INCOMPARABLE; 
    case EQUAL: return INCOMPARABLE;
   };
  #ifdef _DEBUG8_ 
    DebugError("DE100 in TMP_LITERAL::KBCompareEqOU(..)");
    return INCOMPARABLE; 
  #else
   #ifdef _SUPPRESS_WARNINGS_
    return INCOMPARABLE; 
   #endif  
  #endif 
 }; // COMPARISON TMP_LITERAL::KBCompareEqOU(TMP_LITERAL& lit)

inline COMPARISON TMP_LITERAL::KBCompareEqUU(TMP_LITERAL& lit)  
 {
  // compare t1 = t2 with s1 = s2, below "*" means "incomparable"
  CHECK8(IsUnorderedEq(),"A0 in TMP_LITERAL::KBCompareEqUU(..)");
  CHECK8(lit.IsUnorderedEq(),"A1 in TMP_LITERAL::KBCompareEqUU(..)");
 
  COMPARISON cmp = KBCompareEqArgT1S1(lit);
  switch (cmp)
   {
    case GREATER:
     // t1 > s1
     cmp = KBCompareEqArgT1S2(lit); 
     switch (cmp)
      {
       case GREATER:
        // t1 > s1, t1 > s2
        return GREATER;
        
       case LESS:
        // t1 > s1, t1 < s2  : error because s1 * s2
        #ifdef _DEBUG8_ 
         DebugError("DE10 in TMP_LITERAL::KBCompareEqUU(..)");
        #endif
        return INCOMPARABLE;
       
       case INCOMPARABLE:
        // t1 > s1, t1 * s2
        cmp = KBCompareEqArgT2S2(lit); 
        switch (cmp)
         {
          case GREATER:
           // t1 > s1, t1 * s2, t2 > s2
           return GREATER;
            
          case LESS:
           // t1 > s1, t1 * s2, t2 < s2
           return INCOMPARABLE;
           
          case INCOMPARABLE:
           // t1 > s1, t1 * s2, t2 * s2
           return INCOMPARABLE;
           
          case EQUAL:
           // t1 > s1, t1 * s2, t2 = s2
           return GREATER;
         };
        
       case EQUAL:
        // t1 > s1, t1 = s2 : error because s1 * s2
        #ifdef _DEBUG8_ 
         DebugError("DE20 in TMP_LITERAL::KBCompareEqUU(..)");
        #endif
        return INCOMPARABLE;
      };
    case LESS:
     // t1 < s1
     cmp = KBCompareEqArgT2S1(lit); 
     switch (cmp)
      {
       case GREATER:
        // t1 < s1, t2 > s1 : error because t1 * t2
        #ifdef _DEBUG8_ 
         DebugError("DE30 in TMP_LITERAL::KBCompareEqUU(..)");
        #endif
        return INCOMPARABLE;
        
       case LESS:
        // t1 < s1, t2 < s1
        return GREATER;
        
       case INCOMPARABLE:
        // t1 < s1, t2 * s1
        cmp = KBCompareEqArgT2S2(lit); 
        switch (cmp)
         {
          case GREATER:
           // t1 < s1, t2 * s1, t2 > s2
           return INCOMPARABLE;
           
          case LESS:
           // t1 < s1, t2 * s1, t2 < s2
           return LESS;
           
          case INCOMPARABLE:
           // t1 < s1, t2 * s1, t2 * s2
           return INCOMPARABLE;
           
          case EQUAL:
           // t1 < s1, t2 * s1, t2 = s2
           return LESS;
         }; 
        
       case EQUAL:
        // t1 < s1, t2 = s1 : error because t1 * t2
        #ifdef _DEBUG8_ 
         DebugError("DE40 in TMP_LITERAL::KBCompareEqUU(..)");
        #endif
        return INCOMPARABLE;
      };
      
    case INCOMPARABLE:
     // t1 * s1
     cmp = KBCompareEqArgT1S2(lit); 
     switch (cmp)
      {
       case GREATER:
        // t1 * s1, t1 > s2
        cmp = KBCompareEqArgT2S1(lit); 
        switch (cmp)
         {
          case GREATER:
           // t1 * s1, t1 > s2, t2 > s1
           return GREATER; 
           
          case LESS:
           // t1 * s1, t1 > s2, t2 < s1
           return INCOMPARABLE;
           
          case INCOMPARABLE:
           // t1 * s1, t1 > s2, t2 * s1
           return INCOMPARABLE;
           
          case EQUAL:
           // t1 * s1, t1 > s2, t2 = s1
           return GREATER;
         };
          
       case LESS:
        // t1 * s1, t1 < s2
        cmp = KBCompareEqArgT2S1(lit); 
        switch (cmp)
         {
          case GREATER:
           // t1 * s1, t1 < s2, t2 > s1
           return INCOMPARABLE;
          
          case LESS:
           // t1 * s1, t1 < s2, t2 < s1
           return LESS;
            
          case INCOMPARABLE:
           // t1 * s1, t1 < s2, t2 * s1
           cmp = KBCompareEqArgT2S2(lit); 
           switch (cmp)
            {
             case GREATER:
              // t1 * s1, t1 < s2, t2 * s1, t2 > s2 : error because t1 * t2
              #ifdef _DEBUG8_ 
               DebugError("DE50 in TMP_LITERAL::KBCompareEqUU(..)");
              #endif
              return INCOMPARABLE; 
           
             case LESS:
              // t1 * s1, t1 < s2, t2 * s1, t2 < s2
              return LESS;
              
             case INCOMPARABLE:
              // t1 * s1, t1 < s2, t2 * s1, t2 * s2
              return INCOMPARABLE;
              
             case EQUAL:
              // t1 * s1, t1 < s2, t2 * s1, t2 = s2 : error because t1 * t2
              #ifdef _DEBUG8_ 
               DebugError("DE60 in TMP_LITERAL::KBCompareEqUU(..)");
              #endif
              return INCOMPARABLE; 
            };
            
          case EQUAL:
           // t1 * s1, t1 < s2, t2 = s1
           return LESS;
         }; 
         
       case INCOMPARABLE:
        // t1 * s1, t1 * s2
        cmp = KBCompareEqArgT2S1(lit); 
        switch (cmp)
         {
          case GREATER:
           // t1 * s1, t1 * s2, t2 > s1
           cmp = KBCompareEqArgT2S2(lit); 
           switch (cmp)
            {
             case GREATER:
              // t1 * s1, t1 * s2, t2 > s1, t2 > s2
              return GREATER;
              
             case LESS:
              // t1 * s1, t1 * s2, t2 > s1, t2 < s2 : error because s1 * s2
              #ifdef _DEBUG8_ 
               DebugError("DE70 in TMP_LITERAL::KBCompareEqUU(..)");
              #endif
              return INCOMPARABLE; 
              
             case INCOMPARABLE:
              // t1 * s1, t1 * s2, t2 > s1, t2 * s2
              return INCOMPARABLE;
              
             case EQUAL:
              // t1 * s1, t1 * s2, t2 > s1, t2 = s2 : error because s1 * s2
              #ifdef _DEBUG8_ 
               DebugError("DE80 in TMP_LITERAL::KBCompareEqUU(..)");
              #endif
              return INCOMPARABLE; 
            };
           
          case LESS:
           // t1 * s1, t1 * s2, t2 < s1
           return INCOMPARABLE;
           
          case INCOMPARABLE:
           // t1 * s1, t1 * s2, t2 * s1
           return INCOMPARABLE;
           
          case EQUAL:
           // t1 * s1, t1 * s2, t2 = s1
           return INCOMPARABLE;
         }; 
        
       case EQUAL:
        // t1 * s1, t1 = s2
        cmp = KBCompareEqArgT2S1(lit); 
        return cmp;
      };
      
    case EQUAL:
     // t1 = s1
     cmp = KBCompareEqArgT2S2(lit); 
     return cmp;
   };
  #ifdef _DEBUG8_ 
    DebugError("DE100 in TMP_LITERAL::KBCompareEqUU(..)");
    return INCOMPARABLE; 
  #else
   #ifdef _SUPPRESS_WARNINGS_
    return INCOMPARABLE; 
   #endif  
  #endif
   
 }; // COMPARISON TMP_LITERAL::KBCompareEqUU(TMP_LITERAL& lit)  


inline COMPARISON TMP_LITERAL::KBCompareEqArgT1S1(TMP_LITERAL& lit)
 {
  return FLAT_TERM::KBCompare(Arg1(),&eq_vars1,lit.Arg1(),&lit.eq_vars1);
 }; // COMPARISON TMP_LITERAL::KBCompareEqT1S1(TMP_LITERAL& lit)
 
inline COMPARISON TMP_LITERAL::KBCompareEqArgT1S2(TMP_LITERAL& lit)
 {
  return FLAT_TERM::KBCompare(Arg1(),&eq_vars1,lit.Arg2(),&lit.eq_vars2);
 }; // COMPARISON TMP_LITERAL::KBCompareEqT1S2(TMP_LITERAL& lit)

inline COMPARISON TMP_LITERAL::KBCompareEqArgT2S1(TMP_LITERAL& lit)
 {
  return FLAT_TERM::KBCompare(Arg2(),&eq_vars2,lit.Arg1(),&lit.eq_vars1);
 }; // COMPARISON TMP_LITERAL::KBCompareEqT2S1(TMP_LITERAL& lit)   

inline COMPARISON TMP_LITERAL::KBCompareEqArgT2S2(TMP_LITERAL& lit)
 {
  return FLAT_TERM::KBCompare(Arg2(),&eq_vars2,lit.Arg2(),&lit.eq_vars2);
 }; // COMPARISON TMP_LITERAL::KBCompareEqT2S2(TMP_LITERAL& lit)    
   
   
 
inline COMPARISON TMP_LITERAL::Compare(TMP_LITERAL& lit,const VAR_WEIGHTING& var_weighting)
 {
  COMPARISON hd_cmp = HeaderWeighting.Compare(Header(),lit.Header());
  if (hd_cmp != EQUAL) { return hd_cmp; };
  
  // here the headers are equal
  if (Weight() > lit.Weight()) { return GREATER; };
  if (Weight() < lit.Weight()) { return LESS; };
  
  // here the weights are equal
  if (variables.Size() < lit.variables.Size()) { return GREATER; };
  if (variables.Size() > lit.variables.Size()) { return LESS; };
  
  // here both terms have same number of variable occurences
  if (variables.NumOfDifferent() < lit.variables.NumOfDifferent()) { return GREATER; };
  if (variables.NumOfDifferent() > lit.variables.NumOfDifferent()) { return LESS; };
  
  // compare lexicographically
  COMPARISON var_cmp;
  const FLAT_TERM* fst = FlatTerm() + 1;
  const FLAT_TERM* const end_of_fst = FlatTerm()->NextTerm();
  const FLAT_TERM* snd = lit.FlatTerm();
  check_end:
   if (fst == end_of_fst) return INCOMPARABLE;
   snd++;
   if (fst->Symbol() != snd->Symbol())
    {
     if (fst->Symbol().IsVariable()) 
      {
       if (snd->Symbol().IsVariable()) 
        {
         // different variables
         var_cmp = var_weighting.Compare(fst->Symbol().Var(),snd->Symbol().Var());
         if (var_cmp != EQUAL) { return var_cmp; };
        }
       else return LESS; // a variable is less than a functor
      }
     else 
      {
       if (snd->Symbol().IsVariable()) return GREATER;
       // both are functors
       return fst->Symbol().CompareFunctor(snd->Symbol()); 
      }; 
    };
   fst++;
   goto check_end;
 }; // COMPARISON TMP_LITERAL::Compare(TMP_LITERAL& lit,const VAR_WEIGHTING& var_weighting)
 
//**************************************************************************** 
