#define __TMP_TERM_H

#ifndef __TERM_H
 #include "term.h"
#endif

#ifndef __CLAUSE_H
 #include "clause.h"
#endif

#ifndef __STACK_H
 #include "stack.h"
#endif

#ifndef __VARIABLES_H
 #include "variables.h"
#endif

#ifndef __WEIGHTING_H
 #include "weighting.h"
#endif

#ifndef __STATISTICS_H
 #include "statistics.h"
#endif

#ifndef __FLAT_TERM_H
 #include "flat_term.h"
#endif 

extern HEADER_WEIGHTING HeaderWeighting;


/******* debug *******/

#ifndef __CLAUSE_INDEX_H
 #include "clause_index.h"
#endif

extern CLAUSE_INDEX ClauseIndex; 

/*********************/


//**************************************************************************

class OPEN_TERM // for complex terms only!
 {
  private:
   FLAT_TERM::OPEN_WITH_MEMORY flat_term;
 
  private:
   TERM* term;
   TERM* end_of_term;
   long weight;
   long depth;
  private: // data neded only on the stage of collecting a new term
   TERM* curr_hole;
   TERM* curr_end;
   STACK<TERM*,MAX_TERM_DEPTH> holes; // this can be done static!
   STACK<TERM*,MAX_TERM_DEPTH> ends; // this can be done static!
   
   long max_weight;
   long max_depth;
   long curr_depth;
  public: 
   OPEN_TERM() 
    : flat_term(),
      
      holes("MAX_TERM_DEPTH in OPEN_TERM::holes"),
      ends("MAX_TERM_DEPTH in OPEN_TERM::ends"),
      max_weight(MAX_TERM_WEIGHT),
      max_depth(MAX_TERM_DEPTH)
    {};
   
   inline OPEN_TERM& operator=(OPEN_TERM& ot)
    {
     // applicable to already collected terms only!
     term = ot.term;
     end_of_term = ot.end_of_term;
     weight = ot.weight;
     depth = ot.depth;
     return *this;
    }; // OPEN_TERM& operator=(OPEN_TERM& ot)
    
   inline void Reset(TERM* const memory)
    {
     flat_term.Reset();
    
     holes.Reset();
     ends.Reset();
     term = memory;
     end_of_term = memory;
     
     curr_hole = (TERM*)(&term);
     curr_end = curr_hole - 1;
     
     weight = 1;
     depth = 0;
     curr_depth = 0;
    };
   
   inline void SetMaxWeight(const long mw) 
    { 
     if (mw < MAX_TERM_WEIGHT) { max_weight = mw; }
     else { max_weight = MAX_TERM_WEIGHT; };
    };
   inline void SetMaxDepth(const long md) 
    {
     if (md < MAX_TERM_DEPTH) { max_depth = md; }
     else { max_depth = MAX_TERM_DEPTH; };
    };
   
   inline void PushVar(const TERM v);
   inline bool PushFunc(const TERM fun) { return PushFunc(fun,fun.Arity()); };
   inline bool PushFunc(const TERM fun,const int arity);
   inline void MakeConst(const TERM fun,TERM* const memory) // good for pushing propositional literals
    {
     flat_term.Reset();
     flat_term.PushConst(fun);
    
     term = memory;
     *term = fun;
     end_of_term = memory+1;
     weight = 1;
     depth = 0;
    };
   
   inline TERM* Term() const { return term; };
   
   inline const FLAT_TERM* FlatTerm() const { return flat_term.Term(); }; 
   
   inline const TERM* RefToTerm() const { return (TERM*)(&term); };
   inline TERM* EndOfTerm() const { return end_of_term; };
   inline long Weight() const { return weight; };
   inline long Depth() const { return depth; };
 }; // class OPEN_TERM

//*************** Definitions for OPEN_TERM: *******************

inline void OPEN_TERM::PushVar(const TERM v)
 {
  flat_term.PushVar(v);
 
  *curr_hole = v;
  curr_hole = curr_hole->Next();
  check_hole:
   if (curr_hole == curr_end)
    {
     if (holes.Empty()) { return; };
     curr_depth--;
     curr_hole = holes.Pop();
     curr_end = ends.Pop();
     goto check_hole;
    };
 }; // bool OPEN_TERM::PushVar(const TERM v)

inline bool OPEN_TERM::PushFunc(const TERM fun,const int arity)
 {   
  weight += arity;
  if (weight > max_weight) 
   {
    STATISTICS_ON_GENERATED::AnotherTooHeavyClause();
    return false; 
   };
  *end_of_term = fun;
  curr_hole->SetFirst(end_of_term);
  
  if (arity)
   {
    flat_term.PushNonconstFun(fun);
   
   
    curr_depth++;
    if (curr_depth > max_depth) 
     {
      STATISTICS_ON_GENERATED::AnotherTooDeep(); 
      return false;
     };
    if (curr_depth > depth) { depth = curr_depth; };
    holes.Push(curr_hole->Next());
    ends.Push(curr_end);
    curr_hole = end_of_term->Args();
    curr_end = end_of_term;
    end_of_term += arity + 1;
   }
  else // constant 
   {
    flat_term.PushConst(fun);
   
   
    end_of_term++;
    curr_hole = curr_hole->Next();
    check_hole:
     if (curr_hole == curr_end)
      {
       if (holes.Empty()) { return true; };
       curr_depth--;
       curr_hole = holes.Pop();
       curr_end = ends.Pop();
       goto check_hole;
      };
   };
  
  return true;
 }; // bool OPEN_TERM::PushFunc(const TERM fun,const int arity)


//******************************************************************

class TMP_TERM
 {
  private:
   OPEN_TERM term;
   VAR_MULTISET variables;
  public:
   TMP_TERM() 
    : term(),
      variables()
    {};
   
   inline TMP_TERM& operator=(TMP_TERM& tt)
    {
     // applicable to already collected terms only!
     term = tt.term;
     variables = tt.variables;;
     return *this;
    }; // OPEN_TERM& operator=(OPEN_TERM& ot)
    
   inline void Reset(TERM* const memory)
    {
     term.Reset(memory);
     variables.Reset();
    };
   
   inline void SetMaxWeight(const long mw) { term.SetMaxWeight(mw); };
   inline void SetMaxDepth(const long md) { term.SetMaxDepth(md); };
   
   inline void PushVar(const int v)
    {
     variables.Register(v);
     term.PushVar(TERM(TERM_VAR,v));
    };
   inline void PushVar(const TERM v)
    {
     variables.Register(v.Var());
     term.PushVar(v);
    };
   inline bool PushFunc(const TERM fun) 
    {
     return term.PushFunc(fun); 
    };
   inline void MakeConst(const TERM fun,TERM* const memory)
    {
     term.MakeConst(fun,memory);
     variables.Reset();
    };
   
   inline TERM* Term() const { return term.Term(); };
   
   inline const FLAT_TERM* FlatTerm() const { return term.FlatTerm(); };
   
   inline const TERM* RefToTerm() const { return term.RefToTerm(); };
   inline TERM* EndOfTerm() const { return term.EndOfTerm(); };
   inline long Weight() const { return term.Weight(); };
   inline long Depth() const { return term.Depth(); }; 
   
   inline bool Equal(TMP_TERM& t);
   inline bool EqualConst(TMP_TERM& t)
    {
     return (*(Term()) == *(t.Term()));
    };
   inline bool Complementary(TMP_TERM& lit);
   inline bool ComplementaryConst(TMP_TERM& lit)
    {
     return ((Term()->Functor() == lit.Term()->Functor()) 
             && (Term()->Polarity() != lit.Term()->Polarity()));
    };
   
   inline VAR_MULTISET& Variables() { return variables; };
   inline bool Ground() const { return (!(variables.Size())); };
   
   inline COMPARISON KBCompare(TMP_TERM& t)
    {
     //COMPARISON hd_cmp = Term()->CompareHeader(*(t.Term()));
     
     COMPARISON hd_cmp = HeaderWeighting.Compare(*(Term()),*(t.Term()));
     return (hd_cmp == EQUAL) ? KBCompareArgs(t) : hd_cmp;
    };
    
   inline COMPARISON KBCompareArgs(TMP_TERM& t);
   
   inline COMPARISON Compare(TMP_TERM& t,const VAR_WEIGHTING& var_weighting);
   
   inline void RenameVariables(INT_VAR_RENAMING& var_renaming);
 }; // class TMP_TERM
 
//*********** Definitions for TMP_TERM: ***************************
 
inline void TMP_TERM::RenameVariables(INT_VAR_RENAMING& var_renaming)
 {
  if (Variables().Size())
   {
    TERM::ITERATOR iter;
    iter.Reset(Term());
    while (iter.NextVar())
     {
      iter.SymbolRef().RenameVar(var_renaming.Rename(iter.Symbol().Var()));
     };
   };
 }; // void TMP_TERM::RenameVariables(INT_VAR_RENAMING& var_renaming)
 
inline bool TMP_TERM::Equal(TMP_TERM& t)
 {
  if ((Weight() != t.Weight()) 
     || (Depth() != t.Depth()) 
     || (Variables().Size() != t.Variables().Size())
     || (Variables().NumOfDifferent() != t.Variables().NumOfDifferent()))
   {
    return false; 
   };
  
  TERM* fst = Term();
  TERM* snd = t.Term();
  compare:
   if (fst->IsReference())
    {
     if (!(snd->IsReference())) { return false; };
    }
   else
    {
     if (*fst != *snd) { return false; };
    };
   fst++;
   if (fst == EndOfTerm()) { return true; };
   snd++;
   goto compare;
 }; // bool TMP_TERM::Equal(TMP_TERM& t)  
 
inline bool TMP_TERM::Complementary(TMP_TERM& lit) 
 {
  if ((Term()->Functor() != lit.Term()->Functor()) 
     || (Term()->Polarity() == lit.Term()->Polarity())
     || (Weight() != lit.Weight()) 
     || (Depth() != lit.Depth()) 
     || (Variables().Size() != lit.Variables().Size())
     || (Variables().NumOfDifferent() != lit.Variables().NumOfDifferent()))
   {
    return false; 
   };
  TERM* fst = Term() + 1;
  TERM* snd = lit.Term() + 1;
   
  compare_next:
   if (fst == EndOfTerm()) { return true; };
   if (fst->IsReference())
    {
     if (!(snd->IsReference())) { return false; };
    }
   else
    {
     if (*fst != *snd) { return false; };
    };
   fst++;
   snd++;
   goto compare_next; 
 }; // bool TMP_TERM::Complementary(TMP_TERM& lit)  
 
inline COMPARISON TMP_TERM::KBCompareArgs(TMP_TERM& t) 
 {
  if (Weight() == t.Weight())
   {
    if (Variables().Equal(t.Variables()))
     {
      // compare lexicografically
      TERM::ITERATOR fst;
      TERM::ITERATOR snd;
      fst.Reset(Term());
      snd.Reset(t.Term());
      while (fst.Next())
       {
        snd.Next();
        if (fst.Symbol() != snd.Symbol())
         {
          if ((fst.CurrentSymIsVar()) || (snd.CurrentSymIsVar()))
           {
            return INCOMPARABLE;
           }
          else // both are functors
           {
            return fst.Symbol().CompareFunctor(snd.Symbol());
           };   
         };  
       };
      return EQUAL; // absolutely equal arguments
     }
    else { return INCOMPARABLE; };
   }
  else
   {
    if (Weight() > t.Weight())
     {
      if (Variables().GreaterOrEqual(t.Variables())) { return GREATER; }
      else { return INCOMPARABLE; };
     }
    else // Weight() < t.Weight()
     {
      if (t.Variables().GreaterOrEqual(Variables())) { return LESS; }
      else { return INCOMPARABLE; };
     };
   };
 }; // COMPARISON TMP_TERM::KBCompareArgs(TMP_TERM& t) 

inline COMPARISON TMP_TERM::Compare(TMP_TERM& t,const VAR_WEIGHTING& var_weighting)
 {
  //COMPARISON hd_cmp = Term()->CompareHeader(*(t.Term()));
  
  COMPARISON hd_cmp = HeaderWeighting.Compare(*(Term()),*(t.Term()));
  
  if (hd_cmp != EQUAL) { return hd_cmp; };
  
  // here the headers are equal
  if (Weight() > t.Weight()) { return GREATER; };
  if (Weight() < t.Weight()) { return LESS; };
  
  // here the weights are equal
  if (Variables().Size() < t.Variables().Size()) { return GREATER; };
  if (Variables().Size() > t.Variables().Size()) { return LESS; };
  
  // here both terms have same number of variable occurences
  if (Variables().NumOfDifferent() < t.Variables().NumOfDifferent()) { return GREATER; };
  if (Variables().NumOfDifferent() > t.Variables().NumOfDifferent()) { return LESS; };
  
  COMPARISON var_cmp;
  TERM::ITERATOR iter1;
  TERM::ITERATOR iter2;
  iter1.Reset(Term());
  iter2.Reset(t.Term());
  while (iter1.Next())
   {
    iter2.Next();
    if (iter1.Symbol() != iter2.Symbol())
     {
      if (iter1.CurrentSymIsVar())
       {
        if (iter2.CurrentSymIsVar())
         {
          // different variables
          var_cmp = var_weighting.Compare(iter1.Symbol().Var(),iter2.Symbol().Var());
          if (var_cmp != EQUAL) { return var_cmp; };
         }
        else
         {
          return LESS;
         };
       }
      else // first is not a variable
       {
        if (iter2.CurrentSymIsVar())
         {
          return GREATER;
         }
        else // both are not variables
         {
          return iter1.Symbol().CompareFunctor(iter2.Symbol());
         };
       };
     };
   };
  return INCOMPARABLE; 
 }; // COMPARISON TMP_TERM::Compare(TMP_TERM& t,const VAR_WEIGHTING& var_weighting)
 
//*********************************************************************

typedef STACK_WITH_ITERATOR<int,MAX_NUMBER_OF_LITERALS> LIT_NUM_STACK;
typedef STACK_WITH_ITERATOR<TERM*,MAX_NUMBER_OF_LITERALS> LIT_STACK;

const int MAX_NUMBER_OF_TMP_LITERALS = MAX_NUMBER_OF_LITERALS + 1;

const int MAX_NUMBER_OF_RENAMING_LITERALS = MAX_NUMBER_OF_TMP_LITERALS*2;
 
class TMP_TERMLIST
 {
  private:
   TERM memory[MAX_CLAUSE_WEIGHT*2 + MAX_TERM_WEIGHT*2 + MAX_NUMBER_OF_RENAMING_LITERALS];
   TERM* free_memory;
   TMP_TERM term[MAX_NUMBER_OF_TMP_LITERALS + MAX_NUMBER_OF_RENAMING_LITERALS];
   int next_trm;
   int next_renaming_lit;
   long weight;
   INT_VAR_RENAMING var_renaming;
  private: // linked components
   bool component_assigned[MAX_NUMBER_OF_LITERALS];
   LIT_NUM_STACK ground_literals;
   LIT_NUM_STACK nonground_literals[MAX_NUMBER_OF_LITERALS];
   int next_component;
   VAR_MULTISET component_vars;
   bool was_tautology;
  private:
   long max_weight;
   long max_term_weight;
   long max_depth;
   long max_length;
   bool bin_res_negative_selection;
  public:
   TMP_TERMLIST() 
    : var_renaming(),
    
      ground_literals("MAX_NUMBER_OF_LITERALS in TMP_TERMLIST::ground_literals"),
      next_component(0),
      component_vars(),
     
      max_weight(MAX_CLAUSE_WEIGHT),
      max_term_weight(MAX_TERM_WEIGHT),
      max_depth(MAX_TERM_DEPTH),
      max_length(MAX_NUMBER_OF_LITERALS)
    {
     if (max_term_weight > max_weight) { max_term_weight = max_weight; };
     for (int trm = 0; trm < MAX_NUMBER_OF_TMP_LITERALS; trm++)
      {
       term[trm].SetMaxWeight(max_term_weight);
       term[trm].SetMaxDepth(max_depth);
      };
     ground_literals.Init();
     for (int comp = 0; comp < MAX_NUMBER_OF_LITERALS; comp++)
      {
       nonground_literals[comp].Init();
      };
    };
   inline void Reset()
    {
     free_memory = memory;
     next_trm = 0;
     next_renaming_lit = MAX_NUMBER_OF_TMP_LITERALS;
     
     weight = 0L;
     was_tautology = false;
    };
   
   inline long MaxWeight() const { return max_weight; }; 
   
   inline void SetMaxWeight(const long mw) 
    {
     if (mw < MAX_CLAUSE_WEIGHT) { max_weight = mw; }
     else { max_weight = MAX_CLAUSE_WEIGHT; };
     if (max_weight < max_term_weight) 
      { 
       for (int trm = 0; trm < MAX_NUMBER_OF_TMP_LITERALS; trm++)
        {
         term[trm].SetMaxWeight(max_weight);
        };
      }
     else 
      {
       if (max_weight > max_term_weight)
        {
         for (int trm = 0; trm < MAX_NUMBER_OF_TMP_LITERALS; trm++)
          {
           term[trm].SetMaxWeight(max_term_weight);
          };
        };
      };
    };
   inline void SetMaxTermWeight(const long mw) 
    {
     max_term_weight = mw;
     if (max_term_weight < max_weight)
      {
       for (int trm = 0; trm < MAX_NUMBER_OF_TMP_LITERALS; trm++)
        {
         term[trm].SetMaxWeight(max_term_weight);
        };
      }
     else
      {
       for (int trm = 0; trm < MAX_NUMBER_OF_TMP_LITERALS; trm++)
        {
         term[trm].SetMaxWeight(max_weight);
        };
      };
    };
   inline void SetMaxDepth(const long md)
    {
     max_depth = md;
     for (int trm = 0; trm < MAX_NUMBER_OF_TMP_LITERALS; trm++)
      {
       term[trm].SetMaxDepth(max_depth);
      };
    };
   inline void SetMaxLength(const long ml) 
    {
     if (ml < MAX_NUMBER_OF_LITERALS)
      {
       max_length = ml;
      }
     else { max_length = MAX_NUMBER_OF_LITERALS; };
    };
    
   inline void SetNegativeSelection(bool fl) { bin_res_negative_selection = fl; };
    
   inline long Length() const { return next_trm; };
   inline bool Empty() const { return (!(Length())); };
   inline long Weight() const { return weight; };
   inline long Weight(LIT_NUM_STACK& literals) const
    {
     literals.ResetIterator();
     long result = 0L;
     int* lit = literals.Next();
     while (lit) { result += term[*lit].Weight(); lit = literals.Next(); };
     return result;
    };
   
   inline TERM* Term(const int trm) const { return term[trm].Term(); };
   inline const TERM* RefToTerm(const int trm) const { return term[trm].RefToTerm(); };
   inline const FLAT_TERM* FlatTerm(int n) const { return term[n].FlatTerm(); };
   
   
   inline void OpenTerm() 
    {
     term[next_trm].Reset(free_memory);
    };
    
   inline bool CloseTerm()
    {
     // first check for identical or complementary literals
     for (int trm = 0; trm < next_trm; trm++)
      {
       if (term[next_trm].Equal(term[trm])) 
        {
         return true; 
        };
       if (term[next_trm].Complementary(term[trm]))
        {
         STATISTICS_ON_GENERATED::AnotherTautology();
         was_tautology = true;
         return false;
        };
      };
     
     weight += term[next_trm].Weight();
     if (weight > max_weight) 
      {
       STATISTICS_ON_GENERATED::AnotherTooHeavyClause();
       return false;
      };
     if (next_trm >= max_length) 
      {
       STATISTICS_ON_GENERATED::AnotherTooLongClause();
       return false; 
      }; 
     free_memory = term[next_trm].EndOfTerm();
     
     
     /***** debug *****/
      
      //cout << "SHARED "; term[next_trm].FlatTerm()->OutputTerm();
      //TERM* shared = ClauseIndex.Integrate(term[next_trm].FlatTerm());
      //cout << " = "; ::Output(shared); cout << "\n";
      
     
     /*****************/
     
     next_trm++;
     return true;
    };
   
   inline bool PushConst(const TERM fun)
    {
     term[next_trm].MakeConst(fun,free_memory);
     
     // first check for identical or complementary literals
     for (int trm = 0; trm < next_trm; trm++)
      {
       if (term[next_trm].EqualConst(term[trm])) 
        {
         return true; 
        };
       if (term[next_trm].ComplementaryConst(term[trm]))
        {
         STATISTICS_ON_GENERATED::AnotherTautology();
         was_tautology = true;
         return false;
        };
      };
     weight++;
     if (weight > max_weight) 
      {
       STATISTICS_ON_GENERATED::AnotherTooHeavyClause();
       return false;
      };
     if (next_trm >= max_length) 
      {
       STATISTICS_ON_GENERATED::AnotherTooLongClause();
       return false; 
      }; 
     free_memory = term[next_trm].EndOfTerm();
     
     next_trm++;
     return true;
    }; // bool PushConst(const TERM fun) 
   
   inline void RemoveTerm(const int n)
    {
     weight -= term[n].Weight();
     next_trm--;
     if (n < next_trm)
      {
       // copy the term next_trm into the n-th position
       term[n] = term[next_trm];
      };    
    }; // void RemoveTerm(const int n)
    
   inline int PushRenamingLit(const TERM prop_lit)
    {
     term[next_renaming_lit].Reset(free_memory);
     term[next_renaming_lit].PushFunc(prop_lit);
     free_memory = term[next_renaming_lit].EndOfTerm();
     next_renaming_lit++;
     return (next_renaming_lit - 1);
    };
   
   inline bool WasTautology() const { return was_tautology; };
   
   inline void PushVar(const int v) 
    {
     term[next_trm].PushVar(v); 
    };
   inline bool PushFunc(const TERM fun)
    {
     return term[next_trm].PushFunc(fun);
    };
    
   inline void KBOrder(LIT_NUM_STACK& literals,
                       LIT_NUM_STACK& max_literals,
                       LIT_NUM_STACK& nonmax_literals); 
  private:
   inline void Order(LIT_NUM_STACK& literals,LIT_NUM_STACK& ordered,const VAR_WEIGHTING& var_weighting);
  public: 
   inline void Order(LIT_NUM_STACK& literals,const VAR_WEIGHTING& var_weighting)
    {
     if (literals.NumberOfElements() > 1)
      {
       Order(literals,literals,var_weighting);
      };
    };
  
   inline void CreateVarWeighting(LIT_NUM_STACK& literals,VAR_WEIGHTING& var_weighting);
   
   inline void ResetVarRenaming() { var_renaming.Reset(); };
   inline void RenameVariables(LIT_NUM_STACK& literals);
   inline void RenameVariables();
   
  public:
   inline void PushAllLitNums(LIT_NUM_STACK& all_literals,bool& ground)
    {
     ground = true;
     int len = Length();
     for (int i = 0; i < len; i++) 
      {
       ground = (ground && (term[i].Ground()));
       all_literals.Push(i);
      };
    };
    
  public:
   inline void FindLinkedComponents();
   inline int NumOfNongroundComponents() const { return next_component; };
   inline LIT_NUM_STACK& GroundLiterals() { return ground_literals; };
   inline LIT_NUM_STACK& NongroundComponent(const int n) 
    { return nonground_literals[n]; };
  public:
   inline void Output() const;
   inline void Output(LIT_NUM_STACK& lits) const;
   inline void OutputComponents();
 }; // class TMP_TERMLIST
 
//******************* Definitions for TMP_TERMLIST: ***************

inline void TMP_TERMLIST::KBOrder(LIT_NUM_STACK& literals,
                                  LIT_NUM_STACK& max_literals,
                                  LIT_NUM_STACK& nonmax_literals)
 {
  // the stack literals must be nonempty here,
  
  //max_literals.Reset();
  //nonmax_literals.Reset();
  
  if (literals.NumberOfElements() == 1) // nothing to order
   {
    max_literals.Push(literals.Top());
    return;
   };
   
  if (bin_res_negative_selection)
   {
    long max_neg_lit_weight = 0;
    int max_neg_lit_num = -1;
    literals.ResetIterator();
    for (int* neg_lit_num = literals.Next(); neg_lit_num; neg_lit_num = literals.Next())
     { 
      if (Term(*neg_lit_num)->Polarity() == POL_NEG)
       {
        if (term[*neg_lit_num].Weight() > max_neg_lit_weight)
         {
          max_neg_lit_num = *neg_lit_num;
          max_neg_lit_weight = term[max_neg_lit_num].Weight();
         };
       };
     }; 
    if (max_neg_lit_weight) // some negative literal found
     {
      max_literals.Push(max_neg_lit_num);
      literals.ResetIterator();
      for (int* lit_num = literals.Next(); lit_num; lit_num = literals.Next())
       {
        if ((*lit_num) != max_neg_lit_num)
         {
          nonmax_literals.Push(*lit_num);
         };
       }; 
      return; 
     };  
   }; 
   
  // first, find the greatest header 
  literals.ResetIterator();
  TERM max_header = *(Term(*(literals.Next())));
  int* next_lit_num = literals.Next();
  do
   {
    //if (Term(*next_lit_num)->CompareHeader(max_header) == GREATER)
    if (HeaderWeighting.Compare(*(Term(*next_lit_num)),max_header) == GREATER)
     {
      max_header = *(Term(*next_lit_num));
     };
    next_lit_num = literals.Next();
   }
  while (next_lit_num);
  
  if (!(max_header.Arity()))  
   {
    literals.ResetIterator();
    int* next_lit_num = literals.Next();
    do
     {
      if (*(Term(*next_lit_num)) == max_header)
       {
        max_literals.Push(*next_lit_num);
       }
      else
       {
        nonmax_literals.Push(*next_lit_num);
       };
      next_lit_num = literals.Next();
     }
    while (next_lit_num);
    return;
   };
 
  literals.ResetIterator();
  int* lit_num = literals.Next();
  do
   {
    if (*(Term(*lit_num)) == max_header) // can be maximal
     {
      max_literals.ResetBackIterator();
      next_lit_num = max_literals.Previous();
      while (next_lit_num)
       { 
        switch (term[*next_lit_num].KBCompareArgs(term[*lit_num]))
         {
          case GREATER: nonmax_literals.Push(*lit_num); goto next_candidate;
          case LESS: 
           // all literals from max_literals that are LESS than term[*lit_num]
           // must be collected in nonmax_literals
           nonmax_literals.Push(*next_lit_num);
           if (next_lit_num != &(max_literals.Top()))
            {
             *next_lit_num = max_literals.Pop();
            }
           else { max_literals.Pop(); };

           next_lit_num = max_literals.Previous();
           
           while (next_lit_num)
            {
             if (term[*next_lit_num].KBCompareArgs(term[*lit_num]) == LESS)
              {
               nonmax_literals.Push(*next_lit_num);
               if (next_lit_num != &(max_literals.Top()))
                {
                 *next_lit_num = max_literals.Pop();
                }
               else { max_literals.Pop(); };
              };
             next_lit_num = max_literals.Previous();
            };

           max_literals.Push(*lit_num);
           goto next_candidate;
          case INCOMPARABLE: break;
          case EQUAL: break;
         };
        next_lit_num = max_literals.Previous();
       };
      max_literals.Push(*lit_num);
     }
    else { nonmax_literals.Push(*lit_num); };
   
    next_candidate:
     lit_num = literals.Next();
   }
  while (lit_num);

 }; // void TMP_TERMLIST::KBOrder(LIT_NUM_STACK& literals,....                                      

inline void TMP_TERMLIST::Order(LIT_NUM_STACK& literals,LIT_NUM_STACK& ordered,const VAR_WEIGHTING& var_weighting)
 {
  // the stack literals can't contain less than 2 literal numbers here.
  int center = literals.Pop();
  LIT_NUM_STACK smaller("MAX_NUMBER_OF_LITERALS in TMP_TERMLIST::Order():smaller");
  LIT_NUM_STACK bigger("MAX_NUMBER_OF_LITERALS in TMP_TERMLIST::Order():bigger");

  int next_lit;
  do
   {
    next_lit = literals.Pop();
    switch (term[next_lit].Compare(term[center],var_weighting))
     {
      case GREATER: bigger.Push(next_lit); break;
      case LESS: smaller.Push(next_lit); break;
      case INCOMPARABLE: // same as for EQUAL
      case EQUAL: smaller.Push(next_lit); break;
     };   
   }
  while (!(literals.Empty()));

  int n;
  if ((n = smaller.NumberOfElements()))
   {
    if (n > 1) 
     {
      Order(smaller,ordered,var_weighting);
     }
    else // 1 element
     {
      ordered.Push(smaller.Top());
     };
   };
  
  ordered.Push(center);
  
  if ((n = bigger.NumberOfElements()))
   {
    if (n > 1)
     {
      Order(bigger,ordered,var_weighting);
     }
    else // 1 element
     {
      ordered.Push(bigger.Top());
     };
   };
 }; // void TMP_TERMLIST::Order(LIT_NUM_STACK& literals,LIT_NUM_STACK& ordered,const VAR_WEIGHTING var_weighting)

inline void TMP_TERMLIST::CreateVarWeighting(LIT_NUM_STACK& literals,VAR_WEIGHTING& var_weighting)
 {
  var_weighting.Reset();
  literals.ResetIterator();
  int* next_lit = literals.Next();
  while (next_lit)
   {
    var_weighting.AnotherTermVariables(term[*next_lit].Variables());
    next_lit = literals.Next();
   };
 }; // void TMP_TERMLIST::CreateVarWeighting(LIT_NUM_STACK& literals,VAR_WEIGHTING& var_weighting)

inline void TMP_TERMLIST::FindLinkedComponents()
 {
  ground_literals.Reset();
  next_component = 0; 
  nonground_literals[0].Reset();
   
  for (int i = 0; i < Length(); i++) { component_assigned[i] = false; };
  
  for (int lit1 = 0; lit1 < Length(); lit1++)
   {
    if (!(component_assigned[lit1]))
     {
      component_assigned[lit1] = true;
      if (term[lit1].Variables().Size()) // nonground literal
       {
        nonground_literals[next_component].Push(lit1);
        component_vars = term[lit1].Variables();
         
        next_iteration: 
         bool another_literal = false;
         for (int lit2 = lit1 + 1; lit2 < Length(); lit2++)
          {
           if ((!(component_assigned[lit2])) 
               && (term[lit2].Variables().IntersectionNonempty(component_vars)))
            {
             another_literal = true;
             component_assigned[lit2] = true;
             nonground_literals[next_component].Push(lit2);
             component_vars.Add(term[lit2].Variables());
            };
          };
         if (another_literal) { goto next_iteration; }; 
         next_component++;
         nonground_literals[next_component].Reset();
       }
      else
       {
        ground_literals.Push(lit1);
       };
     };
   };
 }; // void TMP_TERMLIST::FindLinkedComponents()

inline void TMP_TERMLIST::RenameVariables(LIT_NUM_STACK& literals)
 {
  literals.ResetBackIterator();
  int* lit = literals.Previous();
  while (lit)
   {
    term[*lit].RenameVariables(var_renaming);
    lit = literals.Previous();
   };
 }; // void TMP_TERMLIST::RenameVariables(LIT_NUM_STACK& literals)

inline void TMP_TERMLIST::RenameVariables()
 {
  for (int lit = 0; lit < Length(); lit++)
   {
    term[lit].RenameVariables(var_renaming);
   };
 }; // void TMP_TERMLIST::RenameVariables()

inline void TMP_TERMLIST::Output() const
 {
  if (Empty())
   {
    ::Output("# ");
    ::Output(Weight());
    return;
   };
  for (int trm = 0; trm < next_trm; trm++)
   {
    ::Output(Term(trm));
    if (trm + 1 < next_trm)
     {
      ::Output(" \\/ ");
     };
   }; 
  ::Output(" "); 
  ::Output(Weight());   
 }; // void TMP_TERMLIST::Output() const
    
inline void TMP_TERMLIST::Output(LIT_NUM_STACK& lits) const
 {
  lits.ResetBackIterator();
  ::Output("{");
  for (int* lit = lits.Previous();lit;lit = lits.Previous()) 
   {
    ::Output(Term(*lit)); ::Output(" ");
   };
  ::Output("}");
 }; // void TMP_TERMLIST::Output(LIT_NUM_STACK& lits) const
    
inline void TMP_TERMLIST::OutputComponents()
 {
  Output(ground_literals);
  int comp = 0;
  while (comp < next_component)
   {
    ::Output(" + ");
    Output(nonground_literals[comp]);
    comp++;
   };
 }; // void TMP_TERMLIST::OutputComponents()
 
//********************************************************************* 













