#define __UNIFICATION_H
#ifndef __DEBUG_H
 #include "debug.h"
#endif 
#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif
#ifndef __IU_COMMAND_H
 #include "iu_command.h"
#endif

//**************************************************************************

class UNIFICATION
 {
  public: 
   static inline bool Unify(PREFIX_SYM* t1,PREFIX_SYM* t2)
    {
     // no bound variables in t1,t2!
     if (t1->Head().IsVariable())
      {
       CHECK1(!(t1->Head().Var()->Binding()),"A10 in UNIFICATION::Unify(PREFIX_SYM* t1,PREFIX_SYM* t2)");
       
       if (t2->Head().IsVariable())
        {
         // both are variables
         CHECK1(!(t2->Head().Var()->Binding()),"A20 in UNIFICATION::Unify(PREFIX_SYM* t1,PREFIX_SYM* t2)");
         GlobalSubst.MarkSection();
         if (t1->Head() != t2->Head())
          GlobalSubst.Bind(t1->Head().Var(),(void*)(t2->Head().Var()));
         return true;
        };
        
       // t1 is a variable, t2 is complex 
       if (OCCUR_CHECK::Occurs(t1->Head().Var(),t2)) return false;
       GlobalSubst.MarkSection();
       GlobalSubst.Bind(t1->Head().Var(),(void*)t2);
       return true;
      }
     else
      {
       if (t2->Head().IsVariable())
        {
         CHECK1(!(t2->Head().Var()->Binding()),"A30 in UNIFICATION::Unify(PREFIX_SYM* t1,PREFIX_SYM* t2)");
         
         // t2 is a variable, t1 is complex 
         if (OCCUR_CHECK::Occurs(t2->Head().Var(),t1)) return false;
         GlobalSubst.MarkSection();
         GlobalSubst.Bind(t2->Head().Var(),(void*)t1);
         return true; 
        }
       else // both complex
        return UnifyComplex(t1,t2);
      };
    }; // bool Unify(PREFIX_SYM* t1,PREFIX_SYM* t2)
       
   static inline bool UnifyComplex(PREFIX_SYM* const word1,PREFIX_SYM* const word2)
    {
     DCONTROL1("Z1 in UNIFICATION::UnifyComplex(..)");
    
     // word1 and word2 must represent complex terms
     GlobalSubst.MarkSection();
     if (UnifyComplex_(word1,word2)) { return true; }
     else { GlobalSubst.UnbindLastSection(); return false; };
    }; // bool UnifyComplex(PREFIX_SYM* const word1,PREFIX_SYM* const word2)
    
   static inline bool UnifyComplex_(PREFIX_SYM* const word1,PREFIX_SYM* const word2)
    // word1 and word2 must represent complex terms
    {
     
     //static long call_n = 0; // debug
     //call_n++; // debug
     //cout << call_n << "\n"; // debug

    // if (call_n == 2612)
    //  { 
    //   PREFIX::OutputWord(word1); cout << "\n"; // debug
     // };


     CHECK1(word1,"null argument 1 in UNIFICATION::UnifyComplex_()");
     CHECK1(word2,"null argument 2 in UNIFICATION::UnifyComplex_()");
     CHECK1(word1->Head().IsFunctor(),"strange argument 1 in UNIFICATION::UnifyComplex_()");
     CHECK1(word2->Head().IsFunctor(),"strange argument 2 in UNIFICATION::UnifyComplex_()");
     CHECK1(word1->Bracket(),"null bracket 1 in UNIFICATION::UnifyComplex_()");
     CHECK1(word2->Bracket(),"null bracket 2 in UNIFICATION::UnifyComplex_()");
     CHECK1((word1->Bracket() > word1) && (word1->Bracket() - word1 <= MAX_TMP_TERM_WEIGHT),
            "strange bracket 1 in UNIFICATION::UnifyComplex_()");
     CHECK1((word2->Bracket() > word2) && (word2->Bracket() - word2 <= MAX_TMP_TERM_WEIGHT),
            "strange bracket 2 in UNIFICATION::UnifyComplex_()");       
     
     CHECK2(PREFIX::CorrectComplexTerm(word1),"A1 in UNIFICATION::UnifyComplex_()"); 
     CHECK2(PREFIX::CorrectComplexTerm(word2),"A1 in UNIFICATION::UnifyComplex_()"); 
     
     if (word1->Head() != word2->Head()) { return false; };

     PREFIX::ITER iter1;
     PREFIX::ITER iter2;
     iter1.Reset(word1);
     iter2.Reset(word2);
     iter1.SkipSym();
     iter2.SkipSym();
     
     next_sym:
      if (iter1.Empty())
       {
        CHECK0(iter2.Empty(),"nonempty <iter2> in UNIFICATION::UnifyComplex_()"); 
        CHECK2(PREFIX::Equal(word1,word2),"wrong result in UNIFICATION::UnifyComplex_()"); 
        return true;
       }     
      else
       {
        if (iter1.CurrVar())
         {
          if (iter2.CurrVar()) // both are variables
           {
            if (iter1.CurrVar() != iter2.CurrVar()) // different variables
             {
              GlobalSubst.Bind(iter1.CurrVar(),(void*)iter2.CurrVar());
             };
            iter1.SkipSym();
            iter2.SkipSym();
            goto next_sym;
           }
          else
           {
            PREFIX::CorrectBracket(iter2.CurrSubterm());
            
            CHECK2(PREFIX::CorrectComplexTerm(iter2.CurrSubterm()),"B1 in UNIFICATION::UnifyComplex_()"); 
            
            if (OCCUR_CHECK::Occurs(iter1.CurrVar(),iter2.CurrSubterm()))
             {
              return false;
             }
            else
             {
              GlobalSubst.Bind(iter1.CurrVar(),(void*)iter2.CurrSubterm());
              iter1.SkipSym();
              iter2.SkipTerm();
              goto next_sym;
             };
           };
         }
        else // first is a complex term
         {
          if (iter2.CurrVar())
           {
            PREFIX::CorrectBracket(iter1.CurrSubterm());
            
            CHECK2(PREFIX::CorrectComplexTerm(iter1.CurrSubterm()),"B2 in UNIFICATION::UnifyComplex_()"); 
            
            if (OCCUR_CHECK::Occurs(iter2.CurrVar(),iter1.CurrSubterm()))
             {
              return false;
             }
            else
             {
              GlobalSubst.Bind(iter2.CurrVar(),(void*)iter1.CurrSubterm());
              iter1.SkipTerm();
              iter2.SkipSym();
              goto next_sym;
             };
           }
          else // both are nonvariable
           {
            if (iter1.CurrSubterm()->Head() == iter2.CurrSubterm()->Head())
             {
              iter1.SkipSym();
              iter2.SkipSym();
              goto next_sym;
             }
            else
             {
              return false;
             };
           };
         };
       };
    }; // bool UnifyComplex_(PREFIX_SYM* word1,PREFIX_SYM* word2)
   
   static inline bool UnifyPlain(PREFIX_SYM* const word,PREFIX_SYM* const pl_word)
    {
     // word must represent a complex term,
     // pl_word must represent a "plain" complex term,
     // no double variable occurences in pl_word
     // We also assume that all brackets are set inside pl_word.
     GlobalSubst.MarkSection();
     if (UnifyPlain_(word,pl_word)) { return true; }
     else { GlobalSubst.UnbindLastSection(); return false; };
     
    }; // bool UnifyPlain(PREFIX_SYM* const word,PREFIX_SYM* const pl_word)
   
   static inline bool UnifyPlain_(PREFIX_SYM* const word,PREFIX_SYM* const pl_word)
    // word must represent a complex term,
    // pl_word must represent a "plain" complex term,
    // We also assume that all brackets are set inside pl_word.
    {
    
     if (word->Head() != pl_word->Head()) { return false; };

     PREFIX::ITER iter1;
     PREFIX::SIMPLE_ITER iter2;
     iter1.Reset(word);
     iter2.Reset(pl_word);
     iter1.SkipSym();
     iter2.SkipSym();
     check_iter:
      if (iter2.Empty()) { return true; };
      if (iter1.CurrVar()) // free variable
       {
        if (iter2.CurrPos()->Head().IsVariable())
         {
          GlobalSubst.Bind(iter1.CurrVar(),(void*)(iter2.CurrPos()->Head().Var()));
          iter2.SkipSym();
         }
        else
         {
          GlobalSubst.Bind(iter1.CurrVar(),(void*)(iter2.CurrPos()));
          iter2.SkipTerm();
         };
        iter1.SkipSym();
        goto check_iter;
       }
      else // first is a complex term
       {
        if (iter2.CurrPos()->Head().IsVariable())
         {
          PREFIX::CorrectBracket(iter1.CurrSubterm());
          GlobalSubst.Bind(iter2.CurrPos()->Head().Var(),iter1.CurrSubterm());
          iter1.SkipTerm();
          iter2.SkipSym();
          goto check_iter;
         }
        else // both are complex, compare headers
         {
          if (iter1.CurrSubterm()->Head() != iter2.CurrPos()->Head())
           {
            return false;
           };
          iter1.SkipSym();
          iter2.SkipSym();
          goto check_iter;
         };
       };
    }; // bool UnifyPlain_(PREFIX_SYM* const word,PREFIX_SYM* const pl_word)
    
   //*********************************************************************
   static inline bool MatchGroundPlain(PREFIX_SYM* const word,PREFIX_SYM* const gp_word)
    // word must represent a complex term,
    // gp_word must represent a ground and "plain" term,
    // "plain" means that there are no bound variables in the term.
    // We also assume that all brackets are set inside gp_word.
    {
     GlobalSubst.MarkSection();
     if (MatchGroundPlain_(word,gp_word)) { return true; }
     else { GlobalSubst.UnbindLastSection(); return false; };
    }; // bool MatchGroundPlain(PREFIX_SYM* const word,PREFIX_SYM* const gp_word)
   
   static inline bool MatchGroundPlain_(PREFIX_SYM* const word,PREFIX_SYM* const gp_word)
    // word must represent a complex term,
    // gp_word must represent a ground and "plain" term,
    // "plain" means that there are no bound variables in the term.
    // We also assume that all brackets are set inside gp_word.
    {
     if (word->Head() != gp_word->Head()) { return false; };

     VAR* var;
     PREFIX::ITER iter1;
     PREFIX::SIMPLE_ITER iter2;
     iter1.Reset(word);
     iter2.Reset(gp_word);
     iter1.SkipSym();
     iter2.SkipSym();
     next_sym:
      if (iter1.Empty())
       {
        return true;
       }
      else
       {
        var = iter1.CurrVar();
        if (var)
         {
          GlobalSubst.Bind(var,(void*)iter2.CurrPos());
          iter1.SkipSym();
          iter2.SkipTerm();
          goto next_sym;
         }
        else // complex
         {
          if (iter1.CurrSubterm()->Head() == iter2.CurrPos()->Head())
           {
            iter1.SkipSym();
            iter2.SkipSym();
            goto next_sym;
           }
          else
           {
            return false;
           };
         };
       }; 
    }; // bool MatchGroundPlain_(PREFIX_SYM* word,PREFIX_SYM* gp_word)
 }; // class UNIFICATION

//*********************************************************************

 
//***************** Compiled unification: *******************************

inline bool Unify(IU_COMMAND* code,PREFIX_SYM* word)
 {
 
  //cout << "\n-\n"; PREFIX::OutputWord(word); cout << "\n+\n"; // debug
  
  //cout << "********************\n"; // debug
  //code->OutputCode(); // debug
  //cout << "********************\n"; // debug
 
 
  DCONTROL1("Z1 in Unify(IU_COMMAND* const code,PREFIX_SYM* const word)");
 
  // all brackets in word must be set correctly 
  
  //return UNIFICATION::Unify(code->Term(),word); // debug
  
  if ((code->Symbol() != word->Head()) // different headers
      || ((!(code->AllVars())) // the code represents a ground term
           && (PREFIX::Weight(code->Term()) < PREFIX::Weight(word))))
   {
    return false; 
   };  
  
  void* bind;
  bool free_var;
  
  GlobalSubst.MarkSection(); 
   
  IU_COMMAND* current_command = code + 1;
  PREFIX::ITER iter;
  iter.Reset(word);
  iter.SkipSym(); // skipping the header
  
  perform_command:

   //DCONTROL0("X1");
   
   //current_command->Output(code); cout << "\n"; // debug
   
   switch (current_command->Tag())
    {
     case IU_COMMAND::START: break; // error
     case IU_COMMAND::VAR: 
      bind = current_command->Symbol().Var()->Unref(free_var);
      if (free_var)
       {
        if (iter.CurrVar()) // both are free variables
         {
          if ((VAR*)bind != iter.CurrVar())
           {
            GlobalSubst.Bind(((VAR*)bind),(void*)(iter.CurrVar()));       
           };
          iter.SkipSym();
         }
        else // iter.CurrSubterm() is a complex term
         {
          if (OCCUR_CHECK::Occurs((VAR*)bind,iter.CurrSubterm()))
           {
            // cycle
            goto failure;
           }
          else
           {
            GlobalSubst.Bind(((VAR*)bind),(void*)iter.CurrSubterm());
            iter.SkipTerm();
           };
         };
       }
      else // bind is a complex term
       {
        if (iter.CurrVar()) 
         {
          if (OCCUR_CHECK::Occurs(iter.CurrVar(),(PREFIX_SYM*)bind))
           {
            // cycle
            goto failure;
           };
          GlobalSubst.Bind(iter.CurrVar(),bind);
          iter.SkipSym();
         }
        else // both terms are complex, unify them
         {
          if (UNIFICATION::UnifyComplex_(iter.CurrSubterm(),(PREFIX_SYM*)bind))
           {
            iter.SkipTerm();
           }
          else { goto failure; };
         };
       };
      current_command++;
      goto perform_command;

     case IU_COMMAND::VAR_FIRST_OCC:
      if (iter.CurrVar()) // both are free variables
       {
        GlobalSubst.Bind(current_command->Symbol().Var(),(void*)(iter.CurrVar()));
        iter.SkipSym();
       }
      else // iter.CurrSubterm() is a complex term
       {
        GlobalSubst.Bind(current_command->Symbol().Var(),(void*)(iter.CurrSubterm()));
        iter.SkipTerm();
       };
      current_command++;
      goto perform_command;
      
     case IU_COMMAND::FUNC:
      if (iter.CurrVar())
       {
        if (OCCUR_CHECK::Occurs(iter.CurrVar(),current_command->Term()))
         {
          // cycle
          goto failure;
         };
        GlobalSubst.Bind(iter.CurrVar(),(void*)(current_command->Term()));
        current_command = current_command->NextCommand();
       }
      else // iter.CurrSubterm() is a complex term, compare headers
       {
        if (current_command->Symbol() != iter.CurrSubterm()->Head())
         {
          // clash of functors
          goto failure;
         };
        current_command++; 
       };
      iter.SkipSym();
      goto perform_command; 

     case IU_COMMAND::FUNC_PLAIN:
      if (iter.CurrVar())
       {
        GlobalSubst.Bind(iter.CurrVar(),(void*)(current_command->Term()));
        current_command = current_command->NextCommand();
       }
      else
       {
        if (current_command->Symbol() != iter.CurrSubterm()->Head())
         {
          // clash of functors
          goto failure;
         };
        current_command++;
       };
      iter.SkipSym();
      goto perform_command;
      
     case IU_COMMAND::FUNC_GROUND:
      
      //DCONTROL0("FUNC_GROUND"); // debug
     
      if (iter.CurrVar())
       {
        GlobalSubst.Bind(iter.CurrVar(),(void*)(current_command->Term()));
        current_command = current_command->NextCommand();
       }
      else
       {
        if ((PREFIX::Weight(current_command->Term()) < PREFIX::Weight(iter.CurrSubterm()))
            || (current_command->Symbol() != iter.CurrSubterm()->Head()))
         {
          goto failure;
         };
        current_command++;
       };
      iter.SkipSym(); 
      goto perform_command;
      
     case IU_COMMAND::CONST:
     
      //DCONTROL0("CONST"); // debug
     
      if (iter.CurrVar())
       {
        GlobalSubst.Bind(iter.CurrVar(),(void*)(current_command->Term()));
       }
      else 
       {
        if (current_command->Symbol() != iter.CurrSubterm()->Head())
         {
          // clash of functors
          goto failure;
         };
       };
      
      iter.SkipSym(); 
      current_command++; 
      goto perform_command;
       
     case IU_COMMAND::END: 
      
      //DCONTROL0("END"); // debug
      
      return true;
     
    };

  RuntimeError(" in Unify(IU_COMMAND* const code,PREFIX_SYM* const word)");
  
  failure:
   GlobalSubst.UnbindLastSection();
   return false;
 }; // bool Unify(IU_COMMAND* code,PREFIX_SYM* word)

//******************************************************************************

 
