
#include "vampire.h"

#ifndef __SIGNATURE_H
 #include "signature.h"
#endif
#ifndef __CLAUSE_SET_H
 #include "clause_set.h"
#endif
#ifndef __KEPT_H
 #include "kept.h"
#endif
#ifndef __ANCESTORS_H
 #include "ancestors.h"
#endif
#ifndef __CLAUSE_H
 #include "clause.h"
#endif
#ifndef __OPTIONS_H
 #include "options.h"
#endif
#ifndef __SHARING_H
 #include "sharing.h"
#endif
#ifndef __CLAUSE_INDEX_H
 #include "clause_index.h"
#endif
#ifndef __OUTPUT_H
 #include "output.h"
#endif
#ifndef __READCMD_H
 #include "readcmd.h"
#endif

#include <signal.h>
//#include "signal.h" 

#ifndef __TIMER_H
 #include "timer.h"
#endif
#ifndef __STATISTICS_H
 #include "statistics.h"
#endif
#ifndef __VARIABLES_H
 #include "variables.h"
#endif
#ifndef __PREFIX_H
 #include "prefix.h"
#endif

void OutputPrefix(void* word)
 {
  PREFIX::OutputTerm((PREFIX_SYM*)word);
 };

#ifndef __OR_INDEX_H
 #include "or_index.h"
#endif
#ifndef __SPLITTING_H
 #include "splitting.h"
#endif
#ifndef __RESOLVENT_H
 #include "resolvent.h"
#endif
#ifndef __HYPERRESOLUTION_H
 #include "hyperresolution.h"
#endif
#ifndef __BINARY_RESOLUTION_H
 #include "binary_resolution.h"
#endif
#ifndef __SUPERPOSITION_H
 #include "superposition.h"
#endif
#ifndef __PATH_INDEX_H
 #include "path_index.h"
#endif
#ifndef __BS_MACHINE_H
 #include "bs_machine.h"
#endif 
#ifndef __FORW_UEQ_SIMPLIFICATION_H
 #include "forw_ueq_simplification.h"
#endif 
#ifndef __BACK_DEMODULATION_H
 #include "back_demodulation.h"
#endif
#ifndef __PROBLEM_PROFILE_H
 #include "problem_profile.h"
#endif

TIMER OldBSTimer;
TIMER NewBSTimer;
TIMER EqClauseTimer; 
#ifndef __FORWARD_SUBSUMPTION_H
 #include "forward_subsumption.h"
#endif 
#ifndef __WEIGHTING_H 
 #include "weighting.h"
#endif

HEADER_WEIGHT_LIST* HWList = (HEADER_WEIGHT_LIST*)0;

HEADER_WEIGHTING HeaderWeighting;

//******************** debug ****************************

void SpellPointer(void* ptr)
 {
  //((PI_C_LIST*)ptr)->Output();
  //::Output(*(((long*)ptr)-1));
  ::Output(*((*((long**)ptr))-1));
 }; // void SpellPointer(void* ptr)


//*******************************************************


/***************** Options: **********************************/

// automaticly chosen strategy
bool strat_auto;
float auto_first_attempt; // share of max_time

// inference rules
bool hyper_resolution;
bool static_splitting;
bool dynamic_splitting;
bool paramodulation;
  
// ordering 
bool positive_is_greater;
bool bin_res_negative_selection;

// preprocessing options
bool prepro_clauses;
bool input_remove_subsumed;
bool prepro_remove_subsumed;
bool prepro_elim_def;
  
// strategy switches
bool forward_subsumption;
//bool simplify_resolvents;
bool simplify_by_fs;
bool simplify_forward_by_unit_eq;
bool simplify_backward_by_unit_eq;
bool simplify_by_eq_res;
bool backward_subsumption;
bool remove_children_of_subsumed;
bool subsume_shorter;
bool set_of_support;
bool hyper_dinamic_lhs_ordering;
bool hyper_unit_conflict;
bool estimate_resources;
bool weight_incrementation;

// numeric options
long pick_given_ratio;
long max_weight;
long max_literal_weight;
long max_length;
long max_depth;

long internal_tick_frequency;
long first_resource_check; 

// proof search limits and available resources
float time_red_mark;
float max_memory; // in Kb
float memory_red_mark;
long max_gen;
long max_kept;
long max_used;

// interface options
 // general output options
bool output_to_file;
bool tabulate_to_file;
char* output_file_name;
char* tabulation_file_name;
long tabulated_file_name_width;
 // options report
bool report_main_options_only;
bool report_options_with_default_values;
 // statistics report
bool statistics_extended_kept_report;
 // progress monitoring
bool show_used;
bool show_kept;
bool show_gen;
bool report_incremented_max_weight;
 // start report
bool start_report_problem_file_name;
bool start_report_version_date;
bool start_report_date;
bool start_report_options;
bool start_report_input_clauses;
bool start_report_preprocessed_clauses;
 // termination report
bool termination_report_problem_file_name;
bool termination_report_proof;
bool termination_report_statistics;
bool termination_report_options;

/*************************************************************/

// state parameters

extern long cl_time_limit; 
float max_time;
extern bool cl_output_proof;

char* problemFileName;
long lastUsed = 0L;	     // the number of clauses made used
long lastKept = 0L;	     // the number of clauses made kept
long lastInput = 0L;
long lastPreprocessed = 0L;
long backSubsumed = 0L; // the number of back subsumed clauses
//int is_busy = 0;
bool check_time = false;
bool check_memory = false;

long current_max_weight;

long backward_simplified = 0L;

//******************* MAIN OBJECTS: *******************************

PROBLEM_PROFILE ProblemProfile;
SIGNATURE Signature;

CLAUSE_SET Input;
CLAUSE_SET SupportedInput;
CLAUSE_SET UnsupportedInput;
CLAUSE_SET PreprocessedSupportedInput;
CLAUSE_SET PreprocessedUnsupportedInput;

KEPT Kept; 
CLAUSE_SET ReservedKept;
CLAUSE_SET Used;
CLAUSE_SET BackwardSubsumedClauses;
CLAUSE_SET Trash;

TIMER MainTimer;
TIMER LoopTimer;

SHARING_INDEX GlobalSharingIndex; 
OR_INDEX ORIndex;
CLAUSE_INDEX ClauseIndex;


SPLITTING_INDEX SplittingIndex;
 
RESOLVENT Resolvent; 
 
//BIN_ORD_RES OrderedResolution;
//HYPERRESOLUTION HyperResolution;
//SUPERPOSITION Superposition;

PATH_INDEX PathIndex;
BS_MACHINE BSMachine(&PathIndex); 

FS ForwardSubsumption;


FUEQ_SIMPLIFICATION FSimplificationByUEq;
BACK_DEMOD BackDemodulation(&GlobalSharingIndex);

CLAUSE_QUEUE QueueForUnitConflict;


//******************************************************************* 
 
Clause* history = 0;
//Clause* first;

bool backup_shared = true;            

// interface options 
int show_post = 0;
int show_back = 0;
int show_prepro = 0; 

Clause* refutation;


//*********************************************************************



void IntegrateInputClause(Clause* const cl);
void NewUsed(Clause*);
int RetentionTest();
void OutputRefutation();
void OutputSimpleStatistics();
void ReportStatistics();
void ReportMemory();
void ReportOptions();
void ReleaseMemory();
char* UsedRule();
void TabulateResult(char res,const char* const short_expl);
void DestroyKeptClause(Clause* c);
void MakeReserved(Clause* cl);
void ReportWeightChange();
void TryUnitConflict();

//********************* Loop termination:  **************************

enum LOOP_TERMINATION_REASON
 {
  LTR_UNCKNOWN,
  LTR_USER_INT,
  LTR_REFUTATION,
  LTR_TIME,
  LTR_EMPTY_KEPT,
  LTR_UNPROVABLE,
  LTR_LIMIT
 };

LOOP_TERMINATION_REASON loop_termination_reason = LTR_UNCKNOWN;
const char* loop_termination_expl = "Terminated by uncknown reason.";
const char* loop_termination_short_expl = "uncknown";
const char* loop_termination_parameter_name = "uncknown parameter";

void TerminateLoop(const LOOP_TERMINATION_REASON reason,
                   const char* const expl,
                   const char* const short_expl,
                   const char* const parameter)
 {
  loop_termination_reason = reason;
  loop_termination_expl = expl;
  loop_termination_short_expl = short_expl;
  loop_termination_parameter_name = parameter;
 }; // void TerminateLoop(const LOOP_TERMINATION_REASON,const char* const short_expl)

void TerminateLoop(const LOOP_TERMINATION_REASON reason)
 {
  switch (reason)
   {
    case LTR_UNCKNOWN: 
     TerminateLoop(reason,"Terminated by uncknown reason.","uncknown  ","");
     break;
    case LTR_USER_INT:
     TerminateLoop(reason,"Terminated by the user.","user int  ","");
     break;
    case LTR_REFUTATION: 
     TerminateLoop(reason,"Proof found. Thanks to Tanya!","refutation",""); 
     break;
    case LTR_TIME: 
     TerminateLoop(reason,"Time limit exhausted.","time limit",""); 
     break;
    case LTR_EMPTY_KEPT: 
     TerminateLoop(reason,"Proof not found.Try to reset options.","empty kept",""); 
     break;
    case LTR_UNPROVABLE: 
     TerminateLoop(reason,"Unprovable.","unprovable",""); 
     break;
    case LTR_LIMIT:
     TerminateLoop(reason,"Parameter limit exhausted: ","uncknown  ","uncknown parameter"); 
     break;
    default: RuntimeError(" in TerminateLoop()"); 
   };
 }; // void TerminateLoop(const LOOP_TERMINATION_REASON reason)  

void TerminateProgram()
 {
  Output(loop_termination_expl);
  if (loop_termination_reason == LTR_LIMIT) 
   {
    Output(loop_termination_parameter_name);
   };
  Output('\n');
  if (termination_report_problem_file_name)
   {
    Output("Problem was: ");
    Output(problemFileName);
    Output("\n"); 
   };
  switch (loop_termination_reason)
   {
    case LTR_REFUTATION: 
     if (termination_report_proof || cl_output_proof)
      {
       OutputRefutation();
      };
     TabulateResult('+',loop_termination_short_expl);
     break;
    case LTR_UNPROVABLE: 

     TabulateResult('-',loop_termination_short_expl);
     break;
    default:
     TabulateResult('0',loop_termination_short_expl);
   };
  if (termination_report_statistics)
   {
    ReportStatistics();
   };
  if (termination_report_options)
   {
    ReportOptions();
   };
  Exit("\n");
 }; // TerminateProgram()

void ReportRefutation(Clause* c)
 {
  if (show_used) { NewUsed(c); };
  Used.Insert(c);
  refutation = c;
  TerminateLoop(LTR_REFUTATION);
  TerminateProgram(); 
 }; // void ReportRefutation(Clause* c)

//********************************************************************

void InitializeCtrlC();

void ProcessCtrlC(int)
 {
  MainTimer.Stop();
  LoopTimer.Stop();
  Output("\n");
  
  ReportStatistics();

  //FSimplificationByUEq.OutputIndex(); // debug
  
  #ifdef _DEBUG_
   cout << "CONTROL POINT:\n";
   REPORT_CONTROL_POINT();
   cout << "\n";
  #endif  
  //ForwardSubsumption.OutputCode(); // debug
  
  
  Output("\nContinue?(y/n)");
  char a;
  cin >> a;
  if ((a == 'n') || (a == 'N'))
   {
    TerminateLoop(LTR_USER_INT);
    TerminateProgram();
   }
  else
   {
    Output("\nContinuing proof search...\n");
    MainTimer.Start();
    LoopTimer.Start();
    InitializeCtrlC();
   };
 };

void InitializeCtrlC()
 {
  signal(SIGINT,ProcessCtrlC);
 };

char* UsedRule()
 {
  if (hyper_resolution)
   {
    if (positive_is_greater)
     {
      return "neg.hyper.";
     }
    else
     {
      return "pos.hyper.";
     };
     
   }
  else
   {
    if (positive_is_greater)
     {
      return "binary-   ";
     }
    else
     {
      return "binary+   ";
     };
   };
 };

//****************** Resources: ******************************************

inline float AvailableTime()
 {
  return (max_time - MainTimer.Elapsed());
 };

float loop_time_red_mark;
float loop_avail_time;
float loop_used_time;

inline bool LoopTimeInRed()
 {
  return (loop_used_time > loop_time_red_mark);
 };

inline void ResetLoopTimeCount(const float time) 
 {
  loop_avail_time = time;
  if (loop_avail_time <= 0) 
   {
    TerminateLoop(LTR_TIME);
    TerminateProgram();
   }
  else
   {
    loop_time_red_mark = loop_avail_time*time_red_mark;
    LoopTimer.Reset();
    LoopTimer.Start();
   };
 };


inline bool LoopTimeExceeded()
 {
  return (LoopTimer.Elapsed() > loop_avail_time);
 };

float loop_start_memory;
float loop_used_memory;
float loop_avail_memory;

inline float UsedMemory() // in Kb
 {
  return (float)(Memory::Occupied()/256); // in Kb
 }; 

inline void ResetLoopMemoryCount()
 {
  loop_start_memory = UsedMemory();
  loop_avail_memory = max_memory - loop_start_memory;
  loop_used_memory = 0.0;
 };

 
inline bool MemoryInRed()
 {
  return (UsedMemory() > memory_red_mark);
 };

inline float UsageSpeed(const float usedResource)
 {
  return (usedResource/lastUsed);
 };

inline long ReachableKeptEstimate(const float totalResource,const float usedResource)
 {
  return (long)((totalResource - usedResource)/(UsageSpeed(usedResource)));
 };

long num_of_reachable = MAXLONG;
 

inline void EstimateResources()
 {

  if (!estimate_resources) 
   {
    return;
   };
   
  /**************
  Clause* c; 
  
  while (MemoryInRed())
   {
    c = Kept.TheWorstClause();
    if (c)
     {
      Kept.Remove(c);
      DestroyKeptClause(c); // this is wrong, c can be "used" even here (as simplifying clause) 
      
      STATISTICS_ON_GENERATED::AnotherPurgedFromKept();
     }
    else { return; };
   }; 
  ****************/
  
   loop_used_memory = UsedMemory() - loop_start_memory;
   if (loop_used_memory <= 0) // possible!
    {
     loop_used_memory = 0.001;
    };
   
   loop_used_time = LoopTimer.Elapsed();
  
   long moved_from_kept = Kept.UsedAndRemoved();
 
   /******************************/ 
   long r1 = (long)(((loop_avail_time - loop_used_time)*moved_from_kept)/loop_used_time);
   long r2 = (long)(((loop_avail_memory - loop_used_memory)*moved_from_kept)/loop_used_memory);
 
   if (r1 > r2)
    {
     num_of_reachable = r2;
    }
   else 
    {
     num_of_reachable = r1;
    };
   /********************************/
  
  
   DOUT(loop_used_time); 
   DOUT(' '); 
   DOUT(moved_from_kept/loop_used_time); 
   DOUT('\n');
    
  
   //num_of_reachable = r1;
  
   //num_of_reachable = (long)(((loop_avail_time - loop_used_time)*moved_from_kept)/loop_used_time);
  
   Kept.SetNumOfReachable(num_of_reachable);
   
   long maw = Kept.MaxAcceptedWeight();
  
   Resolvent.SetMaxWeight(maw);
   
   //ReportWeightChange();
   
 }; // void EstimateResources()


void OutputRefutation() { OutputInference(refutation); };



void DestroyUsedClause(Clause* c)
 {
  DCONTROL4("BEGIN in DestroyUsedClause(Clause* c)");
  if (forward_subsumption)
   {
    ForwardSubsumption.Remove(c);
   };
   
  DCONTROL4("A1 in DestroyUsedClause(Clause* c)");
   
  ORIndex.Remove(c);
  
  DCONTROL4("A2 in DestroyUsedClause(Clause* c)");
  
  if (backward_subsumption)
   {
    PathIndex.Remove(c);
   }; 
  
  
  if (simplify_forward_by_unit_eq)
   {
    FSimplificationByUEq.Remove(c);
   }; 

  if (simplify_backward_by_unit_eq)
   {
    BackDemodulation.RemoveFromIndex(c);
   };  

  DCONTROL4("A3 in DestroyUsedClause(Clause* c)");
  
  ClauseIndex.Remove(c,false);
  
  
  DCONTROL4("A4 in DestroyUsedClause(Clause* c)");
  
  c->SetPrevious(history);
  c->SetStore(CS_HISTORY);
  history = c;
  
  DCONTROL4("END in DestroyUsedClause(Clause* c)");
 };

void DestroyKeptClause(Clause* c)
 {
  CHECK4(c->Store() == CS_TRASH_KEPT,"XX1 in DestroyKeptClause(Clause* c)");
 
  DCONTROL4("BEGIN in DestroyKeptClause(Clause* c)");
  
  if (forward_subsumption)
   {
    ForwardSubsumption.Remove(c);
   };
   
  DCONTROL4("A1 in DestroyKeptClause(Clause* c)");
  
  if (backward_subsumption)
   {
    PathIndex.Remove(c);
   }; 

  if (simplify_forward_by_unit_eq)
   {
    FSimplificationByUEq.Remove(c);
   }; 

  if (simplify_backward_by_unit_eq)
   {
    BackDemodulation.RemoveFromIndex(c);
   }; 

  DCONTROL4("A2 in DestroyKeptClause(Clause* c)");
  
  ClauseIndex.Remove(c,true);
  
  DCONTROL4("A3 in DestroyKeptClause(Clause* c)");
  
  c->Destroy();
  
  DCONTROL4("END in DestroyKeptClause(Clause* c)");
 };

void DestroyKeptSimplifier(Clause* c)
 {
  CHECK4(c->Store() == CS_TRASH_KEPT_SIMPLIFIER,"XX1 in DestroyKeptSimplifier(Clause* c)");
 
  DCONTROL4("BEGIN in DestroyKeptSimplifier(Clause* c)");
  
  
  if (forward_subsumption)
   {
    ForwardSubsumption.Remove(c);
   };
  DCONTROL4("A1 in DestroyKeptSimplifier(Clause* c)"); 
  if (backward_subsumption)
   {
    PathIndex.Remove(c);
   };

  if (simplify_forward_by_unit_eq)
   {
    FSimplificationByUEq.Remove(c);
   }; 

  if (simplify_backward_by_unit_eq)
   {
    BackDemodulation.RemoveFromIndex(c);
   }; 
   
  DCONTROL4("A2 in DestroyKeptSimplifier(Clause* c)"); 
   
  ClauseIndex.Remove(c,false);
  
  DCONTROL4("A3 in DestroyKeptSimplifier(Clause* c)"); 
  
  c->SetPrevious(history);
  c->SetStore(CS_HISTORY);
  history = c;
  DCONTROL4("END in DestroyKeptSimplifier(Clause* c)"); 
 }; 


void MakeReserved(Clause* cl)
 {
  // temporary version
  switch (cl->Store())
   {
    case CS_KEPT:
     cl->SetStore(CS_KEPT_RESERVED);
     ReservedKept.Insert(cl);
     break;
    case CS_KEPT_SIMPLIFIER: 
     cl->SetStore(CS_KEPT_SIMPLIFIER_RESERVED);
     ReservedKept.Insert(cl);    
     break;
    #ifdef _DEBUG_
     default: DebugError("A0 in MakeReserved(Clause* cl)"); return; 
    #else
     #ifdef _SUPRESS_WARNINGS_
      default: break;
     #endif
    #endif
   };
 };

inline void ReservedIntoKept(Clause* cl)
 {
  switch (cl->Store())
   {
    case CS_KEPT_RESERVED:
     cl->SetStore(CS_KEPT);
     break;
    case CS_KEPT_SIMPLIFIER_RESERVED:
     cl->SetStore(CS_KEPT_SIMPLIFIER);
     break;
    #ifdef _DEBUG_
     default: DebugError("A0 in ReservedIntoKept(Clause* cl)"); return; 
    #else
     #ifdef _SUPRESS_WARNINGS_
      default: break;
     #endif
    #endif
   };
  Kept.Insert(cl);
 };

void ReservedIntoKept()
 {
  while (!ReservedKept.Empty()) ReservedIntoKept(ReservedKept.Pop());
 };



//************** Trash: ******************************************

 
void RecycleTrash()
 {
  Clause* c;
  bool something_removed = (bool)Trash.Size();
  while (!(Trash.Empty()))
   {
    c = Trash.Pop();
    switch (c->Store())
     {
      case CS_TRASH_KEPT : DestroyKeptClause(c); break;
      case CS_TRASH_KEPT_SIMPLIFIER : DestroyKeptSimplifier(c); break;
      case CS_TRASH_USED : DestroyUsedClause(c); break;
      #ifdef _DEBUG_
       default :
       DebugError("A10 in RecycleTrash()");
      #endif
     };
   };
  if (something_removed) { EstimateResources(); };
 }; // void RecycleTrash()

//**********************************************************************

void NewKept(Clause* c)
 {
  Output("* ");
  Output(c);
  Output('\n');
 };  // NewKept


void NewUsed(Clause* c)
 {
  Output(c);
  Output('\n');
 };  // NewUsed()

//*****************************************************************************

inline Clause* FSubsume(Clause* cl)
 {
  return Resolvent.FSubsume(cl);
 }; // Clause* FSubsume(Clause* cl)

void SeparateSubsumed(CLAUSE_SET& cset,CLAUSE_SET& subsumed)
 {
  // FS index must be empty here
  
  CLAUSE_SET tmp_cset;
  Clause* cl;
  while (!(cset.Empty()))
   {
    cl = cset.Pop();
    if (FSubsume(cl)) { subsumed.Insert(cl); }
    else
     {
      ForwardSubsumption.Integrate(cl);
      DCONTROL0("A7 in void SeparateSubsumed(..)");
      tmp_cset.Insert(cl);
      DCONTROL0("A8 in void SeparateSubsumed(..)");
     };
   };
  DCONTROL0("A10 in void SeparateSubsumed(..)"); 
  
  //ForwardSubsumption.OutputCode(); // debug
  
  ForwardSubsumption.Destroy();
  DCONTROL0("A11 in void SeparateSubsumed(..)"); 
  while (!(tmp_cset.Empty()))
   {
    DCONTROL0("A12 in void SeparateSubsumed(..)"); 
    cl = tmp_cset.Pop();
    DCONTROL0("A13 in void SeparateSubsumed(..)"); 
    if (FSubsume(cl)) 
     {
      DCONTROL0("A14 in void SeparateSubsumed(..)"); 
      subsumed.Insert(cl); 
      DCONTROL0("A15 in void SeparateSubsumed(..)"); 
     }
    else
     {
      DCONTROL0("A17 in void SeparateSubsumed(..)"); 
      ForwardSubsumption.Integrate(cl); 
      DCONTROL0("A19 in void SeparateSubsumed(..)"); 
      cset.Insert(cl);
      DCONTROL0("A20 in void SeparateSubsumed(..)"); 
     };
   };
  ForwardSubsumption.Destroy();
  DCONTROL0("END in void SeparateSubsumed(..)");
 }; // void SeparateSubsumed(CLAUSE_SET& cset,CLAUSE_SET& subsumed)

void RegisterInputClause(Clause* cl)
 {
  lastInput++;
  cl->SetNumber(lastInput);
  Input.Insert(cl);
 }; // void IntegrateInputClause(Clause* cl)

bool UnsupportedInputClause(Clause* const c)
 {
  //temporary version
  return (c->Ancestors()->Background() == ANCESTORS::AXIOM);
 };

void SeparateUnsupportedInput()
 { 
  Clause* cl;
  SupportedInput.Purge();
  UnsupportedInput.Purge();
  Input.ResetIterator();
  while (!(Input.Empty())) 
   {
    cl = Input.Pop();
    if ((!hyper_resolution) && (set_of_support) && (UnsupportedInputClause(cl)))
     {
      UnsupportedInput.Insert(cl);     
     }
    else
     {
      SupportedInput.Insert(cl);
     };
    cl = Input.Next();
   };
 };

void RegisterPreprocessedSupported(Clause* cl)
 {
  lastPreprocessed++;
  cl->SetNumber(lastPreprocessed);
  PreprocessedSupportedInput.Insert(cl);
 }; // void RegisterPreprocessedSupported(Clause* cl)

void RegisterPreprocessedUnsupported(Clause* cl)
 {
  lastPreprocessed++;
  cl->SetNumber(lastPreprocessed);
  PreprocessedUnsupportedInput.Insert(cl);
 }; // void RegisterPreprocessedUnsupported(Clause* cl)

void PreprocessInputClauses()
 {
  Resolvent.SetStaticSplitting(static_splitting);
 
  lastPreprocessed = lastInput; 
 
  SplittingIndex.Reset();
  
  Clause* cl;
  SupportedInput.ResetIterator();
  while ((cl = SupportedInput.Next()))
   {
    Resolvent.PreprocessSupported(cl);
   };
  UnsupportedInput.ResetIterator();
  while ((cl = UnsupportedInput.Next()))
   {
    Resolvent.PreprocessUnsupported(cl);
   };

  CLAUSE_SET PreprocessedInputSubsumed; 
  if ((prepro_clauses) && (prepro_remove_subsumed))
   {
    SeparateSubsumed(PreprocessedSupportedInput,PreprocessedInputSubsumed); 
    while (!(PreprocessedInputSubsumed.Empty()))
     {
      PreprocessedInputSubsumed.Pop()->Destroy();
     };
    SeparateSubsumed(PreprocessedUnsupportedInput,PreprocessedInputSubsumed); 
    while (!(PreprocessedInputSubsumed.Empty()))
     {
      PreprocessedInputSubsumed.Pop()->Destroy();
     };
   };
 }; // void PreprocessInputClauses()

void IntegrateIntoUsed(Clause* const cl)
 {
  DCONTROL0("BEGIN in IntegrateIntoUsed(Clause* const cl)");
  ++lastUsed;
   if (lastUsed > max_used) 
    {
     TerminateLoop(LTR_LIMIT,"Parameter limit exhausted: ","max_used  ","max_used");
     TerminateProgram();
    };
    
   cl->SetStore(CS_USED); 
   
   DCONTROL0("X10 in IntegrateIntoUsed(Clause* const cl)");
  
   Used.Insert(cl);

   DCONTROL0("X20 in IntegrateIntoUsed(Clause* const cl)"); 
 
   ORIndex.Integrate(cl);
 
   DCONTROL0("X30 in IntegrateIntoUsed(Clause* const cl)");  

   if (show_used) { NewUsed(cl); };

   DCONTROL0("END in IntegrateIntoUsed(Clause* const cl)");
 }; // void IntegrateIntoUsed(Clause* const cl)


inline void MakeShared(Clause* cl)
 {
  // for unshared input or preprocessed clauses
  STACK<TERM*,MAX_NUMBER_OF_LITERALS> literals("MAX_NUMBER_OF_LITERALS in MakeShared(Clause* cl):literals");
  for (TERM* lit = cl->Literals()->Args(); !(lit->IsNull()); lit = lit->Next())
   {
    literals.Push(lit->First());
   };
  ClauseIndex.Integrate(cl);
  while (!(literals.Empty())) literals.Pop()->Destroy();
 }; // void MakeShared(Clause* cl)


void InputToKept()
 {
  Kept.Reset();
  Clause* c;
  
  if (hyper_resolution)
   {
    // all input clauses must be in PreprocessedSupportedInput
      
    /***** new version: *****/ 
    CLAUSE_SET facts;
    lastKept = lastPreprocessed;
    while (!(PreprocessedSupportedInput.Empty()))
     {
      DCONTROL0("E20 in InputToKept()");    
      
      c = PreprocessedSupportedInput.Pop();
       
      DCONTROL0("E30 in InputToKept()");     

      if (HyperResolution.NewRule(c,true)) // true means that the rule is compiled
       {
        DCONTROL0("E40 in InputToKept()");     
    
        c->SetStore(CS_HYPER_RULE);

        DCONTROL0("E50 in InputToKept()");
       
        //cout << "MAKE SHARED\n"; ::Output(c); cout << "\n"; // debug
 
        MakeShared(c); 

        //cout << "DONE\n"; // debug
              

        DCONTROL0("E60 in InputToKept()");
       }
      else // fact
       {
        DCONTROL0("E70 in InputToKept()");
        facts.Insert(c);
        DCONTROL0("E80 in InputToKept()");
       };
     }; 
    
    DCONTROL0("E100 in InputToKept()");  

    while (!facts.Empty())
     {
      c = facts.Pop();

      if (Resolvent.LoadClause(c))
       {
        Resolvent.PushAncestor(c);        
        Resolvent.SetBackground(ANCESTORS::RESOL);
        Kept.MaxPriorityOn();
        Resolvent.ProcessCollected(); 
        Kept.MaxPriorityOff();
       }  
      else 
       {
	Output("WARNING: a preprocessed input clause rejected:\n");
	Output(c);
        Output("\n");
       }; 
     };

    TryUnitConflict();      

     /************************/

    //lastKept = lastPreprocessed;
   }
  else // binary resolution
   {
    // push supported into Kept
    lastKept = lastPreprocessed;
    while (!(PreprocessedSupportedInput.Empty()))
     {
      c = PreprocessedSupportedInput.Pop();

      if (Resolvent.LoadClause(c))
       {
        Resolvent.PushAncestor(c);        
        Resolvent.SetBackground(ANCESTORS::RESOL);
        Kept.MaxPriorityOn();
        Resolvent.ProcessCollected(); 
        Kept.MaxPriorityOff();
       }  
      else 
       {
	Output("WARNING: a preprocessed input clause rejected:\n");
	Output(c);
        Output("\n");
       };
     };

    // push unsupported into Used
    while (!(PreprocessedUnsupportedInput.Empty()))
     {
      c = PreprocessedUnsupportedInput.Pop();
      MakeShared(c);
      IntegrateIntoUsed(c);
     };
   };
 }; // void InputToKept()
 

void InternalTickEvent(const long ticks)
 {
  if (ticks >= first_resource_check)
   {
    EstimateResources();
   };
 }; // void InternalTickEvent(const long ticks)

void InsertIntoKept(Clause* c)
 {
  DCONTROL0("BEGIN in InsertIntoKept(..)");
  
  //cout << "AGA\n";
  //Output(c); cout << "\n"; 
  
  Kept.Insert(c);
  
  DCONTROL0("A0 in InsertIntoKept(..)");
  
  ++lastKept;

  if (lastKept > max_kept) 
   {
    TerminateLoop(LTR_LIMIT,"Parameter limit exhausted: ","max_kept  ","max_kept");
    TerminateProgram();
   };

  DCONTROL0("A1 in InsertIntoKept(..)");
  
  c->SetNumber(lastKept);
  
  DCONTROL0("A2 in InsertIntoKept(..)");
  
  c->SetStore(CS_KEPT);

  DCONTROL0("A3 in InsertIntoKept(..)");
 
  if (show_kept) { NewKept(c); };
  
  DCONTROL0("END in InsertIntoKept(..)"); 
 };  // InsertIntoKept()

inline void TestForRefutation(Clause* const cl)
 {
  if (cl->Empty())
   {
    ++lastKept;
    cl->SetNumber(lastKept);
    ReportRefutation(cl);
   };
 }; // void TestForRefutation(Clause* const cl)


void TestForUnitConflict(Clause* const cl)
 {
  if (!cl->Active()) return;
  
  //::Output(cl); cout << "\n"; // debug  

  if (hyper_resolution)
   {
    
    Clause* conflicting_clause = HyperResolution.UnitConflictTest(cl);

    if (conflicting_clause) // refutation found
     {
      Resolvent.Reset();
      Resolvent.SetBackground(ANCESTORS::RESOL);
      Resolvent.PushAncestor(cl);
      Resolvent.PushAncestor(conflicting_clause);
      ReportRefutation(Resolvent.MakeEmptyClause());
     };
   };
 }; // void TestForUnitConflict(Clause* cl)


void TryUnitConflict()
 {
  Clause* query;
  while (QueueForUnitConflict.Dequeue(query)) TestForUnitConflict(query);
 }; // void TryUnitConflict()


void AnotherBackwardSubsumed(Clause* const cl)
 {   
  if (!(cl->Trash()))
   {
     //cl->MarkAsSimplifier(); // debug    

    switch (cl->Store())
     {
      case CS_KEPT: Kept.Remove(cl); break;
      case CS_KEPT_SIMPLIFIER: Kept.Remove(cl); break;
      case CS_KEPT_RESERVED: ReservedKept.Remove(cl); break;    
      case CS_KEPT_SIMPLIFIER_RESERVED: ReservedKept.Remove(cl); break;
      case CS_USED: Used.Remove(cl); break;
      #ifdef _DEBUG_
       default: DebugError("A0 in AnotherBackwardSubsumed(Clause* const cl)"); return;
      #else
       #ifdef _SUPRESS_WARNINGS_
        default: return;
       #endif
      #endif
     };
    cl->MakeTrash();
    BackwardSubsumedClauses.Insert(cl);
   }; 
 }; // AnotherBackwardSubsumed(Clause* const cl)

void AnotherBackwardSimplified(Clause* simplified_cl,Clause* simplifier)
 {  
  CHECK0(!simplified_cl->Trash(),"A0 in AnotherBackwardSimplified(Clause* simplified_cl,Clause* simplifier)");
  CHECK0(!simplifier->Trash(),"A1 in AnotherBackwardSimplified(Clause* simplified_cl,Clause* simplifier)");
  CHECK0(simplified_cl != simplifier,"A2 in AnotherBackwardSimplified(Clause* simplified_cl,Clause* simplifier)");
  #ifdef _SUPPRESS_WARNINGS_
   simplifier = simplifier;
  #endif

   //Output("CLAUSE "); Output(simplifier); Output("\n"); // debug
   //Output("SIMPLIFIES "); Output(simplified_cl); Output("\n\n\n"); // debug


  backward_simplified++;
  simplified_cl->MarkAsSimplifier(); 
  switch (simplified_cl->Store())
   {
    case CS_KEPT: Kept.Remove(simplified_cl); break;
    case CS_KEPT_SIMPLIFIER: Kept.Remove(simplified_cl); break;
    case CS_KEPT_RESERVED: ReservedKept.Remove(simplified_cl); break;    
    case CS_KEPT_SIMPLIFIER_RESERVED: ReservedKept.Remove(simplified_cl); break;
    case CS_USED: Used.Remove(simplified_cl); break;
    #ifdef _DEBUG_
     default: DebugError("A0 in AnotherBackwardSimplified(Clause* simplified_cl,Clause* simplifier)"); return;
    #else
     #ifdef _SUPRESS_WARNINGS_
      default: return;
     #endif
    #endif
   };    
  simplified_cl->MakeTrash();
  Trash.Insert(simplified_cl);
 }; // void AnotherBackwardSimplified(Clause* simplified_cl,Clause* simplifier)

void AnotherObsolete(Clause* cl)
 {
  if (cl->Trash()) return;

  //::Output("OBSOLETE "); ::Output(cl); ::Output("\n"); // debug

  switch (cl->Store())
   {
    case CS_KEPT: Kept.Remove(cl); break;
    case CS_KEPT_SIMPLIFIER: Kept.Remove(cl); break;
    case CS_KEPT_RESERVED: ReservedKept.Remove(cl); break;    
    case CS_KEPT_SIMPLIFIER_RESERVED: ReservedKept.Remove(cl); break;
    case CS_USED: Used.Remove(cl); break;
    #ifdef _DEBUG_
     default: DebugError("A0 in AnotherObsolete(Clause* cl)"); return;
    #else
     #ifdef _SUPRESS_WARNINGS_
      default: return;
     #endif
    #endif
   };
  cl->MakeTrash();  
  Trash.Insert(cl);
 }; // void AnotherObsolete(Clause* cl)

inline bool IsObsolete(Clause* cl)
 {
  return cl->ContainsEliminated(ProblemProfile.EliminationTable());  
 }; // bool IsObsolete(Clause* cl) 


void PurgeObsolete()
 {
  Used.ResetIterator();
  for (Clause* c = Used.Next(); c; c = Used.Next())
   {
    if (IsObsolete(c)) AnotherObsolete(c);
   };   
   
  Kept.ResetIterator();
  for (Clause* c = Kept.IterNext(); c; c = Kept.IterNext())
   {
    if (IsObsolete(c)) AnotherObsolete(c);
   }; 
 }; // void PurgeObsolete()


inline void BackwardSimplification()
 {  
  if (simplify_backward_by_unit_eq)
   {
    BackDemodulation.Simplify();
   };  
 }; // void BackwardSimplification()


inline void BackwardSubsumedIntoTrash(const Clause* subsuming_cl)
 {
  Clause* subsumed_cl;
  bool kept_destroyed = false;
  
  DCONTROL4("A1 in void BackwardSubsumedIntoTrash(const Clause* subsuming_cl)");
  
  while (!(BackwardSubsumedClauses.Empty()))
   {
    DCONTROL4("A2 in void BackwardSubsumedIntoTrash(const Clause* subsuming_cl)");
    
    subsumed_cl = BackwardSubsumedClauses.Pop();

    
    //Output("\n\nCLAUSE   "); Output(subsuming_cl); Output("\n"); // debug
    //Output("SUBSUMES "); Output(subsumed_cl); Output("\n\n"); // debug


    
    DCONTROL4("A3 in void BackwardSubsumedIntoTrash(const Clause* subsuming_cl)");
    
    if (subsumed_cl->Store() == CS_TRASH_KEPT)
     {
      DCONTROL4("A4 in void BackwardSubsumedIntoTrash(const Clause* subsuming_cl)");
      kept_destroyed = true;
      
      DestroyKeptClause(subsumed_cl);
      DCONTROL4("A5 in void BackwardSubsumedIntoTrash(const Clause* subsuming_cl)");
     }
    else // was in Used or a simplifier
     {
      CHECK4((subsumed_cl->Store() == CS_TRASH_USED) || (subsumed_cl->Store() == CS_TRASH_KEPT_SIMPLIFIER),
             "XXX1 in BackwardSubsumedIntoTrash(..)");
     
      DCONTROL4("A6 in void BackwardSubsumedIntoTrash(const Clause* subsuming_cl)");
      if (remove_children_of_subsumed)
       {
        Used.ResetIterator();
        for (Clause* c = Used.Next(); c; c = Used.Next())
         {
          if ((c->IsAChildOf(subsumed_cl)) && (c != subsuming_cl))
           {
            AnotherBackwardSubsumed(c);
           };
         };
       
        Kept.ResetIterator();
        for (Clause* c = Kept.IterNext(); c; c = Kept.IterNext())
         {
          if ((c->IsAChildOf(subsumed_cl)) && (c != subsuming_cl))
           {
            AnotherBackwardSubsumed(c);
           };
         };
       };
      DCONTROL4("A7 in void BackwardSubsumedIntoTrash(const Clause* subsuming_cl)"); 
      Trash.Insert(subsumed_cl);
      DCONTROL4("A8 in void BackwardSubsumedIntoTrash(const Clause* subsuming_cl)");
     };
   };
  DCONTROL4("A9 in void BackwardSubsumedIntoTrash(const Clause* subsuming_cl)"); 
  if (kept_destroyed) { EstimateResources(); };
 }; // void BackwardSubsumedIntoTrash(const Clause* subsuming_cl)



inline void BackwardSubsumption(Clause* const cl)
 {     
  if (backward_subsumption) 
   {       
    
     //NewBSTimer.Start();
     BSMachine.BSubsume(cl);
     //NewBSTimer.Stop();
    
   
    //OldBSTimer.Start(); // debug
    
    //STREE::BackSubsume(cl,subsume_shorter); 
    
    //OldBSTimer.Stop(); // debug
    
    DCONTROL4("A1 in BackwardSubsumption(Clause* const cl)");
    
    
    BackwardSubsumedIntoTrash(cl);
    
    DCONTROL4("A2 in BackwardSubsumption(Clause* const cl)");
   };
 }; // BackwardSubsumption(Clause* cl)



void RegisterNewResolvent(Clause* cl)
 {
  DCONTROL0("BEGIN in RegisterNewResolvent(..)");

  // cl must be already integrated into ClauseIndex
  TestForRefutation(cl);
  
  DCONTROL0("A1 in RegisterNewResolvent(..)"); 


  // TestForUnitConflict(cl); // bug!!!!
  

  
  DCONTROL0("A2 in RegisterNewResolvent(..)"); 
  
  if ((estimate_resources) && (LoopTimeInRed())) 
   {
    STATISTICS_ON_GENERATED::AnotherRejectedByKept(); 
    DCONTROL0("END1 in RegisterNewResolvent(..)"); 
    return; 
   };
   
  DCONTROL0("A3 in RegisterNewResolvent(..)");  
  InsertIntoKept(cl); 

  
  DCONTROL0("A4 in RegisterNewResolvent(..)"); 
  
  BackwardSubsumption(cl);
  
  DCONTROL0("A5 in RegisterNewResolvent(..)"); 
  
  if (backward_subsumption)
   { 
    PathIndex.Integrate(cl);
   };
   
  DCONTROL0("A6 in RegisterNewResolvent(..)"); 
   
  if (forward_subsumption) 
   {    
    
     //cout << "AGA  "; Output(cl); cout << "\n";

    ForwardSubsumption.Integrate(cl);
   };

  DCONTROL0("A7 in RegisterNewResolvent(..)"); 

  if (simplify_forward_by_unit_eq)
   {
    FSimplificationByUEq.Integrate(cl);
   }; 
 
  /*********/

  if (simplify_backward_by_unit_eq)
   {
    if (BackDemodulation.EnqueueQuery(cl)) cl->MarkAsSimplifier();
    BackDemodulation.IntegrateIntoIndex(cl); 
   };

  /*********/

  if (hyper_resolution && hyper_unit_conflict && cl->Unit()) 
   {
    cl->MarkAsSimplifier();
    QueueForUnitConflict.Enqueue(cl);
   };


  DCONTROL0("END2 in RegisterNewResolvent(..)"); 
 }; // void RegisterNewResolvent(Clause* cl)


void DestroyPreprocessedClauses()
 {
  lastPreprocessed = 0;
  PreprocessedSupportedInput.DestroyAllClauses();
  PreprocessedUnsupportedInput.DestroyAllClauses();
 }; // void DestroyPreprocessedClauses()

void ReleaseMemory()
 {
  DCONTROL0("BEGIN in ReleaseMemory()");
  BackwardSubsumedClauses.DestroyAllClauses();
  DCONTROL0("C10 in ReleaseMemory()");
  Trash.DestroyAllClauses();
  DCONTROL0("C20 in ReleaseMemory()");
  Clause* dummy_clause;
  while (QueueForUnitConflict.Dequeue(dummy_clause)) {};  

  DCONTROL0("C30 in ReleaseMemory()"); 
  PathIndex.Destroy();
  
  DCONTROL0("C40 in ReleaseMemory()"); 
  ClauseIndex.Destroy();
  DCONTROL0("C50 in ReleaseMemory()"); 
  GlobalSharingIndex.Destroy();
  DCONTROL0("C60 in ReleaseMemory()"); 
  Used.DestroyAllClauses();
  DCONTROL0("C70 in ReleaseMemory()"); 
  Kept.DestroyAllClauses();
  DCONTROL0("C80 in ReleaseMemory()"); 
  ReservedKept.DestroyAllClauses();
  DCONTROL0("C90 in ReleaseMemory()");
  ForwardSubsumption.Destroy();
  DCONTROL0("C100 in ReleaseMemory()"); 
  ORIndex.Destroy();
  DCONTROL0("END in ReleaseMemory()"); 
 };


bool Complete()
 {
  return !(STATISTICS_ON_GENERATED::CompletenessViolated()); 
 };  // Complete()

void OutputSimpleStatistics()
 {
  Output("time spent (in seconds) "); Output(MainTimer.Elapsed()); Output("\n");
  ReportMemory();
  Output("last value of pick_given_ratio "); Output(pick_given_ratio); Output("\n");
  Output("last value of max_weight "); Output(current_max_weight); Output("\n");
  Output("max. weight accepted by Kept "); Output(Kept.MaxAcceptedWeight()); Output("\n");
  Output("backward subsumed "); Output(BackwardSubsumedClauses.Total()); Output("\n");
  Output("backward simplified "); Output(backward_simplified); Output("\n");
 }; // OutputSimpleStatistics()

void ReportWeightChange()
 {
  if (Resolvent.PreviousMaxWeight() != Resolvent.MaxWeight())
   {
    Output("MAX.WEIGHT "); Output(Resolvent.PreviousMaxWeight());
    Output("  --->  "); Output(Resolvent.MaxWeight()); Output("\n");
   }; 
 };

void ReportStatistics()
 {
  Output("******************** Statistics: ***********************\n");
  OutputSimpleStatistics();

  STATISTICS_ON_GENERATED::Output();
  
  
  Output("*********** Statistics on Used: ***************\n");
  Output("        Total:          "); Output(Used.Total()); Output("\n");
  Output("        Currently kept: "); Output(Used.Size()); Output("\n");
  Output("        Removed:        ");  Output(Used.Removed()); Output("\n");
  
  Output("*********** Statistics on Kept: ***************\n");
  Kept.OutputStatistics(statistics_extended_kept_report);

  Output("*************** End of statistics report. ****************\n");
  
  
  //Output("Old BS consumed: "); Output(OldBSTimer.Elapsed()); Output("\n"); // debug
  //Output("New BS consumed: "); Output(NewBSTimer.Elapsed()); Output("\n"); // debug
  //Output("Clause equalization consumed: "); Output(EqClauseTimer.Elapsed()); Output("\n"); // debug

  //ForwardSubsumption.OutputCode(); // debug
 };

void ReportOptions()
 {
  Output("********** Options (");
  if (report_main_options_only) 
   {
    Output("main"); 
    if (!(report_options_with_default_values)) { Output(","); };
   };
  if (!(report_options_with_default_values)) 
   {
    Output("explicitely set");
   }; 
  if ((!(report_main_options_only)) && (report_options_with_default_values))
   { Output("all"); };
  Output("): *********\n");
  OPTION::OutputOptions(report_main_options_only,report_options_with_default_values);
  Output("*********** End of option list. ******************\n");
 }; // void ReportOptions()

void ShowInputClauses()
 {
  if (start_report_input_clauses)
   {
    Output("************************** INPUT CLAUSES: ************************\n");
    Output("SUPPORTED:\n");
    SupportedInput.OutputClauses(); 
    Output("UNSUPPORTED:\n");
    UnsupportedInput.OutputClauses(); 
    Output("\n********************** END OF INPUT CLAUSES. *********************\n");
   };
 }; // void ShowInputClauses()
 
void ReportStart(char* problemFileName)
 {
  TabulateFileName(problemFileName);
  
  if (start_report_problem_file_name)
   { Output("Problem: "); Output(problemFileName); Output("\n"); };
  if (start_report_version_date)
   {
    Output("Vampire version date: "); Output(VAMPIRE_DATE); Output(" ");
   };
  if (start_report_date)
   { 
    time_t startTime = time((time_t*)0);
    Output("Started "); Output(asctime(localtime(&startTime)));
   };
  Output("\n");
  if (start_report_options) 
   {
    ReportOptions(); 
    if (HWList)
     {
      Output("INPUT HEADER WEIGHTING:\n");
      HEADER_WEIGHT_LIST::Output(HWList);
     };
   };
  ShowInputClauses();
 }; // void ReportStart(char* problemFileName)
 
 
void ReportMemory()
 {
  long occup = Memory::Occupied();
  long req = Memory::Requested();
  
  Output("occupied memory ");
  Output(occup/256);
  Output(" Kb (");
  Output(occup);
  Output(" 32-bit words)\n");
  
  Output("memory requested from the system ");
  Output(req/256);
  Output(" Kb (");
  Output(req);
  Output(" 32-bit words)\n");
 };

void TabulateResult(char res,const char* const short_expl)
 {
  TabulateChar(res);
  TabulateTime(MainTimer.Elapsed());
  TabulateMemory((Memory::Requested())/256);
  TabulateString(short_expl);
  TabulateString(UsedRule());
  if (weight_incrementation)
   {
    TabulateString("+w.inc.");
   }
  else
   {
    TabulateString("-w.inc.");
   };
  if (estimate_resources)
   {
    TabulateString("+est.res.");
   }
  else
   {
    TabulateString("-est.res.");
   };
  TabulateLong(current_max_weight);
  TabulateLong(pick_given_ratio);  
  TabulateChar('\n');
 }; // void TabulateResult(char res,const char* const short_expl)


void ShowPreprocessedInputClauses()
 {
  if (start_report_preprocessed_clauses)
   {
    Output("**************** PREPROCESSED INPUT CLAUSES: ************************\n");
    Output("SUPPORTED:\n");
    PreprocessedSupportedInput.OutputClauses(); 
    Output("UNSUPPORTED:\n");
    PreprocessedUnsupportedInput.OutputClauses();
    Output("\n************* END OF PREPROCESSED INPUT CLAUSES. ******************\n");
   };
 }; // void ShowPreprocessedInputClauses()


void ResetEverything()
 {
  lastUsed = 0L;
  lastKept = 0L;
  backward_simplified = 0L;  

  Resolvent.ResetInternalTick(internal_tick_frequency);
  
  STATISTICS_ON_GENERATED::Reset();
  Kept.Reset();
  
  Resolvent.SetMaxLength(max_length);
  Resolvent.SetMaxWeight(current_max_weight);
  
  loop_termination_reason = LTR_UNCKNOWN;
  loop_termination_expl = "Terminated by uncknown reason.";
  loop_termination_short_expl = "uncknown";
  loop_termination_parameter_name = "uncknown parameter";
  
  GlobalSubst.UnbindAll();
 }; // void ResetEverything()

void SetOptions()
 {
  // automaticly chosen strategy
  strat_auto = (bool)(OPTION::Value("strat_auto"));
  auto_first_attempt = ((float)((long)(OPTION::Value("auto_first_attempt"))))/100;
  
  // inference rules
  hyper_resolution = (bool)(OPTION::Value("hyper_resolution"));
  static_splitting = (bool)(OPTION::Value("static_splitting"));
  dynamic_splitting = (bool)(OPTION::Value("dynamic_splitting"));
  paramodulation = (bool)(OPTION::Value("paramodulation"));
  
  // ordering 
  positive_is_greater = (bool)(OPTION::Value("positive_is_greater"));
  bin_res_negative_selection = (bool)(OPTION::Value("bin_res_negative_selection"));
  
  // preprocessing
  
  prepro_clauses = (bool)(OPTION::Value("prepro_clauses"));
  input_remove_subsumed = (bool)(OPTION::Value("input_remove_subsumed"));
  prepro_remove_subsumed = (bool)(OPTION::Value("prepro_remove_subsumed"));
  prepro_elim_def = (bool)(OPTION::Value("prepro_elim_def"));

  // strategy switches
  forward_subsumption = (bool)(OPTION::Value("forward_subsumption"));
  simplify_by_fs = (bool)(OPTION::Value("simplify_by_fs"));
  simplify_forward_by_unit_eq = (bool)(OPTION::Value("simplify_forward_by_unit_eq"));
  simplify_backward_by_unit_eq = (bool)(OPTION::Value("simplify_backward_by_unit_eq"));
  simplify_by_eq_res = (bool)(OPTION::Value("simplify_by_eq_res"));
  backward_subsumption = (bool)(OPTION::Value("backward_subsumption"));
  remove_children_of_subsumed = (bool)(OPTION::Value("remove_children_of_subsumed"));
  subsume_shorter = (bool)(OPTION::Value("subsume_shorter"));
  set_of_support = (bool)(OPTION::Value("set_of_support"));
  hyper_dinamic_lhs_ordering = (bool)(OPTION::Value("hyper_dinamic_lhs_ordering"));
  hyper_unit_conflict = (bool)(OPTION::Value("hyper_unit_conflict"));
  estimate_resources = (bool)(OPTION::Value("estimate_resources"));
  weight_incrementation = (bool)(OPTION::Value("weight_incrementation"));

  // numeric options
  pick_given_ratio = (long)(OPTION::Value("pick_given_ratio"));
  max_weight = (long)(OPTION::Value("max_weight"));
  max_literal_weight = (long)(OPTION::Value("max_literal_weight"));
  max_length = (long)(OPTION::Value("max_length"));
  max_depth = (long)(OPTION::Value("max_depth"));

  internal_tick_frequency = (long)(OPTION::Value("internal_tick_frequency"));
  first_resource_check = (long)(OPTION::Value("first_resource_check"));
  
  
  // proof search limits and available resources
  time_red_mark = ((float)((long)(OPTION::Value("time_red_mark"))))/100;
  max_memory = (float)((long)(OPTION::Value("max_memory")));
  memory_red_mark = max_memory*(((float)((long)(OPTION::Value("memory_red_mark"))))/100);
  max_gen = (long)(OPTION::Value("max_gen"));
  max_kept = (long)(OPTION::Value("max_kept"));
  max_used = (long)(OPTION::Value("max_used"));
  
  // interface options
   // general output options
  output_to_file = (bool)(OPTION::Value("output_to_file"));
  tabulate_to_file = (bool)(OPTION::Value("tabulate_to_file"));
  output_file_name = (char*)(OPTION::Value("output_file_name"));
  tabulation_file_name = (char*)(OPTION::Value("tabulation_file_name"));
  tabulated_file_name_width = (long)(OPTION::Value("tabulated_file_name_width"));
   // options report
  report_main_options_only = (bool)(OPTION::Value("report_main_options_only"));
  report_options_with_default_values = (bool)(OPTION::Value("report_options_with_default_values"));
   
   // statistics report
  statistics_extended_kept_report = (bool)(OPTION::Value("statistics_extended_kept_report"));
 
   // progress monitoring
  show_used = (bool)(OPTION::Value("show_used"));
  show_kept = (bool)(OPTION::Value("show_kept"));
  show_gen = (bool)(OPTION::Value("show_gen")); 
  report_incremented_max_weight = (bool)(OPTION::Value("report_incremented_max_weight")); 

   // start report
  start_report_problem_file_name = (bool)(OPTION::Value("start_report_problem_file_name")); 
  start_report_version_date = (bool)(OPTION::Value("start_report_version_date")); 
  start_report_date = (bool)(OPTION::Value("start_report_date")); 
  start_report_options = (bool)(OPTION::Value("start_report_options"));
  start_report_input_clauses = (bool)(OPTION::Value("start_report_input_clauses"));
  start_report_preprocessed_clauses = (bool)(OPTION::Value("start_report_preprocessed_clauses"));
   // termination report
  termination_report_problem_file_name = (bool)(OPTION::Value("termination_report_problem_file_name"));
  termination_report_proof = (bool)(OPTION::Value("termination_report_proof"));
  termination_report_statistics = (bool)(OPTION::Value("termination_report_statistics"));
  termination_report_options = (bool)(OPTION::Value("termination_report_options"));

   // derive necessary state parameters:
  check_time = (cl_time_limit != MAXLONG);
  check_memory = (((long)OPTION::Value("max_memory")) != MAXLONG); 
 }; // void SetOptions()




void SetDefaultValues()
 {
  // automaticly chosen strategy
  strat_auto = true;
  #ifdef _DEBUG_
   auto_first_attempt = 0.5;
  #else
   auto_first_attempt = 0.1;
  #endif
  
  // inference rules
  hyper_resolution = false;
  static_splitting = true;
  dynamic_splitting = false;
  paramodulation = false;
  
  // ordering 
  positive_is_greater = false;
  bin_res_negative_selection = false;
  
  // preprocessing
  
  prepro_clauses = true;
  input_remove_subsumed = true;
  prepro_remove_subsumed = true;
  prepro_elim_def = false;

  // strategy switches
  forward_subsumption = true;
  simplify_by_fs = false;
  simplify_forward_by_unit_eq = false; 
  simplify_backward_by_unit_eq = false;
  simplify_by_eq_res = false;
  backward_subsumption = false;
  remove_children_of_subsumed = false; 
  subsume_shorter = true;
  set_of_support = false;
  hyper_dinamic_lhs_ordering = true; 
  hyper_unit_conflict = true;
  estimate_resources = true;
  weight_incrementation = false;

  // numeric options
  pick_given_ratio = 3;
  max_weight = MAXLONG;
  max_literal_weight = MAXLONG;
  max_length = MAXLONG;
  max_depth = MAXLONG;

  internal_tick_frequency = 500;
  first_resource_check = 2;
  
  
  // proof search limits and available resources
  time_red_mark = 1.0;
  max_memory = 50000.0;
  memory_red_mark = 0.97;
  max_gen = MAXLONG;
  max_kept = MAXLONG;
  max_used = MAXLONG;
  
  // interface options
   // general output options
  output_to_file = false;
  tabulate_to_file = false;
  output_file_name = "vampire.out";
  tabulation_file_name = "vampire.tab";
  tabulated_file_name_width = 14;
   // options report
  report_main_options_only = true;
  report_options_with_default_values = false;
   
   // statistics report
  statistics_extended_kept_report = false;

   // progress monitoring
  show_used = false;
  show_kept = false;
  show_gen = false; 
  report_incremented_max_weight = false;

   // start report
  start_report_problem_file_name = true;
  start_report_version_date = true;
  start_report_date = true;
  start_report_options = false;
  start_report_input_clauses = false;
  start_report_preprocessed_clauses = false;
   // termination report
  termination_report_problem_file_name = true;
  termination_report_proof = false; 
  termination_report_statistics = false;
  termination_report_options = false;

   // derive necessary state parameters:
  check_time = (cl_time_limit != MAXLONG);
  check_memory = (max_memory != MAXLONG); 
 }; // void SetDefaultValues()


void OutputMainValues()
 {
  // inference rules
  Output("hyper_resolution = "); Output(hyper_resolution); Output("\n");
  Output("static_splitting = "); Output(static_splitting); Output("\n");
  Output("paramodulation   = "); Output(paramodulation); Output("\n");
  
  // ordering 
  Output("positive_is_greater = "); Output(positive_is_greater); Output("\n");  
  Output("bin_res_negative_selection = "); Output(bin_res_negative_selection); Output("\n");  

  // preprocessing
  Output("prepro_elim_def = "); Output(prepro_elim_def); Output("\n"); 
 
  // strategy switches
  Output("forward_subsumption = "); Output(forward_subsumption); Output("\n");  
  Output("simplify_by_fs = "); Output(simplify_by_fs); Output("\n");  
  Output("simplify_forward_by_unit_eq = "); Output(simplify_forward_by_unit_eq); Output("\n");  
  Output("simplify_backward_by_unit_eq = "); Output(simplify_backward_by_unit_eq); Output("\n");  
  Output("backward_subsumption = "); Output(backward_subsumption); Output("\n");  
  Output("estimate_resources = "); Output(estimate_resources); Output("\n"); 
  Output("weight_incrementation = "); Output(weight_incrementation); Output("\n");

  // numeric options
  Output("pick_given_ratio = "); Output(pick_given_ratio); Output("\n");
  Output("max_weight = "); Output(max_weight); Output("\n");
 };




void AdjustTmpClause()
 {
  Resolvent.SetMaxWeight(current_max_weight);
  Resolvent.SetMaxLitWeight(max_literal_weight);
  Resolvent.SetMaxDepth(max_depth);
  Resolvent.SetMaxLength(max_length);
  
  Resolvent.SetForwardSubsumption(forward_subsumption);
  //---TmpClause.SetSubsumeShorter(subsume_shorter);
  
  Resolvent.SetStaticSplitting(static_splitting);
  //---TmpClause.SetDynamicSplitting(dynamic_splitting);
  
  //---TmpClause.SetSimplifyResolvents(simplify_resolvents);
  Resolvent.SetSimplificationByFS(forward_subsumption && simplify_by_fs);
  Resolvent.SetSimplificationByUEq(simplify_forward_by_unit_eq); 
  Resolvent.SetSimplificationByEqRes(simplify_by_eq_res);

  Resolvent.SetNegativeSelection((!hyper_resolution) && bin_res_negative_selection);
  Resolvent.SetEqualityPossible(paramodulation);
  
  Resolvent.ResetInternalTick(internal_tick_frequency);
 }; // void AdjustTmpClause()

void ResetTmpClauseParameters()
 {
  Resolvent.SetMaxWeight(MAX_CLAUSE_WEIGHT);
  Resolvent.SetMaxLitWeight(MAX_TERM_WEIGHT);
  Resolvent.SetMaxDepth(MAX_TERM_DEPTH);
  Resolvent.SetMaxLength(MAX_NUMBER_OF_LITERALS);
 }; // void ResetTmpClauseParameters()

bool IncMaxWeight() 
 {
  if (current_max_weight < MAX_CLAUSE_WEIGHT)
   {
    current_max_weight++;
    return true;
   }
  else { return false; };
 };

bool UseNext()
 {
  DCONTROL0("BEGIN in UseNext()");
  Clause* cl;
  if (hyper_resolution)
   {
    hyper_try_next_fact:
     cl = Kept.Next();
     if (!cl)  // Kept is empty
      {
       TryUnitConflict();
       if (!ReservedKept.Empty()) 
        {
         if (LoopTimeExceeded()) return true;
         ReservedIntoKept();
         EstimateResources();
         goto hyper_try_next_fact;
        };
       return false;
      };
     
     if (cl->MinResWeight() > Resolvent.MaxWeight())
      {
       MakeReserved(cl);
       EstimateResources();
       goto hyper_try_next_fact;
      };
     
     IntegrateIntoUsed(cl);
     if (!HyperResolution.Load(cl)) // no rule applicable to the fact
      goto hyper_try_next_fact;
     HyperResolution.ProcessRule();
     TryUnitConflict();
     RecycleTrash();
     while (HyperResolution.FindNextRule())
      {
       HyperResolution.ProcessRule();
       TryUnitConflict();
       RecycleTrash();
      }; 
     return true; 
   }
  else // binary resolution, probably with superposition
   {
    DCONTROL0("BINARY in UseNext()");
    binary_try_next:     
     cl = Kept.Next();
     DCONTROL0("BIN10 in UseNext()");
     if (!cl)  // Kept is empty
      {
       if (!ReservedKept.Empty()) 
        {
         if (LoopTimeExceeded()) return true;
         ReservedIntoKept();
         EstimateResources();

         DCONTROL0("BIN50 in UseNext()");            

         goto binary_try_next;
        };

       // here ReservedKept.Empty()
       return false;
      };

     DCONTROL0("BIN60 in UseNext()"); 

     if (cl->MinResWeight() > Resolvent.MaxWeight())
      { 
       DCONTROL0("BIN70 in UseNext()");      
  
       MakeReserved(cl);
      
       DCONTROL0("BIN80 in UseNext()"); 
      
       EstimateResources();
       
       DCONTROL0("BIN90 in UseNext()");     

       goto binary_try_next;
      };


     DCONTROL0("BIN100 in UseNext()");     

     IntegrateIntoUsed(cl);
 
     DCONTROL0("BIN110 in UseNext()");
       
     if (paramodulation) 
      { 
       Superposition.Process(cl); 
       BackwardSimplification();
       RecycleTrash();
      };
    
     if (!(cl->Literals()->Args()->First()->IsEquality()))
      {
       OrderedResolution.Load(cl);
       while (OrderedResolution.NextLiteral()) 
        {
         BackwardSimplification();
         RecycleTrash();
        };
      };
     
     
     return true;
   }; 
  
 }; // bool UseNext()

void RunHyper(const bool positive_hyper,const float avail_time)
 {
  DCONTROL0("BEGIN in RunHyper(const bool positive_hyper,const float avail_time)"); 
  hyper_resolution = true;
  positive_is_greater = !(positive_hyper);

  //if (strat_auto) OutputMainValues();  

  TERM::SetPositiveIsGreater(positive_is_greater);
  ORIndex.SetParamodulation(false);

  DCONTROL0("C10 in RunHyper(const bool positive_hyper,const float avail_time)"); 

  PreprocessInputClauses();

  DCONTROL0("C20 in RunHyper(const bool positive_hyper,const float avail_time)"); 

  ShowPreprocessedInputClauses();
  STATISTICS_ON_GENERATED::Reset();
   
  HyperResolution.Orient();
  HYPERRESOLUTION::SetDinamicLHSOrdering(hyper_dinamic_lhs_ordering);
  HyperResolution.SetUnitConfict(hyper_unit_conflict);
  
  Kept.SetPickGivenRatio(pick_given_ratio);
  
  current_max_weight = max_weight;
  
  try_with_current_weight_limit:
   AdjustTmpClause();
   
   //cout << "INPUT TO KEPT\n"; // debug
   DCONTROL0("C30 in RunHyper(const bool positive_hyper,const float avail_time)"); 
   
   InputToKept();

   DCONTROL0("C40 in RunHyper(const bool positive_hyper,const float avail_time)"); 
   //cout << "DONE\n"; // debug

   //HyperResolution.OutputRules(); // debug
   
   ResetLoopTimeCount(avail_time);
   ResetLoopMemoryCount();
   
   //AdjustTmpClause();
   
   for (;;)  // main loop
    {
     //if (!PickFromKept())
     if (!UseNext())
      {
       if (Complete()) 
        {
         TerminateLoop(LTR_UNPROVABLE);
         return;
        };
        
       if ((weight_incrementation) && (IncMaxWeight()))
        {
         HyperResolution.ResetRules(); 
         ReleaseMemory();
         if (report_incremented_max_weight)
          {
           Output("************* Maximal weight of generated resolvents set to ");
           Output(current_max_weight);
           Output(".\n");
          };
         ResetEverything();
         goto try_with_current_weight_limit;
        };
       TerminateLoop(LTR_EMPTY_KEPT);
       return; 
      };
     if (LoopTimeExceeded()) { TerminateLoop(LTR_TIME); return; };
     //MoreResolvents(); 
    };  // exit main loop
 }; // void RunHyper(bool positive_hyper)

void RunBinary(const float avail_time)
 {
  hyper_resolution = false;

  //if (strat_auto) OutputMainValues();

  TERM::SetPositiveIsGreater(positive_is_greater);
  ORIndex.SetParamodulation(paramodulation);

  PreprocessInputClauses();
  ShowPreprocessedInputClauses();
  STATISTICS_ON_GENERATED::Reset();
  
  Kept.SetPickGivenRatio(pick_given_ratio);

  try_with_current_weight_limit:
   
   current_max_weight = max_weight;
   AdjustTmpClause();  


   //Resolvent.SetDiscriminateDefinitions(true); // experimental     

   Resolvent.SetEliminationTable(ProblemProfile.EliminationTable());
   if (prepro_elim_def && ProblemProfile.NumOfEliminatedPred())
    {
     Resolvent.SetEliminatePredicates(true); 
    };

   /*********
   if (prepro_elim_def && ProblemProfile.NumOfEliminatedPred())
    {
     Resolvent.SetEliminatePredicates(true);
     ProblemProfile.PredInfoTable().ResetIter();
     int predicate;
     while (ProblemProfile.PredInfoTable().Next(predicate))
       if (ProblemProfile.PredInfoTable().Info(predicate).Status == PRED_INFO::ELIMINATED)
        Resolvent.Eliminate(predicate);
    };
    *********/

   InputToKept();

   Resolvent.SetEliminatePredicates(false);

   //Resolvent.SetDiscriminateDefinitions(false);  // experimental

   
   ResetLoopTimeCount(avail_time);
   ResetLoopMemoryCount();
   
   //current_max_weight = max_weight;

   //AdjustTmpClause();
   
   DCONTROL0("B1 in RunBinary(..)");
   
   for (;;)  // main loop
    {
     
     DCONTROL0("B2 in RunBinary(..)");
     
     //if (!PickFromKept())
     if (!UseNext())
      {
       DCONTROL0("B3 in RunBinary(..)");
       if (Complete()) 
        {
         TerminateLoop(LTR_UNPROVABLE);
         return;
        };
       if ((weight_incrementation) && (IncMaxWeight()))
        {
         ReleaseMemory();
         if (report_incremented_max_weight)
          {
           Output("************* Maximal weight of generated resolvents set to ");
           Output(current_max_weight);
           Output(".\n");
          };
         ResetEverything();
         goto try_with_current_weight_limit;
        };
       TerminateLoop(LTR_EMPTY_KEPT);
       return;
      };
     DCONTROL0("B4 in RunBinary(..)");
     if (LoopTimeExceeded()) { TerminateLoop(LTR_TIME); return; };

     //MoreResolvents(); 

     /******** experimental *****/
    static bool once_purged = false;
 
    if ((!once_purged) && prepro_elim_def && (Kept.MaxPriorityEmpty())) 
     {
      once_purged = true;
      PurgeObsolete(); 
      RecycleTrash();
     };
   /*********/


     
    };  // exit main loop
 }; // void RunBinary()


void CreateProblemProfile()
 {
  ProblemProfile.Reset();
  Input.ResetIterator(); 
  Clause* c = Input.Next();
  while (c)
   {
    ProblemProfile.Analyse(c);
    c = Input.Next();
   };  
  ProblemProfile.MarkEliminated();
 };


void RunAuto()
 {
  Output("AUTOMATIC MODE.\n");
  //ProblemProfile.Output();
  
  SetDefaultValues(); 
 
  switch (ProblemProfile.Category())
   {
    case PROBLEM_PROFILE::HNE:
     hyper_resolution = true;
     static_splitting = false;
     prepro_elim_def = false;
     pick_given_ratio = 5;
          

     Output("*************** Trying negative hyperresolution...\n");
     RunHyper(false,AvailableTime()*auto_first_attempt); 
    
     Output("***** Negative resolution failed, trying positive hyperresolution...\n");     

     DCONTROL0("C100 in RunAuto()");

     DestroyPreprocessedClauses();

     DCONTROL0("C110 in RunAuto()");

     HyperResolution.ResetRules(); 

     DCONTROL0("C120 in RunAuto()");

     ReleaseMemory();

     DCONTROL0("C130 in RunAuto()");

     max_weight = (long)(OPTION::Value("max_weight"));
     current_max_weight = max_weight;
     
     DCONTROL0("C140 in RunAuto()"); 

     ResetEverything();
     
     DCONTROL0("C150 in RunAuto()");

     RunHyper(true,AvailableTime());
     break;

    case PROBLEM_PROFILE::NNE:
     prepro_elim_def = true;
     simplify_by_fs = (bool)(ProblemProfile.NumOfNotEliminatedPred());
     backward_subsumption = true;     
     pick_given_ratio = 3;

     Output("*************** Trying binary resolution...\n");
     RunBinary(AvailableTime());
     break;

    case PROBLEM_PROFILE::HEQ:
     paramodulation = true;
     prepro_elim_def = true;
     simplify_by_fs = (bool)(ProblemProfile.NumOfNotEliminatedPred());  
     simplify_forward_by_unit_eq = true;      
     simplify_backward_by_unit_eq = true;
     simplify_by_eq_res = true;

     max_memory = 60000.0;
    
     pick_given_ratio = 5;
     
     Output("*************** Trying binary resolution with paramodulation...\n");
     RunBinary(AvailableTime());
     break;

    case PROBLEM_PROFILE::NEQ:
     paramodulation = true;
     prepro_elim_def = true;
     simplify_by_fs = true; 
     simplify_forward_by_unit_eq = true;      
     simplify_backward_by_unit_eq = true;
     simplify_by_eq_res = true;

     max_memory = 60000.0;

     pick_given_ratio = 5;

     Output("*************** Trying binary resolution with paramodulation...\n");
     RunBinary(AvailableTime());
     break;

    case PROBLEM_PROFILE::PEQ:
     // show_kept = true; // debug

     paramodulation = true;
     prepro_elim_def = false;
     simplify_by_fs = false;
     simplify_forward_by_unit_eq = true;      
     simplify_backward_by_unit_eq = true;
     simplify_by_eq_res = true;

     max_memory = 60000.0;

     pick_given_ratio = 5;

     Output("*************** Trying binary resolution with paramodulation...\n");
     RunBinary(AvailableTime());
     break;

    case PROBLEM_PROFILE::UEQ:
     paramodulation = true;
     prepro_elim_def = false;
     simplify_by_fs = false;
     simplify_forward_by_unit_eq = true;      
     simplify_backward_by_unit_eq = true;

     max_memory = 60000.0;

     pick_given_ratio = 5;

     Output("*************** Trying binary resolution with paramodulation...\n");
     RunBinary(AvailableTime());
     break;
   };
 }; // void RunAuto()

void Prove(char* problemFN)
 {
  DCONTROL0("BEGIN in Prove(..)");
  AssertHeaderWeighting(HeaderWeighting,HWList); // debug
  //HeaderWeighting.Output(); // debug
 
  problemFileName = problemFN;
  MainTimer.Start();

  max_time = (float)cl_time_limit;
  SetOptions();

  //ORIndex.SetParamodulation(paramodulation);

  CLAUSE_SET SubsumedInput;
  if (input_remove_subsumed)
   {
    SeparateSubsumed(Input,SubsumedInput);
    while (!(SubsumedInput.Empty()))
     {
      SubsumedInput.Pop()->Destroy();
     };
   };
     
  CreateProblemProfile();
  
  ProblemProfile.Output();

  SeparateUnsupportedInput();

  
  if (strat_auto)
   {
    tabulate_to_file = false;
    tabulation_file_name = "vampire.tab";
    output_to_file = false;
    output_file_name = "vampire.out"; 
   };
    
  AdjustOutput();
  ReportStart(problemFileName);
   
  if (strat_auto)
   {
    RunAuto();
   }
  else
   {
    if (hyper_resolution)
     {
      RunHyper(!(positive_is_greater),AvailableTime());
     }
    else
     {
      RunBinary(AvailableTime());
     };
   };
  TerminateProgram();
};  // Prove()


//*******************************************************************************
