#define __WEIGHTING_H

#ifndef __DEBUG_H
 #include "debug.h"
#endif

#ifndef __VAMPIRE_H
 #include "vampire.h"
#endif

#ifndef __TERM_H
 #include "term.h"
#endif

#ifndef __OUTPUT_H
 #include "output.h"
#endif

//**************************************************************************

class HEADER_WEIGHTING
 {
  private:
   
   //long bullshit1[10000]; // debug

   long weight[MAX_NUMBER_OF_DIFFERENT_HEADERS];

   //long bullshit2[10000]; // debug
   

   long min_weight;
   long max_weight;
  private:
   long& WeightRef(const TERM header)
    {
     CHECK8((header.Functor()*2)+header.Polarity() >= 0,
            "A0 in HEADER_WEIGHTING::WeightRef(const TERM header)");
     CHECK8((header.Functor()*2)+header.Polarity() < MAX_NUMBER_OF_DIFFERENT_HEADERS,
             "A0 in HEADER_WEIGHTING::WeightRef(const TERM header)"); 
     return weight[(header.Functor()*2)+header.Polarity()];
    };
  public:
   HEADER_WEIGHTING() { Reset(); };
   inline void Reset() 
    {
     min_weight = LONG_MAX;
     max_weight = LONG_MIN;
     for (int i = 0; i < MAX_NUMBER_OF_DIFFERENT_HEADERS; i++)
      {
       weight[i] = 0L;
      };
    };
   inline long Weight(const TERM header) { return WeightRef(header); };
   inline void SetWeight(const TERM header,const long w) 
    {
     WeightRef(header) = w;
     if (w > max_weight) { max_weight = w; };
     if (w < min_weight) { min_weight = w; };
    };
   inline long MinWeight() const { return min_weight; };
   inline long MaxWeight() const { return max_weight; };
   inline void MakeMinimal(TERM header)
    {     
     long new_weight = MinWeight() - 1;
     if (!new_weight) { new_weight = -1; };
     SetWeight(header,new_weight);
    };
   inline void MakeMaximal(TERM header)
    {
     long new_weight = MaxWeight() + 1;
     if (!new_weight) { new_weight = 1; };
     SetWeight(header,new_weight);
    };
   inline COMPARISON Compare(const TERM header1,const TERM header2)
    {
     if (Weight(header1) > Weight(header2)) { return GREATER; }
     else
      {
       if (Weight(header1) < Weight(header2)) { return LESS; }
       else { return EQUAL; };
      };
    };
  public: // for debugging
   inline void Output()
    {
     for (int pred = 0; pred < MAX_NUMBER_OF_PREDICATES; pred++)
      {
       TERM h_pos((ARITY)0,(FUNCTOR)pred,POL_POS); 
       TERM h_neg((ARITY)0,(FUNCTOR)pred,POL_NEG); 
       ::Output("weight("); ::Output(h_pos); ::Output(") = ");
       ::Output(Weight(h_pos)); ::Output("   ");
       ::Output("weight("); ::Output(h_neg); ::Output(") = ");
       ::Output(Weight(h_neg)); ::Output("\n");
      };
    };
 }; // class HEADER_WEIGHTING

class HEADER_WEIGHT
 {
  public:
   enum EXTREMUM
    {
     MIN,
     MAX
    };
  private:
   TERM header;
   long weight;
  public:
   HEADER_WEIGHT(const TERM h,const long w) 
    : header(h),
      weight(w)
    {};
   HEADER_WEIGHT(const TERM h,const EXTREMUM e) 
    : header(h)
    {
     if (e == MIN)
      {
       weight = LONG_MIN;
      }
     else
      {
       weight = LONG_MAX;
      };
    };
   TERM Header() const { return header; };
   long Weight() const { return weight; };
   bool Min() const { return (weight == LONG_MIN); };
   bool Max() const { return (weight == LONG_MAX); };
   bool Extremal() const { return (Min() || Max()); };
  public: // for debugging
   inline void Output() const 
    {
     ::Output("weight("); ::Output(header); ::Output(") = ");
     if (Min()) { ::Output("min"); }
     else 
      {
       if (Max()) { ::Output("max"); }
       else { ::Output(weight); };
      }; 
    };
 }; // class HEADER_WEIGHT

class HEADER_WEIGHT_LIST
 {
  private:
   const HEADER_WEIGHT* h_weight;
   HEADER_WEIGHT_LIST* next;
  public:
   HEADER_WEIGHT_LIST(const HEADER_WEIGHT* const h_w,HEADER_WEIGHT_LIST* const n)
    : h_weight(h_w),
      next(n)
     {};
   inline const HEADER_WEIGHT* HWeight() const { return h_weight; };
   inline HEADER_WEIGHT_LIST* Next() const { return next; };
   static inline void Add(const HEADER_WEIGHT* const h_w,HEADER_WEIGHT_LIST*& lst)
    {
     lst = new HEADER_WEIGHT_LIST(h_w,lst);
    }; 
  public: // for debugging
   static inline void Output(const HEADER_WEIGHT_LIST* lst)
    {
     while (lst)
      {
       lst->HWeight()->Output();
       ::Output("\n");
       lst = lst->Next();
      };
    };
 };

inline void AssertHeaderWeighting(HEADER_WEIGHTING& weighting,
                                  const HEADER_WEIGHT_LIST* const lst)
 {
  weighting.Reset();
  weighting.SetWeight(TERM((ARITY)0,(FUNCTOR)(MAX_NUMBER_OF_PREDICATES-1),POL_POS),1);
  const HEADER_WEIGHT_LIST* tmp_lst = lst;
  const HEADER_WEIGHT* h_w;
  while (tmp_lst)
   {
    h_w = tmp_lst->HWeight();
    if (!(h_w->Extremal())) 
     {
      weighting.SetWeight(h_w->Header(),h_w->Weight());
     };
    tmp_lst = tmp_lst->Next();
   };
  
  for (int pred = 0; pred < MAX_NUMBER_OF_PREDICATES; pred++)
   {
    TERM h_pos((ARITY)0,(FUNCTOR)pred,POL_POS);
    TERM h_neg((ARITY)0,(FUNCTOR)pred,POL_NEG);
    if (!(weighting.Weight(h_pos))) 
     {
      weighting.MakeMinimal(h_pos);
     };
    if (!(weighting.Weight(h_neg))) 
     {
      weighting.MakeMaximal(h_neg);
     };
   };  
 
  tmp_lst = lst; 
  while (tmp_lst)
   {
    h_w = tmp_lst->HWeight();
    if (h_w->Min()) 
     {
      weighting.MakeMinimal(h_w->Header());
     }
    else
     {
      if (h_w->Max())
       {
        weighting.MakeMaximal(h_w->Header());
       }; 
     };
    tmp_lst = tmp_lst->Next();
   };
 }; // void AssertHeaderWeighting(HEADER_WEIGHTING& weighting,const HEADER_WEIGHT_LIST* const lst)                            



//**************************************************************************
